@@ -134,8 +134,7 @@ void CL_Disconnect (void)
 
 // stop sounds (especially looping!)
 	S_StopAllSounds (true);
-	BGM_Stop();
-	CDAudio_Stop();
+	BGM_Pause ();
 
 // if running a local server, shut it down
 	if (cls.demoplayback)
@@ -169,6 +168,8 @@ void CL_Disconnect (void)
 void CL_Disconnect_f (void)
 {
 	CL_Disconnect ();
+	BGM_Stop ();
+	CDAudio_Stop ();
 	if (sv.active)
 		Host_ShutdownServer (false);
 }
@@ -240,6 +241,7 @@ void CL_SignonReply (void)
 		break;
 
 	case 4:
+		cl.spawntime = cl.mtime[0];
 		SCR_EndLoadingPlaque ();		// allow normal screen updates
 		break;
 	}
@@ -304,6 +306,44 @@ void CL_PrintEntities_f (void)
 	}
 }
 
+/*
+===============
+CL_IsPlayerEnt
+===============
+*/
+qboolean CL_IsPlayerEnt (const entity_t *ent)
+{
+	return (uintptr_t)ent - (uintptr_t)(cl_entities + 1) < (uintptr_t)cl.maxclients;
+}
+
+/*
+===============
+CL_SetLightstyle
+
+Sets the lightstyle map and updates the length, peak, and average values
+===============
+*/
+void CL_SetLightstyle (int i, const char *str)
+{
+	q_strlcpy (cl_lightstyle[i].map, str, MAX_STYLESTRING);
+	cl_lightstyle[i].length = Q_strlen(cl_lightstyle[i].map);
+	//johnfitz -- save extra info
+	if (cl_lightstyle[i].length)
+	{
+		int j, total = 0;
+		cl_lightstyle[i].peak = 'a';
+		for (j=0; j<cl_lightstyle[i].length; j++)
+		{
+			total += cl_lightstyle[i].map[j] - 'a';
+			cl_lightstyle[i].peak = q_max(cl_lightstyle[i].peak, cl_lightstyle[i].map[j]);
+		}
+		cl_lightstyle[i].average = total / cl_lightstyle[i].length + 'a';
+	}
+	else
+		cl_lightstyle[i].average = cl_lightstyle[i].peak = 'm';
+	//johnfitz
+}
+
 /*
 ===============
 CL_AllocDlight
@@ -326,6 +366,7 @@ dlight_t *CL_AllocDlight (int key)
 				memset (dl, 0, sizeof(*dl));
 				dl->key = key;
 				dl->color[0] = dl->color[1] = dl->color[2] = 1; //johnfitz -- lit support via lordhavoc
+				dl->spawn = cl.time - 0.001;
 				return dl;
 			}
 		}
@@ -335,11 +376,12 @@ dlight_t *CL_AllocDlight (int key)
 	dl = cl_dlights;
 	for (i=0 ; i<MAX_DLIGHTS ; i++, dl++)
 	{
-		if (dl->die < cl.time)
+		if (dl->die < cl.time || dl->spawn > cl.time)
 		{
 			memset (dl, 0, sizeof(*dl));
 			dl->key = key;
 			dl->color[0] = dl->color[1] = dl->color[2] = 1; //johnfitz -- lit support via lordhavoc
+			dl->spawn = cl.time - 0.001;
 			return dl;
 		}
 	}
@@ -348,6 +390,7 @@ dlight_t *CL_AllocDlight (int key)
 	memset (dl, 0, sizeof(*dl));
 	dl->key = key;
 	dl->color[0] = dl->color[1] = dl->color[2] = 1; //johnfitz -- lit support via lordhavoc
+	dl->spawn = cl.time - 0.001;
 	return dl;
 }
 
@@ -369,7 +412,7 @@ void CL_DecayLights (void)
 	dl = cl_dlights;
 	for (i=0 ; i<MAX_DLIGHTS ; i++, dl++)
 	{
-		if (dl->die < cl.time || !dl->radius)
+		if (dl->die < cl.time || dl->spawn > cl.time || !dl->radius)
 			continue;
 
 		dl->radius -= time*dl->decay;
@@ -682,9 +725,7 @@ int CL_ReadFromServer (void)
 	dlight_t	*l; //johnfitz
 	int			i; //johnfitz
 
-
-	cl.oldtime = cl.time;
-	cl.time += host_frametime;
+	CL_AdvanceTime ();
 
 	do
 	{
@@ -720,7 +761,7 @@ int CL_ReadFromServer (void)
 
 	//beams
 	for (i=0, b=cl_beams ; i< MAX_BEAMS ; i++, b++)
-		if (b->model && b->endtime >= cl.time)
+		if (b->model && b->starttime <= cl.time && b->endtime >= cl.time)
 			num_beams++;
 	if (num_beams > 24 && dev_peakstats.beams <= 24)
 		Con_DWarning ("%i beams exceeded standard limit of 24 (max = %d).\n", num_beams, MAX_BEAMS);
@@ -729,7 +770,7 @@ int CL_ReadFromServer (void)
 
 	//dlights
 	for (i=0, l=cl_dlights ; i<MAX_DLIGHTS ; i++, l++)
-		if (l->die >= cl.time && l->radius)
+		if (l->die >= cl.time && l->spawn <= cl.time && l->radius)
 			num_dlights++;
 	if (num_dlights > 32 && dev_peakstats.dlights <= 32)
 		Con_DWarning ("%i dlights exceeded standard limit of 32 (max = %d).\n", num_dlights, MAX_DLIGHTS);
@@ -849,33 +890,51 @@ void CL_Viewpos_f (void)
 	char buf[256];
 	if (cls.state != ca_connected)
 		return;
-#if 0
-	//camera position
+
+	// player position
 	q_snprintf (buf, sizeof (buf),
-		"(%i %i %i) %i %i %i",
-		(int)r_refdef.vieworg[0],
-		(int)r_refdef.vieworg[1],
-		(int)r_refdef.vieworg[2],
-		(int)r_refdef.viewangles[PITCH],
-		(int)r_refdef.viewangles[YAW],
-		(int)r_refdef.viewangles[ROLL]);
-#else
-	//player position
-	q_snprintf (buf, sizeof (buf),
-		"(%i %i %i) %i %i %i",
-		(int)cl_entities[cl.viewentity].origin[0],
-		(int)cl_entities[cl.viewentity].origin[1],
-		(int)cl_entities[cl.viewentity].origin[2],
-		(int)cl.viewangles[PITCH],
-		(int)cl.viewangles[YAW],
-		(int)cl.viewangles[ROLL]
+		"(%.0f %.0f %.0f) %.0f %.0f %.0f",
+		cl_entities[cl.viewentity].origin[0],
+		cl_entities[cl.viewentity].origin[1],
+		cl_entities[cl.viewentity].origin[2],
+		cl.viewangles[PITCH],
+		cl.viewangles[YAW],
+		cl.viewangles[ROLL]
 	);
-#endif
-	Con_Printf ("Viewpos: %s\n", buf);
+	Con_SafePrintf ("Player pos: %s\n", buf);
 
 	if (Cmd_Argc () >= 2 && !q_strcasecmp (Cmd_Argv (1), "copy"))
 		if (SDL_SetClipboardText (buf) < 0)
-			Con_Printf ("Clipboard copy failed: %s\n", SDL_GetError ());
+			Con_SafePrintf ("Clipboard copy failed: %s\n", SDL_GetError ());
+
+	// camera position
+	Con_SafePrintf ("Camera pos: (%.0f %.0f %.0f) %.0f %.0f %.0f\n",
+		r_refdef.vieworg[0],
+		r_refdef.vieworg[1],
+		r_refdef.vieworg[2],
+		r_refdef.viewangles[PITCH],
+		r_refdef.viewangles[YAW],
+		r_refdef.viewangles[ROLL]
+	);
+
+	// sun mangle
+	Con_SafePrintf ("Sun mangle: %.0f %.0f %.0f\n",
+		NormalizeAngle (r_refdef.viewangles[YAW]) + 180.f,
+		r_refdef.viewangles[PITCH],
+		r_refdef.viewangles[ROLL]
+	);
+}
+
+/*
+===============
+CL_Viewpos_Completion_f -- tab completion for the viewpos command
+===============
+*/
+static void CL_Viewpos_Completion_f (const char *partial)
+{
+	if (Cmd_Argc () != 2)
+		return;
+	Con_AddToTabList ("copy", partial, NULL);
 }
 
 /*
@@ -939,6 +998,8 @@ CL_Init
 */
 void CL_Init (void)
 {
+	cmd_function_t *cmd;
+
 	SZ_Alloc (&cls.message, 1024);
 
 	CL_InitInput ();
@@ -986,7 +1047,9 @@ void CL_Init (void)
 	Cmd_AddCommand ("timedemo", CL_TimeDemo_f);
 
 	Cmd_AddCommand ("tracepos", CL_Tracepos_f); //johnfitz
-	Cmd_AddCommand ("viewpos", CL_Viewpos_f); //johnfitz
+	cmd = Cmd_AddCommand ("viewpos", CL_Viewpos_f); //johnfitz
+	if (cmd)
+		cmd->completion = CL_Viewpos_Completion_f;
 
 	Cmd_AddCommand_ServerCommand ("st", CL_SetStat_f);
 	Cmd_AddCommand_ServerCommand ("sts", CL_SetStatString_f);
