@@ -25,6 +25,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 #include "quakedef.h"
 #include "bgmusic.h"
+#include "steam.h"
 
 const char *svc_strings[] =
 {
@@ -97,7 +98,7 @@ const char *svc_strings[] =
 	"svc_backtolobby", // 55
 	"svc_localsound" // 56
 };
-#define	NUM_SVC_STRINGS	(sizeof(svc_strings) / sizeof(svc_strings[0]))
+#define NUM_SVC_STRINGS Q_COUNTOF(svc_strings)
 
 qboolean warn_about_nehahra_protocol; //johnfitz
 
@@ -458,6 +459,7 @@ void CL_ParseUpdate (int bits)
 	entity_t	*ent;
 	int		num;
 	int		skin;
+	int		prevframe;
 
 	if (cls.signon == SIGNONS - 1)
 	{	// first update is the final signon stage
@@ -509,6 +511,7 @@ void CL_ParseUpdate (int bits)
 	else
 		modnum = ent->baseline.modelindex;
 
+	prevframe = ent->frame;
 	if (bits & U_FRAME)
 		ent->frame = MSG_ReadByte ();
 	else
@@ -639,7 +642,9 @@ void CL_ParseUpdate (int bits)
 	// or randomized
 		if (model)
 		{
-			if (model->synctype == ST_RAND)
+			if (model->synctype == ST_FRAMETIME)
+				ent->syncbase = -cl.time;
+			else if (model->synctype == ST_RAND)
 				ent->syncbase = (float)(rand()&0x7fff) / 0x7fff;
 			else
 				ent->syncbase = 0.0;
@@ -652,6 +657,8 @@ void CL_ParseUpdate (int bits)
 		ent->lerpflags |= LERP_RESETANIM; //johnfitz -- don't lerp animation across model changes
 	}
 	//johnfitz
+	else if (model && model->synctype == ST_FRAMETIME && ent->frame != prevframe)
+		ent->syncbase = -cl.time;
 
 	if ( forcelink )
 	{	// didn't have an update last message
@@ -1064,7 +1071,7 @@ void CL_ParseServerMessage (void)
 	int			cmd;
 	int			i;
 	const char		*str; //johnfitz
-	int			total, j, lastcmd; //johnfitz
+	int			lastcmd; //johnfitz
 
 //
 // if recording demos, copy the message out
@@ -1096,6 +1103,7 @@ void CL_ParseServerMessage (void)
 			if (*cl.stuffcmdbuf && net_message.cursize < 512)
 				CL_ParseStuffText("\n");	//there's a few mods that forget to write \ns, that then fuck up other things too. So make sure it gets flushed to the cbuf. the cursize check is to reduce backbuffer overflows that would give a false positive.
 
+			CL_FinishDemoFrame ();
 			return;		// end of message
 		}
 
@@ -1184,23 +1192,7 @@ void CL_ParseServerMessage (void)
 			i = MSG_ReadByte ();
 			if (i >= MAX_LIGHTSTYLES)
 				Sys_Error ("svc_lightstyle > MAX_LIGHTSTYLES");
-			q_strlcpy (cl_lightstyle[i].map, MSG_ReadString(), MAX_STYLESTRING);
-			cl_lightstyle[i].length = Q_strlen(cl_lightstyle[i].map);
-			//johnfitz -- save extra info
-			if (cl_lightstyle[i].length)
-			{
-				total = 0;
-				cl_lightstyle[i].peak = 'a';
-				for (j=0; j<cl_lightstyle[i].length; j++)
-				{
-					total += cl_lightstyle[i].map[j] - 'a';
-					cl_lightstyle[i].peak = q_max(cl_lightstyle[i].peak, cl_lightstyle[i].map[j]);
-				}
-				cl_lightstyle[i].average = total / cl_lightstyle[i].length + 'a';
-			}
-			else
-				cl_lightstyle[i].average = cl_lightstyle[i].peak = 'm';
-			//johnfitz
+			CL_SetLightstyle (i, MSG_ReadString ());
 			break;
 
 		case svc_sound:
@@ -1382,7 +1374,10 @@ void CL_ParseServerMessage (void)
 		//used by the 2021 rerelease
 		case svc_achievement:
 			str = MSG_ReadString();
-			Con_DPrintf("Ignoring svc_achievement (%s)\n", str);
+			if (cls.demoplayback)
+				Con_DPrintf ("Ignoring svc_achievement (%s)\n", str);
+			else if (!Steam_SetAchievement (str))
+				Con_DPrintf ("Couldn't set achievement \"%s\"\n", str);
 			break;
 		case svc_localsound:
 			CL_ParseLocalSound();
