@@ -286,7 +286,7 @@ Cmd_Exec_f
 void Cmd_Exec_f (void)
 {
 	const char *path;
-	char		*f;
+	const char	*f;
 	int			mark;
 
 	if (Cmd_Argc () < 2)
@@ -303,7 +303,7 @@ void Cmd_Exec_f (void)
 	// "exec config.cfg pls" will execute config.cfg
 	if (Cmd_Argc () == 2 && !strcmp (path, "config.cfg"))
 	{
-		f = (char *)COM_LoadHunkFile (CONFIG_NAME, NULL);
+		f = (const char *)COM_LoadHunkFile (CONFIG_NAME, NULL);
 		if (f)
 		{
 			path = CONFIG_NAME;
@@ -311,7 +311,7 @@ void Cmd_Exec_f (void)
 		}
 	}
 
-	f = (char *)COM_LoadHunkFile (path, NULL);
+	f = (const char *)COM_LoadHunkFile (path, NULL);
 	if (!f && !strcmp(Cmd_Argv(1), "default.cfg")) {
 		f = default_cfg;	/* see above.. */
 	}
@@ -778,6 +778,8 @@ cmd_function_t *Cmd_AddCommand2 (const char *cmd_name, xcommand_t function, cmd_
 	if (host_initialized)
 	{
 		cmd = (cmd_function_t *) malloc(sizeof(*cmd) + strlen(cmd_name)+1);
+		if (!cmd)
+			Sys_Error ("Cmd_AddCommand2: out of memory (%s)", cmd_name);
 		cmd->name = strcpy((char*)(cmd + 1), cmd_name);
 		cmd->dynamic = true;
 	}
