@@ -30,7 +30,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include "miniz.h"
 #include "unicode_translit.h"
 
-static char	*largv[MAX_NUM_ARGVS + 1];
+static const char	*largv[MAX_NUM_ARGVS + 1];
 static char	argvdummy[] = " ";
 
 int		safemode;
@@ -55,8 +55,8 @@ static void COM_Path_f (void);
 #define PAK0_CRC_V091		28804	/* id1/pak0.pak - v0.91/0.92, not supported */
 
 THREAD_LOCAL char	com_token[1024];
-int		com_argc;
-char	**com_argv;
+int			com_argc;
+const char	**com_argv;
 
 #define CMDLINE_LENGTH	256		/* johnfitz -- mirrored in cmd.c */
 char	com_cmdline[CMDLINE_LENGTH];
@@ -210,6 +210,17 @@ void Vec_Free (void **pvec)
 	}
 }
 
+void MultiString_Append (char **pvec, const char *str)
+{
+	Vec_Append ((void **)pvec, 1, str, strlen (str) + 1);
+}
+
+void MultiString_AppendN (char **pvec, const char *str, size_t len)
+{
+	Vec_Append ((void **)pvec, 1, str, len);
+	VEC_PUSH (*pvec, '\0');
+}
+
 /*
 ============================================================================
 
@@ -220,9 +231,20 @@ void Vec_Free (void **pvec)
 
 int q_strnaturalcmp (const char *s1, const char *s2)
 {
+	qboolean neg1, neg2, sign1, sign2;
+
 	if (s1 == s2)
 		return 0;
 
+	neg1 = *s1 == '-';
+	neg2 = *s2 == '-';
+	sign1 = neg1 || *s1 == '+';
+	sign2 = neg2 || *s2 == '+';
+
+	// early out if strings start with different signs followed by digits
+	if (neg1 != neg2 && q_isdigit (s1[sign1]) && q_isdigit (s1[sign2]))
+		return neg2 - neg1;
+
 skip_prefix:
 	while (*s1 && !q_isdigit (*s1) && q_toupper (*s1) == q_toupper (*s2))
 	{
@@ -235,7 +257,7 @@ skip_prefix:
 	{
 		const char *begin1 = s1++;
 		const char *begin2 = s2++;
-		int diff;
+		int diff, sign;
 
 		while (*begin1 == '0')
 			begin1++;
@@ -247,17 +269,23 @@ skip_prefix:
 		while (q_isdigit (*s2))
 			s2++;
 
+		sign = neg1 ? -1 : 1;
+
 		diff = (s1 - begin1) - (s2 - begin2);
 		if (diff)
-			return diff;
+			return diff * sign;
 
 		while (begin1 != s1)
 		{
 			diff = *begin1++ - *begin2++;
 			if (diff)
-				return diff;
+				return diff * sign;
 		}
 
+		// We only support negative numbers at the beginning of strings so that
+		// "-2" is sorted before "-1", but "file-2345.ext" *after* "file-1234.ext".
+		neg1 = neg2 = false;
+
 		goto skip_prefix;
 	}
 
@@ -304,44 +332,22 @@ int q_strncasecmp(const char *s1, const char *s2, size_t n)
 	return (int)(c1 - c2);
 }
 
-//spike -- grabbed this from fte, because its useful to me
 char *q_strcasestr(const char *haystack, const char *needle)
 {
-	int c1, c2, c2f;
-	int i;
-	c2f = *needle;
-	if (c2f >= 'a' && c2f <= 'z')
-		c2f -= ('a' - 'A');
-	if (!c2f)
-		return (char*)haystack;
-	while (1)
+	const size_t len = strlen(needle);
+
+	if (!len)
+		return (char *)haystack;
+
+	while (*haystack)
 	{
-		c1 = *haystack;
-		if (!c1)
-			return NULL;
-		if (c1 >= 'a' && c1 <= 'z')
-			c1 -= ('a' - 'A');
-		if (c1 == c2f)
-		{
-			for (i = 1; ; i++)
-			{
-				c1 = haystack[i];
-				c2 = needle[i];
-				if (c1 >= 'a' && c1 <= 'z')
-					c1 -= ('a' - 'A');
-				if (c2 >= 'a' && c2 <= 'z')
-					c2 -= ('a' - 'A');
-				if (!c2)
-					return (char*)haystack;	//end of needle means we found a complete match
-				if (!c1)	//end of haystack means we can't possibly find needle in it any more
-					return NULL;
-				if (c1 != c2)	//mismatch means no match starting at haystack[0]
-					break;
-			}
-		}
-		haystack++;
+		if (!q_strncasecmp(haystack, needle, len))
+			return (char *)haystack;
+
+		++haystack;
 	}
-	return NULL;	//didn't find it
+
+	return NULL;
 }
 
 char *q_strlwr (char *str)
@@ -667,7 +673,6 @@ float Q_atof (const char *str)
 	return val*sign;
 }
 
-
 /*
 ==============================================================================
 
@@ -1043,6 +1048,37 @@ void SZ_Print (sizebuf_t *buf, const char *data)
 
 //============================================================================
 
+/*
+============
+COM_FirstPathSep
+
+Returns a pointer to the first path separator, if any, or the end of the string
+============
+*/
+const char *COM_FirstPathSep (const char *path)
+{
+	while (*path && !Sys_IsPathSep (*path))
+		path++;
+	return path;
+}
+
+/*
+============
+COM_NormalizePath
+
+Replaces all path separators with forward slashes
+============
+*/
+void COM_NormalizePath (char *path)
+{
+	while (*path)
+	{
+		if (Sys_IsPathSep (*path))
+			*path = '/';
+		path++;
+	}
+}
+
 /*
 ============
 COM_SkipPath
@@ -1062,6 +1098,19 @@ const char *COM_SkipPath (const char *pathname)
 	return last;
 }
 
+
+/*
+============
+COM_SkipSpace
+============
+*/
+const char *COM_SkipSpace (const char *str)
+{
+	while (q_isspace (*str))
+		str++;
+	return str;
+}
+
 /*
 ============
 COM_StripExtension
@@ -1253,12 +1302,16 @@ char *COM_TintString (const char *in, char *out, size_t outsize)
 
 /*
 ==============
-COM_Parse
+COM_ParseEx
 
 Parse a token out of a string
+
+The mode argument controls how overflow is handled:
+- CPE_NOTRUNC:		return NULL (abort parsing)
+- CPE_ALLOWTRUNC:	truncate com_token (ignore the extra characters in this token)
 ==============
 */
-const char *COM_Parse (const char *data)
+const char *COM_ParseEx (const char *data, cpe_mode mode)
 {
 	int		c;
 	int		len;
@@ -1310,16 +1363,20 @@ skipwhite:
 				com_token[len] = 0;
 				return data;
 			}
-			com_token[len] = c;
-			len++;
+			if (len < Q_COUNTOF(com_token) - 1)
+				com_token[len++] = c;
+			else if (mode == CPE_NOTRUNC)
+				return NULL;
 		}
 	}
 
 // parse single characters
 	if (c == '{' || c == '}'|| c == '('|| c == ')' || c == '\'' || c == ':')
 	{
-		com_token[len] = c;
-		len++;
+		if (len < Q_COUNTOF(com_token) - 1)
+			com_token[len++] = c;
+		else if (mode == CPE_NOTRUNC)
+			return NULL;
 		com_token[len] = 0;
 		return data+1;
 	}
@@ -1327,9 +1384,11 @@ skipwhite:
 // parse a regular word
 	do
 	{
-		com_token[len] = c;
+		if (len < Q_COUNTOF(com_token) - 1)
+			com_token[len++] = c;
+		else if (mode == CPE_NOTRUNC)
+			return NULL;
 		data++;
-		len++;
 		c = *data;
 		/* commented out the check for ':' so that ip:port works */
 		if (c == '{' || c == '}'|| c == '('|| c == ')' || c == '\''/* || c == ':' */)
@@ -1341,6 +1400,47 @@ skipwhite:
 }
 
 
+/*
+==============
+COM_Parse
+
+Parse a token out of a string
+
+Return NULL in case of overflow
+==============
+*/
+const char *COM_Parse (const char *data)
+{
+	return COM_ParseEx (data, CPE_NOTRUNC);
+}
+
+
+/*
+================
+COM_ParseLine
+================
+*/
+qboolean COM_ParseLine (const char **str, stringview_t *line)
+{
+	const char *p;
+
+	if (!str || !*str)
+		return false;
+
+	p = *str;
+	if (line)
+		line->data = p;
+	while (*p && *p != '\n')
+		p++;
+	if (line)
+		line->len = p - line->data;
+
+	*str = (*p == '\n') ? p + 1 : NULL;
+
+	return true;
+}
+
+
 /*
 ================
 COM_CheckParm
@@ -1399,13 +1499,17 @@ static void COM_CheckRegistered (void)
 		return;
 	}
 
-	Sys_FileRead (h, check, sizeof(check));
+	i = Sys_FileRead (h, check, sizeof(check));
 	COM_CloseFile (h);
+	if (i != (int) sizeof(check))
+		goto corrupt;
 
 	for (i = 0; i < 128; i++)
 	{
 		if (pop[i] != (unsigned short)BigShort (check[i]))
+		{ corrupt:
 			Sys_Error ("Corrupted data file.");
+		}
 	}
 
 	for (i = 0; com_cmdline[i]; i++)
@@ -1475,6 +1579,22 @@ void COM_InitArgv (int argc, char **argv)
 	}
 }
 
+/*
+================
+COM_AddArg
+
+Returns the index of the new argument or -1 on overflow
+================
+*/
+int COM_AddArg (const char *arg)
+{
+	if (com_argc >= MAX_NUM_ARGVS)
+		return -1;
+	com_argv[com_argc] = arg;
+	com_argv[com_argc + 1] = argvdummy;
+	return com_argc++;
+}
+
 /*
 ================
 COM_Init
@@ -1529,7 +1649,7 @@ QUAKE FILESYSTEM
 =============================================================================
 */
 
-THREAD_LOCAL int com_filesize;
+THREAD_LOCAL qfileofs_t com_filesize;
 
 
 //
@@ -1666,6 +1786,43 @@ char *COM_TempSuffix (unsigned seq)
 	return buf;
 }
 
+/*
+============
+COM_DescribeDuration
+
+Describes the given duration, e.g. "3 minutes"
+============
+*/
+void COM_DescribeDuration (char *out, size_t outsize, double seconds)
+{
+	const double SECOND = 1;
+	const double MINUTE = 60 * SECOND;
+	const double HOUR = 60 * MINUTE;
+	const double DAY = 24 * HOUR;
+	const double WEEK = 7 * DAY;
+	const double MONTH = 30.436875 * DAY;
+	const double YEAR = 365.2425 * DAY;
+
+	seconds = fabs (seconds);
+
+	if (seconds < 1)
+		q_strlcpy (out, "moments", outsize);
+	else if (seconds < 60 * SECOND)
+		q_snprintf (out, outsize, "%i second%s", PLURAL (seconds));
+	else if (seconds < 90 * MINUTE)
+		q_snprintf (out, outsize, "%i minute%s", PLURAL (seconds / MINUTE));
+	else if (seconds < DAY)
+		q_snprintf (out, outsize, "%i hour%s", PLURAL (seconds / HOUR));
+	else if (seconds < WEEK)
+		q_snprintf (out, outsize, "%i day%s", PLURAL (seconds / DAY));
+	else if (seconds < MONTH)
+		q_snprintf (out, outsize, "%i week%s", PLURAL (seconds / WEEK));
+	else if (seconds < YEAR)
+		q_snprintf (out, outsize, "%i month%s", PLURAL (seconds / MONTH));
+	else
+		q_snprintf (out, outsize, "%i year%s", PLURAL (seconds / YEAR));
+}
+
 /*
 ============
 COM_CreatePath
@@ -1691,14 +1848,14 @@ void COM_CreatePath (char *path)
 COM_filelength
 ================
 */
-long COM_filelength (FILE *f)
+qfileofs_t COM_filelength (FILE *f)
 {
-	long		pos, end;
+	qfileofs_t	pos, end;
 
-	pos = ftell (f);
-	fseek (f, 0, SEEK_END);
-	end = ftell (f);
-	fseek (f, pos, SEEK_SET);
+	pos = Sys_ftell (f);
+	Sys_fseek (f, 0, SEEK_END);
+	end = Sys_ftell (f);
+	Sys_fseek (f, pos, SEEK_SET);
 
 	return end;
 }
@@ -1795,13 +1952,19 @@ static int COM_FindFile (const char *filename, int *handle, FILE **file,
 		}
 	}
 
-	if (strcmp(COM_FileGetExtension(filename), "pcx") != 0
-		&& strcmp(COM_FileGetExtension(filename), "tga") != 0
-		&& strcmp(COM_FileGetExtension(filename), "lit") != 0
-		&& strcmp(COM_FileGetExtension(filename), "vis") != 0
-		&& strcmp(COM_FileGetExtension(filename), "ent") != 0)
-		Con_DPrintf ("FindFile: can't find %s\n", filename);
-	else	Con_DPrintf2("FindFile: can't find %s\n", filename);
+	if (developer.value)
+	{
+		const char *ext = COM_FileGetExtension (filename);
+
+		if (strcmp(ext, "pcx") != 0 &&
+			strcmp(ext, "tga") != 0 &&
+			strcmp(ext, "lit") != 0 &&
+			strcmp(ext, "vis") != 0 &&
+			strcmp(ext, "ent") != 0)
+			Con_DPrintf ("FindFile: can't find %s\n", filename);
+		else
+			Con_DPrintf2 ("FindFile: can't find %s\n", filename);
+	}
 
 	if (handle)
 		*handle = -1;
@@ -1879,23 +2042,15 @@ Filename are reletive to the quake directory.
 Allways appends a 0 byte.
 ============
 */
-#define	LOADFILE_ZONE		0
-#define	LOADFILE_HUNK		1
-#define	LOADFILE_TEMPHUNK	2
-#define	LOADFILE_CACHE		3
-#define	LOADFILE_STACK		4
-#define	LOADFILE_MALLOC		5
-
-static byte	*loadbuf;
-static cache_user_t *loadcache;
-static int	loadsize;
+#define	LOADFILE_HUNK		0
+#define	LOADFILE_MALLOC		1
 
 byte *COM_LoadFile (const char *path, int usehunk, unsigned int *path_id)
 {
 	int		h;
 	byte	*buf;
 	char	base[32];
-	int		len;
+	int	len, nread;
 
 	buf = NULL;	// quiet compiler warning
 
@@ -1910,22 +2065,7 @@ byte *COM_LoadFile (const char *path, int usehunk, unsigned int *path_id)
 	switch (usehunk)
 	{
 	case LOADFILE_HUNK:
-		buf = (byte *) Hunk_AllocName (len+1, base);
-		break;
-	case LOADFILE_TEMPHUNK:
-		buf = (byte *) Hunk_TempAlloc (len+1);
-		break;
-	case LOADFILE_ZONE:
-		buf = (byte *) Z_Malloc (len+1);
-		break;
-	case LOADFILE_CACHE:
-		buf = (byte *) Cache_Alloc (loadcache, len+1, base);
-		break;
-	case LOADFILE_STACK:
-		if (len < loadsize)
-			buf = loadbuf;
-		else
-			buf = (byte *) Hunk_TempAlloc (len+1);
+		buf = (byte *) Hunk_AllocNameNoFill (len+1, base);
 		break;
 	case LOADFILE_MALLOC:
 		buf = (byte *) malloc (len+1);
@@ -1939,8 +2079,10 @@ byte *COM_LoadFile (const char *path, int usehunk, unsigned int *path_id)
 
 	((byte *)buf)[len] = 0;
 
-	Sys_FileRead (h, buf, len);
+	nread = Sys_FileRead (h, buf, len);
 	COM_CloseFile (h);
+	if (nread != len)
+		Sys_Error ("COM_LoadFile: Error reading %s", path);
 
 	return buf;
 }
@@ -1950,34 +2092,6 @@ byte *COM_LoadHunkFile (const char *path, unsigned int *path_id)
 	return COM_LoadFile (path, LOADFILE_HUNK, path_id);
 }
 
-byte *COM_LoadZoneFile (const char *path, unsigned int *path_id)
-{
-	return COM_LoadFile (path, LOADFILE_ZONE, path_id);
-}
-
-byte *COM_LoadTempFile (const char *path, unsigned int *path_id)
-{
-	return COM_LoadFile (path, LOADFILE_TEMPHUNK, path_id);
-}
-
-void COM_LoadCacheFile (const char *path, struct cache_user_s *cu, unsigned int *path_id)
-{
-	loadcache = cu;
-	COM_LoadFile (path, LOADFILE_CACHE, path_id);
-}
-
-// uses temp hunk if larger than bufsize
-byte *COM_LoadStackFile (const char *path, void *buffer, int bufsize, unsigned int *path_id)
-{
-	byte	*buf;
-
-	loadbuf = (byte *)buffer;
-	loadsize = bufsize;
-	buf = COM_LoadFile (path, LOADFILE_STACK, path_id);
-
-	return buf;
-}
-
 // returns malloc'd memory
 byte *COM_LoadMallocFile (const char *path, unsigned int *path_id)
 {
@@ -2082,7 +2196,7 @@ const char *COM_ParseFloatNewline(const char *buffer, float *value)
 const char *COM_ParseStringNewline(const char *buffer)
 {
 	int i;
-	for (i = 0; i < 1023; i++)
+	for (i = 0; i < countof (com_token) - 1; i++)
 		if (!buffer[i] || q_isspace (buffer[i]))
 			break;
 	memcpy (com_token, buffer, i);
@@ -2111,13 +2225,12 @@ static pack_t *COM_LoadPackFile (const char *packfile)
 	pack_t		*pack;
 	int		packhandle;
 	dpackfile_t	info[MAX_FILES_IN_PACK];
-	unsigned short	crc;
 
 	if (Sys_FileOpenRead (packfile, &packhandle) == -1)
 		return NULL;
 
-	Sys_FileRead (packhandle, (void *)&header, sizeof(header));
-	if (header.id[0] != 'P' || header.id[1] != 'A' || header.id[2] != 'C' || header.id[3] != 'K')
+	if (Sys_FileRead(packhandle, &header, sizeof(header)) != (int) sizeof(header) ||
+	    header.id[0] != 'P' || header.id[1] != 'A' || header.id[2] != 'C' || header.id[3] != 'K')
 		Sys_Error ("%s is not a packfile", packfile);
 
 	header.dirofs = LittleLong (header.dirofs);
@@ -2145,14 +2258,16 @@ static pack_t *COM_LoadPackFile (const char *packfile)
 	newfiles = (packfile_t *) Z_Malloc(numpackfiles * sizeof(packfile_t));
 
 	Sys_FileSeek (packhandle, header.dirofs);
-	Sys_FileRead (packhandle, (void *)info, header.dirlen);
+	if (Sys_FileRead(packhandle, info, header.dirlen) != header.dirlen)
+		Sys_Error ("Error reading %s", packfile);
 
 	// crc the directory to check for modifications
-	CRC_Init (&crc);
-	for (i = 0; i < header.dirlen; i++)
-		CRC_ProcessByte (&crc, ((byte *)info)[i]);
-	if (crc != PAK0_CRC_V106 && crc != PAK0_CRC_V101 && crc != PAK0_CRC_V100)
-		com_modified = true;
+	if (!com_modified)
+	{
+		unsigned short	crc = CRC_Block (info, header.dirlen);
+		if (crc != PAK0_CRC_V106 && crc != PAK0_CRC_V101 && crc != PAK0_CRC_V100)
+			com_modified = true;
+	}
 
 	// parse the directory
 	for (i = 0; i < numpackfiles; i++)
@@ -2237,7 +2352,7 @@ static void COM_AddEnginePak (void)
 COM_AddGameDirectory -- johnfitz -- modified based on topaz's tutorial
 =================
 */
-static void COM_AddGameDirectory (const char *dir)
+void COM_AddGameDirectory (const char *dir)
 {
 	const char *base;
 	int i, j;
@@ -2303,9 +2418,9 @@ static void COM_AddGameDirectory (const char *dir)
 	}
 }
 
-void COM_ResetGameDirectories(char *newgamedirs)
+void COM_ResetGameDirectories(const char *newgamedirs)
 {
-	char *newpath, *path;
+	const char *newpath, *path;
 	searchpath_t *search;
 	//Kill the extra game if it is loaded
 	while (com_searchpaths != com_base_searchpaths)
@@ -2352,11 +2467,125 @@ void COM_ResetGameDirectories(char *newgamedirs)
 //==============================================================================
 //johnfitz -- dynamic gamedir stuff -- modified by QuakeSpasm team.
 //==============================================================================
+
+/*
+=================
+COM_GameDirExists
+
+Checks if a gamedir exists in any basedir
+=================
+*/
+static qboolean COM_GameDirExists (const char *game)
+{
+	int i;
+
+	for (i = 0; i < com_numbasedirs; i++)
+		if (Sys_FileType (va ("%s/%s", com_basedirs[i], game)) == FS_ENT_DIRECTORY)
+			return true;
+
+	return false;
+}
+
+/*
+=================
+COM_ValidateGameDirs
+=================
+*/
+static qboolean COM_ValidateGameDirs (const char *newgamedirs)
+{
+	char	dirscopy[1024];
+	char	*newpath;
+
+	q_strlcpy (dirscopy, newgamedirs, sizeof (dirscopy));
+
+	// parse the individual semicolon-separated gamedirs (e.g. id1;ad;ad_heresp1)
+	for (newpath = dirscopy; newpath && *newpath; )
+	{
+		char *p = strchr (newpath, ';');
+		if (p)
+			*p++ = 0;
+
+		if (!COM_GameDirExists (newpath))
+		{
+			Con_Printf ("No such game directory \"%s\"\n", newpath);
+			return false;
+		}
+
+		newpath = p;
+	}
+
+	return true;
+}
+
+/*
+=================
+COM_SwitchGame
+=================
+*/
+void COM_SwitchGame (const char *paths)
+{
+	extern cvar_t max_edicts;
+	if (!q_strcasecmp(paths, COM_GetGameNames(true)))
+	{
+		Con_Printf("\"game\" is already \"%s\"\n", COM_GetGameNames(true));
+		return;
+	}
+
+	if (!COM_ValidateGameDirs (paths))
+		return;
+
+	Host_WaitForSaveThread ();
+
+	com_modified = true;
+
+	//Kill the server
+	CL_Disconnect ();
+	Host_ShutdownServer(true);
+
+	//Write config file
+	Host_WriteConfiguration ();
+
+	// stop parsing map files before changing file system search paths
+	ExtraMaps_Clear ();
+
+	COM_ResetGameDirectories(paths);
+
+	//clear out and reload appropriate data
+	Cache_Flush ();
+	Mod_ResetAll();
+	Sky_ClearAll();
+	if (!isDedicated)
+	{
+		TexMgr_NewGame ();
+		Draw_NewGame ();
+		R_NewGame ();
+	}
+	ExtraMaps_Init ();
+	Host_Resetdemos ();
+	DemoList_Rebuild ();
+	SaveList_Rebuild ();
+	SkyList_Rebuild ();
+	M_CheckMods ();
+	Cvar_SetQuick (&max_edicts, max_edicts.default_string);
+
+	Con_Printf("\n%s\n\"game\" changed to \"%s\"\n", Con_Quakebar (40), COM_GetGameNames(true));
+
+	VID_Lock ();
+	Cbuf_AddText ("unaliasall\n");
+	Cbuf_AddText ("exec quake.rc\n");
+	Cbuf_AddText ("vid_unlock\n");
+}
+
+
+/*
+=================
+COM_Game_f
+=================
+*/
 static void COM_Game_f (void)
 {
 	if (Cmd_Argc() > 1)
 	{
-		extern cvar_t max_edicts;
 		int i, pri;
 		char paths[1024];
 
@@ -2399,88 +2628,12 @@ static void COM_Game_f (void)
 			}
 		}
 
-		if (!q_strcasecmp(paths, COM_GetGameNames(true)))
-		{
-			Con_Printf("\"game\" is already \"%s\"\n", COM_GetGameNames(true));
-			return;
-		}
-
-		Host_WaitForSaveThread ();
-
-		com_modified = true;
-
-		//Kill the server
-		CL_Disconnect ();
-		Host_ShutdownServer(true);
-
-		//Write config file
-		Host_WriteConfiguration ();
-
-		// stop parsing map files before changing file system search paths
-		ExtraMaps_Clear ();
-
-		COM_ResetGameDirectories(paths);
-
-		//clear out and reload appropriate data
-		Cache_Flush ();
-		Mod_ResetAll();
-		Sky_ClearAll();
-		if (!isDedicated)
-		{
-			TexMgr_NewGame ();
-			Draw_NewGame ();
-			R_NewGame ();
-		}
-		ExtraMaps_Init ();
-		Host_Resetdemos ();
-		DemoList_Rebuild ();
-		SaveList_Rebuild ();
-		SkyList_Rebuild ();
-		M_CheckMods ();
-		Cvar_SetQuick (&max_edicts, max_edicts.default_string);
-
-		Con_Printf("\n%s\n\"game\" changed to \"%s\"\n", Con_Quakebar (40), COM_GetGameNames(true));
-
-		VID_Lock ();
-		Cbuf_AddText ("unaliasall\n");
-		Cbuf_AddText ("exec quake.rc\n");
-		Cbuf_AddText ("vid_unlock\n");
+		COM_SwitchGame (paths);
 	}
 	else //Diplay the current gamedir
 		Con_Printf("\"game\" is \"%s\"\n", COM_GetGameNames(true));
 }
 
-/*
-=================
-COM_IsFileWritable
-=================
-*/
-static qboolean COM_IsFileWritable (const char *path)
-{
-	qboolean exists = false;
-	FILE *f;
-
-	if (!path || !*path)
-		return false;
-
-	f = Sys_fopen (path, "rb");
-	if (f)
-	{
-		exists = true;
-		fclose (f);
-	}
-
-	f = Sys_fopen (path, "ab");
-	if (!f)
-		return false;
-
-	fclose (f);
-	if (!exists)
-		Sys_remove (path);
-
-	return true;
-}
-
 /*
 =================
 COM_SetBaseDir
@@ -2624,6 +2777,206 @@ static void COM_MigrateNightdiveUserFiles (void)
 	VEC_FREE (subdirs);
 }
 
+/*
+=================
+COM_SetBaseDirRec
+
+Looks for a valid basedir in all the ancestors of the supplied path
+Returns the path relative to that basedir if successful, NULL otherwise
+=================
+*/
+static const char *COM_SetBaseDirRec (const char *start)
+{
+	char	buf[MAX_OSPATH];
+	size_t	i, len;
+
+	q_strlcpy (buf, start, sizeof (buf));
+	len = strlen (start);
+
+	for (i = len - 1; i > 1; i--)
+	{
+		if (Sys_IsPathSep (buf[i]))
+		{
+			buf[i] = '\0';
+			if (COM_SetBaseDir (buf))
+				return start + i + 1;
+		}
+	}
+
+	return NULL;
+}
+
+/*
+=================
+COM_MakeRelative
+=================
+*/
+static const char *COM_MakeRelative (const char *basepath, const char *fullpath)
+{
+	for (; *basepath && *fullpath; ++basepath, ++fullpath)
+	{
+		if (Sys_IsPathSep (*basepath) != Sys_IsPathSep (*fullpath))
+			return NULL;
+		if (*basepath != *fullpath)
+			return NULL;
+	}
+
+	while (Sys_IsPathSep (*fullpath))
+		++fullpath;
+
+	return fullpath;
+}
+
+/*
+=================
+COM_PatchCmdLine
+
+Tries to initialize basedir from a single command-line argument
+(either a mod dir or a map/save/demo file)
+
+Returns true if successful, false otherwise
+=================
+*/
+static qboolean COM_PatchCmdLine (const char *fullpath)
+{
+	static char	game[MAX_QPATH];
+	char		qpath[MAX_QPATH];
+	char		printpath[MAX_OSPATH];
+	const char	*relpath;
+	const char	*sep;
+	int			type;
+	int			i;
+
+	// The path (file or directory) must exist
+	type = Sys_FileType (fullpath);
+	if (type == FS_ENT_NONE)
+	{
+		UTF8_ToQuake (printpath, sizeof (printpath), fullpath);
+		Con_SafePrintf ("\"%s\" does not exist\n", printpath);
+		return false;
+	}
+
+	// Look for the corresponding basedir
+	relpath = NULL;
+	for (i = 0; i < com_numbasedirs; i++)
+	{
+		relpath = COM_MakeRelative (com_basedirs[i], fullpath);
+		if (relpath)
+			break;
+	}
+	if (!relpath)
+	{
+		UTF8_ToQuake (printpath, sizeof (printpath), fullpath);
+		Con_SafePrintf ("\"%s\" does not belong to an existing Quake installation\n", printpath);
+		return false;
+	}
+
+	// Game dir is the first component of the relative path
+	sep = COM_FirstPathSep (relpath);
+	if ((uintptr_t)(sep - relpath) >= sizeof (game))
+	{
+		UTF8_ToQuake (printpath, sizeof (printpath), relpath);
+		Con_SafePrintf ("\"%s\" is too long\n", printpath);
+		return false;
+	}
+
+	UTF8_ToQuake (printpath, sizeof (printpath), relpath);
+
+	// Apply game dir
+	if (*sep)
+	{
+		Q_strncpy (game, relpath, (int)(sep - relpath));
+		COM_AddArg ("-game");
+		COM_AddArg (game);
+		relpath = sep + 1;
+	}
+	else if (type == FS_ENT_DIRECTORY)
+	{
+		COM_AddArg ("-game");
+		COM_AddArg (relpath);
+		return true;
+	}
+	else
+	{
+		game[0] = '\0';
+	}
+
+	q_strlcpy (qpath, relpath, sizeof (qpath));
+	COM_NormalizePath (qpath);
+
+	// Check argument type
+	switch (type)
+	{
+	case FS_ENT_DIRECTORY:
+		if (qpath[0])
+		{
+			if (q_strcasecmp (qpath, "maps") == 0)
+			{
+				Cbuf_AddText ("menu_maps\n");
+				return true;
+			}
+			UTF8_ToQuake (printpath, sizeof (printpath), qpath);
+			Con_SafePrintf ("\x02subdir \"%s\" ignored\n", printpath);
+		}
+		return true;
+
+	case FS_ENT_FILE:
+		{
+			const char *ext = COM_FileGetExtension (qpath);
+
+			// Map file
+			if (q_strcasecmp (ext, "bsp") == 0)
+			{
+				if (!game[0])
+				{
+					Con_SafePrintf ("Map \"%s\" not in a mod dir, ignoring.\n", printpath);
+					return false;
+				}
+				if (q_strncasecmp (qpath, "maps/", 5) != 0)
+				{
+					Con_SafePrintf ("Map \"%s\" not in the \"maps\" dir, ignoring.\n", printpath);
+					return false;
+				}
+				memmove (qpath, qpath + 5, strlen (qpath + 5) + 1);
+				Cbuf_AddText (va ("menu_maps \"%s\"\n", qpath));
+				return true;
+			}
+
+			// Save file
+			if (q_strcasecmp (ext, "sav") == 0)
+			{
+				const char *kex = game[0] ? "" : "kex";
+				Cbuf_AddText (va ("load \"%s\" %s\n", qpath, kex));
+				return true;
+			}
+
+			// Demo file
+			if (q_strcasecmp (ext, "dem") == 0)
+			{
+				if (!game[0])
+				{
+					Con_SafePrintf ("Demo \"%s\" not in a mod dir, ignoring.\n", printpath);
+					return false;
+				}
+				Cbuf_AddText (va ("playdemo \"%s\"\n", qpath));
+				return true;
+			}
+
+			break;
+		}
+
+	default:
+		break;
+	}
+
+	if (!game[0])
+		Con_SafePrintf ("File \"%s\" not in a mod dir, ignoring.\n", printpath);
+	else
+		Con_SafePrintf ("Unsupported file type \"%s\", ignoring.\n", printpath);
+
+	return false;
+}
+
 /*
 =================
 COM_InitBaseDir
@@ -2709,6 +3062,10 @@ static void COM_InitBaseDir (void)
 			else if (!Sys_GetSteamQuakeUserDir (com_nightdivedir, sizeof (com_nightdivedir), steamquake.library))
 				com_nightdivedir[0] = '\0';
 		}
+		else
+		{
+			memset (&steamquake, 0, sizeof (steamquake));
+		}
 		if (steam)
 			goto storesetup;
 	}
@@ -2755,11 +3112,21 @@ storesetup:
 	{
 		quakeflavor_t flavor;
 		if (original[0] && remastered[0])
-			flavor = ChooseQuakeFlavor ();
+		{
+			if (COM_CheckParm ("-prefremaster") || COM_CheckParm ("-remaster") || COM_CheckParm ("-remastered"))
+				flavor = QUAKE_FLAVOR_REMASTERED;
+			else if (COM_CheckParm ("-preforiginal") || COM_CheckParm ("-original"))
+				flavor = QUAKE_FLAVOR_ORIGINAL;
+			else
+				flavor = ChooseQuakeFlavor ();
+		}
 		else
 			flavor = remastered[0] ? QUAKE_FLAVOR_REMASTERED : QUAKE_FLAVOR_ORIGINAL;
 		q_strlcpy (path, flavor == QUAKE_FLAVOR_REMASTERED ? remastered : original, sizeof (path));
 
+		if (steamquake.appid)
+			Steam_Init (&steamquake);
+
 		if (COM_SetBaseDir (path))
 		{
 			if (!Sys_GetAltUserPrefDir (flavor == QUAKE_FLAVOR_REMASTERED, com_userprefdir, sizeof (com_userprefdir)))
@@ -2800,6 +3167,49 @@ storesetup:
 	);
 }
 
+/*
+=================
+COM_ChooseStartArgFlavor
+
+Checks if the supplied path belongs to the user dir
+of a specific Quake flavor (original/remastered)
+and adds the corresponding command-line argument if needed
+=================
+*/
+static void COM_ChooseStartArgFlavor (const char *startarg)
+{
+	steamgame_t steamquake;
+	char steampath[MAX_OSPATH];
+	char userdir[MAX_OSPATH];
+
+	if (Sys_GetAltUserPrefDir (true, userdir, sizeof (userdir)) && COM_MakeRelative (userdir, startarg))
+	{
+		COM_AddArg ("-prefremaster");
+		return;
+	}
+
+	if (Sys_GetAltUserPrefDir (false, userdir, sizeof (userdir)) && COM_MakeRelative (userdir, startarg))
+	{
+		COM_AddArg ("-preforiginal");
+		return;
+	}
+
+	if (Steam_FindGame (&steamquake, QUAKE_STEAM_APPID) &&
+		Steam_ResolvePath (steampath, sizeof (steampath), &steamquake) &&
+		Sys_GetSteamQuakeUserDir (userdir, sizeof (userdir), steamquake.library) &&
+		COM_MakeRelative (userdir, startarg))
+	{
+		COM_AddArg ("-prefremaster");
+		return;
+	}
+
+	if (Sys_GetGOGQuakeEnhancedUserDir (userdir, sizeof (userdir)) && COM_MakeRelative (userdir, startarg))
+	{
+		COM_AddArg ("-prefremaster");
+		return;
+	}
+}
+
 /*
 =================
 COM_InitFilesystem
@@ -2808,18 +3218,26 @@ COM_InitFilesystem
 void COM_InitFilesystem (void) //johnfitz -- modified based on topaz's tutorial
 {
 	int i;
-	const char *p;
+	const char *p, *startarg;
 
 	Cvar_RegisterVariable (&registered);
 	Cvar_RegisterVariable (&cmdline);
 	Cmd_AddCommand ("path", COM_Path_f);
 	Cmd_AddCommand ("game", COM_Game_f); //johnfitz
 
-	COM_InitBaseDir ();
+	startarg = (com_argc == 2 && Sys_FileType (com_argv[1]) != FS_ENT_NONE) ? com_argv[1] : NULL;
+	if (startarg)
+		COM_ChooseStartArgFlavor (startarg);
+
+	if (!startarg || !COM_SetBaseDirRec (startarg))
+		COM_InitBaseDir ();
 
 	if (host_parms->userdir != host_parms->basedir)
 		COM_AddBaseDir (host_parms->userdir);
 
+	if (startarg)
+		COM_PatchCmdLine (startarg);
+
 	i = COM_CheckParm ("-basegame");
 	if (i)
 	{	//-basegame:
@@ -2851,6 +3269,8 @@ void COM_InitFilesystem (void) //johnfitz -- modified based on topaz's tutorial
 	com_base_searchpaths = com_searchpaths;
 	COM_ResetGameDirectories("");
 
+	Modlist_Init ();
+
 	// add mission pack requests (only one should be specified)
 	if (COM_CheckParm ("-rogue"))
 		COM_AddGameDirectory ("rogue");
@@ -2868,6 +3288,8 @@ void COM_InitFilesystem (void) //johnfitz -- modified based on topaz's tutorial
 		p = com_argv[i + 1];
 		if (!*p || !strcmp(p, ".") || strstr(p, "..") || strstr(p, "/") || strstr(p, "\\") || strstr(p, ":"))
 			Sys_Error ("gamedir should be a single directory name, not a path\n");
+		if (!COM_GameDirExists (p))
+			Sys_Error ("No such game directory \"%s\"", p);
 		com_modified = true;
 		if (p != NULL)
 			COM_AddGameDirectory (p);
@@ -3119,7 +3541,7 @@ LOC_LoadFile
 qboolean LOC_LoadFile (const char *file)
 {
 	char path[1024];
-	int i,lineno;
+	int i,lineno,warnings;
 	char *cursor;
 
 	SDL_RWops *rw = NULL;
@@ -3220,6 +3642,7 @@ fail:			mz_zip_reader_end(&archive);
 	if ((unsigned char)(cursor[0]) == 0xEF && (unsigned char)(cursor[1]) == 0xBB && (unsigned char)(cursor[2]) == 0xBF)
 		cursor += 3;
 
+	warnings = 0;
 	lineno = 0;
 	while (*cursor)
 	{
@@ -3330,6 +3753,19 @@ fail:			mz_zip_reader_end(&archive);
 
 			UTF8_ToQuake (value, strlen (value) + 1, value);
 
+			// Log entries with unprintable characters if developer is set to 2 or higher
+			if (developer.value >= 2.f && strchr (value, QCHAR_BOX))
+			{
+				int trim = (int) strlen (value);
+				// trim trailing newlines
+				while (trim > 0 && value[trim - 1] == '\n')
+					--trim;
+				// print header before first entry
+				if (!warnings++)
+					Sys_Printf ("Entries with unprintable characters:\n");
+				Sys_Printf ("   %d. %s = \"%.*s\"\n", warnings, line, trim, value);
+			}
+
 			entry = &localization.entries[localization.numentries++];
 			entry->key = line;
 			entry->value = value;
@@ -3339,6 +3775,9 @@ fail:			mz_zip_reader_end(&archive);
 			*cursor++ = 0; // terminate line and advance to next
 	}
 
+	if (developer.value >= 2.f && warnings > 0)
+		Sys_Printf ("%d strings with unprintable characters\n", warnings);
+
 	// hash all entries
 
 	localization.numindices = localization.numentries * 2; // 50% load factor
@@ -3373,7 +3812,7 @@ fail:			mz_zip_reader_end(&archive);
 		}
 	}
 
-	Con_Printf("Loaded %d strings from '%s'\n", localization.numentries, file);
+	Con_SafePrintf ("[skipnotify]Loaded %d strings from '%s'\n", localization.numentries, file);
 
 	return true;
 }
@@ -3461,8 +3900,7 @@ void LOC_LanguageCompletion_f (cvar_t *cvar, const char *partial)
 	size_t i;
 
 	for (i = 0; i < countof (knownlangs); i++)
-		if (Con_Match (knownlangs[i][1], partial))
-			Con_AddToTabList (knownlangs[i][1], partial, NULL);
+		Con_AddToTabList (knownlangs[i][1], partial, NULL);
 }
 
 /*
@@ -3656,7 +4094,7 @@ size_t LOC_Format (const char *format, const char* (*getarg_fn) (int idx, void*
 
 static const uint32_t 	utf8_maxcode[4] = {0x7F, 0x7FF, 0xFFFF, 0x10FFFF}; // 1/2/3/4 bytes
 
-static uint8_t			unicode_translit[65536];
+static char				unicode_translit[65536][2];
 static qboolean			unicode_translit_init = false;
 
 static const uint32_t qchar_to_unicode[256] =
@@ -3682,6 +4120,31 @@ static const uint32_t qchar_to_unicode[256] =
       ----------------------------------------------------------------------------------------------------------------------------------
 */};
 
+/*
+==================
+UTF8_CodePointLength
+
+Returns the number of bytes needed to encode the codepoint
+using UTF-8 (max 4), or 0 for an invalid code point
+==================
+*/
+size_t UTF8_CodePointLength (uint32_t codepoint)
+{
+	if (codepoint < 0x80)
+		return 1;
+
+	if (codepoint < 0x800)
+		return 2;
+
+	if (codepoint < 0x10000)
+		return 3;
+
+	if (codepoint < 0x110000)
+		return 4;
+
+	return 0;
+}
+
 /*
 ==================
 UTF8_WriteCodePoint
@@ -3805,14 +4268,30 @@ uint32_t UTF8_ReadCodePoint (const char **src)
 UTF8_FromQuake
 
 Converts a string from Quake encoding to UTF-8
+
+Returns the number of written characters (including the NUL terminator)
+if a valid output buffer is provided (dst is non-NULL, maxbytes > 0),
+or the total amount of space necessary to encode the entire src string
+if dst is NULL and maxbytes is 0.
 ==================
 */
-void UTF8_FromQuake (char *dst, size_t maxbytes, const char *src)
+size_t UTF8_FromQuake (char *dst, size_t maxbytes, const char *src)
 {
 	size_t i, j, written;
 
 	if (!maxbytes)
-		return;
+	{
+		if (dst)
+			return 0; // error
+		for (i = 0, j = 0; src[i]; i++)
+		{
+			uint32_t codepoint = qchar_to_unicode[(unsigned char) src[i]];
+			if (codepoint)
+				j += UTF8_CodePointLength (codepoint);
+		}
+		return j + 1; // include terminator
+	}
+
 	--maxbytes;
 
 	for (i = 0, j = 0; j < maxbytes && src[i]; i++)
@@ -3826,7 +4305,9 @@ void UTF8_FromQuake (char *dst, size_t maxbytes, const char *src)
 		j += written;
 	}
 
-	dst[j] = '\0';
+	dst[j++] = '\0';
+
+	return j;
 }
 
 /*
@@ -3835,43 +4316,95 @@ UTF8_ToQuake
 
 Transliterates a string from UTF-8 to Quake encoding
 
-Note: only single-character transliterations are used for now,
-mainly to remove diacritics
+Returns the number of written characters (including the NUL terminator)
+if a valid output buffer is provided (dst is non-NULL, maxbytes > 0),
+or the total amount of space necessary to encode the entire src string
+if dst is NULL and maxbytes is 0.
+
+Note: only single-character and two-character transliterations are used
+so that the string is never expanded during the process.
 ==================
 */
-void UTF8_ToQuake (char *dst, size_t maxbytes, const char *src)
+size_t UTF8_ToQuake (char *dst, size_t maxbytes, const char *src)
 {
-	size_t i;
+	size_t i, j;
+	uint32_t cp;
 
 	if (!unicode_translit_init)
 	{
-		// single-character transliterations
+		// precomputed single-character/two-character transliterations
 		for (i = 0; i < countof (unicode_translit_src); i++)
-			unicode_translit[unicode_translit_src[i][0]] = (uint8_t) unicode_translit_src[i][1];
+		{
+			unicode_translit[unicode_translit_src[i].code][0] = unicode_translit_src[i].remap[0];
+			unicode_translit[unicode_translit_src[i].code][1] = unicode_translit_src[i].remap[1];
+		}
 
 		// Quake-specific characters: we process the list in reverse order
 		// so that codepoints used for both colored and non-colored qchars
 		// end up being remapped to the non-colored versions
 		// Note: 0 is not included
 		for (i = countof (qchar_to_unicode) - 1; i > 0; i--)
+		{
 			if (qchar_to_unicode[i] >= 128 && qchar_to_unicode[i] < countof (unicode_translit))
-				unicode_translit[qchar_to_unicode[i]] = (uint8_t) i;
+			{
+				unicode_translit[qchar_to_unicode[i]][0] = (char) i;
+				unicode_translit[qchar_to_unicode[i]][1] = '\0';
+			}
+		}
 
 		// map ASCII characters to themselves
 		for (i = 0; i < 128; i++)
-			unicode_translit[i] = (uint8_t) i;
+		{
+			unicode_translit[i][0] = (char) i;
+			unicode_translit[i][1] = '\0';
+		}
+
+		// Map all other characters to QCHAR_BOX (unknown character)
+		for (i = 0; i < Q_COUNTOF (unicode_translit); i++)
+		{
+			if (!unicode_translit[i][0])
+			{
+				unicode_translit[i][0] = QCHAR_BOX;
+				unicode_translit[i][1] = '\0';
+			}
+		}
 
 		unicode_translit_init = true;
 	}
 
 	if (!maxbytes)
-		return;
+	{
+		if (dst)
+			return 0; // error
+
+		// Determine necessary output buffer size
+		for (i = 0, j = 0; *src; i++)
+		{
+			// ASCII fast path
+			while (*src && (byte)*src < 0x80)
+			{
+				src++;
+				j++;
+			}
+
+			if (!*src)
+				break;
+
+			// A codepoint maps to either one or two Quake characters
+			cp = UTF8_ReadCodePoint (&src);
+			if (cp < Q_COUNTOF (unicode_translit))
+				j += unicode_translit[cp][1] != '\0' ? 2 : 1;
+			else
+				j++;
+		}
+
+		return j + 1; // include terminator
+	}
+
 	--maxbytes;
 
 	for (i = 0; i < maxbytes && *src; i++)
 	{
-		uint32_t cp;
-
 		// ASCII fast path
 		while (*src && i < maxbytes && (byte)*src < 0x80)
 			dst[i++] = *src++;
@@ -3882,15 +4415,17 @@ void UTF8_ToQuake (char *dst, size_t maxbytes, const char *src)
 		cp = UTF8_ReadCodePoint (&src);
 		if (cp < countof (unicode_translit))
 		{
-			cp = unicode_translit[cp];
-			if (!cp)
-				cp = QCHAR_BOX;
+			char c0 = unicode_translit[cp][0];
+			char c1 = unicode_translit[cp][1];
+			dst[i] = c0;
+			if (c1 && i + 1 < maxbytes)
+				dst[++i] = c1;
 		}
 		else
-			cp = QCHAR_BOX;
-
-		dst[i] = (uint8_t) cp;
+			dst[i] = QCHAR_BOX;
 	}
 
 	dst[i++] = '\0';
+
+	return i;
 }
