@@ -34,12 +34,18 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include <unistd.h>
 #endif
 
+extern qboolean	keydown[256];
+
 int 		con_linewidth;
 
 float		con_cursorspeed = 4;
 
 #define		CON_TEXTSIZE (1024 * 1024) //ericw -- was 65536. johnfitz -- new default size
 #define		CON_MINSIZE  16384 //johnfitz -- old default, now the minimum size
+#define		CON_MARGIN   1
+
+#define		CON_SCROLL_ZONE			(CHARSIZE * 2)
+#define		CON_MAX_SCROLL_SPEED	32.f
 
 int		con_buffersize; //johnfitz -- user can now override default
 
@@ -51,11 +57,67 @@ int		con_current;		// where next message will be printed
 int		con_x;				// offset in current line for next print
 char		*con_text = NULL;
 
+static float	con_scrollspeed;
+static float	con_scrolldelta;
+
+typedef struct
+{
+	int			line;
+	int			col;
+} conofs_t;
+
+typedef struct
+{
+	const char	*path;
+	conofs_t	begin;
+	conofs_t	end;
+} conlink_t;
+
+typedef struct
+{
+	conofs_t		begin;
+	conofs_t		end;
+} conselection_t;
+
+typedef enum
+{
+	// Used for link hover/clicking:
+	// - picks the character that contains the cursor
+	// - rejects areas outside the visible console region
+	CT_INSIDE,
+
+	// Used for text selection:
+	// - picks the closest edge horizontally, on whichever line contains the cursor vertically
+	// - clamps to the margins of the visible console region
+	CT_NEAREST,
+} contest_t;		// Console hit testing mode
+
+typedef enum
+{
+	CMS_NOTPRESSED,
+	CMS_PRESSED,
+	CMS_DRAGGING,
+} conmouse_t;
+
+static conlink_t	**con_links = NULL;
+static conlink_t	*con_hotlink = NULL;
+
+static const double		DOUBLECLICK_TIME = 0.5;
+
+static double			con_mouseclickdelay = 0.0;
+static int				con_mouseclicks = 0;
+static conmouse_t		con_mousestate = CMS_NOTPRESSED;
+static conselection_t	con_mouseselection;
+static conselection_t	con_selection;
+static int				con_clickx;
+static int				con_clicky;
+
 cvar_t		con_notifytime = {"con_notifytime","3",CVAR_NONE};	//seconds
 cvar_t		con_logcenterprint = {"con_logcenterprint", "1", CVAR_NONE}; //johnfitz
 cvar_t		con_notifycenter = {"con_notifycenter", "0", CVAR_ARCHIVE};
 cvar_t		con_notifyfade = {"con_notifyfade", "0", CVAR_ARCHIVE};
 cvar_t		con_notifyfadetime = {"con_notifyfadetime", "0.5", CVAR_ARCHIVE};
+cvar_t		con_maxcols = {"con_maxcols", "0", CVAR_ARCHIVE};
 
 char		con_lastcenterstring[1024]; //johnfitz
 
@@ -65,11 +127,572 @@ double		con_times[NUM_CON_TIMES];	// realtime time the line was generated
 
 int			con_vislines;
 
-qboolean	con_debuglog = false;
-
 qboolean	con_initialized;
 
 
+/*
+================
+Con_GetLine
+================
+*/
+static const char *Con_GetLine (int line)
+{
+	return con_text + (line%con_totallines)*con_linewidth;
+}
+
+/*
+================
+Con_StrLen
+================
+*/
+static size_t Con_StrLen (int line)
+{
+	const char *text;
+	size_t len;
+	if (line > con_current)
+		return 0;
+	text = Con_GetLine (line);
+	len = con_linewidth;
+	while (len > 0 && (char)(text[len - 1] & 0x7f) == ' ')
+		len--;
+	return len;
+}
+
+static void Con_ScreenToCanvas (int x, int y, int *outx, int *outy)
+{
+	drawtransform_t	transform;
+	float			px, py;
+
+	// screen space to [-1..1]
+	px = (x - glx) * 2.f / (float) glwidth - 1.f;
+	py = (y - gly) * 2.f / (float) glheight - 1.f;
+	py = -py;
+
+	// [-1..1] to console canvas
+	Draw_GetCanvasTransform (CANVAS_CONSOLE, &transform);
+	px = (px - transform.offset[0]) / transform.scale[0];
+	py = (py - transform.offset[1]) / transform.scale[1];
+	x = (int) (px + 0.5f);
+	y = (int) (py + 0.5f);
+
+	*outx = x;
+	*outy = y;
+}
+
+/*
+================
+Con_CanvasToOffset
+
+Converts canvas coordinates to a console offset
+Returns true if the offset is inside the visible portion of the console
+================
+*/
+static qboolean Con_CanvasToOffset (int x, int y, conofs_t *ofs, contest_t testmode)
+{
+	qboolean ret = true;
+
+// Start from the bottom of the console
+	y = vid.conheight - y;
+
+// Apply rounding
+	if (testmode == CT_NEAREST)
+		x += 4;
+
+// pixels to characters
+	x >>= 3;
+	y >>= 3;
+
+// apply margins and scrolling
+	x -= CON_MARGIN;
+	y -= 2;
+
+	if (testmode == CT_INSIDE)
+	{
+		if (x < 0 || x >= con_linewidth)
+			ret = false;
+		if (y < 0 || y >= con_vislines)
+			ret = false;
+		if (con_backscroll && y < 2)
+			ret = false;
+	}
+	else
+	{
+		// Allow the cursor to move one character past the end of the line
+		// by clamping to con_linewidth instead of con_linewidth - 1
+		x = CLAMP (0, x, con_linewidth);
+
+		// Enable selecting the entire bottom line by allowing the cursor
+		// to move to the beginning of the line below it (line -1)
+		y = CLAMP (-1, y, con_vislines);
+		if (y < 0)
+			x = 0;
+
+		// Enable selecting the entire line above the backscroll cutoff by
+		// allowing the cursor to move to the beginning of the line below it
+		if (con_backscroll && y < 2)
+		{
+			x = 0;
+			y = 1;
+		}
+	}
+
+	y += con_backscroll;
+	y = con_current - y;
+
+	ofs->line = y;
+	ofs->col = x;
+
+	return ret;
+}
+
+/*
+================
+Con_ScreenToOffset
+
+Converts screen (pixel) coordinates to a console offset
+Returns true if the offset is inside the visible portion of the console
+================
+*/
+static qboolean Con_ScreenToOffset (int x, int y, conofs_t *ofs, contest_t testmode)
+{
+	Con_ScreenToCanvas (x, y, &x, &y);
+	return Con_CanvasToOffset (x, y, ofs, testmode);
+}
+
+/*
+================
+Con_OfsCompare
+
+Performs a three-way comparison on console offsets
+================
+*/
+static int Con_OfsCompare (const conofs_t *lhs, const conofs_t *rhs)
+{
+	if (lhs->line != rhs->line)
+		return lhs->line - rhs->line;
+	return lhs->col - rhs->col;
+}
+
+/*
+================
+Con_OfsInRange
+
+Checks if an offset is within a half-open range
+================
+*/
+static qboolean Con_OfsInRange (const conofs_t *ofs, const conofs_t *begin, const conofs_t *end)
+{
+	return Con_OfsCompare (ofs, begin) >= 0 && Con_OfsCompare (ofs, end) < 0;
+}
+
+/*
+================
+Con_GetCurrentRange
+================
+*/
+static void Con_GetCurrentRange (conofs_t *begin, conofs_t *end)
+{
+	begin->line = con_current - con_totallines + 1;
+	begin->col = 0;
+	end->line = con_current + 1;
+	end->col = 0;
+}
+
+/*
+================
+Con_IntersectRanges
+================
+*/
+static qboolean Con_IntersectRanges (conofs_t *begin, conofs_t *end, const conofs_t *selbegin, const conofs_t *selend)
+{
+	if (Con_OfsCompare (selend, begin) <= 0)
+		return false;
+	if (Con_OfsCompare (end, selbegin) <= 0)
+		return false;
+
+	if (Con_OfsCompare (begin, selbegin) < 0)
+		*begin = *selbegin;
+	if (Con_OfsCompare (selend, end) < 0)
+		*end = *selend;
+
+	return true;
+}
+
+/*
+================
+Con_GetLinkAtOfs
+
+Returns the link at the given offset, if any, or NULL otherwise
+================
+*/
+static conlink_t *Con_GetLinkAtOfs (const conofs_t *ofs)
+{
+	size_t lo, hi;
+
+// find the first link that ends after the offset
+	lo = 0;
+	hi = VEC_SIZE (con_links);
+	while (lo < hi)
+	{
+		size_t mid = (lo + hi) / 2;
+		if (Con_OfsCompare (ofs, &con_links[mid]->end) >= 0)
+			lo = mid + 1;
+		else
+			hi = mid;
+	}
+
+	if (lo == VEC_SIZE (con_links))
+		return NULL;
+
+	if (Con_OfsCompare (ofs, &con_links[lo]->begin) >= 0)
+		return con_links[lo];
+
+	return NULL;
+}
+
+/*
+================
+Con_GetLinkAtPixel
+
+Returns the link at the given pixel coordinates, if any, or NULL otherwise
+================
+*/
+static conlink_t *Con_GetLinkAtPixel (int x, int y)
+{
+	conofs_t ofs;
+	if (!Con_ScreenToOffset (x, y, &ofs, CT_INSIDE))
+		return NULL;
+	return Con_GetLinkAtOfs (&ofs);
+}
+
+/*
+================
+Con_SetHotLink
+
+Changes the current hot link and updates the mouse cursor
+================
+*/
+static void Con_SetHotLink (conlink_t *link)
+{
+	if (link == con_hotlink)
+		return;
+	con_hotlink = link;
+}
+
+/*
+================
+Con_ClearSelection
+================
+*/
+static void Con_ClearSelection (void)
+{
+	memset (&con_selection, 0, sizeof (con_selection));
+}
+
+/*
+================
+Con_HasSelection
+================
+*/
+static qboolean Con_HasSelection (void)
+{
+	return Con_OfsCompare (&con_selection.begin, &con_selection.end) != 0;
+}
+
+/*
+================
+Con_SelectAll
+================
+*/
+void Con_SelectAll (void)
+{
+	Con_GetCurrentRange (&con_selection.begin, &con_selection.end);
+	while (Con_HasSelection () && Con_StrLen (con_selection.begin.line) == 0)
+		con_selection.begin.line++;
+}
+
+/*
+================
+Con_GetNormalizedSelection
+================
+*/
+static qboolean Con_GetNormalizedSelection (conofs_t *begin, conofs_t *end)
+{
+	conofs_t	*selbegin = &con_selection.begin;
+	conofs_t	*selend = &con_selection.end;
+	conofs_t	tbegin, tend;
+
+	if (Con_OfsCompare (selbegin, selend) > 0)
+	{
+		conofs_t *tmp = selbegin;
+		selbegin = selend;
+		selend = tmp;
+	}
+	*begin = *selbegin;
+	*end = *selend;
+
+	Con_GetCurrentRange (&tbegin, &tend);
+
+	return Con_IntersectRanges (begin, end, &tbegin, &tend);
+}
+
+/*
+================
+Con_TestWordBoundary
+
+Returns:
+     < 0 if on a word boundary and non-whitespace characters are to the left
+       0 if not on a word boundary
+     > 0 if on a word boundary and non-whitespace characters are to the right
+================
+*/
+static int Con_TestWordBoundary (int pos, const char *text, int len)
+{
+	if (pos <= 0)
+		return 1;
+	if (pos >= len)
+		return -1;
+	return q_isspace (text[pos - 1] & 0x7f) - q_isspace (text[pos] & 0x7f);
+}
+
+int IntSign (int i)
+{
+	if (i < 0)
+		return -1;
+	if (i > 0)
+		return 1;
+	return i;
+}
+
+/*
+================
+Con_ApplyMouseSelection
+================
+*/
+static void Con_ApplyMouseSelection (void)
+{
+	const char	*line;
+	int			len;
+
+	con_selection = con_mouseselection;
+
+	line = Con_GetLine (con_selection.begin.line);
+	len = (int) Con_StrLen (con_selection.begin.line);
+
+	// Clamp starting point to the end of the actual content (one character past it)
+	// so that double-clicking beyond the end of the line selects the last word
+	con_selection.begin.col = q_min (con_selection.begin.col, len);
+
+	// Special case: if we're selecting whole words, the initial click was on a word boundary,
+	// and the current selection hasn't advanced towards the actual content (either left or right),
+	// then we nudge the starting point by one character so that the word adjacent to the initial click
+	// is always selected.
+	if (con_mouseclicks == 2)
+	{
+		int boundary = IntSign (Con_TestWordBoundary (con_selection.begin.col, line, len));
+		int dir = IntSign (Con_OfsCompare (&con_selection.end, &con_selection.begin));
+		if (boundary && boundary != dir)
+			con_selection.begin.col += boundary;
+	}
+
+	// Swap begin/end if necessary
+	if (Con_OfsCompare (&con_selection.begin, &con_selection.end) > 0)
+	{
+		conofs_t tmp = con_selection.begin;
+		con_selection.begin = con_selection.end;
+		con_selection.end = tmp;
+	}
+
+	// Selecting character by character? Nothing left to do
+	if (con_mouseclicks <= 1)
+		return;
+
+	// Quadruple click: select whole buffer
+	if (con_mouseclicks >= 4)
+	{
+		Con_SelectAll ();
+		return;
+	}
+
+	// Triple click: select whole lines
+	if (con_mouseclicks == 3)
+	{
+		con_selection.begin.col = 0;
+		con_selection.end.col = 0;
+		con_selection.end.line = q_min (con_selection.end.line, con_current) + 1;
+		return;
+	}
+
+	// Double click: select whole words
+
+	// Move begin marker to the first word boundary to its left
+	line = Con_GetLine (con_selection.begin.line);
+	len = (int) Con_StrLen (con_selection.begin.line);
+	while (!Con_TestWordBoundary (con_selection.begin.col, line, len))
+		--con_selection.begin.col;
+
+	// Move end marker to the first word boundary to its right
+	if (con_selection.end.line <= con_current)
+	{
+		line = Con_GetLine (con_selection.end.line);
+		len = (int) Con_StrLen (con_selection.end.line);
+		while (!Con_TestWordBoundary (con_selection.end.col, line, len))
+			++con_selection.end.col;
+	}
+}
+
+/*
+================
+Con_SetMouseState
+================
+*/
+static void Con_SetMouseState (conmouse_t state)
+{
+	int x, y;
+	conofs_t pos;
+
+	if (con_mousestate == state)
+		return;
+
+	switch (state)
+	{
+	case CMS_PRESSED:
+		SDL_GetMouseState (&x, &y);
+		Con_ScreenToCanvas (x, y, &con_clickx, &con_clicky);
+		Con_CanvasToOffset (con_clickx, con_clicky, &pos, CT_NEAREST);
+
+		if (con_mouseclicks == 0 || con_mouseclickdelay >= DOUBLECLICK_TIME || Con_OfsCompare (&pos, &con_mouseselection.end) != 0)
+			con_mouseclicks = 1;
+		else
+			con_mouseclicks++;
+		con_mouseclickdelay = 0.0;
+		con_mouseselection.begin = con_mouseselection.end = pos;
+
+		Con_ApplyMouseSelection ();
+
+		if (con_mouseclicks >= 2)
+			VID_SetMouseCursor (MOUSECURSOR_IBEAM);
+		break;
+
+	case CMS_DRAGGING:
+		Con_SetHotLink (NULL);
+		VID_SetMouseCursor (MOUSECURSOR_IBEAM);
+		break;
+
+	case CMS_NOTPRESSED:
+		if (con_mousestate != CMS_DRAGGING && con_hotlink && !Sys_Explore (con_hotlink->path))
+			S_LocalSound ("misc/menu2.wav");
+		con_scrolldelta = 0.f;
+		con_scrollspeed = 0.f;
+		break;
+
+	default:
+		break;
+	}
+
+	con_mousestate = state;
+	Con_ForceMouseMove ();
+}
+
+/*
+================
+Con_Mousemove
+Mouse movement callback
+================
+*/
+void Con_Mousemove (int x, int y)
+{
+	if (con_mousestate == CMS_NOTPRESSED)
+	{
+		conofs_t ofs;
+		qboolean inside = Con_ScreenToOffset (x, y, &ofs, CT_INSIDE);
+		Con_SetHotLink (Con_GetLinkAtPixel (x, y));
+		VID_SetMouseCursor (con_hotlink ? MOUSECURSOR_HAND : inside ? MOUSECURSOR_IBEAM : MOUSECURSOR_DEFAULT);
+	}
+	else
+	{
+		int cx, cy, delta;
+		float frac;
+
+		Con_ScreenToCanvas (x, y, &cx, &cy);
+		Con_CanvasToOffset (cx, cy, &con_mouseselection.end, CT_NEAREST);
+		Con_ApplyMouseSelection ();
+		if (Con_OfsCompare (&con_mouseselection.begin, &con_mouseselection.end) != 0)
+			Con_SetMouseState (CMS_DRAGGING);
+
+		// Compute distance inside the auto-scroll range
+		delta = cy + con_vislines / 2 - vid.conheight;
+		if (abs (delta) < con_vislines / 2 - CON_SCROLL_ZONE)
+			delta = 0;
+		else
+			delta -= IntSign (delta) * (con_vislines / 2 - CON_SCROLL_ZONE);
+		delta = CLAMP (-CON_SCROLL_ZONE, delta, CON_SCROLL_ZONE);
+
+		if (delta < 0)
+		{
+			// If the initial click was close to the top (inside the scroll range),
+			// we don't want to immediately start scrolling.
+			// Once we've started scrolling we gradually relax the restriction.
+			// NOTE: This is not an issue on the bottom because of the input line and margin.
+			int moved = cy - con_clicky;
+			int scrolled = q_min (con_mouseselection.end.line - con_mouseselection.begin.line, 0) * CHARSIZE;
+			delta = q_max (delta, moved + scrolled / 4);
+			delta = q_min (delta, 0);
+		}
+
+		// Compute scroll speed
+		frac = delta / (float) CON_SCROLL_ZONE;
+		frac *= fabs (frac); // quadratic easing
+		con_scrollspeed = -CON_MAX_SCROLL_SPEED * frac;
+		if (!delta)
+			con_scrolldelta = 0.0f;
+	}
+}
+
+/*
+================
+Con_ForceMouseMove
+================
+*/
+void Con_ForceMouseMove (void)
+{
+	int x, y;
+	SDL_GetMouseState (&x, &y);
+	Con_Mousemove (x, y);
+}
+
+/*
+================
+Con_UpdateMouseState
+================
+*/
+static void Con_UpdateMouseState (void)
+{
+	if (key_dest != key_console)
+	{
+		Con_SetHotLink (NULL);
+		Con_SetMouseState (CMS_NOTPRESSED);
+		Con_ClearSelection ();
+		return;
+	}
+
+	if (!keydown[K_MOUSE1])
+		Con_SetMouseState (CMS_NOTPRESSED);
+	else if (con_mousestate == CMS_NOTPRESSED)
+		Con_SetMouseState (CMS_PRESSED);
+
+	con_mouseclickdelay += host_rawframetime;
+
+	// Handle auto-scrolling
+	con_scrolldelta += con_scrollspeed * host_rawframetime;
+	if (fabs (con_scrolldelta) >= 1.0f)
+	{
+		int lines = (int)con_scrolldelta;
+		Con_Scroll (lines);
+		con_scrolldelta -= lines;
+	}
+}
+
 /*
 ================
 Con_Quakebar -- johnfitz -- returns a bar of the desired length, but never wider than the console
@@ -116,6 +739,8 @@ void Con_ToggleConsole_f (void)
 		key_linepos = 1;
 		con_backscroll = 0; //johnfitz -- toggleconsole should return you to the bottom of the scrollback
 		history_line = edit_line; //johnfitz -- it should also return you to the bottom of the command history
+		key_tabhint[0] = '\0';			// clear tab hint
+		Con_SetHotLink (NULL);
 
 		if (cls.state == ca_connected)
 		{
@@ -144,9 +769,69 @@ Con_Clear_f
 */
 static void Con_Clear_f (void)
 {
+	size_t i;
+
 	if (con_text)
 		Q_memset (con_text, ' ', con_buffersize); //johnfitz -- con_buffersize replaces CON_TEXTSIZE
+
 	con_backscroll = 0; //johnfitz -- if console is empty, being scrolled up is confusing
+
+	Con_SetHotLink (NULL);
+	for (i = 0; i < VEC_SIZE (con_links); i++)
+		free (con_links[i]);
+	VEC_CLEAR (con_links);
+}
+
+/*
+================
+Con_CopySelectionToClipboard
+================
+*/
+qboolean Con_CopySelectionToClipboard (void)
+{
+	conofs_t selbegin, selend;
+	conofs_t cursor, eol;
+	char *qtext = NULL;
+	char *utf8 = NULL;
+	size_t maxsize;
+
+	S_LocalSound ("misc/menu2.wav");
+
+	// Get forward selection range
+	if (!Con_GetNormalizedSelection (&selbegin, &selend))
+		return false;
+
+	// Iterate through all lines in the selection
+	for (cursor = selbegin; Con_OfsCompare (&cursor, &selend) <= 0; cursor.line++, cursor.col = 0)
+	{
+		const char *text = Con_GetLine (cursor.line);
+		eol.line = cursor.line;
+		eol.col = Con_StrLen (cursor.line);
+		if (cursor.line == selend.line)
+			eol.col = q_min (eol.col, selend.col);
+		Vec_Append ((void **)&qtext, 1, text + cursor.col, eol.col - cursor.col);
+		if (eol.line != selend.line)
+			VEC_PUSH (qtext, '\n');
+	}
+	VEC_PUSH (qtext, '\0');
+
+	// Convert to UTF-8
+	maxsize = UTF8_FromQuake (NULL, 0, qtext);
+	utf8 = (char *) malloc (maxsize);
+	if (!utf8)
+		Sys_Error ("Con_CopySelectionToClipboard: out of memory on %" SDL_PRIu64 " bytes", (uint64_t)maxsize);
+	UTF8_FromQuake (utf8, maxsize, qtext);
+
+	// Copy the UTF-8 text to clipboard
+	SDL_SetClipboardText (utf8);
+
+	// Clean up temporary buffers
+	free (utf8);
+	VEC_FREE (qtext);
+
+	Con_ClearSelection ();
+
+	return true;
 }
 
 /*
@@ -166,7 +851,6 @@ static void Con_Dump_f (void)
 	q_strlcpy (relname, Cmd_Argc () >= 2 ? Cmd_Argv (1) : "condump.txt", sizeof (relname));
 	COM_AddExtension (relname, ".txt", sizeof (relname));
 	q_snprintf (name, sizeof(name), "%s/%s", com_gamedir, relname);
-	COM_CreatePath (name);
 	f = Sys_fopen (name, "w");
 	if (!f)
 	{
@@ -205,7 +889,9 @@ static void Con_Dump_f (void)
 	}
 
 	fclose (f);
-	Con_Printf ("Dumped console text to %s.\n", relname);
+	Con_SafePrintf ("Dumped console text to ");
+	Con_LinkPrintf (name, "%s", relname);
+	Con_SafePrintf (".\n");
 }
 
 /*
@@ -249,6 +935,13 @@ static void Con_MessageMode2_f (void)
 }
 
 
+void Con_RecalcOffset (conofs_t *ofs, int oldnumlines)
+{
+	ofs->col = q_min (ofs->col, con_linewidth);
+	ofs->line += con_totallines - 1 - oldnumlines;
+}
+
+
 /*
 ================
 Con_CheckResize
@@ -262,7 +955,7 @@ void Con_CheckResize (void)
 	char	*tbuf; //johnfitz -- tbuf no longer a static array
 	int mark; //johnfitz
 
-	width = (vid.conwidth >> 3) - 2; //johnfitz -- use vid.conwidth instead of vid.width
+	width = (vid.conwidth >> 3) - CON_MARGIN*2; //johnfitz -- use vid.conwidth instead of vid.width
 
 	if (width == con_linewidth)
 		return;
@@ -298,6 +991,13 @@ void Con_CheckResize (void)
 
 	Hunk_FreeToLowMark (mark); //johnfitz
 
+	for (i = 0; i < (int) VEC_SIZE (con_links); i++)
+	{
+		conlink_t *link = con_links[i];
+		Con_RecalcOffset (&link->begin, con_current);
+		Con_RecalcOffset (&link->end, con_current);
+	}
+
 	Con_ClearNotify ();
 
 	con_backscroll = 0;
@@ -305,6 +1005,33 @@ void Con_CheckResize (void)
 }
 
 
+/*
+================
+Con_Scroll
+================
+*/
+void Con_Scroll (int lines)
+{
+	if (!lines)
+		return;
+
+	con_backscroll += lines;
+
+	if (lines > 0)
+	{
+		if (con_backscroll > con_totallines - (vid.height>>3) - 1)
+			con_backscroll = con_totallines - (vid.height>>3) - 1;
+	}
+	else
+	{
+		if (con_backscroll < 0)
+			con_backscroll = 0;
+	}
+
+	Con_ForceMouseMove ();
+}
+
+
 /*
 ================
 Con_Init
@@ -325,7 +1052,7 @@ void Con_Init (void)
 		con_buffersize = CON_TEXTSIZE;
 	//johnfitz
 
-	con_text = (char *) Hunk_AllocName (con_buffersize, "context");//johnfitz -- con_buffersize replaces CON_TEXTSIZE
+	con_text = (char *) Hunk_AllocNameNoFill (con_buffersize, "context");//johnfitz -- con_buffersize replaces CON_TEXTSIZE
 	Q_memset (con_text, ' ', con_buffersize);//johnfitz -- con_buffersize replaces CON_TEXTSIZE
 	con_linewidth = -1;
 
@@ -343,6 +1070,7 @@ void Con_Init (void)
 	Cvar_RegisterVariable (&con_notifyfade);
 	Cvar_RegisterVariable (&con_notifyfadetime);
 	Cvar_RegisterVariable (&con_logcenterprint); //johnfitz
+	Cvar_RegisterVariable (&con_maxcols);
 
 	Cmd_AddCommand ("toggleconsole", Con_ToggleConsole_f);
 	Cmd_AddCommand ("messagemode", Con_MessageMode_f);
@@ -387,7 +1115,7 @@ static void Con_Print (const char *txt)
 	int		c, l;
 	static int	cr;
 	int		mask;
-	qboolean	boundary;
+	qboolean	boundary, skipnotify;
 
 	//con_backscroll = 0; //johnfitz -- better console scrolling
 
@@ -406,6 +1134,12 @@ static void Con_Print (const char *txt)
 		mask = 0;
 
 	boundary = true;
+	skipnotify = false;
+	if (!Q_strncmp (txt, "[skipnotify]", 12))
+	{
+		skipnotify = true;
+		txt += 12;
+	}
 
 	while ( (c = *txt) )
 	{
@@ -440,7 +1174,7 @@ static void Con_Print (const char *txt)
 			Con_Linefeed ();
 		// mark time for transparent overlay
 			if (con_current >= 0)
-				con_times[con_current % NUM_CON_TIMES] = realtime;
+				con_times[con_current % NUM_CON_TIMES] = skipnotify ? 0 : realtime;
 		}
 
 		switch (c)
@@ -474,6 +1208,10 @@ static int	log_fd = -1;			// log file descriptor
 /*
 ================
 Con_DebugLog
+
+Writes msg to log if -condebug was specified on the command line
+
+Note: msg is expected to be in UTF-8, not Quake charset
 ================
 */
 void Con_DebugLog(const char *msg)
@@ -481,7 +1219,31 @@ void Con_DebugLog(const char *msg)
 	if (log_fd == -1)
 		return;
 
-	write(log_fd, msg, strlen(msg));
+	if (write(log_fd, msg, strlen(msg)) < 0)
+	{
+		close (log_fd);
+		log_fd = -1;
+		fprintf (stderr, "Error writing to log file\n");
+	}
+}
+
+
+/*
+================
+Con_StripControlPrefixes
+================
+*/
+static const char *Con_StripControlPrefixes (const char *txt)
+{
+	// colored text
+	if (txt[0] == 1 || txt[0] == 2)
+		txt++;
+
+	// [skipnotify]
+	if (!Q_strncmp (txt, "[skipnotify]", 12))
+		txt += 12;
+
+	return txt;
 }
 
 
@@ -504,11 +1266,7 @@ void Con_Printf (const char *fmt, ...)
 	va_end (argptr);
 
 // also echo to debugging console
-	Sys_Printf ("%s", msg);
-
-// log all messages to file
-	if (con_debuglog)
-		Con_DebugLog(msg);
+	Sys_Printf ("%s", Con_StripControlPrefixes (msg));
 
 	if (!con_initialized)
 		return;
@@ -620,6 +1378,58 @@ void Con_DPrintf2 (const char *fmt, ...)
 }
 
 
+/*
+==================
+Con_LinkPrintf
+
+Prints text that opens a link when clicked
+==================
+*/
+void Con_LinkPrintf (const char *addr, const char *fmt, ...)
+{
+	conlink_t	*link;
+	size_t		len;
+	va_list		argptr;
+	char		msg[MAXPRINTMSG];
+	char		*text;
+
+	len = strlen (addr);
+	link = (conlink_t *) malloc (sizeof (conlink_t) + len + 1);
+	if (!link)
+		Sys_Error ("Con_LinkPrintf: out of memory on %" SDL_PRIu64 " bytes", (uint64_t)(sizeof (conlink_t) + len + 1));
+	
+	memcpy (link + 1, addr, len + 1);
+	link->path			= (const char *)(link + 1);
+	link->begin.line	= con_current;
+	link->begin.col		= con_x;
+	link->end			= link->begin;
+
+	va_start (argptr, fmt);
+	q_vsnprintf (msg, sizeof(msg), fmt, argptr);
+	va_end (argptr);
+
+	Con_SafePrintf ("\x02%s", msg);
+
+	link->end.line	= con_current;
+	link->end.col	= con_x;
+	VEC_PUSH (con_links, link);
+
+// Because of wrapping our text might actually start on the next line, so we skip leading spaces
+	text = con_text + (link->begin.line % con_totallines)*con_linewidth + link->begin.col;
+	while (Con_OfsCompare (&link->begin, &link->end) < 0)
+	{
+		if ((*text & 0x7f) != ' ')
+			break;
+		text++;
+		if (++link->begin.col == con_linewidth)
+		{
+			link->begin.col = 0;
+			link->begin.line++;
+		}
+	}
+}
+
+
 /*
 ==================
 Con_SafePrintf
@@ -732,7 +1542,6 @@ typedef struct tab_s
 tab_t	*tablist;
 
 //defs from elsewhere
-extern qboolean	keydown[256];
 extern	cmd_function_t	*cmd_functions;
 #define	MAX_ALIAS_NAME	32
 typedef struct cmdalias_s
@@ -761,7 +1570,10 @@ void Con_AddToTabList (const char *name, const char *partial, const char *type)
 	tab_t	*t,*insert;
 	char	*i_bash, *i_bash2;
 	const char *i_name, *i_name2;
-	int		len, mark;
+	int		namelen, typelen, mark;
+
+	if (!Con_Match (name, partial))
+		return;
 
 	if (!*bash_partial && bash_singlematch)
 	{
@@ -798,11 +1610,16 @@ void Con_AddToTabList (const char *name, const char *partial, const char *type)
 	}
 
 	mark = Hunk_LowMark ();
-	len = strlen (name);
-	t = (tab_t *) Hunk_AllocName (sizeof (tab_t) + len + 1, "tablist");
-	memcpy (t + 1, name, len + 1);
+	namelen = (int) strlen (name) + 1;
+	typelen = type ? (int) strlen (type) + 1 : 0;
+	t = (tab_t *) Hunk_AllocName (sizeof (tab_t) + namelen + typelen, "tablist");
 	t->name = (const char *) (t + 1);
-	t->type = type;
+	memcpy ((char *) t->name, name, namelen);
+	if (type)
+	{
+		t->type = t->name + namelen;
+		memcpy ((char *) t->type, type, typelen);
+	}
 	t->count = 1;
 
 	if (!tablist) //create list
@@ -909,34 +1726,14 @@ static qboolean CompleteFileList (const char *partial, void *param)
 {
 	filelist_item_t *file, **list = (filelist_item_t **) param;
 	for (file = *list; file; file = file->next)
-		if (Con_Match (file->name, partial))
-			Con_AddToTabList (file->name, partial, NULL);
+		Con_AddToTabList (file->name, partial, NULL);
 	return true;
 }
 
-static qboolean CompleteClassnames (const char *partial, void *unused)
+static qboolean CompleteFileListSingle (const char *partial, void *param)
 {
-	extern edict_t *sv_player;
-	qcvm_t	*oldvm;
-	edict_t	*ed;
-	int		i;
-
-	if (!sv.active)
-		return true;
-	PR_PushQCVM (&sv.qcvm, &oldvm);
-
-	for (i = 1, ed = NEXT_EDICT (qcvm->edicts); i < qcvm->num_edicts; i++, ed = NEXT_EDICT (ed))
-	{
-		const char *name;
-		if (ed == sv_player || ed->free || !ed->v.classname)
-			continue;
-		name = PR_GetString (ed->v.classname);
-		if (*name && Con_Match (name, partial))
-			Con_AddToTabList (name, partial, "#");
-	}
-
-	PR_PopQCVM (oldvm);
-
+	if (Cmd_Argc () < 3)
+		CompleteFileList (partial, param);
 	return true;
 }
 
@@ -944,13 +1741,14 @@ static qboolean CompleteBindKeys (const char *partial, void *unused)
 {
 	int i;
 
+	// fall back to default tab completion after 1st arg (key name)
 	if (Cmd_Argc () > 2)
 		return false;
 
 	for (i = 0; i < MAX_KEYS; i++)
 	{
 		const char *name = Key_KeynumToString (i);
-		if (strcmp (name, "<UNKNOWN KEYNUM>") != 0 && Con_Match (name, partial))
+		if (strcmp (name, "<UNKNOWN KEYNUM>") != 0)
 			Con_AddToTabList (name, partial, keybindings[i]);
 	}
 
@@ -961,12 +1759,16 @@ static qboolean CompleteUnbindKeys (const char *partial, void *unused)
 {
 	int i;
 
+	// disable completion after 1st arg (key name)
+	if (Cmd_Argc () > 2)
+		return true;
+
 	for (i = 0; i < MAX_KEYS; i++)
 	{
 		if (keybindings[i])
 		{
 			const char *name = Key_KeynumToString (i);
-			if (strcmp (name, "<UNKNOWN KEYNUM>") != 0 && Con_Match (name, partial))
+			if (strcmp (name, "<UNKNOWN KEYNUM>") != 0)
 				Con_AddToTabList (name, partial, keybindings[i]);
 		}
 	}
@@ -983,22 +1785,20 @@ typedef struct arg_completion_type_s
 
 static const arg_completion_type_t arg_completion_types[] =
 {
-	{ "map",					CompleteFileList,		&extralevels },
-	{ "changelevel",			CompleteFileList,		&extralevels },
+	{ "map",					CompleteFileListSingle,	&extralevels },
+	{ "changelevel",			CompleteFileListSingle,	&extralevels },
 	{ "game",					CompleteFileList,		&modlist },
-	{ "record",					CompleteFileList,		&demolist },
-	{ "playdemo",				CompleteFileList,		&demolist },
-	{ "timedemo",				CompleteFileList,		&demolist },
-	{ "load",					CompleteFileList,		&savelist },
-	{ "save",					CompleteFileList,		&savelist },
-	{ "sky",					CompleteFileList,		&skylist },
-	{ "r_showbboxes_filter",	CompleteClassnames,		NULL },
+	{ "record",					CompleteFileListSingle,	&demolist },
+	{ "playdemo",				CompleteFileListSingle,	&demolist },
+	{ "timedemo",				CompleteFileListSingle,	&demolist },
+	{ "load",					CompleteFileListSingle,	&savelist },
+	{ "save",					CompleteFileListSingle,	&savelist },
+	{ "sky",					CompleteFileListSingle,	&skylist },
 	{ "bind",					CompleteBindKeys,		NULL },
 	{ "unbind",					CompleteUnbindKeys,		NULL },
 };
 
-static const int num_arg_completion_types =
-	sizeof(arg_completion_types)/sizeof(arg_completion_types[0]);
+static const int num_arg_completion_types = Q_COUNTOF(arg_completion_types);
 
 /*
 ============
@@ -1022,9 +1822,11 @@ static void BuildTabList (const char *partial)
 	if (Cmd_Argc () >= 2)
 	{
 		cvar = Cvar_FindVar (Cmd_Argv (0));
-		if (cvar && cvar->completion)
+		if (cvar)
 		{
-			cvar->completion (cvar, partial);
+		// cvars can only have one argument
+			if (Cmd_Argc () == 2 && cvar->completion)
+				cvar->completion (cvar, partial);
 			return;
 		}
 
@@ -1067,12 +1869,93 @@ static void BuildTabList (const char *partial)
 			Con_AddToTabList (alias->name, partial, "alias");
 }
 
+/*
+============
+Con_FormatTabMatch
+============
+*/
+static void Con_FormatTabMatch (const tab_t *t, char *dst, size_t dstsize)
+{
+	char tinted[MAXCMDLINE];
+
+	COM_TintSubstring (t->name, bash_partial, tinted, sizeof (tinted));
+
+	if (!t->type)
+		q_strlcpy (dst, tinted, dstsize);
+	else if (t->type[0] == '#' && !t->type[1])
+		q_snprintf (dst, dstsize, "%s (%d)", tinted, t->count);
+	else
+		q_snprintf (dst, dstsize, "%s (%s)", tinted, t->type);
+}
+
+/*
+============
+Con_PrintTabList
+============
+*/
+static void Con_PrintTabList (void)
+{
+	char	buf[MAXCMDLINE];
+	int		i, maxlen, cols, matches, total;
+	tab_t	*t;
+
+// determine maximum item length
+	matches = maxlen = 0;
+	t = tablist;
+	do
+	{
+		Con_FormatTabMatch (t, buf, sizeof (buf));
+		total = (int) strlen (buf);
+		maxlen = q_max (maxlen, total);
+		t = t->next;
+		++matches;
+	} while (t != tablist);
+
+// determine number of columns
+	if (!maxlen)
+		return;
+	maxlen += 3;										// indent
+	maxlen = q_max (maxlen, 8);							// min width
+	maxlen = (maxlen + 3) & ~3;							// round up to multiple of 4
+	cols = q_max (con_linewidth, maxlen) / maxlen;
+	if (con_maxcols.value >= 1.f)
+		cols = q_min (cols, (int) con_maxcols.value);	// apply user limit
+	if (matches < 6)									// single column if fewer than 6 matches
+		cols = 1;
+
+// print all matches
+	Con_SafePrintf("\n");
+	i = total = 0;
+	t = tablist;
+	do
+	{
+		Con_FormatTabMatch (t, buf, sizeof (buf));
+		if (++i == cols)
+		{
+			i = 0;
+			Con_SafePrintf ("   %s\n", buf);
+		}
+		else
+			Con_SafePrintf ("   %*s", -(maxlen-3), buf);
+		if (t->type && t->type[0] == '#' && !t->type[1])
+			total += t->count;
+		t = t->next;
+	} while (t != tablist);
+	if (i != 0)
+		Con_SafePrintf ("\n");
+
+	if (total > 0)
+		Con_SafePrintf ("   %d unique matches (%d total)\n", matches, total);
+
+	Con_SafePrintf("\n");
+}
+
 /*
 ============
 Con_TabComplete -- johnfitz
 ============
 */
-void Con_TabComplete (void)
+void Con_TabComplete (tabcomplete_t mode)
 {
 	char	partial[MAXCMDLINE];
 	const char	*match;
@@ -1080,6 +1963,16 @@ void Con_TabComplete (void)
 	tab_t		*t;
 	int		mark, i;
 
+	key_tabhint[0] = '\0';
+	if (mode == TABCOMPLETE_AUTOHINT)
+	{
+		key_tabpartial[0] = '\0';
+
+	// only show completion hint when the cursor is at the end of the line
+		if ((size_t)key_linepos >= sizeof (key_lines[edit_line]) || key_lines[edit_line][key_linepos])
+			return;
+	}
+
 // if editline is empty, return
 	if (key_lines[edit_line][1] == 0)
 		return;
@@ -1111,36 +2004,9 @@ void Con_TabComplete (void)
 		if (!tablist)
 			return;
 
-		// print list if length > 1
-		if (tablist->next != tablist)
-		{
-			int matches = 0;
-			int total = 0;
-			t = tablist;
-			Con_SafePrintf("\n");
-			do
-			{
-				char tinted[MAXCMDLINE];
-				COM_TintSubstring (t->name, bash_partial, tinted, sizeof (tinted));
-				if (t->type)
-				{
-					if (t->type[0] == '#' && !t->type[1])
-					{
-						Con_SafePrintf("   %s (%d)\n", tinted, t->count);
-						total += t->count;
-					}
-					else
-						Con_SafePrintf("   %s (%s)\n", tinted, t->type);
-				}
-				else
-					Con_SafePrintf("   %s\n", tinted);
-				t = t->next;
-				++matches;
-			} while (t != tablist);
-			if (total > 0)
-				Con_Printf ("   %d unique matches (%d total)\n", matches, total);
-			Con_SafePrintf("\n");
-		}
+		// print list if length > 1 and action is user-initiated
+		if (tablist->next != tablist && mode == TABCOMPLETE_USER)
+			Con_PrintTabList ();
 
 	//	match = tablist->name;
 	// First time, just show maximum matching chars -- S.A.
@@ -1167,6 +2033,17 @@ void Con_TabComplete (void)
 		} while (t != tablist);
 	}
 
+	if (mode == TABCOMPLETE_AUTOHINT)
+	{
+		size_t len = strlen (partial);
+		match = q_strcasestr (match, partial);
+		if (match && match[len])
+			q_strlcpy (key_tabhint, match + len, sizeof (key_tabhint));
+		Hunk_FreeToLowMark (mark);
+		key_tabpartial[0] = '\0';
+		return;
+	}
+
 // insert new match into edit line
 	q_strlcpy (partial, match, MAXCMDLINE); //first copy match string
 	q_strlcat (partial, key_lines[edit_line] + key_linepos, MAXCMDLINE); //then add chars after cursor
@@ -1192,6 +2069,8 @@ void Con_TabComplete (void)
 	// the changelevel command. the line below "fixes" it, although I'm not sure about
 	// the reason, yet, neither do I know any possible side effects of it:
 		c = key_lines[edit_line] + key_linepos;
+
+		Con_TabComplete (TABCOMPLETE_AUTOHINT);
 	}
 }
 
@@ -1314,7 +2193,8 @@ extern	qpic_t *pic_ovr, *pic_ins; //johnfitz -- new cursor handling
 
 void Con_DrawInput (void)
 {
-	int	i, ofs;
+	const char *workline = key_lines[edit_line];
+	int	i, ofs, len;
 
 	if (key_dest != key_console && !con_forcedup)
 		return;		// don't draw anything
@@ -1325,9 +2205,20 @@ void Con_DrawInput (void)
 	else
 		ofs = 0;
 
+	len = strlen (workline);
+
 // draw input string
-	for (i = 0; key_lines[edit_line][i+ofs] && i < con_linewidth; i++)
-		Draw_Character ((i+1)<<3, vid.conheight - 16, key_lines[edit_line][i+ofs]);
+	for (i = 0; i+ofs < len; i++)
+		Draw_Character ((i+1)<<3, vid.conheight - 16, workline[i+ofs]);
+
+// draw tab completion hint
+	if (key_tabhint[0])
+	{
+		GL_SetCanvasColor (1.0f, 1.0f, 1.0f, 0.75f);
+		for (i = 0; key_tabhint[i] && i+1+len-ofs < con_linewidth+CON_MARGIN*2; i++)
+			Draw_Character ((i+1+len-ofs)<<3, vid.conheight - 16, key_tabhint[i] | 0x80);
+		GL_SetCanvasColor (1.0f, 1.0f, 1.0f, 1.0f);
+	}
 
 // johnfitz -- new cursor handling
 	if (!((int)((realtime-key_blinktime)*con_cursorspeed) & 1))
@@ -1337,6 +2228,39 @@ void Con_DrawInput (void)
 	}
 }
 
+/*
+================
+Con_DrawSelectionHighlight
+================
+*/
+static void Con_DrawSelectionHighlight (int x, int y, int line, float alpha)
+{
+	conofs_t	selbegin, selend;
+	conofs_t	begin, end;
+	size_t		len;
+
+	if (!Con_GetNormalizedSelection (&selbegin, &selend))
+		return;
+
+	len = Con_StrLen (line);
+	begin.line = line;
+	begin.col = 0;
+	end.line = line;
+	end.col = len;
+
+	// Highlight line ends (as in Notepad, Visual Studio etc.)
+	if (end.line != selend.line && end.col == len)
+		end.col++;
+
+	// ...unless we would end up overlapping the console margin
+	end.col = q_min (end.col, con_linewidth);
+
+	if (!Con_IntersectRanges (&begin, &end, &selbegin, &selend))
+		return;
+
+	Draw_Fill (x + begin.col*8, y, (end.col-begin.col)*8, 8, 220, alpha);
+}
+
 /*
 ================
 Con_DrawConsole -- johnfitz -- heavy revision
@@ -1345,11 +2269,14 @@ Draws the console with the solid background
 The typing input line at the bottom should only be drawn if typing is allowed
 ================
 */
-void Con_DrawConsole (int lines, qboolean drawinput)
+void Con_DrawConsole (int lines, qboolean drawbg, qboolean drawinput)
 {
 	int	i, x, y, j, sb, rows;
 	const char	*text;
-	char	ver[32];
+	qboolean forced;
+	float alpha;
+
+	Con_UpdateMouseState ();
 
 	if (lines <= 0)
 		return;
@@ -1358,7 +2285,14 @@ void Con_DrawConsole (int lines, qboolean drawinput)
 	GL_SetCanvas (CANVAS_CONSOLE);
 
 // draw the background
-	Draw_ConsoleBackground ();
+	if (drawbg)
+		Draw_ConsoleBackground ();
+
+// fade out during live previews for console options when there's no active game
+	forced = con_forcedup && M_WantsConsole (&alpha);
+	if (!forced)
+		alpha = 1.f;
+	GL_PushCanvasColor (1.f, 1.f, 1.f, alpha);
 
 // draw the buffer text
 	rows = (con_vislines +7)/8;
@@ -1372,9 +2306,30 @@ void Con_DrawConsole (int lines, qboolean drawinput)
 		if (j < 0)
 			j = 0;
 		text = con_text + (j % con_totallines)*con_linewidth;
+		Con_DrawSelectionHighlight (8, y, j, alpha);
+	}
 
+	y = vid.conheight - (rows+2)*8; // +2 for input and version lines
+	for (i = con_current - rows + 1; i <= con_current - sb; i++, y += 8)
+	{
+		conofs_t ofs;
+		j = i - con_backscroll;
+		if (j < 0)
+			j = 0;
+		text = con_text + (j % con_totallines)*con_linewidth;
+		ofs.line = j;
 		for (x = 0; x < con_linewidth; x++)
-			Draw_Character ( (x + 1)<<3, y, text[x]);
+		{
+			char c = text[x];
+			ofs.col = x;
+			if (con_hotlink && Con_OfsInRange (&ofs, &con_hotlink->begin, &con_hotlink->end))
+			{
+				if (keydown[K_MOUSE1])
+					c &= 0x7f;
+				Draw_Character ((x + 1)<<3, y + 2, '_' | (c & 0x80));
+			}
+			Draw_Character ((x + 1)<<3, y, c);
+		}
 	}
 
 // draw scrollback arrows
@@ -1391,10 +2346,10 @@ void Con_DrawConsole (int lines, qboolean drawinput)
 		Con_DrawInput ();
 
 //draw version number in bottom right
-	y += 8;
-	q_snprintf (ver, sizeof(ver), CONSOLE_TITLE_STRING);
-	for (x = 0; x < (int)strlen(ver); x++)
-		Draw_Character ((con_linewidth - strlen(ver) + x + 2)<<3, y, ver[x] /*+ 128*/);
+	text = CONSOLE_TITLE_STRING;
+	M_PrintWhite (vid.conwidth - (strlen (text) << 3), vid.conheight - 8, text);
+
+	GL_PopCanvasColor ();
 }
 
 
@@ -1458,7 +2413,6 @@ void LOG_Init (quakeparms_t *parms)
 		return;
 	}
 
-	con_debuglog = true;
 	Con_DebugLog (va("LOG started on: %s \n", session));
 
 }
