@@ -80,6 +80,7 @@ cvar_t	r_simd = {"r_simd","1",CVAR_ARCHIVE};
 #endif
 cvar_t	r_alphasort = {"r_alphasort","1",CVAR_ARCHIVE};
 cvar_t	r_oit = {"r_oit","1",CVAR_ARCHIVE};
+cvar_t	r_dither = {"r_dither", "1.0", CVAR_ARCHIVE};
 
 cvar_t	gl_finish = {"gl_finish","0",CVAR_NONE};
 cvar_t	gl_clear = {"gl_clear","1",CVAR_NONE};
@@ -92,7 +93,7 @@ cvar_t	r_clearcolor = {"r_clearcolor","2",CVAR_ARCHIVE};
 cvar_t	r_flatlightstyles = {"r_flatlightstyles", "0", CVAR_NONE};
 cvar_t	r_lerplightstyles = {"r_lerplightstyles", "1", CVAR_ARCHIVE}; // 0=off; 1=skip abrupt transitions; 2=always lerp
 cvar_t	gl_fullbrights = {"gl_fullbrights", "1", CVAR_ARCHIVE};
-cvar_t	gl_farclip = {"gl_farclip", "16384", CVAR_ARCHIVE};
+cvar_t	gl_farclip = {"gl_farclip", "65536", CVAR_ARCHIVE};
 cvar_t	gl_overbright_models = {"gl_overbright_models", "1", CVAR_ARCHIVE};
 cvar_t	r_oldskyleaf = {"r_oldskyleaf", "0", CVAR_NONE};
 cvar_t	r_drawworld = {"r_drawworld", "1", CVAR_NONE};
@@ -100,6 +101,10 @@ cvar_t	r_showtris = {"r_showtris", "0", CVAR_NONE};
 cvar_t	r_showbboxes = {"r_showbboxes", "0", CVAR_NONE};
 cvar_t	r_showbboxes_think = {"r_showbboxes_think", "0", CVAR_NONE}; // 0=show all; 1=thinkers only; -1=non-thinkers only
 cvar_t	r_showbboxes_health = {"r_showbboxes_health", "0", CVAR_NONE}; // 0=show all; 1=healthy only; -1=non-healthy only
+cvar_t	r_showbboxes_links = {"r_showbboxes_links", "3", CVAR_NONE}; // 0=off; 1=outgoing only; 2=incoming only; 3=incoming+outgoing
+cvar_t	r_showbboxes_targets = {"r_showbboxes_targets", "1", CVAR_NONE};
+cvar_t	r_showfields = {"r_showfields", "0", CVAR_NONE};
+cvar_t	r_showfields_align = {"r_showfields_align", "1", CVAR_ARCHIVE}; // 0=entity pos; 1=bottom-right
 cvar_t	r_lerpmodels = {"r_lerpmodels", "1", CVAR_ARCHIVE};
 cvar_t	r_lerpmove = {"r_lerpmove", "1", CVAR_ARCHIVE};
 cvar_t	r_nolerp_list = {"r_nolerp_list", "progs/flame.mdl,progs/flame2.mdl,progs/braztall.mdl,progs/brazshrt.mdl,progs/longtrch.mdl,progs/flame_pyre.mdl,progs/v_saw.mdl,progs/v_xfist.mdl,progs/h2stuff/newfire.mdl", CVAR_NONE};
@@ -108,6 +113,8 @@ cvar_t	r_noshadow_list = {"r_noshadow_list", "progs/flame2.mdl,progs/flame.mdl,p
 extern cvar_t	r_vfog;
 extern cvar_t	vid_fsaa;
 //johnfitz
+extern cvar_t	r_softemu_dither_screen;
+extern cvar_t	r_softemu_dither_texture;
 
 cvar_t	gl_zfix = {"gl_zfix", "1", CVAR_ARCHIVE}; // QuakeSpasm z-fighting fix
 
@@ -308,6 +315,8 @@ void GL_DeleteFrameBuffers (void)
 //
 //==============================================================================
 
+static const float NOISESCALE = 9.f / 255.f;
+
 extern GLuint gl_palette_lut;
 extern GLuint gl_palette_buffer[2];
 
@@ -319,12 +328,14 @@ GL_PostProcess
 void GL_PostProcess (void)
 {
 	int palidx, variant;
+	float dither;
 	if (!GL_NeedsPostprocess ())
 		return;
 
 	GL_BeginGroup ("Postprocess");
 
 	palidx =  GLPalette_Postprocess ();
+	dither = (softemu == SOFTEMU_FINE) ? NOISESCALE * r_dither.value * r_softemu_dither_screen.value : 0.f;
 
 	GL_BindFramebufferFunc (GL_FRAMEBUFFER, 0);
 	glViewport (glx, gly, glwidth, glheight);
@@ -336,7 +347,7 @@ void GL_PostProcess (void)
 	GL_BindNative (GL_TEXTURE1, GL_TEXTURE_3D, gl_palette_lut);
 	GL_BindBufferRange (GL_SHADER_STORAGE_BUFFER, 0, gl_palette_buffer[palidx], 0, 256 * sizeof (GLuint));
 	if (variant != 2) // some AMD drivers optimize out the uniform in variant #2
-		GL_Uniform3fFunc (0, vid_gamma.value, q_min(2.0f, q_max(1.0f, vid_contrast.value)), 1.f/r_refdef.scale);
+		GL_Uniform4fFunc (0, vid_gamma.value, q_min(2.0f, q_max(1.0f, vid_contrast.value)), 1.f/r_refdef.scale, dither);
 
 	glDrawArrays (GL_TRIANGLES, 0, 3);
 
@@ -558,6 +569,43 @@ void GL_DepthRange (zrange_t range)
 	}
 }
 
+/*
+=============
+R_GetAlphaMode
+=============
+*/
+alphamode_t R_GetAlphaMode (void)
+{
+	if (r_oit.value)
+		return ALPHAMODE_OIT;
+	return r_alphasort.value ? ALPHAMODE_SORTED : ALPHAMODE_BASIC;
+}
+
+/*
+=============
+R_GetEffectiveAlphaMode
+=============
+*/
+alphamode_t R_GetEffectiveAlphaMode (void)
+{
+	if (map_checks.value)
+		return ALPHAMODE_BASIC;
+	return R_GetAlphaMode ();
+}
+
+/*
+=============
+R_SetAlphaMode
+=============
+*/
+void R_SetAlphaMode (alphamode_t mode)
+{
+	Cvar_SetValueQuick (&r_oit, mode == ALPHAMODE_OIT);
+	if (mode != ALPHAMODE_OIT)
+		Cvar_SetValueQuick (&r_alphasort, mode == ALPHAMODE_SORTED);
+}
+
+
 //==============================================================================
 //
 // SETUP FRAME
@@ -587,7 +635,7 @@ static void R_SortEntities (void)
 	int i, j, pass;
 	int bins[1 << (MODSORT_BITS/2)];
 	int typebins[mod_numtypes*2];
-	qboolean alphasort = r_alphasort.value && !r_oit.value;
+	alphamode_t alphamode = R_GetEffectiveAlphaMode ();
 
 	if (!r_drawentities.value)
 		cl_numvisedicts = 0;
@@ -614,7 +662,7 @@ static void R_SortEntities (void)
 		entity_t *ent = cl_visedicts[i];
 		qboolean translucent = !ENTALPHA_OPAQUE (ent->alpha);
 
-		if (translucent && alphasort)
+		if (translucent && alphamode == ALPHAMODE_SORTED)
 		{
 			float dist, delta;
 			vec3_t mins, maxs;
@@ -628,7 +676,7 @@ static void R_SortEntities (void)
 			dist = sqrt (dist);
 			visedict_keys[i] = ~CLAMP (0, (int)dist, MODSORT_MASK);
 		}
-		else if (translucent && !r_oit.value)
+		else if (translucent && alphamode != ALPHAMODE_OIT)
 		{
 			// Note: -1 (0xfffff) for non-static entities (firstleaf=0),
 			// so they are sorted after static ones
@@ -835,7 +883,7 @@ GL_NeedsPostprocess
 */
 qboolean GL_NeedsPostprocess (void)
 {
-	return vid_gamma.value != 1.f || vid_contrast.value != 1.f || softemu || r_oit.value;
+	return vid_gamma.value != 1.f || vid_contrast.value != 1.f || softemu || R_GetEffectiveAlphaMode () == ALPHAMODE_OIT;
 }
 
 /*
@@ -920,6 +968,28 @@ void R_SetupView (void)
 	r_framedata.eyepos[1] = r_refdef.vieworg[1];
 	r_framedata.eyepos[2] = r_refdef.vieworg[2];
 	r_framedata.time = cl.time;
+	if (softemu == SOFTEMU_COARSE)
+	{
+		r_framedata.screendither = NOISESCALE * r_dither.value * r_softemu_dither_screen.value;
+		r_framedata.texturedither = NOISESCALE * r_dither.value * r_softemu_dither_texture.value;
+
+		// r_fullbright replaces the actual lightmap texture with a 2x2 50% grey one.
+		// Since texture-space dithering is applied on a scale of 1/16 of a lightmap texel,
+		// this would lead to massively overscaled dithering patterns, so we disable
+		// texture-space dithering in this case.
+		if (r_fullbright_cheatsafe)
+			r_framedata.texturedither = 0.f;
+	}
+	else if (softemu == SOFTEMU_OFF)
+	{
+		r_framedata.screendither = r_dither.value * (1.f/255.f);
+		r_framedata.texturedither = 0.f;
+	}
+	else // FINE (screen-space dithering applied during postprocessing), or BANDED (no dithering)
+	{
+		r_framedata.screendither = 0.f;
+		r_framedata.texturedither = 0.f;
+	}
 
 	Fog_SetupFrame (); //johnfitz
 	Sky_SetupFrame ();
@@ -942,13 +1012,15 @@ void R_SetupView (void)
 	if (r_waterwarp.value)
 	{
 		int contents = Mod_PointInLeaf (r_origin, cl.worldmodel)->contents;
-		if (contents == CONTENTS_WATER || contents == CONTENTS_SLIME || contents == CONTENTS_LAVA || cl.forceunderwater)
+		qboolean forced = M_ForcedUnderwater ();
+		if (contents == CONTENTS_WATER || contents == CONTENTS_SLIME || contents == CONTENTS_LAVA || cl.forceunderwater || forced)
 		{
+			double t = forced ? realtime : cl.time;
 			if (r_waterwarp.value > 1.f)
 			{
 				//variance is a percentage of width, where width = 2 * tan(fov / 2) otherwise the effect is too dramatic at high FOV and too subtle at low FOV.  what a mess!
-				r_fovx = atan(tan(DEG2RAD(r_refdef.fov_x) / 2) * (0.97 + sin(cl.time * 1.5) * 0.03)) * 2 / M_PI_DIV_180;
-				r_fovy = atan(tan(DEG2RAD(r_refdef.fov_y) / 2) * (1.03 - sin(cl.time * 1.5) * 0.03)) * 2 / M_PI_DIV_180;
+				r_fovx = atan(tan(DEG2RAD(r_refdef.fov_x) / 2) * (0.97 + sin(t * 1.5) * 0.03)) * 2 / M_PI_DIV_180;
+				r_fovy = atan(tan(DEG2RAD(r_refdef.fov_y) / 2) * (1.03 - sin(t * 1.5) * 0.03)) * 2 / M_PI_DIV_180;
 			}
 			else
 			{
@@ -1116,7 +1188,7 @@ static void R_FlushDebugGeometry (void)
 		GLbyte	*ofs;
 
 		GL_UseProgram (glprogs.debug3d);
-		GL_SetState (GLS_BLEND_OPAQUE | GLS_NO_ZTEST | GLS_NO_ZWRITE | GLS_CULL_NONE | GLS_ATTRIBS(2));
+		GL_SetState (GLS_BLEND_ALPHA | GLS_NO_ZTEST | GLS_NO_ZWRITE | GLS_CULL_NONE | GLS_ATTRIBS(2));
 
 		GL_Upload (GL_ARRAY_BUFFER, debugverts, sizeof (debugverts[0]) * numdebugverts, &buf, &ofs);
 		GL_BindBuffer (GL_ARRAY_BUFFER, buf);
@@ -1217,38 +1289,163 @@ static void R_EmitWireBox (const vec3_t mins, const vec3_t maxs, uint32_t color)
 	R_AddDebugGeometry (v, countof (v), boxidx, countof (boxidx));
 }
 
+/*
+================
+R_EmitArrow
+================
+*/
+static void R_EmitArrow (const vec3_t from, const vec3_t to, uint32_t color)
+{
+	float	frac, len;
+	vec3_t	center, dir, side, tmp;
+
+	R_EmitLine (from, to, color);
+
+	VectorSubtract (to, from, dir);
+	len = VectorNormalize (dir);
+	if (len < 1e-2f)
+	{
+		VectorCopy (vup, dir);
+		VectorCopy (vright, side);
+	}
+	else
+	{
+		VectorSubtract (from, r_origin, tmp);
+		CrossProduct (dir, tmp, side);
+		VectorNormalize (side);
+	}
+
+	frac = realtime - floor (realtime);
+	VectorLerp (from, to, frac, center);
+
+	VectorMA (center, 8.f, side, tmp);
+	VectorMA (tmp, -8.f, dir, tmp);
+	R_EmitLine (tmp, center, color);
+
+	VectorMA (tmp, -16.f, side, tmp);
+	R_EmitLine (tmp, center, color);
+}
+
+/*
+================
+R_EmitEdictLink
+================
+*/
+static void R_EmitEdictLink (const edict_t *from, const edict_t *to, showbboxflags_t flags)
+{
+	vec3_t vec_from, vec_to;
+
+	if (!flags)
+		return;
+
+	VectorCopy (from->v.origin, vec_from);
+	if (!VectorCompare (from->v.mins, from->v.maxs))
+	{
+		VectorMA (vec_from, 0.5f, from->v.mins, vec_from);
+		VectorMA (vec_from, 0.5f, from->v.maxs, vec_from);
+	}
+
+	VectorCopy (to->v.origin, vec_to);
+	if (!VectorCompare (to->v.mins, to->v.maxs))
+	{
+		VectorMA (vec_to, 0.5f, to->v.mins, vec_to);
+		VectorMA (vec_to, 0.5f, to->v.maxs, vec_to);
+	}
+
+	if (flags == SHOWBBOX_LINK_BOTH)
+		R_EmitLine (vec_from, vec_to, 0x7f7f3f7f);
+	else if (flags == SHOWBBOX_LINK_OUTGOING)
+		R_EmitArrow (vec_from, vec_to, 0x7f7f3f3f);
+	else if (flags == SHOWBBOX_LINK_INCOMING)
+		R_EmitArrow (vec_to, vec_from, 0x7f3f3f7f);
+}
+
 /*
 ================
 R_ShowBoundingBoxesFilter
 
-r_showbboxes_filter artifact =trigger_secret
+r_showbboxes_filter artifact =trigger_secret #42
 ================
 */
 char r_showbboxes_filter_strings[MAXCMDLINE];
+qboolean r_showbboxes_filter_byindex;
 
 static qboolean R_ShowBoundingBoxesFilter (edict_t *ed)
 {
+	char entnum[16] = "";
+	const char *classname = NULL;
+	const char *filter_p = r_showbboxes_filter_strings;
+
 	if (!r_showbboxes_filter_strings[0])
 		return true;
 
+	if (r_showbboxes_filter_byindex)
+		q_snprintf (entnum, sizeof (entnum), "%d", NUM_FOR_EDICT (ed));
+
 	if (ed->v.classname)
+		classname = PR_GetString (ed->v.classname);
+
+	for (filter_p = r_showbboxes_filter_strings; *filter_p; filter_p += strlen (filter_p) + 1)
 	{
-		const char *classname = PR_GetString (ed->v.classname);
-		const char *str = r_showbboxes_filter_strings;
-		qboolean is_allowed = false;
-		while (*str && !is_allowed)
+		if (*filter_p == '#')
 		{
-			if (*str == '=')
-				is_allowed = !strcmp (classname, str + 1);
-			else
-				is_allowed = strstr (classname, str) != NULL;
-			str += strlen (str) + 1;
+			if (!strcmp (entnum, filter_p + 1))
+				return true;
+			continue;
 		}
-		return is_allowed;
+
+		if (!classname)
+			continue;
+
+		if (*filter_p == '=')
+		{
+			if (!strcmp (classname, filter_p + 1))
+				return true;
+			continue;
+		}
+
+		if (strstr (classname, filter_p) != NULL)
+			return true;
 	}
+
 	return false;
 }
 
+static edict_t **bbox_edicts = NULL;		// all edicts shown by r_showbboxes & co
+edict_t **bbox_linked = NULL;				// focused edict, followed by edicts linked from/to it
+
+/*
+================
+R_AddHighlightedEntity
+================
+*/
+static void R_AddHighlightedEntity (edict_t *ed, showbboxflags_t flags)
+{
+	if (ed->showbboxframe != r_framecount)
+	{
+		ed->showbboxframe = r_framecount;
+		ed->showbboxflags = SHOWBBOX_LINK_NONE;
+		VEC_PUSH (bbox_edicts, ed);
+	}
+
+	if (!(ed->showbboxflags & flags) && (int)r_showbboxes_links.value & flags)
+	{
+		VEC_PUSH (bbox_linked, ed);
+		ed->showbboxflags |= flags;
+	}
+}
+
+/*
+================
+R_ClearBoundingBoxes
+================
+*/
+void R_ClearBoundingBoxes (void)
+{
+	VEC_CLEAR (bbox_edicts);
+	VEC_CLEAR (bbox_linked);
+}
+
 /*
 ================
 R_ShowBoundingBoxes -- johnfitz
@@ -1261,12 +1458,19 @@ static void R_ShowBoundingBoxes (void)
 	extern		edict_t *sv_player;
 	byte		*pvs;
 	vec3_t		mins,maxs;
-	edict_t		*ed;
-	int			i, mode;
+	edict_t		*ed, *focused;
+	int			i, j, mode;
 	uint32_t	color;
 	qcvm_t 		*oldvm;	//in case we ever draw a scene from within csqc.
+	float		dist, bestdist, extend;
+	vec3_t		rcpdelta;
 
-	if (!r_showbboxes.value || cl.maxclients > 1 || !r_drawentities.value || !sv.active)
+	VEC_CLEAR (bbox_edicts);
+	VEC_CLEAR (bbox_linked);
+	focused = NULL;
+
+	mode = abs ((int)r_showbboxes.value);
+	if ((!mode && !r_showfields.value) || cl.maxclients > 1 || !r_drawentities.value || !sv.active)
 		return;
 
 	GL_BeginGroup ("Show bounding boxes");
@@ -1275,8 +1479,8 @@ static void R_ShowBoundingBoxes (void)
 	PR_SwitchQCVM(NULL);
 	PR_SwitchQCVM(&sv.qcvm);
 
-	mode = abs ((int)r_showbboxes.value);
-	if (mode >= 2)
+	// Use PVS if r_showbboxes >= 2, or if r_showbboxes is 0 (which means r_showfields is active)
+	if (mode >= 2 || mode == 0)
 	{
 		vec3_t org;
 		VectorAdd (sv_player->v.origin, sv_player->v.view_ofs, org);
@@ -1285,6 +1489,12 @@ static void R_ShowBoundingBoxes (void)
 	else
 		pvs = NULL;
 
+	// Compute ray reciprocal delta
+	for (i = 0; i < 3; i++)
+		rcpdelta[i] = 1.f / (gl_farclip.value * vpn[i]);
+
+	// Iterate over all server entities
+	bestdist = FLT_MAX;
 	for (i=1, ed=NEXT_EDICT(qcvm->edicts) ; i<qcvm->num_edicts ; i++, ed=NEXT_EDICT(ed))
 	{
 		if (ed == sv_player || ed->free)
@@ -1296,9 +1506,23 @@ static void R_ShowBoundingBoxes (void)
 		if (r_showbboxes_health.value && (ed->v.health <= 0) == (r_showbboxes_health.value > 0))
 			continue;
 
+		// Compute bounding box (16 units wide for point entities)
+		extend = VectorCompare (ed->v.mins, ed->v.maxs) ? 8.f : 0.f;
+		for (j = 0; j < 3; j++)
+		{
+			mins[j] = ed->v.origin[j] + ed->v.mins[j] - extend;
+			maxs[j] = ed->v.origin[j] + ed->v.maxs[j] + extend;
+		}
+
+		// Frustum culling
+		if (R_CullBox (mins, maxs))
+			continue;
+
+		// Classname or edict num filter
 		if (!R_ShowBoundingBoxesFilter(ed))
 			continue;
 
+		// PVS filter
 		if (pvs)
 		{
 			qboolean inpvs =
@@ -1310,28 +1534,115 @@ static void R_ShowBoundingBoxes (void)
 				continue;
 		}
 
-		if (r_showbboxes.value > 0.f)
+		// Keep track of the closest bounding box intersecting the center ray
+		// Note: if we're inside the box (dist == 0), we ignore this entity
+		if (RayVsBox (r_origin, rcpdelta, mins, maxs, &dist) && dist > 0.f && dist < bestdist)
+		{
+			bestdist = dist;
+			focused = ed;
+		}
+
+		// Add edict to list
+		R_AddHighlightedEntity (ed, SHOWBBOX_LINK_NONE);
+	}
+
+	if (focused)
+		VEC_PUSH (bbox_linked, focused);
+
+	if (focused && r_showbboxes_links.value)
+	{
+		// Find outgoing links (entity field references other than .chain)
+		if ((int)r_showbboxes_links.value & SHOWBBOX_LINK_OUTGOING)
+		{
+			for (i = 0; i < qcvm->numentityfields; i++)
+			{
+				eval_t *val = (eval_t *)((char *)&focused->v + qcvm->entityfieldofs[i]);
+				if (qcvm->entityfieldofs[i] == offsetof (entvars_t, chain) || !val->edict)
+					continue;
+				ed = PROG_TO_EDICT (val->edict);
+				if (ed == focused || ed->free || ed == sv_player)
+					continue;
+				R_AddHighlightedEntity (ed, SHOWBBOX_LINK_OUTGOING);
+			}
+		}
+
+		// Inspect all other edicts to find incoming links
+		// (either entity field references or target/targetname matches)
+		if ((int)r_showbboxes_links.value & SHOWBBOX_LINK_INCOMING || r_showbboxes_targets.value)
+		{
+			const char *focus_target = PR_GetString (focused->v.target);
+			const char *focus_targetname = PR_GetString (focused->v.targetname);
+
+			for (i=1, ed=NEXT_EDICT(qcvm->edicts) ; i<qcvm->num_edicts ; i++, ed=NEXT_EDICT(ed))
+			{
+				if (ed == sv_player || ed->free || ed == focused)
+					continue;
+
+				// Check target/targetname matches
+				if (r_showbboxes_targets.value && (*focus_target || *focus_targetname))
+				{
+					const char *target = PR_GetString (ed->v.target);
+					const char *targetname = PR_GetString (ed->v.targetname);
+
+					if (*focus_targetname && !strcmp (focus_targetname, target))
+						R_AddHighlightedEntity (ed, SHOWBBOX_LINK_INCOMING);
+					if (*focus_target && !strcmp (focus_target, targetname))
+						R_AddHighlightedEntity (ed, SHOWBBOX_LINK_OUTGOING);
+				}
+
+				// Check for entity field references (other than .chain)
+				if ((int)r_showbboxes_links.value & SHOWBBOX_LINK_INCOMING)
+				{
+					for (j = 0; j < qcvm->numentityfields; j++)
+					{
+						eval_t *val = (eval_t *)((char *)&ed->v + qcvm->entityfieldofs[j]);
+						if (qcvm->entityfieldofs[i] == offsetof (entvars_t, chain) || !val->edict)
+							continue;
+						if (PROG_TO_EDICT (val->edict) == focused)
+							R_AddHighlightedEntity (ed, SHOWBBOX_LINK_INCOMING);
+					}
+				}
+			}
+		}
+
+		// Draw all links
+		for (j = 0; j < (int) VEC_SIZE (bbox_linked); j++)
+			R_EmitEdictLink (focused, bbox_linked[j], bbox_linked[j]->showbboxflags);
+	}
+
+	// Draw all the matching edicts
+	for (i = 0; i < (int) VEC_SIZE (bbox_edicts); i++)
+	{
+		ed = bbox_edicts[i];
+
+		if (ed == focused)
+			color = 0xffffffff;
+		else if (ed->showbboxflags)
+			color = 0xaaaaaaaa;
+		else if (r_showbboxes.value > 0.f)
 		{
 			int modelindex = (int)ed->v.modelindex;
-			color = 0xff800080;
+			color = 0x7f800080;
 			if (modelindex >= 0 && modelindex < MAX_MODELS && sv.models[modelindex])
 			{
 				switch (sv.models[modelindex]->type)
 				{
-					case mod_brush:  color = 0xffff8080; break;
-					case mod_alias:  color = 0xff408080; break;
-					case mod_sprite: color = 0xff4040ff; break;
+					case mod_brush:  color = 0x7fff8080; break;
+					case mod_alias:  color = 0x7f408080; break;
+					case mod_sprite: color = 0x7f4040ff; break;
 					default:
 						break;
 				}
 			}
 			if (ed->v.health > 0)
-				color = 0xff0000ff;
+				color = 0x7f0000ff;
 		}
+		else if (r_showbboxes.value < 0.f)
+			color = 0x7fffffff;
 		else
-			color = 0xffffffff;
+			color = 0x5f7f7f7f;
 
-		if (ed->v.mins[0] == ed->v.maxs[0] && ed->v.mins[1] == ed->v.maxs[1] && ed->v.mins[2] == ed->v.maxs[2])
+		if (VectorCompare (ed->v.mins, ed->v.maxs))
 		{
 			//point entity
 			R_EmitWirePoint (ed->v.origin, color);
@@ -1345,6 +1656,8 @@ static void R_ShowBoundingBoxes (void)
 		}
 	}
 
+	VEC_CLEAR (bbox_edicts);
+
 	PR_SwitchQCVM(NULL);
 	PR_SwitchQCVM(oldvm);
 
@@ -1420,7 +1733,7 @@ static void R_BeginTranslucency (void)
 
 	GL_BeginGroup ("Translucent objects");
 
-	if (r_oit.value)
+	if (R_GetEffectiveAlphaMode () == ALPHAMODE_OIT)
 	{
 		GL_BindFramebufferFunc (GL_FRAMEBUFFER, framesetup.oit_fbo);
 		GL_ClearBufferfvFunc (GL_COLOR, 0, zeroes);
@@ -1440,7 +1753,7 @@ R_EndTranslucency
 */
 static void R_EndTranslucency (void)
 {
-	if (r_oit.value)
+	if (R_GetEffectiveAlphaMode () == ALPHAMODE_OIT)
 	{
 		GL_BeginGroup  ("OIT resolve");
 
@@ -1521,6 +1834,7 @@ void R_WarpScaleView (void)
 	qboolean msaa = framebufs.scene.samples > 1;
 	qboolean needwarpscale;
 	GLuint fbodest;
+	double t;
 
 	if (!GL_NeedsSceneEffects ())
 		return;
@@ -1566,14 +1880,15 @@ void R_WarpScaleView (void)
 	GL_UseProgram (glprogs.warpscale[water_warp]);
 	GL_SetState (GLS_BLEND_OPAQUE | GLS_NO_ZTEST | GLS_NO_ZWRITE | GLS_CULL_NONE | GLS_ATTRIBS(0));
 
-	GL_Uniform4fFunc (0, smax, tmax, water_warp ? 1.f/256.f : 0.f, cl.time);
+	t = M_ForcedUnderwater () ? realtime : cl.time;
+	GL_Uniform4fFunc (0, smax, tmax, water_warp ? 1.f/256.f : 0.f, (float)t);
 	if (v_blend[3] && gl_polyblend.value && !softemu)
 		GL_Uniform4fvFunc (1, 1, v_blend);
 	else
 		GL_Uniform4fFunc (1, 0.f, 0.f, 0.f, 0.f);
 	GL_BindNative (GL_TEXTURE0, GL_TEXTURE_2D, msaa ? framebufs.resolved_scene.color_tex : framebufs.scene.color_tex);
-	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, water_warp ? GL_LINEAR : GL_NEAREST);
-	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, water_warp ? GL_LINEAR : GL_NEAREST);
+	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, water_warp && msaa ? GL_LINEAR : GL_NEAREST);
+	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, water_warp && msaa ? GL_LINEAR : GL_NEAREST);
 
 	glDrawArrays (GL_TRIANGLES, 0, 3);
 
