@@ -38,6 +38,10 @@ extern cvar_t r_showtris;
 extern cvar_t r_showbboxes;
 extern cvar_t r_showbboxes_think;
 extern cvar_t r_showbboxes_health;
+extern cvar_t r_showbboxes_links;
+extern cvar_t r_showbboxes_targets;
+extern cvar_t r_showfields;
+extern cvar_t r_showfields_align;
 extern cvar_t r_lerpmodels;
 extern cvar_t r_lerpmove;
 extern cvar_t r_nolerp_list;
@@ -46,6 +50,7 @@ extern cvar_t r_noshadow_list;
 extern cvar_t gl_zfix; // QuakeSpasm z-fighting fix
 extern cvar_t r_alphasort;
 extern cvar_t r_oit;
+extern cvar_t r_dither;
 
 #if defined(USE_SIMD)
 extern cvar_t r_simd;
@@ -54,6 +59,9 @@ qboolean use_simd;
 
 extern gltexture_t *playertextures[MAX_SCOREBOARD]; //johnfitz
 
+extern char r_showbboxes_filter_strings[MAXCMDLINE];
+extern qboolean r_showbboxes_filter_byindex;
+
 /*
 ====================
 R_ShowbboxesFilter_f
@@ -61,11 +69,10 @@ R_ShowbboxesFilter_f
 */
 static void R_ShowbboxesFilter_f (void)
 {
-	extern char r_showbboxes_filter_strings[MAXCMDLINE];
-
 	if (Cmd_Argc () >= 2)
 	{
 		int i, len, ofs;
+		r_showbboxes_filter_byindex = false;
 		for (i = 1, ofs = 0; i < Cmd_Argc (); i++)
 		{
 			const char *arg = Cmd_Argv (i);
@@ -77,6 +84,7 @@ static void R_ShowbboxesFilter_f (void)
 				Con_Warning ("overflow at \"%s\"\n", arg);
 				break;
 			}
+			r_showbboxes_filter_byindex |= (arg[0] == '#');
 			memcpy (&r_showbboxes_filter_strings[ofs], arg, len);
 			ofs += len;
 		}
@@ -97,6 +105,59 @@ static void R_ShowbboxesFilter_f (void)
 	}
 }
 
+/*
+====================
+R_ShowbboxesFilter_Completion_f -- tab completion for r_showbboxes_filter
+====================
+*/
+static void R_ShowbboxesFilter_Completion_f (const char *partial)
+{
+	extern edict_t *sv_player;
+	extern edict_t **bbox_linked;
+	qcvm_t	*oldvm;
+	edict_t	*ed;
+	int		i;
+
+	if (!sv.active)
+		return;
+
+	PR_PushQCVM (&sv.qcvm, &oldvm);
+
+	if (*partial == '#')
+	{
+		for (i = 0; i < (int) VEC_SIZE (bbox_linked); i++)
+		{
+			ed = bbox_linked[i];
+			Con_AddToTabList (va ("#%d", NUM_FOR_EDICT (ed)), partial, PR_GetString (ed->v.classname));
+		}
+	}
+	else
+	{
+		for (i = 1, ed = NEXT_EDICT (qcvm->edicts); i < qcvm->num_edicts; i++, ed = NEXT_EDICT (ed))
+		{
+			const char *name;
+			if (ed == sv_player || ed->free || !ed->v.classname)
+				continue;
+			name = PR_GetString (ed->v.classname);
+			if (*name)
+				Con_AddToTabList (name, partial, "#");
+		}
+	}
+
+	PR_PopQCVM (oldvm);
+}
+
+/*
+====================
+R_ShowbboxesFilterClear_f
+====================
+*/
+static void R_ShowbboxesFilterClear_f (void)
+{
+	r_showbboxes_filter_strings[0] = '\0';
+	r_showbboxes_filter_byindex = false;
+}
+
 /*
 ====================
 GL_Fullbrights_f -- johnfitz
@@ -224,11 +285,14 @@ R_Init
 */
 void R_Init (void)
 {
-	extern cvar_t gl_finish;
+	cmd_function_t *cmd;
 
 	Cmd_AddCommand ("timerefresh", R_TimeRefresh_f);
 	Cmd_AddCommand ("pointfile", R_ReadPointFile_f);
-	Cmd_AddCommand ("r_showbboxes_filter", R_ShowbboxesFilter_f);
+	cmd = Cmd_AddCommand ("r_showbboxes_filter", R_ShowbboxesFilter_f);
+	if (cmd)
+		cmd->completion = R_ShowbboxesFilter_Completion_f;
+	Cmd_AddCommand ("r_showbboxes_filter_clear", R_ShowbboxesFilterClear_f);
 
 	Cvar_RegisterVariable (&r_norefresh);
 	Cvar_RegisterVariable (&r_lightmap);
@@ -249,6 +313,7 @@ void R_Init (void)
 	Cvar_RegisterVariable (&r_pos);
 	Cvar_RegisterVariable (&r_alphasort);
 	Cvar_RegisterVariable (&r_oit);
+	Cvar_RegisterVariable (&r_dither);
 
 	Cvar_RegisterVariable (&gl_finish);
 	Cvar_RegisterVariable (&gl_clear);
@@ -268,6 +333,10 @@ void R_Init (void)
 	Cvar_RegisterVariable (&r_showbboxes);
 	Cvar_RegisterVariable (&r_showbboxes_think);
 	Cvar_RegisterVariable (&r_showbboxes_health);
+	Cvar_RegisterVariable (&r_showbboxes_links);
+	Cvar_RegisterVariable (&r_showbboxes_targets);
+	Cvar_RegisterVariable (&r_showfields);
+	Cvar_RegisterVariable (&r_showfields_align);
 	Cvar_RegisterVariable (&gl_farclip);
 	Cvar_RegisterVariable (&gl_fullbrights);
 	Cvar_SetCallback (&gl_fullbrights, GL_Fullbrights_f);
@@ -310,8 +379,10 @@ void R_TranslatePlayerSkin (int playernum)
 
 	//FIXME: if gl_nocolors is on, then turned off, the textures may be out of sync with the scoreboard colors.
 	if (!gl_nocolors.value)
+	{
 		if (playertextures[playernum])
 			TexMgr_ReloadImage (playertextures[playernum], top, bottom);
+	}
 }
 
 /*
@@ -394,6 +465,7 @@ static void R_ParseWorldspawn (void)
 		return; // error
 	if (com_token[0] != '{')
 		return; // error
+
 	while (1)
 	{
 		data = COM_Parse(data);
@@ -407,7 +479,7 @@ static void R_ParseWorldspawn (void)
 			q_strlcpy(key, com_token, sizeof(key));
 		while (key[0] && key[strlen(key)-1] == ' ') // remove trailing spaces
 			key[strlen(key)-1] = 0;
-		data = COM_Parse(data);
+		data = COM_ParseEx(data, CPE_ALLOWTRUNC);
 		if (!data)
 			return; // error
 		q_strlcpy(value, com_token, sizeof(value));
@@ -454,6 +526,11 @@ void R_NewMap (void)
 	Sky_NewMap (); //johnfitz -- skybox in worldspawn
 	Fog_NewMap (); //johnfitz -- global fog in worldspawn
 	R_ParseWorldspawn (); //ericw -- wateralpha, lavaalpha, telealpha, slimealpha in worldspawn
+
+	// Load pointfile if map has no vis data and either developer mode is on or the game was started from a map editing tool
+	if (developer.value || map_checks.value)
+		if (!cl.worldmodel->visdata && COM_FileExists (va ("maps/%s.pts", cl.mapname), NULL))
+			Cbuf_AddText ("pointfile\n");
 }
 
 /*
@@ -489,10 +566,6 @@ void R_TimeRefresh_f (void)
 	Con_Printf ("%lf seconds (%.1lf fps)\n", time, 128/time);
 }
 
-void D_FlushCaches (void)
-{
-}
-
 static GLuint current_array_buffer;
 static GLuint current_element_array_buffer;
 static GLuint current_shader_storage_buffer;
@@ -652,7 +725,7 @@ This must be called if you do anything that could make the cached bindings
 invalid (e.g. manually binding, destroying the context).
 ====================
 */
-void GL_ClearBufferBindings ()
+void GL_ClearBufferBindings (void)
 {
 	int i;
 
