@@ -24,6 +24,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 // screen.c -- master for refresh, status bar, console, chat, notify, etc
 
 #include "quakedef.h"
+#include "steam.h"
 #include <time.h>
 
 /*
@@ -81,6 +82,8 @@ float		scr_conlines;		// lines of console to display
 //johnfitz -- new cvars
 cvar_t		scr_menuscale = {"scr_menuscale", "1", CVAR_ARCHIVE};
 cvar_t		scr_menubgalpha = {"scr_menubgalpha", "0.7", CVAR_ARCHIVE};
+cvar_t		scr_menubgstyle = {"scr_menubgstyle", "-1", CVAR_ARCHIVE};
+cvar_t		scr_centerprintbg = {"scr_centerprintbg", "0", CVAR_ARCHIVE}; // 0 = off; 1 = text box; 2 = menu box; 3 = menu strip
 cvar_t		scr_sbarscale = {"scr_sbarscale", "1", CVAR_ARCHIVE};
 cvar_t		scr_sbaralpha = {"scr_sbaralpha", "0.75", CVAR_ARCHIVE};
 cvar_t		scr_conwidth = {"scr_conwidth", "0", CVAR_ARCHIVE};
@@ -88,12 +91,14 @@ cvar_t		scr_conscale = {"scr_conscale", "1", CVAR_ARCHIVE};
 cvar_t		scr_crosshairscale = {"scr_crosshairscale", "1", CVAR_ARCHIVE};
 cvar_t		scr_pixelaspect = {"scr_pixelaspect", "1", CVAR_ARCHIVE};
 cvar_t		scr_showfps = {"scr_showfps", "0", CVAR_ARCHIVE};
+cvar_t		scr_showspeed = {"scr_showspeed", "0", CVAR_ARCHIVE};
 cvar_t		scr_clock = {"scr_clock", "0", CVAR_ARCHIVE};
 //johnfitz
 cvar_t		scr_usekfont = {"scr_usekfont", "0", CVAR_NONE}; // 2021 re-release
 
 cvar_t		scr_hudstyle = {"hudstyle", "2", CVAR_ARCHIVE};
 cvar_t		cl_screenshotname = {"cl_screenshotname", "screenshots/%map%_%date%_%time%", CVAR_ARCHIVE};
+cvar_t		scr_demobar_timeout = {"scr_demobar_timeout", "1", CVAR_ARCHIVE};
 
 cvar_t		scr_viewsize = {"viewsize","100", CVAR_ARCHIVE};
 cvar_t		scr_fov = {"fov","90",CVAR_ARCHIVE};	// 10 - 170
@@ -108,12 +113,19 @@ cvar_t		scr_printspeed = {"scr_printspeed","8",CVAR_NONE};
 cvar_t		gl_triplebuffer = {"gl_triplebuffer", "1", CVAR_ARCHIVE};
 
 cvar_t		cl_gun_fovscale = {"cl_gun_fovscale","1",CVAR_ARCHIVE}; // Qrack
+cvar_t		cl_gun_x = {"cl_gun_x","0",CVAR_ARCHIVE};
+cvar_t		cl_gun_y = {"cl_gun_y","0",CVAR_ARCHIVE};
+cvar_t		cl_gun_z = {"cl_gun_z","0",CVAR_ARCHIVE};
 
 extern	char	crosshair_char;
 extern	cvar_t	crosshair;
 extern	cvar_t	con_notifyfade;
 extern	cvar_t	con_notifyfadetime;
 
+extern	edict_t	**bbox_linked;
+extern	cvar_t	r_showfields;
+extern	cvar_t	r_showfields_align;
+
 qboolean	scr_initialized;		// ready to draw
 
 qpic_t		*scr_net;
@@ -130,6 +142,8 @@ float		scr_disabled_time;
 
 int	scr_tileclear_updates = 0; //johnfitz
 
+hudstyle_t	hudstyle;
+
 void SCR_ScreenShot_f (void);
 
 /*
@@ -144,6 +158,7 @@ char		scr_centerstring[1024];
 float		scr_centertime_start;	// for slow victory printing
 float		scr_centertime_off;
 int			scr_center_lines;
+int			scr_center_maxcols;
 int			scr_erase_lines;
 int			scr_erase_center;
 
@@ -157,7 +172,16 @@ for a few moments
 */
 void SCR_CenterPrint (const char *str) //update centerprint data
 {
-	strncpy (scr_centerstring, str, sizeof(scr_centerstring)-1);
+	int cols;
+
+	q_strlcpy (scr_centerstring, str, sizeof (scr_centerstring));
+	if (!scr_centerstring[0])
+	{
+		scr_center_lines = 0;
+		scr_center_maxcols = 0;
+		return;
+	}
+
 	scr_centertime_off = scr_centertime.value;
 	scr_centertime_start = cl.time;
 	if (!cl.intermission)
@@ -165,12 +189,72 @@ void SCR_CenterPrint (const char *str) //update centerprint data
 
 // count the number of lines for centering
 	scr_center_lines = 1;
+	scr_center_maxcols = 0;
 	str = scr_centerstring;
+	cols = 0;
 	while (*str)
 	{
 		if (*str == '\n')
+		{
 			scr_center_lines++;
+			scr_center_maxcols = q_max (scr_center_maxcols, cols);
+			cols = -1; // compensate the following ++
+		}
 		str++;
+		cols++;
+	}
+	scr_center_maxcols = q_max (scr_center_maxcols, cols);
+}
+
+static void SCR_DrawCenterStringBG (int y, float alpha)
+{
+	const char *str;
+	int i, len, lines, x;
+
+	if (cl.intermission || q_min (scr_center_lines, scr_center_maxcols) <= 0 || alpha <= 0.f)
+		return;
+
+	// skip leading empty lines (might be there just to reposition the text)
+	str = scr_centerstring;
+	while (*str == '\n')
+	{
+		str++;
+		y += CHARSIZE;
+	}
+
+	// skip trailing empty lines
+	len = (int) strlen (str);
+	while (len > 0 && str[len - 1] == '\n')
+		--len;
+
+	// count remaining lines
+	for (i = 0, lines = 1; i < len; i++)
+		if (str[i] == '\n')
+			lines++;
+
+	// draw the background
+	switch ((int)scr_centerprintbg.value)
+	{
+	case 1:
+		len = (scr_center_maxcols + 3) & ~1;
+		x = (320 - len * 8) / 2;
+		GL_PushCanvasColor (1.f, 1.f, 1.f, alpha * scr_menubgalpha.value);
+		M_DrawTextBox (x - 8, y - 12, len, lines + 1);
+		GL_PopCanvasColor ();
+		break;
+
+	case 2:
+		len = scr_center_maxcols + 2;
+		x = (320 - len* 8) / 2;
+		Draw_PartialFadeScreen (x, x + len * 8, y - 4, y + lines * 8 + 4, alpha);
+		break;
+
+	case 3:
+		Draw_PartialFadeScreen (glcanvas.left, glcanvas.right, y - 4, y + lines * 8 + 4, alpha);
+		break;
+
+	default:
+		return;
 	}
 }
 
@@ -181,7 +265,7 @@ void SCR_DrawCenterString (void) //actually do the drawing
 	int		j;
 	int		x, y;
 	int		remaining;
-	float	alpha;
+	float	alpha, forcedalpha;
 
 	GL_SetCanvas (CANVAS_MENU); //johnfitz
 
@@ -198,7 +282,10 @@ void SCR_DrawCenterString (void) //actually do the drawing
 		alpha = fade ? q_min (scr_centertime_off / fade, 1.f) : 1.f;
 	}
 
-	GL_SetCanvasColor (1.f, 1.f, 1.f, alpha);
+	if (M_ForcedCenterPrint (&forcedalpha))
+		alpha *= forcedalpha * forcedalpha;
+
+	GL_PushCanvasColor (1.f, 1.f, 1.f, alpha);
 
 	scr_erase_center = 0;
 	start = scr_centerstring;
@@ -210,6 +297,8 @@ void SCR_DrawCenterString (void) //actually do the drawing
 	if (crosshair.value && scr_viewsize.value < 130)
 		y -= 8;
 
+	SCR_DrawCenterStringBG (y, alpha);
+
 	do
 	{
 	// scan the width of the line
@@ -232,7 +321,7 @@ void SCR_DrawCenterString (void) //actually do the drawing
 		start++;		// skip the \n
 	} while (1);
 
-	GL_SetCanvasColor (1.f, 1.f, 1.f, 1.f);
+	GL_PopCanvasColor ();
 }
 
 void SCR_CheckDrawCenterString (void)
@@ -244,10 +333,13 @@ void SCR_CheckDrawCenterString (void)
 
 	if (scr_centertime_off <= 0 && !cl.intermission)
 		return;
-	if (key_dest != key_game)
-		return;
-	if (cl.paused) //johnfitz -- don't show centerprint during a pause
-		return;
+	if (!M_ForcedCenterPrint (NULL))
+	{
+		if (key_dest != key_game)
+			return;
+		if (cl.paused) //johnfitz -- don't show centerprint during a pause
+			return;
+	}
 
 	SCR_DrawCenterString ();
 }
@@ -299,7 +391,8 @@ SCR_UpdateZoom
 */
 void SCR_UpdateZoom (void)
 {
-	float delta = cl.zoomdir * scr_zoomspeed.value * (cl.time - cl.oldtime);
+	float speed = scr_zoomspeed.value > 0.f ? scr_zoomspeed.value : 1e6;
+	float delta = cl.zoomdir * speed * (cl.time - cl.oldtime);
 	if (!delta)
 		return;
 	cl.zoom += delta;
@@ -398,7 +491,7 @@ static void SCR_CalcRefdef (void)
 	scale = CLAMP (1.0f, scr_sbarscale.value, (float)glwidth / 320.0f);
 	scale *= (float) vid.height / vid.guiheight;
 
-	if (size >= 120 || cl.intermission || scr_sbaralpha.value < 1 || scr_hudstyle.value >= 1 || cl.qcvm.extfuncs.CSQC_DrawHud) //johnfitz -- scr_sbaralpha.value
+	if (size >= 120 || cl.intermission || scr_sbaralpha.value < 1 || hudstyle != HUD_CLASSIC || cl.qcvm.extfuncs.CSQC_DrawHud) //johnfitz -- scr_sbaralpha.value
 		sb_lines = 0;
 	else if (size >= 110)
 		sb_lines = 24 * scale;
@@ -494,6 +587,20 @@ void SCR_PixelAspect_f (cvar_t *cvar)
 	VID_RecalcInterfaceSize ();
 }
 
+/*
+==================
+SCR_HUDStyle_f
+
+Updates hudstyle variable and invalidates refdef when scr_hudstyle changes
+==================
+*/
+void SCR_HUDStyle_f (cvar_t *cvar)
+{
+	int val = (int) cvar->value;
+	hudstyle = (hudstyle_t) CLAMP (0, val, (int) HUD_COUNT - 1);
+	vid.recalc_refdef = 1;
+}
+
 //============================================================================
 
 /*
@@ -517,6 +624,8 @@ void SCR_Init (void)
 	//johnfitz -- new cvars
 	Cvar_RegisterVariable (&scr_menuscale);
 	Cvar_RegisterVariable (&scr_menubgalpha);
+	Cvar_RegisterVariable (&scr_menubgstyle);
+	Cvar_RegisterVariable (&scr_centerprintbg);
 	Cvar_RegisterVariable (&scr_sbarscale);
 	Cvar_SetCallback (&scr_sbaralpha, SCR_Callback_refdef);
 	Cvar_RegisterVariable (&scr_sbaralpha);
@@ -526,16 +635,18 @@ void SCR_Init (void)
 	Cvar_RegisterVariable (&scr_conscale);
 	Cvar_RegisterVariable (&scr_crosshairscale);
 	Cvar_RegisterVariable (&scr_showfps);
+	Cvar_RegisterVariable (&scr_showspeed);
 	Cvar_RegisterVariable (&scr_clock);
-	Cvar_RegisterVariable (&scr_hudstyle);
 	Cvar_RegisterVariable (&cl_screenshotname);
+	Cvar_RegisterVariable (&scr_demobar_timeout);
 	//johnfitz
 	Cvar_RegisterVariable (&scr_usekfont); // 2021 re-release
 	Cvar_SetCallback (&scr_fov, SCR_Callback_refdef);
 	Cvar_SetCallback (&scr_fov_adapt, SCR_Callback_refdef);
 	Cvar_SetCallback (&scr_zoomfov, SCR_Callback_refdef);
 	Cvar_SetCallback (&scr_viewsize, SCR_Callback_refdef);
-	Cvar_SetCallback (&scr_hudstyle, SCR_Callback_refdef);
+	Cvar_SetCallback (&scr_hudstyle, SCR_HUDStyle_f);
+	Cvar_RegisterVariable (&scr_hudstyle);
 	Cvar_RegisterVariable (&scr_fov);
 	Cvar_RegisterVariable (&scr_fov_adapt);
 	Cvar_RegisterVariable (&scr_zoomfov);
@@ -548,6 +659,9 @@ void SCR_Init (void)
 	Cvar_RegisterVariable (&scr_printspeed);
 	Cvar_RegisterVariable (&gl_triplebuffer);
 	Cvar_RegisterVariable (&cl_gun_fovscale);
+	Cvar_RegisterVariable (&cl_gun_x);
+	Cvar_RegisterVariable (&cl_gun_y);
+	Cvar_RegisterVariable (&cl_gun_z);
 
 	Cmd_AddCommand ("scr_autoscale",SCR_AutoScale_f);
 
@@ -613,7 +727,7 @@ void SCR_DrawFPS (void)
 		else
 			sprintf (st, "%.2f ms", 1000.f / lastfps);
 		x = 320 - (strlen(st)<<3);
-		if (scr_hudstyle.value >= 1)
+		if (hudstyle != HUD_CLASSIC)
 		{
 			x = 320 - 16 - (strlen(st)<<3);
 			y = 8;
@@ -632,6 +746,55 @@ void SCR_DrawFPS (void)
 	}
 }
 
+/*
+==============
+SCR_DrawSpeed
+==============
+*/
+void SCR_DrawSpeed (void)
+{
+    if (cl.intermission || CL_InCutscene () || scr_viewsize.value >= 130)
+		return;
+
+	const float show_speed_interval_value = 0.05f;
+	static float maxspeed = 0, display_speed = -1;
+	static double lastrealtime = 0;
+	float speed;
+	vec3_t vel;
+
+	if (lastrealtime > realtime)
+	{
+		lastrealtime = 0;
+		display_speed = -1;
+		maxspeed = 0;
+	}
+
+	VectorCopy (cl.velocity, vel);
+	vel[2] = 0;
+	speed = VectorLength (vel);
+
+	if (speed > maxspeed)
+		maxspeed = speed;
+
+	if (scr_showspeed.value)
+	{
+		if (display_speed >= 0)
+		{
+			char str[12];
+			sprintf (str, "%d", (int) display_speed);
+			GL_SetCanvas (CANVAS_CROSSHAIR);
+			Draw_String (-(int)strlen(str)*4, 4, str);
+		}
+	}
+
+	if (realtime - lastrealtime >= show_speed_interval_value)
+	{
+		lastrealtime = realtime;
+		display_speed = maxspeed;
+		maxspeed = 0;
+	}
+}
+
 /*
 ==============
 SCR_DrawClock -- johnfitz
@@ -654,7 +817,7 @@ void SCR_DrawClock (void)
 		return;
 
 	//draw it
-	if (scr_hudstyle.value < 1)
+	if (hudstyle == HUD_CLASSIC)
 	{
 		GL_SetCanvas (CANVAS_BOTTOMRIGHT);
 		Draw_String (320 - (strlen(str)<<3), 200 - 8, str);
@@ -668,6 +831,141 @@ void SCR_DrawClock (void)
 	scr_tileclear_updates = 0;
 }
 
+/*
+==============
+SCR_PrintMirrored
+==============
+*/
+static void SCR_PrintMirrored (int x, int y, const char *str)
+{
+	x += strlen (str) * 8;
+	while (*str)
+	{
+		Draw_CharacterEx (x, y, -8, 8, 0x80 ^ (*str++));
+		x -= 8;
+	}
+}
+
+/*
+==============
+SCR_DrawDemoControls
+==============
+*/
+void SCR_DrawDemoControls (void)
+{
+	static const int	TIMEBAR_CHARS = 38;
+	static float		prevspeed = 1.0f;
+	static float		prevbasespeed = 1.0f;
+	static float		showtime = 1.0f;
+	int					i, len, x, y, min, sec;
+	float				frac;
+	const char			*str;
+	char				name[31]; // size chosen to avoid overlap with side text
+
+	if (!cls.demoplayback || scr_demobar_timeout.value < 0.f)
+	{
+		showtime = 0.f;
+		return;
+	}
+
+	// Determine for how long the demo playback info should be displayed
+	if (cls.demospeed != prevspeed || cls.basedemospeed != prevbasespeed ||			// speed/base speed changed
+		fabs (cls.demospeed) > cls.basedemospeed ||									// fast forward/rewind
+		!scr_demobar_timeout.value)													// controls always shown
+	{
+		prevspeed = cls.demospeed;
+		prevbasespeed = cls.basedemospeed;
+		showtime = scr_demobar_timeout.value > 0.f ? scr_demobar_timeout.value : 1.f;
+	}
+	else
+	{
+		showtime -= host_rawframetime;
+		if (showtime < 0.f)
+		{
+			showtime = 0.f;
+			return;
+		}
+	}
+
+	// Approximate the fraction of the demo that's already been played back
+	// based on the current file offset and total demo size
+	// Note: we need to take into account the starting offset for pak files
+	frac = (Sys_ftell (cls.demofile) - cls.demofilestart) / (double)cls.demofilesize;
+	frac = CLAMP (0.f, frac, 1.f);
+
+	if (cl.intermission)
+	{
+		GL_SetCanvas (CANVAS_MENU);
+		y = LERP (glcanvas.bottom, glcanvas.top, 0.125f) + 8;
+	}
+	else
+	{
+		GL_SetCanvas (CANVAS_SBAR);
+		y = glcanvas.bottom - 68;
+	}
+	x = (glcanvas.left + glcanvas.right) / 2 - TIMEBAR_CHARS / 2 * 8;
+
+	// Draw status box background
+	GL_SetCanvasColor (1.f, 1.f, 1.f, scr_sbaralpha.value);
+	M_DrawTextBox (x - 8, y - 8, TIMEBAR_CHARS, 1);
+	GL_SetCanvasColor (1.f, 1.f, 1.f, 1.f);
+
+	// Print playback status on the left (paused/playing/fast-forward/rewind)
+	// Note: character #13 works well as a forward symbol, but Alkaline 1.2 changes it to a disk.
+	// If we have a custom conchars texture we switch to a safer alternative, the '>' character.
+	if (!cls.demospeed)
+		str = "II";
+	else if (fabs (cls.demospeed) > 1.f)
+		str = custom_conchars ? ">>" : "\xD\xD";
+	else 
+		str = custom_conchars ? ">" : "\xD";
+	if (cls.demospeed >= 0.f)
+		M_Print (x, y, str);
+	else
+		SCR_PrintMirrored (x, y, str);
+
+	// Print base playback speed on the right
+	if (!cls.basedemospeed)
+		str = "";
+	else if (fabs (cls.basedemospeed) >= 1.f)
+		str = va ("%gx", fabs (cls.basedemospeed));
+	else
+		str = va ("1/%gx", 1.f / fabs (cls.basedemospeed));
+	M_Print (x + (TIMEBAR_CHARS - strlen (str)) * 8, y, str);
+
+	// Print demo name in the center
+	COM_StripExtension (COM_SkipPath (cls.demofilename), name, sizeof (name));
+	x = (glcanvas.left + glcanvas.right) / 2;
+	M_Print (x - strlen (name) * 8 / 2, y, name);
+
+	// Draw seek bar rail
+	x = (glcanvas.left + glcanvas.right) / 2 - TIMEBAR_CHARS / 2 * 8;
+	y -= 8;
+	Draw_Character (x - 8, y, 128);
+	for (i = 0; i < TIMEBAR_CHARS; i++)
+		Draw_Character (x + i * 8, y, 129);
+	Draw_Character (x + i * 8, y, 130);
+
+	// Draw seek bar cursor
+	x += (TIMEBAR_CHARS - 1) * 8 * frac;
+	Draw_Character (x, y, 131);
+
+	// Print current time above the cursor
+	y -= 11;
+	sec = (int) cl.time;
+	min = sec / 60;
+	sec %= 60;
+	str = va ("%i:%02i", min, sec);
+	x -= (strchr (str, ':') - str) * 8; // align ':' with cursor
+	len = strlen (str);
+	// M_DrawTextBox effectively rounds width up to a multiple of 2,
+	// so if our length is odd we pad by half a character on each side
+	GL_SetCanvasColor (1.f, 1.f, 1.f, scr_sbaralpha.value);
+	M_DrawTextBox (x - 8 - (len & 1) * 8 / 2, y - 8, len + (len & 1), 1);
+	GL_SetCanvasColor (1.f, 1.f, 1.f, 1.f);
+	Draw_String (x, y, str);
+}
+
 /*
 ==============
 SCR_DrawDevStats
@@ -769,6 +1067,7 @@ DrawPause
 void SCR_DrawPause (void)
 {
 	qpic_t	*pic;
+	float	alpha;
 
 	if (!cl.paused)
 		return;
@@ -776,11 +1075,23 @@ void SCR_DrawPause (void)
 	if (!scr_showpause.value)		// turn off for screenshots
 		return;
 
+	if (M_ForcedCenterPrint (&alpha))
+	{
+		alpha = 1.f - alpha;
+		if (alpha <= 0.f)
+			return;
+	}
+	else
+		alpha = 1.f;
+
 	GL_SetCanvas (CANVAS_MENU); //johnfitz
+	GL_PushCanvasColor (1.f, 1.f, 1.f, alpha);
 
 	pic = Draw_CachePic ("gfx/pause.lmp");
 	Draw_Pic ( (320 - pic->width)/2, (240 - 48 - pic->height)/2, pic); //johnfitz -- stretched menus
 
+	GL_PopCanvasColor ();
+
 	scr_tileclear_updates = 0; //johnfitz
 }
 
@@ -812,7 +1123,6 @@ SCR_DrawSaving
 void SCR_DrawSaving (void)
 {
 	int x, y;
-	const char *str;
 
 	if (!Host_IsSaving ())
 		return;
@@ -821,7 +1131,7 @@ void SCR_DrawSaving (void)
 
 	x = 320 - 16 - draw_disc->width;
 	y = 8;
-	if (scr_hudstyle.value >= 1 && scr_viewsize.value < 130)
+	if (hudstyle != HUD_CLASSIC && scr_viewsize.value < 130)
 	{
 		if (scr_clock.value) y += 8;
 		if (scr_showfps.value) y += 8;
@@ -830,12 +1140,6 @@ void SCR_DrawSaving (void)
 	}
 
 	Draw_Pic (x, y, draw_disc);
-	y += draw_disc->height/2 - 8;
-
-	str = "Saving";
-	Draw_String (x - 4 - strlen (str)*8, y, str);
-	str = sv.lastsave;
-	Draw_String (x - 4 - strlen (str)*8, y+8, str);
 }
 
 /*
@@ -845,13 +1149,318 @@ SCR_DrawCrosshair -- johnfitz
 */
 void SCR_DrawCrosshair (void)
 {
-	if (!crosshair.value || scr_viewsize.value >= 130)
+	if (cl.intermission || CL_InCutscene () || !crosshair.value || scr_viewsize.value >= 130)
 		return;
 
 	GL_SetCanvas (CANVAS_CROSSHAIR);
 	Draw_Character(-4, -4, crosshair_char);
 }
 
+typedef struct
+{
+	float	scale[2];
+	float	offset[2];
+} canvasmap_t;
+
+/*
+==============
+SCR_SetupProjToCanvasMap
+
+Computes mapping from NDC coordinates (-1..1) to canvas coordinates
+Note: canvas Y increases from top to bottom, GL Y is bottom to top
+==============
+*/
+static void SCR_SetupProjToCanvasMap (canvasmap_t *remap)
+{
+	float	xmin, ymin, width, height;
+
+	width = r_refdef.vrect.width / (float)glwidth * (glcanvas.right - glcanvas.left);
+	height = r_refdef.vrect.height / (float)glheight * (glcanvas.top - glcanvas.bottom);
+	xmin = glcanvas.left + (r_refdef.vrect.x / (float)glwidth) * (glcanvas.right - glcanvas.left);
+	ymin = glcanvas.bottom - ((glheight - r_refdef.vrect.y - r_refdef.vrect.height) / (float)glheight) * (glcanvas.bottom - glcanvas.top);
+
+	remap->scale[0]  = width * 0.5f;
+	remap->offset[0] = width * 0.5f + xmin;
+	remap->scale[1]  = height * 0.5f;
+	remap->offset[1] = height * 0.5f + ymin;
+}
+
+/*
+==============
+SCR_ProjToCanvas
+==============
+*/
+static void SCR_ProjToCanvas (const vec3_t proj, const canvasmap_t *remap, float *outx, float *outy)
+{
+	*outx = proj[0] * remap->scale[0] + remap->offset[0];
+	*outy = proj[1] * remap->scale[1] + remap->offset[1];
+}
+
+/*
+==============
+SCR_GetEntityBottom
+==============
+*/
+static void SCR_GetEntityBottom (const edict_t *ed, vec3_t pos)
+{
+	pos[0] = ed->v.origin[0] + (ed->v.mins[0] + ed->v.maxs[0]) * 0.5f;
+	pos[1] = ed->v.origin[1] + (ed->v.mins[1] + ed->v.maxs[1]) * 0.5f;
+	pos[2] = ed->v.origin[2] + ed->v.mins[2];
+
+	// If it's a point entity, move anchor point down by 8 units to avoid overlapping debug visualization
+	if (VectorCompare (ed->v.mins, ed->v.maxs))
+		pos[2] -= 8.f;
+}
+
+/*
+==============
+SCR_GetEntityCenter
+==============
+*/
+static void SCR_GetEntityCenter (const edict_t *ed, vec3_t pos)
+{
+	if (!VectorCompare (ed->v.mins, ed->v.maxs))
+	{
+		pos[0] = ed->v.origin[0] + (ed->v.mins[0] + ed->v.maxs[0]) * 0.5f;
+		pos[1] = ed->v.origin[1] + (ed->v.mins[1] + ed->v.maxs[1]) * 0.5f;
+		pos[2] = ed->v.origin[2] + (ed->v.mins[2] + ed->v.maxs[2]) * 0.5f;
+	}
+	else
+	{
+		VectorCopy (ed->v.origin, pos);
+	}
+}
+
+/*
+==============
+SCR_ClipToFrustum
+
+Clips the position to the view frustum towards a reference point.
+==============
+*/
+static void SCR_ClipToFrustum (vec3_t pos, const vec3_t ref)
+{
+	int i;
+
+	for (i = 0; i < 4; i++)
+	{
+		mplane_t *plane = &frustum[i];
+		float dist = DotProduct (plane->normal, pos) - plane->dist;
+		if (dist < 0.f)
+		{
+			float ref_dist = DotProduct (plane->normal, ref) - plane->dist;
+			if (ref_dist != dist)
+				VectorLerp (pos, ref, -dist / (ref_dist - dist), pos);
+		}
+	}
+}
+
+/*
+==============
+SCR_DrawKeyValueOverlay
+==============
+*/
+static void SCR_DrawKeyValueOverlay (float x, float y, const char *multistr, const vec3_t bgcolor)
+{
+	int			ofs, numlines, keylen, vallen, maxkey, maxval;
+	const char	*key, *val;
+
+	// Compute column widths and total number of lines
+	for (ofs = numlines = maxkey = maxval = 0; ofs < (int) VEC_SIZE (multistr); numlines++)
+	{
+		key = multistr + ofs;
+		keylen = (int) strlen (key);
+		ofs += keylen + 1;
+
+		val = multistr + ofs;
+		vallen = (int) strlen (val);
+		ofs += vallen + 1;
+
+		if (keylen == 0)
+		{
+			// empty key, value is printed centered
+			keylen = vallen * CHARSIZE / 2;
+			maxkey = q_max (maxkey, keylen);
+			maxval = q_max (maxval, keylen);
+		}
+		else
+		{
+			// key/value pair, we leave half a character on each side for spacing
+			maxkey = q_max (maxkey, keylen * CHARSIZE + CHARSIZE / 2);
+			maxval = q_max (maxval, vallen * CHARSIZE + CHARSIZE / 2);
+		}
+	}
+
+	// Try to keep as much text as possible on screen
+	if (y > glcanvas.bottom - numlines * CHARSIZE)
+		y = glcanvas.bottom - numlines * CHARSIZE;
+	if (y < glcanvas.top)
+		y = glcanvas.top;
+	if (x > glcanvas.right - maxval)
+		x = glcanvas.right - maxval;
+	if (x < glcanvas.left + maxkey)
+		x = glcanvas.left + maxkey;
+
+	// Draw background box
+	Draw_FillEx (x - maxkey, y, maxkey + maxval, numlines * CHARSIZE, bgcolor, 0.75f);
+
+	// Print each line
+	for (ofs = 0; ofs < (int) VEC_SIZE (multistr); y += CHARSIZE)
+	{
+		key = multistr + ofs;
+		keylen = (int) strlen (key);
+		ofs += keylen + 1;
+
+		val = multistr + ofs;
+		vallen = (int) strlen (val);
+		ofs += vallen + 1;
+
+		if (keylen == 0)
+		{
+			// empty key, center-print the value
+			Draw_StringEx (x - vallen * CHARSIZE / 2, y, CHARSIZE, val);
+		}
+		else
+		{
+			// key/value pair, leave half a character on each side for spacing
+			Draw_StringEx (x - CHARSIZE / 2 - keylen * CHARSIZE, y, CHARSIZE, key);
+			Draw_StringEx (x + CHARSIZE / 2, y, CHARSIZE, val);
+		}
+	}
+}
+
+/*
+==============
+SCR_DrawEdictInfo
+
+Show info for the highlighted entity with r_showfields/r_showbboxes
+==============
+*/
+
+static char *scr_edictoverlaystrings = NULL;
+
+void SCR_DrawEdictInfo (void)
+{
+	char		tinted[1024];
+	int			i;
+	float		x, y;
+	canvasmap_t	proj2canvas;
+	vec3_t		crosshair, focus, anchor, proj, bgcolor;
+	edict_t		*ed;
+
+	if (VEC_SIZE (bbox_linked) == 0)
+		return;
+
+	GL_SetCanvas (CANVAS_BOTTOMRIGHT);
+	SCR_SetupProjToCanvasMap (&proj2canvas);
+
+	PR_SwitchQCVM (&sv.qcvm);
+
+	VectorMA (r_origin, 8.f, vpn, crosshair);
+	SCR_GetEntityCenter (bbox_linked[0], focus);
+	SCR_ClipToFrustum (focus, crosshair);
+
+	// Show edict numbers and classnames for all highlighted entities.
+	// Note: reversed order, so that the focused entity is drawn last,
+	// on top of all the others.
+	for (i = (int) VEC_SIZE (bbox_linked) - 1; i >= 0; i--)
+	{
+		ed = bbox_linked[i];
+
+		// Compute anchor point
+		if (i == 0)
+		{
+			// With r_showfields_align 0 the field overlay tracks the focused entity,
+			// so we disable the simple one (number + classname) to avoid overlap.
+			if (!r_showfields_align.value)
+				continue;
+			//SCR_GetEntityBottom (ed, anchor);
+			SCR_GetEntityCenter (ed, anchor);
+			SCR_ClipToFrustum (anchor, crosshair);
+		}
+		else
+		{
+			SCR_GetEntityCenter (ed, anchor);
+			SCR_ClipToFrustum (anchor, focus);
+		}
+		ProjectVector (anchor, r_matviewproj, proj);
+		SCR_ProjToCanvas (proj, &proj2canvas, &x, &y);
+
+		// Simple overlay: centered edict number and classname (if not empty)
+		VEC_CLEAR (scr_edictoverlaystrings);
+		MultiString_Append (&scr_edictoverlaystrings, "");
+		MultiString_Append (&scr_edictoverlaystrings, va ("edict %d", NUM_FOR_EDICT (ed)));
+		if (ed->v.classname)
+		{
+			MultiString_Append (&scr_edictoverlaystrings, "");
+			MultiString_Append (&scr_edictoverlaystrings, PR_GetString (ed->v.classname));
+		}
+
+		// Set background color based on link type
+		switch (ed->showbboxflags)
+		{
+		default:
+		case SHOWBBOX_LINK_NONE:		VectorSet (bgcolor, 0.f,	0.f,	0.f); break;
+		case SHOWBBOX_LINK_INCOMING:	VectorSet (bgcolor, 0.25f,	0.125f,	0.125f); break;
+		case SHOWBBOX_LINK_OUTGOING:	VectorSet (bgcolor, 0.125f,	0.125f,	0.25f); break;
+		case SHOWBBOX_LINK_BOTH:		VectorSet (bgcolor, 0.25f,	0.125f,	0.25f); break;
+		}
+
+		SCR_DrawKeyValueOverlay (x, y, scr_edictoverlaystrings, bgcolor);
+	}
+
+	// Show all relevant entity fields for the highlighted entity (first in bbox_linked)
+	if (r_showfields.value)
+	{
+		ed = bbox_linked[0];
+
+		// Compute anchor point
+		SCR_GetEntityBottom (ed, anchor);
+		ProjectVector (anchor, r_matviewproj, proj);
+		SCR_ProjToCanvas (proj, &proj2canvas, &x, &y);
+
+		// r_showfields_align 1 locks the overlay to the bottom-right
+		if (r_showfields_align.value)
+		{
+			x = glcanvas.right;
+			y = glcanvas.bottom;
+		}
+
+		// Build overlay text: first 2 lines are entity number and classname
+		VEC_CLEAR (scr_edictoverlaystrings);
+		MultiString_Append (&scr_edictoverlaystrings, "Edict");
+		MultiString_Append (&scr_edictoverlaystrings, va ("%d", NUM_FOR_EDICT (ed)));
+		COM_TintString ("classname", tinted, sizeof (tinted));
+		MultiString_Append (&scr_edictoverlaystrings, tinted);
+		MultiString_Append (&scr_edictoverlaystrings, PR_GetString (ed->v.classname));
+
+		// Add all relevant fields, excluding classname (already added to the header)
+		for (i = 1; i < qcvm->progs->numfielddefs; i++)
+		{
+			stringview_t	line;
+			const char		*val;
+			ddef_t			*d = &qcvm->fielddefs[i];
+
+			if (d->ofs*4 == offsetof (entvars_t, classname) || !ED_IsRelevantField (ed, d))
+				continue;
+
+			COM_TintString (PR_GetString (d->s_name), tinted, sizeof (tinted));
+			val = ED_FieldValueString (ed, d);
+
+			while (COM_ParseLine (&val, &line))
+			{
+				MultiString_Append (&scr_edictoverlaystrings, tinted);
+				MultiString_AppendN (&scr_edictoverlaystrings, line.data, line.len);
+				tinted[0] = ' ';
+				tinted[1] = '\0';
+			}
+		}
+
+		SCR_DrawKeyValueOverlay (x, y, scr_edictoverlaystrings, rgb_black);
+	}
+
+	PR_SwitchQCVM (NULL);
+}
 
 
 //=============================================================================
@@ -882,13 +1491,17 @@ void SCR_SetUpToDrawConsole (void)
 		scr_conlines = glheight; //full screen //johnfitz -- glheight instead of vid.height
 		scr_con_current = scr_conlines;
 	}
-	else if (key_dest == key_console)
+	else if (key_dest == key_console || M_WantsConsole (NULL))
 		scr_conlines = glheight/2; //half screen //johnfitz -- glheight instead of vid.height
 	else
 		scr_conlines = 0; //none visible
 
 	timescale = (host_timescale.value > 0) ? host_timescale.value : 1; //johnfitz -- timescale
-	conspeed = (scr_conspeed.value > 0) ? scr_conspeed.value : 1e6f;
+	conspeed = (scr_conspeed.value > 0 && !cls.timedemo) ? scr_conspeed.value : 1e6f;
+
+	// make sure the console isn't too slow to come down if we're forcing it open to preview a specific option
+	if (key_dest == key_menu)
+		conspeed = q_max (conspeed, 2000.f);
 
 	if (scr_conlines < scr_con_current)
 	{
@@ -921,7 +1534,10 @@ void SCR_DrawConsole (void)
 {
 	if (scr_con_current)
 	{
-		Con_DrawConsole (scr_con_current, true);
+		if (key_dest != key_menu || (!con_forcedup || M_WantsConsole (NULL)))
+			Con_DrawConsole (scr_con_current, true, key_dest == key_console);
+		else
+			Draw_ConsoleBackground ();
 		clearconsole = 0;
 	}
 	else
@@ -1109,18 +1725,14 @@ static void SCR_ScreenShot_Usage (void)
 
 /*
 ==================
-SCR_ScreenShot_f -- johnfitz -- rewritten to use Image_WriteTGA
+SCR_ScreenShot_f
 ==================
 */
 void SCR_ScreenShot_f (void)
 {
 	byte	*buffer;
 	char	ext[4];
-	char	basename[MAX_OSPATH];
-	char	imagename[MAX_OSPATH];
-	char	checkname[MAX_OSPATH];
 	int		i, quality;
-	qboolean	ok, has_vars;
 
 	Q_strncpy (ext, "png", sizeof(ext));
 
@@ -1149,41 +1761,6 @@ void SCR_ScreenShot_f (void)
 		return;
 	}
 
-// find a file name to save it to
-	has_vars = SCR_ExpandVariables (cl_screenshotname.string, basename, sizeof (basename));
-	if (!basename[0])
-		q_strlcpy (basename, SCREENSHOT_PREFIX, sizeof (basename));
-
-	if (!has_vars)
-		goto append_index;
-
-	q_snprintf (imagename, sizeof (imagename), "%s.%s", basename, ext);
-	q_snprintf (checkname, sizeof (checkname), "%s/%s", com_gamedir, imagename);
-	if (Sys_FileType (checkname) != FS_ENT_NONE) // base name already used, try appending an index
-	{
-	append_index:
-		// append underscore if basename ends with a digit
-		i = (int) strlen (basename);
-		if (i && i + 1 < (int) countof (basename) && (unsigned int)(basename[i - 1] - '0') < 10u)
-		{
-			basename[i] = '_';
-			basename[i + 1] = '\0';
-		}
-
-		for (i = has_vars; i < 10000; i++)
-		{
-			q_snprintf (imagename, sizeof (imagename), "%s%04i.%s", basename, i, ext);
-			q_snprintf (checkname, sizeof (checkname), "%s/%s", com_gamedir, imagename);
-			if (Sys_FileType (checkname) == FS_ENT_NONE)
-				break;	// file doesn't exist
-		}
-		if (i == 10000)
-		{
-			Con_Printf ("SCR_ScreenShot_f: Couldn't find an unused filename\n");
-			return;
-		}
-	}
-
 //get data
 	if (!(buffer = (byte *) malloc(glwidth*glheight*3)))
 	{
@@ -1205,20 +1782,68 @@ void SCR_ScreenShot_f (void)
 	glReadPixels (glx, gly, glwidth, glheight, GL_RGB, GL_UNSIGNED_BYTE, buffer);
 
 // now write the file
-	if (!q_strncasecmp (ext, "png", sizeof(ext)))
-		ok = Image_WritePNG (imagename, buffer, glwidth, glheight, 24, false);
-	else if (!q_strncasecmp (ext, "tga", sizeof(ext)))
-		ok = Image_WriteTGA (imagename, buffer, glwidth, glheight, 24, false);
-	else if (!q_strncasecmp (ext, "jpg", sizeof(ext)))
-		ok = Image_WriteJPG (imagename, buffer, glwidth, glheight, 24, quality, false);
-	else
-		ok = false;
+	if (!Steam_SaveScreenshot (buffer, glwidth, glheight))
+	{
+		char		basename[MAX_OSPATH];
+		char		imagename[MAX_OSPATH];
+		char		checkname[MAX_OSPATH];
+		qboolean	ok, has_vars;
 
-	UTF8_ToQuake (basename, sizeof (basename), imagename);
-	if (ok)
-		Con_Printf ("Wrote %s\n", basename);
-	else
-		Con_Printf ("SCR_ScreenShot_f: Couldn't create %s\n", basename);
+	// find a file name to save it to
+		has_vars = SCR_ExpandVariables (cl_screenshotname.string, basename, sizeof (basename));
+		if (!basename[0])
+			q_strlcpy (basename, SCREENSHOT_PREFIX, sizeof (basename));
+
+		if (!has_vars)
+			goto append_index;
+
+		q_snprintf (imagename, sizeof (imagename), "%s.%s", basename, ext);
+		q_snprintf (checkname, sizeof (checkname), "%s/%s", com_gamedir, imagename);
+		if (Sys_FileType (checkname) != FS_ENT_NONE) // base name already used, try appending an index
+		{
+		append_index:
+			// append underscore if basename ends with a digit
+			i = (int) strlen (basename);
+			if (i && i + 1 < (int) countof (basename) && (unsigned int)(basename[i - 1] - '0') < 10u)
+			{
+				basename[i] = '_';
+				basename[i + 1] = '\0';
+			}
+
+			for (i = has_vars; i < 10000; i++)
+			{
+				q_snprintf (imagename, sizeof (imagename), "%s%04i.%s", basename, i, ext);
+				q_snprintf (checkname, sizeof (checkname), "%s/%s", com_gamedir, imagename);
+				if (Sys_FileType (checkname) == FS_ENT_NONE)
+					break;	// file doesn't exist
+			}
+			if (i == 10000)
+			{
+				Con_Printf ("SCR_ScreenShot_f: Couldn't find an unused filename\n");
+				free (buffer);
+				return;
+			}
+		}
+
+		if (!q_strncasecmp (ext, "png", sizeof(ext)))
+			ok = Image_WritePNG (imagename, buffer, glwidth, glheight, 24, false);
+		else if (!q_strncasecmp (ext, "tga", sizeof(ext)))
+			ok = Image_WriteTGA (imagename, buffer, glwidth, glheight, 24, false);
+		else if (!q_strncasecmp (ext, "jpg", sizeof(ext)))
+			ok = Image_WriteJPG (imagename, buffer, glwidth, glheight, 24, quality, false);
+		else
+			ok = false;
+
+		UTF8_ToQuake (basename, sizeof (basename), imagename);
+		if (ok)
+		{
+			Con_SafePrintf ("Wrote ");
+			Con_LinkPrintf (va("%s/%s", com_gamedir, imagename), "%s", basename);
+			Con_SafePrintf ("\n");
+		}
+		else
+			Con_Printf ("SCR_ScreenShot_f: Couldn't create %s\n", basename);
+	}
 
 	free (buffer);
 }
@@ -1246,7 +1871,9 @@ void SCR_BeginLoadingPlaque (void)
 	if (key_dest != key_console)
 	{
 		Con_ClearNotify ();
-		scr_con_current = 0;
+		// don't reset console when advancing to a new demo while previewing a console option
+		if (!M_WantsConsole (NULL))
+			scr_con_current = 0;
 		scr_drawloading = true;
 	}
 
@@ -1473,22 +2100,25 @@ void SCR_UpdateScreen (void)
 			Draw_ConsoleBackground ();
 		else
 			Sbar_Draw ();
-		Draw_FadeScreen ();
+		Draw_FadeScreen (1.f);
 		SCR_DrawNotifyString ();
 	}
 	else if (scr_drawloading) //loading
 	{
 		SCR_DrawLoading ();
 		Sbar_Draw ();
+		M_Draw ();
 	}
 	else if (cl.intermission == 1 && key_dest == key_game) //end of level
 	{
 		Sbar_IntermissionOverlay ();
+		SCR_DrawDemoControls ();
 	}
 	else if (cl.intermission == 2 && key_dest == key_game) //end of episode
 	{
 		Sbar_FinaleOverlay ();
 		SCR_CheckDrawCenterString ();
+		SCR_DrawDemoControls ();
 	}
 	else
 	{
@@ -1500,6 +2130,9 @@ void SCR_UpdateScreen (void)
 		Sbar_Draw ();
 		SCR_DrawDevStats (); //johnfitz
 		SCR_DrawClock (); //johnfitz
+		SCR_DrawDemoControls ();
+		SCR_DrawSpeed ();
+		SCR_DrawEdictInfo ();
 		SCR_DrawConsole ();
 		M_Draw ();
 		SCR_DrawFPS (); //johnfitz
