@@ -159,16 +159,22 @@ static const char warpscale_fragment_shader[] =
 ////////////////////////////////////////////////////////////////
 
 #define NOISE_FUNCTIONS \
-"// Interleaved gradient noise - Jorge Jimenez\n"\
-"// http://www.iryoku.com/next-generation-post-processing-in-call-of-duty-advanced-warfare \n"\
-"float ignoise01(vec2 p)\n"\
+"// ALU-only 16x16 Bayer matrix\n"\
+"float bayer01(ivec2 coord)\n"\
 "{\n"\
-"	return fract(52.9829189 * fract(dot(p, vec2(0.06711056, 0.00583715))));\n"\
+"	coord &= 15;\n"\
+"	coord.y ^= coord.x;\n"\
+"	uint v = uint(coord.y | (coord.x << 8));	// 0  0  0  0 | x3 x2 x1 x0 |  0  0  0  0 | y3 y2 y1 y0\n"\
+"	v = (v ^ (v << 2)) & 0x3333;				// 0  0 x3 x2 |  0  0 x1 x0 |  0  0 y3 y2 |  0  0 y1 y0\n"\
+"	v = (v ^ (v << 1)) & 0x5555;				// 0 x3  0 x2 |  0 x1  0 x0 |  0 y3  0 y2 |  0 y1  0 y0\n"\
+"	v |= v >> 7;								// 0 x3  0 x2 |  0 x1  0 x0 | x3 y3 x2 y2 | x1 y1 x0 y0\n"\
+"	v = bitfieldReverse(v) >> 24;				// 0  0  0  0 |  0  0  0  0 | y0 x0 y1 x1 | y2 x2 y3 x3\n"\
+"	return float(v) * (1.0/256.0);\n"\
 "}\n"\
 "\n"\
-"float ignoise(vec2 p)\n"\
+"float bayer(ivec2 coord)\n"\
 "{\n"\
-"	return ignoise01(p) - 0.5;\n"\
+"	return bayer01(coord) - 0.5;\n"\
 "}\n"\
 "\n"\
 "// Hash without Sine\n"\
@@ -216,13 +222,9 @@ SOFTWARE.*/\
 "	return x;\n"\
 "}\n"\
 "\n"\
-"#define DITHER_NOISE(uv) tri(ignoise01(uv))\n"\
+"#define DITHER_NOISE(uv) tri(bayer01(ivec2(uv)))\n"\
 "#define SCREEN_SPACE_NOISE() DITHER_NOISE(floor(gl_FragCoord.xy)+0.5)\n"\
-"#define SUPPRESS_BANDING() \\\n"\
-"	(gl_NumSamples > 1 ?\\\n"\
-"		ignoise(floor(gl_FragCoord.xy) + 0.5) * (1.5/255.) :\\\n"\
-"		ignoise(gl_FragCoord.xy) * (1.5/255.))\n"\
-"#define PAL_NOISESCALE (12./255.)\n"\
+"#define SUPPRESS_BANDING() bayer(ivec2(gl_FragCoord.xy))\n"\
 
 ////////////////////////////////////////////////////////////////
 
@@ -242,7 +244,7 @@ static const char postprocess_fragment_shader[] =
 PALETTE_BUFFER
 NOISE_FUNCTIONS
 "\n"
-"layout(location=0) uniform vec3 Params;\n"
+"layout(location=0) uniform vec4 Params;\n"
 "\n"
 "layout(location=0) out vec4 out_fragcolor;\n"
 "\n"
@@ -251,11 +253,12 @@ NOISE_FUNCTIONS
 "	float gamma = Params.x;\n"
 "	float contrast = Params.y;\n"
 "	float scale = Params.z;\n"
+"	float dither = Params.w;\n"
 "	out_fragcolor = texelFetch(GammaTexture, ivec2(gl_FragCoord), 0);\n"
 "#if PALETTIZE == 1\n"
 "	vec2 noiseuv = floor(gl_FragCoord.xy * scale) + 0.5;\n"
 "	out_fragcolor.rgb = sqrt(out_fragcolor.rgb);\n"
-"	out_fragcolor.rgb += DITHER_NOISE(noiseuv) * PAL_NOISESCALE;\n"
+"	out_fragcolor.rgb += DITHER_NOISE(noiseuv) * dither;\n"
 "	out_fragcolor.rgb *= out_fragcolor.rgb;\n"
 "#endif // PALETTIZE == 1\n"
 "#if PALETTIZE\n"
@@ -282,6 +285,8 @@ NOISE_FUNCTIONS
 "	vec4	SkyFog;\n"\
 "	vec3	WindDir;\n"\
 "	float	WindPhase;\n"\
+"	float	ScreenDither;\n"\
+"	float	TextureDither;\n"\
 "	vec3	EyePos;\n"\
 "	float	Time;\n"\
 "	float	ZLogScale;\n"\
@@ -321,7 +326,12 @@ NOISE_FUNCTIONS
 "\n"\
 "float GetLightStyle(int index)\n"\
 "{\n"\
-"	return index < " QS_STRINGIFY (MAX_LIGHTSTYLES) " ? LightStyles[index] : 1.0;\n"\
+"	float result;\n"\
+"	if (index < " QS_STRINGIFY (MAX_LIGHTSTYLES) ")\n"\
+"		result = LightStyles[index];\n"\
+"	else\n"\
+"		result = 1.0;\n"\
+"	return result;\n"\
 "}\n"\
 "\n"\
 
@@ -712,16 +722,16 @@ OIT_OUTPUT (out_fragcolor)
 "	farblend *= farblend;\n"
 "	out_fragcolor.rgb = sqrt(out_fragcolor.rgb);\n"
 "	float luma = dot(out_fragcolor.rgb, vec3(.25, .625, .125));\n"
-"	float nearnoise = tri(whitenoise01(lmuv * lmsize)) * luma;\n"
-"	float farnoise = Fog.w > 0. ? SCREEN_SPACE_NOISE() : 0.;\n"
-"	out_fragcolor.rgb += mix(nearnoise, farnoise, farblend) * PAL_NOISESCALE;\n"
+"	float nearnoise = tri(whitenoise01(lmuv * lmsize)) * luma * TextureDither;\n"
+"	float farnoise = Fog.w > 0. ? SCREEN_SPACE_NOISE() * ScreenDither : 0.;\n"
+"	out_fragcolor.rgb += mix(nearnoise, farnoise, farblend);\n"
 "	out_fragcolor.rgb *= out_fragcolor.rgb;\n"
 "#endif // DITHER == 1\n"
 "#if DITHER >= 2\n"
 "	// nuke extra precision in 10-bit framebuffer\n"
 "	out_fragcolor.rgb = floor(out_fragcolor.rgb * 255. + 0.5) * (1./255.);\n"
 "#elif DITHER == 0\n"
-"	out_fragcolor.rgb += SUPPRESS_BANDING ();\n"
+"	out_fragcolor.rgb += SUPPRESS_BANDING() * ScreenDither;\n"
 "#endif\n"
 "}\n";
 
@@ -795,11 +805,11 @@ OIT_OUTPUT (out_fragcolor)
 "	if (Fog.w > 0.)\n"
 "	{\n"
 "		out_fragcolor.rgb = sqrt(out_fragcolor.rgb);\n"
-"		out_fragcolor.rgb += SCREEN_SPACE_NOISE() * PAL_NOISESCALE;\n"
+"		out_fragcolor.rgb += SCREEN_SPACE_NOISE() * ScreenDither;\n"
 "		out_fragcolor.rgb *= out_fragcolor.rgb;\n"
 "	}\n"
 "#else\n"
-"	out_fragcolor.rgb += SUPPRESS_BANDING();\n"
+"	out_fragcolor.rgb += SUPPRESS_BANDING() * ScreenDither;\n"
 "#endif\n"
 "}\n";
 
@@ -889,7 +899,7 @@ NOISE_FUNCTIONS
 "	result.rgb = mix(result.rgb, layer.rgb, layer.a);\n"
 "	result.rgb = mix(result.rgb, SkyFog.rgb, SkyFog.a);\n"
 "	out_fragcolor = result;\n"
-"	out_fragcolor.rgb += SUPPRESS_BANDING();\n"
+"	out_fragcolor.rgb += SUPPRESS_BANDING() * ScreenDither;\n"
 "}\n";
 
 ////////////////////////////////////////////////////////////////
@@ -953,10 +963,10 @@ NOISE_FUNCTIONS
 "	out_fragcolor.rgb = mix(out_fragcolor.rgb, SkyFog.rgb, SkyFog.a);\n"
 "#if DITHER\n"
 "	out_fragcolor.rgb = sqrt(out_fragcolor.rgb);\n"
-"	out_fragcolor.rgb += SCREEN_SPACE_NOISE() * PAL_NOISESCALE;\n"
+"	out_fragcolor.rgb += SCREEN_SPACE_NOISE() * ScreenDither;\n"
 "	out_fragcolor.rgb *= out_fragcolor.rgb;\n"
 "#else\n"
-"	out_fragcolor.rgb += SUPPRESS_BANDING();\n"
+"	out_fragcolor.rgb += SUPPRESS_BANDING() * ScreenDither;\n"
 "#endif\n"
 "}\n";
 
@@ -992,6 +1002,7 @@ static const char sky_boxside_fragment_shader[] =
 NOISE_FUNCTIONS
 "\n"
 "layout(location=2) uniform vec4 Fog;\n"
+"layout(location=3) uniform float ScreenDither;\n"
 "\n"
 "layout(location=0) in vec3 in_dir;\n"
 "layout(location=1) in vec2 in_uv;\n"
@@ -1004,10 +1015,10 @@ NOISE_FUNCTIONS
 "	out_fragcolor.rgb = mix(out_fragcolor.rgb, Fog.rgb, Fog.w);\n"
 "#if DITHER\n"
 "	out_fragcolor.rgb = sqrt(out_fragcolor.rgb);\n"
-"	out_fragcolor.rgb += SCREEN_SPACE_NOISE() * PAL_NOISESCALE;\n"
+"	out_fragcolor.rgb += SCREEN_SPACE_NOISE() * ScreenDither;\n"
 "	out_fragcolor.rgb *= out_fragcolor.rgb;\n"
 "#else\n"
-"	out_fragcolor.rgb += SUPPRESS_BANDING();\n"
+"	out_fragcolor.rgb += SUPPRESS_BANDING() * ScreenDither;\n"
 "#endif\n"
 "}\n";
 
@@ -1033,6 +1044,7 @@ NOISE_FUNCTIONS
 "	mat4	ViewProj;\n"\
 "	vec3	EyePos;\n"\
 "	vec4	Fog;\n"\
+"	float	ScreenDither;\n"\
 "	InstanceData instances[];\n"\
 "};\n"\
 
@@ -1041,27 +1053,51 @@ NOISE_FUNCTIONS
 static const char alias_vertex_shader[] =
 ALIAS_INSTANCE_BUFFER
 "\n"
-"layout(std430, binding=2) restrict readonly buffer PoseBuffer\n"
-"{\n"
-"	uvec2 PackedPosNor[];\n"
-"};\n"
-"\n"
-"layout(std430, binding=3) restrict readonly buffer UVBuffer\n"
-"{\n"
-"	vec2 TexCoords[];\n"
-"};\n"
-"\n"
-"struct Pose\n"
+"struct PoseVertex\n"
 "{\n"
 "	vec3 pos;\n"
 "	vec3 nor;\n"
 "};\n"
 "\n"
-"Pose GetPose(uint index)\n"
-"{\n"
-"	uvec2 data = PackedPosNor[index + gl_VertexID];\n"
-"	return Pose(vec3((data.xxx >> uvec3(0, 8, 16)) & 255), unpackSnorm4x8(data.y).xyz);\n"
-"}\n"
+"#if MD5\n"
+"	layout(location=0) in vec3 in_pos;\n"
+"	layout(location=1) in vec4 in_nor;\n"
+"	layout(location=2) in vec2 in_uv;\n"
+"	layout(location=3) in vec4 in_weights;\n"
+"	layout(location=4) in ivec4 in_indices;\n"
+"\n"
+"	layout(std430, binding=2) restrict readonly buffer PoseBuffer\n"
+"	{\n"
+"		layout(row_major) mat4x3 BonePoses[];\n"
+"	};\n"
+"\n"
+"	PoseVertex GetPoseVertex(uint pose)\n"
+"	{\n"
+"		mat4x3 anim = BonePoses[pose + in_indices.x] * in_weights.x;\n"
+"		anim += BonePoses[pose + in_indices.y] * in_weights.y;\n"
+"		if (in_weights.z + in_weights.w > 0.0)\n"
+"		{\n"
+"			anim += BonePoses[pose + in_indices.z] * in_weights.z;\n"
+"			anim += BonePoses[pose + in_indices.w] * in_weights.w;\n"
+"		}\n"
+"		return PoseVertex((anim * vec4(in_pos, 1.0)).xyz, (anim * vec4(in_nor.xyz, 0.0)).xyz);\n"
+"	}\n"
+"\n"
+"#else\n"
+"	layout(location=0) in vec2 in_uv;\n"
+"\n"
+"	layout(std430, binding=2) restrict readonly buffer BlendShapeBuffer\n"
+"	{\n"
+"		uvec2 PackedPosNor[];\n"
+"	};\n"
+"\n"
+"	PoseVertex GetPoseVertex(uint pose)\n"
+"	{\n"
+"		uvec2 data = PackedPosNor[pose + gl_VertexID];\n"
+"		return PoseVertex(vec3((data.xxx >> uvec3(0, 8, 16)) & 255), unpackSnorm4x8(data.y).xyz);\n"
+"	}\n"
+"\n"
+"#endif // MD5\n"
 "\n"
 "float r_avertexnormal_dot(vec3 vertexnormal, vec3 dir) // from MH \n"
 "{\n"
@@ -1084,17 +1120,17 @@ ALIAS_INSTANCE_BUFFER
 "void main()\n"
 "{\n"
 "	InstanceData inst = instances[gl_InstanceID];\n"
-"	out_texcoord = TexCoords[gl_VertexID];\n"
-"	Pose pose1 = GetPose(inst.Pose1);\n"
-"	Pose pose2 = GetPose(inst.Pose2);\n"
+"	out_texcoord = in_uv;\n"
+"	PoseVertex pose1 = GetPoseVertex(inst.Pose1);\n"
+"	PoseVertex pose2 = GetPoseVertex(inst.Pose2);\n"
 "	mat4x3 worldmatrix = transpose(mat3x4(inst.WorldMatrix[0], inst.WorldMatrix[1], inst.WorldMatrix[2]));\n"\
 "	vec3 lerpedVert = (worldmatrix * vec4(mix(pose1.pos, pose2.pos, inst.Blend), 1.0)).xyz;\n"
 "	gl_Position = ViewProj * vec4(lerpedVert, 1.0);\n"
 "	out_pos = lerpedVert - EyePos;\n"
-"	vec3 shadevector;\n"
-"	shadevector.x = normalize(worldmatrix[0].xyz).x;\n"
-"	shadevector.y = normalize(worldmatrix[1].xyz).x;\n"
-"	shadevector.z = normalize(worldmatrix[2].xyz).x;\n"
+"	// transform world X and Z axes to local space\n"
+"	mat3 orientation = mat3(normalize(worldmatrix[0].xyz), normalize(worldmatrix[1].xyz), normalize(worldmatrix[2].xyz));\n"
+"	orientation = transpose(orientation);\n"
+"	vec3 shadevector = (orientation[0] + orientation[2]) / sqrt(2.0);\n"
 "	float dot1 = r_avertexnormal_dot(pose1.nor, shadevector);\n"
 "	float dot2 = r_avertexnormal_dot(pose2.nor, shadevector);\n"
 "	out_color = clamp(inst.LightColor * vec4(vec3(mix(dot1, dot2, inst.Blend)), 1.0), 0.0, 1.0);\n"
@@ -1148,16 +1184,16 @@ OIT_OUTPUT (out_fragcolor)
 "	fog = clamp(fog, 0.0, 1.0);\n"
 "	result.rgb = mix(Fog.rgb, result.rgb, fog);\n"
 "	out_fragcolor = result;\n"
-"#if MODE == " QS_STRINGIFY (ALIASSHADER_DITHER) "\n"
+"#if MODE == " QS_STRINGIFY (ALIASSHADER_DITHER) " || MODE == " QS_STRINGIFY (ALIASSHADER_NOPERSP) "\n"
 "	// Note: sign bit is used as overbright flag\n"
 "	if (abs(Fog.w) > 0.)\n"
 "	{\n"
 "		out_fragcolor.rgb = sqrt(out_fragcolor.rgb);\n"
-"		out_fragcolor.rgb += SCREEN_SPACE_NOISE() * PAL_NOISESCALE;\n"
+"		out_fragcolor.rgb += SCREEN_SPACE_NOISE() * ScreenDither;\n"
 "		out_fragcolor.rgb *= out_fragcolor.rgb;\n"
 "	}\n"
 "#else\n"
-"	out_fragcolor.rgb += SUPPRESS_BANDING();\n"
+"	out_fragcolor.rgb += SUPPRESS_BANDING() * ScreenDither;\n"
 "#endif\n"
 "}\n";
 
@@ -1207,11 +1243,11 @@ NOISE_FUNCTIONS
 "	if (Fog.w > 0.)\n"
 "	{\n"
 "		out_fragcolor.rgb = sqrt(out_fragcolor.rgb);\n"
-"		out_fragcolor.rgb += SCREEN_SPACE_NOISE() * PAL_NOISESCALE;\n"
+"		out_fragcolor.rgb += SCREEN_SPACE_NOISE() * ScreenDither;\n"
 "		out_fragcolor.rgb *= out_fragcolor.rgb;\n"
 "	}\n"
 "#else\n"
-"	out_fragcolor.rgb += SUPPRESS_BANDING();\n"
+"	out_fragcolor.rgb += SUPPRESS_BANDING() * ScreenDither;\n"
 "#endif\n"
 "}\n";
 
@@ -1281,11 +1317,11 @@ OIT_OUTPUT (out_fragcolor)
 "	if (Fog.w > 0.)\n"
 "	{\n"
 "		out_fragcolor.rgb = sqrt(out_fragcolor.rgb);\n"
-"		out_fragcolor.rgb += SCREEN_SPACE_NOISE() * PAL_NOISESCALE;\n"
+"		out_fragcolor.rgb += SCREEN_SPACE_NOISE() * ScreenDither;\n"
 "		out_fragcolor.rgb *= out_fragcolor.rgb;\n"
 "	}\n"
 "#else\n"
-"	out_fragcolor.rgb += SUPPRESS_BANDING();\n"
+"	out_fragcolor.rgb += SUPPRESS_BANDING() * ScreenDither;\n"
 "#endif\n"
 "}\n";
 
@@ -1492,46 +1528,45 @@ DRAW_ELEMENTS_INDIRECT_COMMAND
 "	uint vis[];\n"
 "};\n"
 "\n"
-"struct Leaf\n"
+"struct MarkSurface\n"
 "{\n"
-"	vec3	mins;\n"
-"	uint	firstsurf;\n"
-"	vec3	maxs;\n"
-"	uint	surfcountsky; // bit 0=sky; bits 1..31=surfcount\n"
+"	uint packedleafsky; // bit 0=sky; 1..31=leafindex\n"
+"	uint surfindex;\n"
 "};\n"
-"\n"
-"layout(std430, binding=4) restrict readonly buffer LeafBuffer\n"
+";\n"
+"layout(std430, binding=4) restrict readonly buffer MarkSurfaceBuffer\n"
 "{\n"
-"	Leaf leaves[];\n"
-"};\n"
-"\n"
-"layout(std430, binding=5) restrict readonly buffer MarkSurfBuffer\n"
-"{\n"
-"	int marksurf[];\n"
+"	MarkSurface marksurfs[];\n"
 "};\n"
 "\n"
 "struct Surface\n"
 "{\n"
 "	vec4	plane;\n"
+"	vec3	mins;\n"
 "	uint	framecount;\n"
+"	vec3	maxs;\n"
 "	uint	texnum;\n"
 "	uint	numedges;\n"
 "	uint	firstvert;\n"
+"	uint	_pad0;\n"
+"	uint	_pad1;\n"
 "};\n"
 "\n"
 "// Same issue as above\n"
-"layout(std430, binding=6) restrict buffer SurfaceBuffer\n"
+"layout(std430, binding=5) restrict buffer SurfaceBuffer\n"
 "{\n"
 "	//Surface surfaces[];\n"
 "	uvec4 rawsurfaces[];\n"
 "};\n"
 "\n"
 "#define SURF_PLANE(base)			uintBitsToFloat(rawsurfaces[base])\n"
-"#define SURF_FRAMECOUNT(base)		rawsurfaces[base + 1].x\n"
-"#define SURF_TEXNUM(base)			rawsurfaces[base + 1].y\n"
-"#define SURF_NUMEDGES(base)		rawsurfaces[base + 1].z\n"
-"#define SURF_FIRSTVERT(base)		rawsurfaces[base + 1].w\n"
-"#define SIZEOF_SURFACE				2 /* uvec4s */\n"
+"#define SURF_MINS(base)			uintBitsToFloat(rawsurfaces[base + 1].xyz)\n"
+"#define SURF_FRAMECOUNT(base)		rawsurfaces[base + 1].w\n"
+"#define SURF_MAXS(base)			uintBitsToFloat(rawsurfaces[base + 2].xyz)\n"
+"#define SURF_TEXNUM(base)			rawsurfaces[base + 2].w\n"
+"#define SURF_NUMEDGES(base)		rawsurfaces[base + 3].x\n"
+"#define SURF_FIRSTVERT(base)		rawsurfaces[base + 3].y\n"
+"#define SIZEOF_SURFACE				4 /* uvec4s */\n"
 "\n"
 "layout(std140, binding=1) uniform FrameCullUBO\n"
 "{\n"
@@ -1544,49 +1579,60 @@ DRAW_ELEMENTS_INDIRECT_COMMAND
 "void main()\n"
 "{\n"
 "	uint thread_id = gl_GlobalInvocationID.x;\n"
-"	if (thread_id >= leaves.length())\n"
+"	if (thread_id >= marksurfs.length())\n"
 "		return;\n"
-"	uint visible = vis[thread_id >> 5u] & (1u << (thread_id & 31u));\n"
+"	MarkSurface mark = marksurfs[thread_id];\n"
+"\n"
+"	// sky culling: when r_oldskyleaf is 0, surfaces inside a sky leaf are skipped\n"
+"	if ((mark.packedleafsky & 1u) > oldskyleaf)\n"
+"		return;\n"
+"\n"
+"	// vis culling\n"
+"	uint leaf = mark.packedleafsky >> 1u;\n"
+"	uint visible = vis[leaf >> 5u] & (1u << (leaf & 31u));\n"
 "	if (visible == 0u)\n"
 "		return;\n"
 "\n"
-"	Leaf leaf = leaves[thread_id];\n"
-"	uint i, j;\n"
-"	for (i = 0u; i < 4u; i++)\n"
+"	uint surfbase = mark.surfindex * uint(SIZEOF_SURFACE);\n"
+"\n"
+"	// backface culling\n"
+"	vec4 surfplane = SURF_PLANE(surfbase);\n"
+"	if (dot(surfplane.xyz, vieworg) < surfplane.w)\n"
+"		return;\n"
+"\n"
+"	// frustum culling\n"
+"	vec3 mins = SURF_MINS(surfbase);\n"
+"	vec3 maxs = SURF_MAXS(surfbase);\n"
+"	for (uint i = 0u; i < 4u; i++)\n"
 "	{\n"
 "		vec4 plane = frustum[i];\n"
 "		vec3 v;\n"
-"		v.x = plane.x < 0.f ? leaf.mins.x : leaf.maxs.x;\n"
-"		v.y = plane.y < 0.f ? leaf.mins.y : leaf.maxs.y;\n"
-"		v.z = plane.z < 0.f ? leaf.mins.z : leaf.maxs.z;\n"
+"		v.x = plane.x < 0.0 ? mins.x : maxs.x;\n"
+"		v.y = plane.y < 0.0 ? mins.y : maxs.y;\n"
+"		v.z = plane.z < 0.0 ? mins.z : maxs.z;\n"
 "		if (dot(plane.xyz, v) < plane.w)\n"
 "			return;\n"
 "	}\n"
 "\n"
-"	if ((leaf.surfcountsky & 1u) > oldskyleaf)\n"
+"	// surfaces can appear in multiple leaves\n"
+"	// check if this is the first time this surface has passed culling this frame\n"
+"	if (atomicExchange(SURF_FRAMECOUNT(surfbase), framecount) == framecount)\n"
 "		return;\n"
-"	uint surfcount = leaf.surfcountsky >> 1u;\n"
-"	vec3 campos = vieworg;\n"
-"	for (i = 0u; i < surfcount; i++)\n"
+"\n"
+"	// surface is visible, append its triangles to the index buffer\n"
+"	// and update the draw command corresponding to its texture number\n"
+"	uint texnum = SURF_TEXNUM(surfbase);\n"
+"	uint numedges = SURF_NUMEDGES(surfbase);\n"
+"	uint firstvert = SURF_FIRSTVERT(surfbase);\n"
+"	uint cmdbase = texnum * uint(SIZEOF_CMD);\n"
+"	// some bsps out there have faces with < 2 edges, which would cause underflow below\n"
+"	numedges = max(numedges, 2u);\n"
+"	uint ofs = CMD_FIRST_INDEX(cmdbase) + atomicAdd(CMD_COUNT(cmdbase), 3u * (numedges - 2u));\n"
+"	for (uint i = 2u; i < numedges; i++)\n"
 "	{\n"
-"		uint surfindex = uint(marksurf[leaf.firstsurf + i]);\n"
-"		uint surfbase = surfindex * uint(SIZEOF_SURFACE);\n"
-"		vec4 surfplane = SURF_PLANE(surfbase);\n"
-"		if (dot(surfplane.xyz, campos) < surfplane.w)\n"
-"			continue;\n"
-"		if (atomicExchange(SURF_FRAMECOUNT(surfbase), framecount) == framecount)\n"
-"			continue;\n"
-"		uint texnum = SURF_TEXNUM(surfbase);\n"
-"		uint numedges = SURF_NUMEDGES(surfbase);\n"
-"		uint firstvert = SURF_FIRSTVERT(surfbase);\n"
-"		uint cmdbase = texnum * uint(SIZEOF_CMD);\n"
-"		uint ofs = CMD_FIRST_INDEX(cmdbase) + atomicAdd(CMD_COUNT(cmdbase), 3u * (numedges - 2u));\n"
-"		for (j = 2u; j < numedges; j++)\n"
-"		{\n"
-"			indices[ofs++] = firstvert;\n"
-"			indices[ofs++] = firstvert + j - 1u;\n"
-"			indices[ofs++] = firstvert + j;\n"
-"		}\n"
+"		indices[ofs++] = firstvert;\n"
+"		indices[ofs++] = firstvert + i - 1u;\n"
+"		indices[ofs++] = firstvert + i;\n"
 "	}\n"
 "}\n";
 
