@@ -74,7 +74,7 @@ void Sky_LoadTexture (qmodel_t *mod, texture_t *mt)
 
 	if (mt->width != 256 || mt->height != 128)
 	{
-		Con_Warning ("Sky texture %s is %d x %d, expected 256 x 128\n", mt->name, mt->width, mt->height);
+		Con_DWarning ("Sky texture %s is %d x %d, expected 256 x 128\n", mt->name, mt->width, mt->height);
 		if (mt->width < 2 || mt->height < 1)
 			return;
 	}
@@ -184,10 +184,10 @@ void Sky_LoadTextureQ64 (qmodel_t *mod, texture_t *mt)
 
 /*
 =================
-Sky_ClearWind
+Skywind_Clear
 =================
 */
-static void Sky_ClearWind (void)
+static void Skywind_Clear (void)
 {
 	if (!skybox)
 		return;
@@ -199,17 +199,20 @@ static void Sky_ClearWind (void)
 
 /*
 =================
-Sky_LoadWind_f
+Skywind_Load_f
 =================
 */
-static void Sky_LoadWind_f (void)
+static void Skywind_Load_f (void)
 {
 	char relname[MAX_QPATH];
 	char *buf;
 	const char *data;
 
 	if (!skybox)
+	{
+		Con_Printf ("No skybox loaded\n");
 		return;
+	}
 
 	q_snprintf (relname, sizeof (relname), "gfx/env/%s" SKYWIND_CFG, skybox->name);
 	buf = (char *) COM_LoadMallocFile (relname, NULL);
@@ -225,11 +228,11 @@ static void Sky_LoadWind_f (void)
 
 	if (strcmp (com_token, "skywind") != 0)
 	{
-		Con_Printf ("Sky_LoadWind_f: first token must be 'skywind'.\n");
+		Con_Printf ("Skywind_Load_f: first token must be 'skywind'.\n");
 		goto done;
 	}
 
-	Sky_ClearWind ();
+	Skywind_Clear ();
 
 	if ((data = COM_Parse (data)) != NULL)
 		skybox->wind_dist = CLAMP (-2.0, atof (com_token), 2.0);
@@ -249,17 +252,20 @@ done:
 
 /*
 =================
-Sky_SaveWind_f
+Skywind_Save_f
 =================
 */
-static void Sky_SaveWind_f (void)
+static void Skywind_Save_f (void)
 {
 	char relname[MAX_QPATH];
 	char path[MAX_OSPATH];
 	FILE *f;
 
 	if (!skybox)
+	{
+		Con_Printf ("No skybox loaded\n");
 		return;
+	}
 
 	q_snprintf (relname, sizeof (relname), "gfx/env/%s" SKYWIND_CFG, skybox->name);
 	q_snprintf (path, sizeof (path), "%s/%s", com_gamedir, relname);
@@ -281,15 +287,81 @@ static void Sky_SaveWind_f (void)
 
 	fclose (f);
 
-	Con_Printf ("Wrote '%s'.\n", relname);
+	Con_SafePrintf ("Wrote ");
+	Con_LinkPrintf (path, "%s", relname);
+	Con_SafePrintf ("\n");
 }
 
 /*
 =================
-Sky_WindCommand_f
+Skywind_LookDir_f
 =================
 */
-static void Sky_WindCommand_f (void)
+static void Skywind_LookDir_f (void)
+{
+	if (cls.state != ca_connected)
+		return;
+
+	if (!skybox)
+	{
+		Con_Printf ("No skybox loaded\n");
+		return;
+	}
+
+	// invert view direction so that clouds move towards the player, not away from them
+	skybox->wind_yaw = fmod (cl.viewangles[YAW] + 180.0, 360.0);
+	skybox->wind_pitch = -cl.viewangles[PITCH];
+
+	// first argument, if present, overrides the loop duration (default: 30 seconds)
+	if (Cmd_Argc () >= 2)
+		skybox->wind_period = atof (Cmd_Argv (1));
+	else if (!skybox->wind_period)
+		skybox->wind_period = 30.f;
+
+	// second argument, if present, overrides the amplitude of the movement (default: 1.0)
+	if (Cmd_Argc () >= 3)
+		skybox->wind_dist = CLAMP (-2.0, atof (Cmd_Argv (2)), 2.0);
+	else if (!skybox->wind_dist)
+		skybox->wind_dist = 1.f;
+}
+
+/*
+=================
+Skywind_Rotate_f
+=================
+*/
+static void Skywind_Rotate_f (void)
+{
+	if (cls.state != ca_connected)
+		return;
+
+	if (!skybox)
+	{
+		Con_Printf ("No skybox loaded\n");
+		return;
+	}
+
+	if (Cmd_Argc () < 2)
+	{
+		Con_Printf (
+			"usage:\n"
+			"   %s <yawdelta> [pitchdelta]\n",
+			Cmd_Argv (0)
+		);
+		return;
+	}
+
+	skybox->wind_yaw = fmod (skybox->wind_yaw + atof (Cmd_Argv (1)), 360.0);
+	if (Cmd_Argc () >= 3)
+		skybox->wind_pitch = fmod (skybox->wind_pitch + atof (Cmd_Argv (2)) + 90.0, 180.0) - 90.0;
+}
+
+/*
+=================
+Skywind_f
+=================
+*/
+static void Skywind_f (void)
 {
 	if (cls.state != ca_connected)
 		return;
@@ -319,38 +391,6 @@ static void Sky_WindCommand_f (void)
 		return;
 	}
 
-	if (!q_strcasecmp (Cmd_Argv (1), "save"))
-	{
-		Sky_SaveWind_f ();
-		return;
-	}
-
-	if (!q_strcasecmp (Cmd_Argv (1), "load"))
-	{
-		Sky_LoadWind_f ();
-		return;
-	}
-
-	if (!q_strcasecmp (Cmd_Argv (1), "rotate"))
-	{
-		if (Cmd_Argc () < 3)
-		{
-			Con_Printf (
-				"usage:\n"
-				"   %s %s <value>\n",
-				Cmd_Argv (0),
-				Cmd_Argv (1)
-			);
-			return;
-		}
-
-		skybox->wind_yaw = fmod (skybox->wind_yaw + atof (Cmd_Argv (2)), 360.0);
-		if (Cmd_Argc () >= 4)
-			skybox->wind_pitch = fmod (skybox->wind_pitch + atof (Cmd_Argv (3)) + 90.0, 180.0) - 90.0;
-
-		return;
-	}
-
 	skybox->wind_dist = CLAMP (-2.0, atof (Cmd_Argv (1)), 2.0);
 	if (Cmd_Argc () >= 3)
 		skybox->wind_yaw = fmod (atof (Cmd_Argv (2)), 360.0);
@@ -365,13 +405,14 @@ static void Sky_WindCommand_f (void)
 Sky_LoadSkyBox
 ==================
 */
-const char	*suf[6] = {"rt", "bk", "lf", "ft", "up", "dn"};
+static const char *const suf[6] = {"rt", "bk", "lf", "ft", "up", "dn"};
 void Sky_LoadSkyBox (const char *name)
 {
 	int			i, mark, width[6], height[6], samesize, numloaded;
 	char		filename[MAX_OSPATH];
 	byte		*data[6];
 	skybox_t	newsky;
+	enum srcformat	fmt;
 
 	if (skybox && strcmp(skybox->name, name) == 0)
 		return; //no change
@@ -398,9 +439,12 @@ void Sky_LoadSkyBox (const char *name)
 	for (i = 0, numloaded = 0, samesize = 0; i < 6; i++)
 	{
 		q_snprintf (filename, sizeof(filename), "gfx/env/%s%s", name, suf[i]);
-		data[i] = Image_LoadImage (filename, &width[i], &height[i]);
+		data[i] = Image_LoadImage (filename, &width[i], &height[i], &fmt);
 		if (data[i])
 		{
+			if (fmt != SRC_RGBA)
+				Sys_Error ("Bad format %i for skybox side %s", fmt, filename);
+
 			numloaded++;
 			if (width[i] != height[i])
 				samesize = -1;
@@ -469,7 +513,7 @@ void Sky_LoadSkyBox (const char *name)
 	VEC_PUSH (skybox_list, newsky);
 	skybox = &skybox_list[VEC_SIZE (skybox_list) - 1];
 
-	Sky_LoadWind_f ();
+	Skywind_Load_f ();
 }
 
 /*
@@ -499,6 +543,8 @@ void Sky_ClearAll (void)
 	for (i = 0, count = VEC_SIZE (skybox_list); i < count; i++)
 		Sky_FreeSkyBox (&skybox_list[i]);
 	VEC_CLEAR (skybox_list);
+
+	Cvar_SetQuick (&r_skyfog, r_skyfog.default_string);
 }
 
 /*
@@ -539,7 +585,7 @@ void Sky_NewMap (void)
 			q_strlcpy(key, com_token, sizeof(key));
 		while (key[0] && key[strlen(key)-1] == ' ') // remove trailing spaces
 			key[strlen(key)-1] = 0;
-		data = COM_Parse(data);
+		data = COM_ParseEx(data, CPE_ALLOWTRUNC);
 		if (!data)
 			return; // error
 		q_strlcpy(value, com_token, sizeof(value));
@@ -604,9 +650,11 @@ void Sky_Init (void)
 	Cvar_SetCallback (&r_skyfog, R_SetSkyfog_f);
 
 	Cmd_AddCommand ("sky",Sky_SkyCommand_f);
-	Cmd_AddCommand ("skywind",Sky_WindCommand_f);
-	Cmd_AddCommand ("skywind_save",Sky_SaveWind_f);
-	Cmd_AddCommand ("skywind_load",Sky_LoadWind_f);
+	Cmd_AddCommand ("skywind",Skywind_f);
+	Cmd_AddCommand ("skywind_save",Skywind_Save_f);
+	Cmd_AddCommand ("skywind_load",Skywind_Load_f);
+	Cmd_AddCommand ("skywind_lookdir",Skywind_LookDir_f);
+	Cmd_AddCommand ("skywind_rotate",Skywind_Rotate_f);
 }
 
 //==============================================================================
@@ -676,6 +724,7 @@ void Sky_DrawSkyBox (void)
 	GL_UniformMatrix4fvFunc (0, 1, GL_FALSE, r_matviewproj);
 	GL_Uniform3fvFunc (1, 1, r_refdef.vieworg);
 	GL_Uniform4fvFunc (2, 1, fog);
+	GL_Uniform1fFunc (3, r_framedata.screendither);
 
 	for (i = 0; i < 6; i++)
 	{
@@ -755,7 +804,7 @@ Sky_IsAnimated
 */
 qboolean Sky_IsAnimated (void)
 {
-	return r_skywind.value != 0.f && skybox && skybox->wind_dist > 0.f;
+	return r_skywind.value != 0.f && skybox && skybox->wind_dist != 0.f;
 }
 
 /*
