@@ -39,6 +39,11 @@ static const struct {
 
 cvar_t			r_softemu = {"r_softemu", "0", CVAR_ARCHIVE};
 cvar_t			r_softemu_metric = {"r_softemu_metric", "-1", CVAR_ARCHIVE};
+cvar_t			r_softemu_lightmap_banding = {"r_softemu_lightmap_banding", "-1", CVAR_ARCHIVE};
+cvar_t			r_softemu_mdl_warp = {"r_softemu_mdl_warp", "-1", CVAR_ARCHIVE};
+cvar_t			r_softemu_dither_screen = {"r_softemu_dither_screen", "1.0", CVAR_ARCHIVE};
+cvar_t			r_softemu_dither_texture = {"r_softemu_dither_texture", "1.0", CVAR_ARCHIVE};
+
 static cvar_t	gl_max_size = {"gl_max_size", "0", CVAR_NONE};
 static cvar_t	gl_picmip = {"gl_picmip", "0", CVAR_NONE};
 cvar_t			gl_lodbias = {"gl_lodbias", "auto", CVAR_ARCHIVE };
@@ -55,9 +60,10 @@ static int numgltextures;
 static gltexture_t	*active_gltextures, *free_gltextures;
 gltexture_t		*notexture, *nulltexture, *whitetexture, *greytexture, *blacktexture;
 
+unsigned int d_8to24table_opaque[256];			//standard palette with alpha 255 for all colors
 unsigned int d_8to24table[256];					//standard palette, 255 is transparent
 unsigned int d_8to24table_fbright[256];			//fullbright palette, 0-223 are black (for additive blending)
-unsigned int d_8to24table_alphabright[256];		//palette with lighting mask in alpha channel (0=fullbright, 1=lit)
+unsigned int d_8to24table_alphabright[256];		//palette with lighting mask in alpha channel (0=fullbright, 255=lit)
 unsigned int d_8to24table_fbright_fence[256];	//fullbright palette, for fence textures
 unsigned int d_8to24table_nobright[256];		//nobright palette, 224-255 are black (for additive blending)
 unsigned int d_8to24table_nobright_fence[256];	//nobright palette, for fence textures
@@ -75,33 +81,21 @@ static void GL_DeleteTexture (gltexture_t *texture);
 ================================================================================
 */
 
-typedef struct
+const glmode_t glmodes[NUM_GLMODES] =
 {
-	int	magfilter;
-	int	minfilter;
-	const char  *name;
-} glmode_t;
-static const glmode_t glmodes[] = {
-	{GL_NEAREST, GL_NEAREST,		"GL_NEAREST"},
-	{GL_NEAREST, GL_NEAREST_MIPMAP_NEAREST,	"GL_NEAREST_MIPMAP_NEAREST"},
-	{GL_NEAREST, GL_NEAREST_MIPMAP_LINEAR,	"GL_NEAREST_MIPMAP_LINEAR"},
-	{GL_LINEAR,  GL_LINEAR,			"GL_LINEAR"},
-	{GL_LINEAR,  GL_LINEAR_MIPMAP_NEAREST,	"GL_LINEAR_MIPMAP_NEAREST"},
-	{GL_LINEAR,  GL_LINEAR_MIPMAP_LINEAR,	"GL_LINEAR_MIPMAP_LINEAR"},
+	{GL_NEAREST, GL_NEAREST,					"GL_NEAREST",					"Classic"},
+	{GL_NEAREST, GL_NEAREST_MIPMAP_NEAREST,		"GL_NEAREST_MIPMAP_NEAREST",	"Classic"},
+	{GL_NEAREST, GL_NEAREST_MIPMAP_LINEAR,		"GL_NEAREST_MIPMAP_LINEAR",		"Classic"},
+	{GL_LINEAR,  GL_LINEAR,						"GL_LINEAR",					"Smooth"},
+	{GL_LINEAR,  GL_LINEAR_MIPMAP_NEAREST,		"GL_LINEAR_MIPMAP_NEAREST",		"Smooth"},
+	{GL_LINEAR,  GL_LINEAR_MIPMAP_LINEAR,		"GL_LINEAR_MIPMAP_LINEAR",		"Smooth"},
 };
-#define NUM_GLMODES (int)(sizeof(glmodes)/sizeof(glmodes[0]))
 static int glmode_idx = 2; /* nearest with linear mips */
 
 static GLuint gl_samplers[NUM_GLMODES * 2]; // x2: nomip + mip
 
-typedef struct texfilter_s
-{
-	int		mode;
-	float	anisotropy;
-	float	lodbias;
-} texfilter_t;
+texfilter_t gl_texfilter;
 
-static texfilter_t gl_texfilter;
 
 /*
 ===============
@@ -148,10 +142,8 @@ TexMgr_DescribeTextureModes_f -- report available texturemodes
 static void TexMgr_DescribeTextureModes_f (void)
 {
 	int i;
-
 	for (i = 0; i < NUM_GLMODES; i++)
 		Con_SafePrintf ("   %2i: %s\n", i + 1, glmodes[i].name);
-
 	Con_Printf ("%i modes\n", i);
 }
 
@@ -210,6 +202,19 @@ static void TexMgr_SetFilterModes (gltexture_t *glt)
 	}
 }
 
+/*
+===============
+TexMgr_TextureMode_Completion_f -- tab completion for gl_texturemode
+===============
+*/
+static void TexMgr_TextureMode_Completion_f (cvar_t *cvar, const char *partial)
+{
+	int i;
+
+	for (i = 0; i < NUM_GLMODES; i++)
+		Con_AddToTabList (glmodes[i].name, partial, NULL);
+}
+
 /*
 ===============
 TexMgr_TextureMode_f -- called when gl_texturemode changes
@@ -307,6 +312,16 @@ static void TexMgr_ForceFilterUpdate (void)
 	gl_texfilter.anisotropy = -1.f;
 }
 
+/*
+===============
+TexMgr_UsesFilterOverride
+===============
+*/
+qboolean TexMgr_UsesFilterOverride (void)
+{
+	return softemu >= SOFTEMU_COARSE || r_softemu_lightmap_banding.value > 0.f;
+}
+
 /*
 ===============
 TexMgr_ApplySettings -- called at the beginning of each frame
@@ -321,7 +336,8 @@ void TexMgr_ApplySettings (void)
 	gl_texfilter.anisotropy	= CLAMP (1.f, gl_texture_anisotropy.value, gl_max_anisotropy);
 	gl_texfilter.lodbias	= lodbias;
 
-	if (softemu >= SOFTEMU_COARSE)
+	// if softemu is either 2 & 3 or r_softemu_lightmap_banding is > 0 we override the filtering mode, unless it's GL_NEAREST
+	if (gl_texfilter.mode != 0 && TexMgr_UsesFilterOverride ())
 	{
 		const float SOFTEMU_ANISOTROPY = 8.f;
 		gl_texfilter.mode = 2; // nearest with linear mips
@@ -345,6 +361,19 @@ void TexMgr_ApplySettings (void)
 	Sbar_Changed (); //sbar graphics need to be redrawn with new filter mode
 }
 
+/*
+===============
+TexMgr_Imagelist_Completion_f -- tab completion for imagelist/imagedump
+===============
+*/
+static void TexMgr_Imagelist_Completion_f (const char *partial)
+{
+	gltexture_t	*glt;
+
+	for (glt = active_gltextures; glt; glt = glt->next)
+		Con_AddToTabList (glt->name, partial, NULL);
+}
+
 /*
 ===============
 TexMgr_Imagelist_f -- report loaded textures
@@ -408,6 +437,7 @@ TexMgr_Imagedump_f -- dump all current textures to TGA files
 static void TexMgr_Imagedump_f (void)
 {
 	char tganame[MAX_OSPATH], tempname[MAX_OSPATH], dirname[MAX_OSPATH];
+	const char *reldirname = "imagedump";
 	const char *filter = NULL;
 	int count = 0;
 	gltexture_t	*glt;
@@ -417,7 +447,7 @@ static void TexMgr_Imagedump_f (void)
 	if (Cmd_Argc () >= 2)
 		filter = Cmd_Argv (1);
 
-	q_snprintf(dirname, sizeof(dirname), "%s/imagedump", com_gamedir);
+	q_snprintf(dirname, sizeof(dirname), "%s/%s", com_gamedir, reldirname);
 
 	glPixelStorei (GL_PACK_ALIGNMENT, 1);/* for widths that aren't a multiple of 4 */
 
@@ -437,6 +467,8 @@ static void TexMgr_Imagedump_f (void)
 
 		GL_Bind (GL_TEXTURE0, glt);
 		buffer = (byte *) malloc(glt->width * glt->height * glt->depth * channels);
+		if (!buffer)
+			Sys_Error ("TexMgr_Imagedump_f: out of memory (%dx%dx%d %d bpp)", glt->width, glt->height, glt->depth, channels * 8);
 
 		if (glt->flags & TEXPREF_CUBEMAP)
 		{
@@ -461,9 +493,11 @@ static void TexMgr_Imagedump_f (void)
 	}
 
 	if (filter)
-		Con_Printf ("dumped %i textures containing '%s' to %s\n", count, filter, dirname);
+		Con_SafePrintf ("dumped %i textures containing '%s' to ", count, filter);
 	else
-		Con_Printf ("dumped %i textures to %s\n", count, dirname);
+		Con_SafePrintf ("dumped %i textures to ", count);
+	Con_LinkPrintf (va ("%s/", dirname), "%s", reldirname);
+	Con_SafePrintf (".\n");
 }
 
 /*
@@ -510,10 +544,9 @@ TexMgr_CompressTextures_f -- called when gl_compress_textures changes
 */
 void TexMgr_CompressTextures_f (cvar_t *var)
 {
-	qboolean compress = var->value != 0.f;
 	gltexture_t	*glt;
 
-	Con_SafePrintf ("Using %s textures\n", "uncompressed" + 2 * compress);
+	Con_SafePrintf ("Using %s textures\n", var->value ? "compressed" : "uncompressed");
 
 	// In an attempt to reduce VRAM fragmentation, instead of unloading and reloading
 	// each texture sequentially, we first unload them all, then reload them
@@ -710,15 +743,15 @@ void TexMgr_LoadPalette (void)
 		Sys_Error ("Couldn't load gfx/palette.lmp");
 
 	mark = Hunk_LowMark ();
-	pal = (byte *) Hunk_Alloc (768);
+	pal = (byte *) Hunk_AllocNoFill (768);
 	if (fread (pal, 768, 1, f) != 1)
-		Sys_Error ("TexMgr_LoadPalette: read error");
+		Sys_Error ("Failed reading gfx/palette.lmp");
 	fclose(f);
 
 	COM_FOpenFile ("gfx/colormap.lmp", &f, NULL);
 	if (!f)
 		Sys_Error ("Couldn't load gfx/colormap.lmp");
-	colormap = (byte *) Hunk_Alloc (256 * 64);
+	colormap = (byte *) Hunk_AllocNoFill (256 * 64);
 	if (fread (colormap, 256 * 64, 1, f) != 1)
 		Sys_Error ("TexMgr_LoadPalette: colormap read error");
 	fclose(f);
@@ -749,7 +782,7 @@ void TexMgr_LoadPalette (void)
 	src = pal;
 	for (i = 0; i < 256; i++, src += 3)
 	{
-		SetColor (&d_8to24table[i], src[0], src[1], src[2], 255);
+		SetColor (&d_8to24table_opaque[i], src[0], src[1], src[2], 255);
 		if (GetBit (is_fullbright, i))
 		{
 			SetColor (&d_8to24table_alphabright[i],	src[0], src[1], src[2], 0);
@@ -759,10 +792,12 @@ void TexMgr_LoadPalette (void)
 		else
 		{
 			SetColor (&d_8to24table_alphabright[i],	src[0], src[1], src[2], 255);
-			SetColor (&d_8to24table_fbright[i],		0, 0, 0, 0);
+			SetColor (&d_8to24table_fbright[i],		0, 0, 0, 255);
 			SetColor (&d_8to24table_nobright[i],	src[0], src[1], src[2], 255);
 		}
 	}
+
+	memcpy(d_8to24table, d_8to24table_opaque, 256*4);
 	((byte *) &d_8to24table[255]) [3] = 0; //standard palette, 255 is transparent
 
 	//fullbright palette, for fence textures
@@ -808,6 +843,7 @@ void TexMgr_Init (void)
 	static byte greytexture_data[16] = {127,127,127,255,127,127,127,255,127,127,127,255,127,127,127,255}; //50% grey
 	static byte blacktexture_data[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}; //black
 	extern texture_t *r_notexture_mip, *r_notexture_mip2;
+	cmd_function_t	*cmd;
 
 	// init texture list
 	free_gltextures = (gltexture_t *) Hunk_AllocName (MAX_GLTEXTURES * sizeof(gltexture_t), "gltextures");
@@ -829,13 +865,23 @@ void TexMgr_Init (void)
 	gl_texturemode.string = glmodes[glmode_idx].name;
 	Cvar_RegisterVariable (&gl_texturemode);
 	Cvar_SetCallback (&gl_texturemode, &TexMgr_TextureMode_f);
+	Cvar_SetCompletion (&gl_texturemode, &TexMgr_TextureMode_Completion_f);
 	Cvar_RegisterVariable (&gl_lodbias);
 	Cvar_SetCallback (&gl_lodbias, TexMgr_LodBias_f);
 	Cvar_RegisterVariable (&r_softemu);
 	Cvar_SetCallback (&r_softemu, TexMgr_SoftEmu_f);
+	Cvar_RegisterVariable (&r_softemu_lightmap_banding);
+	Cvar_SetCallback (&r_softemu_lightmap_banding, TexMgr_SoftEmu_f);
+	Cvar_RegisterVariable (&r_softemu_mdl_warp);
+	Cvar_RegisterVariable (&r_softemu_dither_screen);
+	Cvar_RegisterVariable (&r_softemu_dither_texture);
 	Cmd_AddCommand ("gl_describetexturemodes", &TexMgr_DescribeTextureModes_f);
-	Cmd_AddCommand ("imagelist", &TexMgr_Imagelist_f);
-	Cmd_AddCommand ("imagedump", &TexMgr_Imagedump_f);
+	cmd = Cmd_AddCommand ("imagelist", &TexMgr_Imagelist_f);
+	if (cmd)
+		cmd->completion = TexMgr_Imagelist_Completion_f;
+	cmd = Cmd_AddCommand ("imagedump", &TexMgr_Imagedump_f);
+	if (cmd)
+		cmd->completion = TexMgr_Imagelist_Completion_f;
 
 	// poll max size from hardware
 	glGetIntegerv (GL_MAX_TEXTURE_SIZE, &gl_max_texture_size);
@@ -897,8 +943,7 @@ int TexMgr_PadConditional (int s)
 {
 	if (s < TexMgr_SafeTextureSize(s))
 		return TexMgr_Pad(s);
-	else
-		return s;
+	return s;
 }
 
 /*
@@ -996,65 +1041,6 @@ static unsigned *TexMgr_MipMapH (unsigned *data, int width, int height, int dept
 	return data;
 }
 
-/*
-================
-TexMgr_ResampleTexture -- bilinear resample
-================
-*/
-static unsigned *TexMgr_ResampleTexture (unsigned *in, int inwidth, int inheight, qboolean alpha)
-{
-	byte *nwpx, *nepx, *swpx, *sepx, *dest;
-	unsigned xfrac, yfrac, x, y, modx, mody, imodx, imody, injump, outjump;
-	unsigned *out;
-	int i, j, outwidth, outheight;
-
-	if (inwidth == TexMgr_Pad(inwidth) && inheight == TexMgr_Pad(inheight))
-		return in;
-
-	outwidth = TexMgr_Pad(inwidth);
-	outheight = TexMgr_Pad(inheight);
-	out = (unsigned *) Hunk_Alloc(outwidth*outheight*4);
-
-	xfrac = ((inwidth-1) << 16) / (outwidth-1);
-	yfrac = ((inheight-1) << 16) / (outheight-1);
-	y = outjump = 0;
-
-	for (i = 0; i < outheight; i++)
-	{
-		mody = (y>>8) & 0xFF;
-		imody = 256 - mody;
-		injump = (y>>16) * inwidth;
-		x = 0;
-
-		for (j = 0; j < outwidth; j++)
-		{
-			modx = (x>>8) & 0xFF;
-			imodx = 256 - modx;
-
-			nwpx = (byte *)(in + (x>>16) + injump);
-			nepx = nwpx + 4;
-			swpx = nwpx + inwidth*4;
-			sepx = swpx + 4;
-
-			dest = (byte *)(out + outjump + j);
-
-			dest[0] = (nwpx[0]*imodx*imody + nepx[0]*modx*imody + swpx[0]*imodx*mody + sepx[0]*modx*mody)>>16;
-			dest[1] = (nwpx[1]*imodx*imody + nepx[1]*modx*imody + swpx[1]*imodx*mody + sepx[1]*modx*mody)>>16;
-			dest[2] = (nwpx[2]*imodx*imody + nepx[2]*modx*imody + swpx[2]*imodx*mody + sepx[2]*modx*mody)>>16;
-			if (alpha)
-				dest[3] = (nwpx[3]*imodx*imody + nepx[3]*modx*imody + swpx[3]*imodx*mody + sepx[3]*modx*mody)>>16;
-			else
-				dest[3] = 255;
-
-			x += xfrac;
-		}
-		outjump += outwidth;
-		y += yfrac;
-	}
-
-	return out;
-}
-
 /*
 ===============
 TexMgr_AlphaEdgeFix
@@ -1198,7 +1184,7 @@ static unsigned *TexMgr_8to32 (byte *in, int pixels, unsigned int *usepal)
 	int i;
 	unsigned *out, *data;
 
-	out = data = (unsigned *) Hunk_Alloc(pixels*4);
+	out = data = (unsigned *) Hunk_AllocNoFill (pixels*4);
 
 	for (i = 0; i < pixels; i++)
 		*out++ = usepal[*in++];
@@ -1221,7 +1207,7 @@ static byte *TexMgr_PadImageW (byte *in, int width, int height, byte padbyte)
 
 	outwidth = TexMgr_Pad(width);
 
-	out = data = (byte *) Hunk_Alloc(outwidth*height);
+	out = data = (byte *) Hunk_AllocNoFill(outwidth*height);
 
 	for (i = 0; i < height; i++)
 	{
@@ -1250,7 +1236,7 @@ static byte *TexMgr_PadImageH (byte *in, int width, int height, byte padbyte)
 	srcpix = width * height;
 	dstpix = width * TexMgr_Pad(height);
 
-	out = data = (byte *) Hunk_Alloc(dstpix);
+	out = data = (byte *) Hunk_AllocNoFill(dstpix);
 
 	for (i = 0; i < srcpix; i++)
 		*out++ = *in++;
@@ -1400,7 +1386,7 @@ static void TexMgr_LoadImage8 (gltexture_t *glt, byte *data)
 	// choose palette and padbyte
 	if (glt->flags & TEXPREF_ALPHABRIGHT)
 	{
-		usepal = gl_fullbrights.value ? d_8to24table_alphabright : d_8to24table;
+		usepal = gl_fullbrights.value ? d_8to24table_alphabright : d_8to24table_opaque;
 		padbyte = 0;
 	}
 	else if (glt->flags & TEXPREF_FULLBRIGHT)
@@ -1606,7 +1592,9 @@ void TexMgr_ReloadImage (gltexture_t *glt, int shirt, int pants)
 {
 	byte	translation[256];
 	byte	*src, *dst, *data = NULL, *translated;
-	int	mark, size, i;
+	int		mark, size, i;
+	enum srcformat fmt = glt->source_format;
+
 //
 // get source data
 //
@@ -1615,6 +1603,7 @@ void TexMgr_ReloadImage (gltexture_t *glt, int shirt, int pants)
 	if (glt->source_file[0] && glt->source_offset) {
 		//lump inside file
 		FILE *f;
+		int sz;
 		COM_FOpenFile(glt->source_file, &f, NULL);
 		if (!f) goto invalid;
 		fseek (f, glt->source_offset, SEEK_CUR);
@@ -1626,16 +1615,16 @@ void TexMgr_ReloadImage (gltexture_t *glt, int shirt, int pants)
 		else if (glt->source_format == SRC_LIGHTMAP) {
 			size *= lightmap_bytes;
 		}
-		data = (byte *) Hunk_Alloc (size);
-		if (fread (data, size, 1, f) != 1)
-		{
-			fclose (f);
-			goto invalid;
-		}
+		data = (byte *) Hunk_AllocNoFill (size);
+		sz = (int) fread (data, 1, size, f);
 		fclose (f);
+		if (sz != size) {
+			Hunk_FreeToLowMark(mark);
+			Host_Error("Read error for %s", glt->name);
+		}
 	}
 	else if (glt->source_file[0] && !glt->source_offset) {
-		data = Image_LoadImage (glt->source_file, (int *)&glt->source_width, (int *)&glt->source_height); //simple file
+		data = Image_LoadImage (glt->source_file, (int *)&glt->source_width, (int *)&glt->source_height, &fmt); //simple file
 	}
 	else if (!glt->source_file[0] && glt->source_offset) {
 		data = (byte *) glt->source_offset; //image in memory
@@ -1646,6 +1635,9 @@ invalid:	Con_Printf ("TexMgr_ReloadImage: invalid source for %s\n", glt->name);
 		return;
 	}
 
+	if (fmt != glt->source_format)
+		Sys_Error ("Texture %s changed format from %i to %i", glt->name, glt->source_format, fmt);
+
 	glt->width = glt->source_width;
 	glt->height = glt->source_height;
 //
@@ -1695,7 +1687,7 @@ invalid:	Con_Printf ("TexMgr_ReloadImage: invalid source for %s\n", glt->name);
 
 		//translate texture
 		size = glt->width * glt->height;
-		dst = translated = (byte *) Hunk_Alloc (size);
+		dst = translated = (byte *) Hunk_AllocNoFill (size);
 		src = data;
 
 		for (i = 0; i < size; i++)
