@@ -65,6 +65,9 @@ static qboolean	vid_initialized = false;
 
 static SDL_Window	*draw_context;
 static SDL_GLContext	gl_context;
+static SDL_Cursor		*cursor_arrow;
+static SDL_Cursor		*cursor_hand;
+static SDL_Cursor		*cursor_ibeam;
 
 static qboolean	vid_locked = false; //johnfitz
 static qboolean vid_changed = false;
@@ -168,14 +171,74 @@ extern cvar_t host_maxfps;
 extern cvar_t scr_showfps;
 extern cvar_t scr_pixelaspect;
 
+//==========================================================================
+//
+//  Mouse cursors
+//
+//==========================================================================
+
+static void VID_InitMouseCursors (void)
+{
+	cursor_arrow = SDL_CreateSystemCursor (SDL_SYSTEM_CURSOR_ARROW);
+	cursor_hand = SDL_CreateSystemCursor (SDL_SYSTEM_CURSOR_HAND);
+	cursor_ibeam = SDL_CreateSystemCursor (SDL_SYSTEM_CURSOR_IBEAM);
+}
+
+static void VID_FreeMouseCursors (void)
+{
+	SDL_FreeCursor (cursor_arrow);
+	SDL_FreeCursor (cursor_hand);
+	SDL_FreeCursor (cursor_ibeam);
+	cursor_arrow = NULL;
+	cursor_hand = NULL;
+	cursor_ibeam = NULL;
+}
+
+void VID_SetMouseCursor (mousecursor_t cursor)
+{
+	switch (cursor)
+	{
+	case MOUSECURSOR_DEFAULT:
+		SDL_SetCursor (cursor_arrow);
+		return;
+
+	case MOUSECURSOR_HAND:
+		SDL_SetCursor (cursor_hand);
+		return;
+
+	case MOUSECURSOR_IBEAM:
+		SDL_SetCursor (cursor_ibeam);
+		return;
+
+	default:
+		return;
+	}
+}
+
 //==========================================================================
 //
 //  HARDWARE GAMMA -- johnfitz
 //
 //==========================================================================
 
+#define MIN_GAMMA		0.1f
+
 static int fsaa;
 
+/*
+================
+VID_Gamma_f -- called when gamma changes
+================
+*/
+static void VID_Gamma_f (cvar_t *var)
+{
+	if (var->value < MIN_GAMMA)
+	{
+		Con_SafePrintf ("%s %g is too low, clamping to %g\n", var->name, var->value, MIN_GAMMA);
+		Cvar_SetValueQuick (var, MIN_GAMMA);
+	}
+}
+
 /*
 ================
 VID_Gamma_Init -- call on init
@@ -184,6 +247,7 @@ VID_Gamma_Init -- call on init
 static void VID_Gamma_Init (void)
 {
 	Cvar_RegisterVariable (&vid_gamma);
+	Cvar_SetCallback (&vid_gamma, VID_Gamma_f);
 	Cvar_RegisterVariable (&vid_contrast);
 }
 
@@ -515,7 +579,7 @@ static qboolean VID_SetMode (int width, int height, int refreshrate, qboolean fu
 // fix the leftover Alt from any Alt-Tab or the like that switched us away
 	ClearAllStates ();
 
-	Con_SafePrintf ("Video mode: %dx%dx%d Z%d S%d %dHz\n",
+	Con_SafePrintf ("Video mode: %d x %d %dbit Z%d S%d %dHz\n",
 				VID_GetCurrentWidth(),
 				VID_GetCurrentHeight(),
 				VID_GetCurrentBPP(),
@@ -696,7 +760,7 @@ VID_Test -- johnfitz -- like vid_restart, but asks for confirmation after switch
 */
 static void VID_Test (void)
 {
-	int old_width, old_height, old_refreshrate, old_bpp, old_fullscreen;
+	int old_width, old_height, old_refreshrate, old_fullscreen;
 
 	if (vid_locked || !vid_changed)
 		return;
@@ -706,7 +770,6 @@ static void VID_Test (void)
 	old_width = VID_GetCurrentWidth();
 	old_height = VID_GetCurrentHeight();
 	old_refreshrate = VID_GetCurrentRefreshRate();
-	old_bpp = VID_GetCurrentBPP();
 	old_fullscreen = VID_GetFullscreen() ? true : false;
 
 	VID_Restart ();
@@ -767,8 +830,7 @@ static void GL_Info_Completion_f (const char *partial)
 	for (i = 0; i < gl_num_extensions; i++)
 	{
 		const char *ext = (const char *) GL_GetStringiFunc (GL_EXTENSIONS, i);
-		if (Con_Match (ext, partial))
-			Con_AddToTabList (ext, partial, NULL);
+		Con_AddToTabList (ext, partial, NULL);
 	}
 }
 
@@ -838,10 +900,10 @@ static qboolean GL_FindExtension (const char *name)
 GL_BeginGroup
 =============
 */
-static qboolean gldebug = false;
+static qboolean glmarkers = false;
 void GL_BeginGroup (const char *name)
 {
-	if (gldebug)
+	if (glmarkers)
 		GL_PushDebugGroupFunc (GL_DEBUG_SOURCE_APPLICATION, 0, -1, name);
 }
 
@@ -852,7 +914,7 @@ GL_EndGroup
 */
 void GL_EndGroup (void)
 {
-	if (gldebug)
+	if (glmarkers)
 		GL_PopDebugGroupFunc ();
 }
 
@@ -954,11 +1016,14 @@ static void GL_CheckExtensions (void)
 {
 	GL_InitFunctions (gl_core_functions, true);
 
+	if (COM_CheckParm ("-glmarkers"))
+		glmarkers = true;
+
 #ifdef NDEBUG
 	if (COM_CheckParm("-gldebug"))
 #endif
 	{
-		gldebug = true;
+		glmarkers = true;
 		GL_DebugMessageCallbackFunc (&GL_DebugCallback, NULL);
 		glEnable (GL_DEBUG_OUTPUT);
 		glEnable (GL_DEBUG_OUTPUT_SYNCHRONOUS);
@@ -1043,8 +1108,6 @@ static void GL_SetStateEx (unsigned mask, unsigned force)
 
 	if (diff & GLS_MASK_BLEND)
 	{
-		extern cvar_t r_oit;
-
 		switch (mask & GLS_MASK_BLEND)
 		{
 			default:
@@ -1052,7 +1115,7 @@ static void GL_SetStateEx (unsigned mask, unsigned force)
 				glBlendFunc(GL_ONE, GL_ZERO);
 				break;
 			case GLS_BLEND_ALPHA_OIT:
-				if (r_oit.value)
+				if (R_GetEffectiveAlphaMode () == ALPHAMODE_OIT)
 				{
 					GL_BlendFunciFunc(0, GL_ONE, GL_ONE); // accum
 					GL_BlendFunciFunc(1, GL_ZERO, GL_ONE_MINUS_SRC_COLOR); // revealage
@@ -1206,10 +1269,10 @@ static void GL_Init (void)
 	gl_version = (const char *) glGetString (GL_VERSION);
 	glGetIntegerv (GL_NUM_EXTENSIONS, &gl_num_extensions);
 
-	Con_SafePrintf ("GL_VENDOR: %s\n", gl_vendor);
+	Con_SafePrintf ("GL_VENDOR:   %s\n", gl_vendor);
 	Con_SafePrintf ("GL_RENDERER: %s\n", gl_renderer);
-	Con_SafePrintf ("GL_VERSION: %s\n", gl_version);
-	
+	Con_SafePrintf ("GL_VERSION:  %s\n", gl_version);
+
 	if (gl_version == NULL || sscanf(gl_version, "%d.%d", &gl_version_major, &gl_version_minor) < 2)
 	{
 		gl_version_major = 0;
@@ -1326,6 +1389,11 @@ void	VID_Shutdown (void)
 {
 	if (vid_initialized)
 	{
+		VID_FreeMouseCursors();
+		SDL_GL_DeleteContext(gl_context);
+		gl_context = NULL;
+		SDL_DestroyWindow(draw_context);
+		draw_context = NULL;
 		SDL_QuitSubSystem(SDL_INIT_VIDEO);
 		draw_context = NULL;
 		gl_context = NULL;
@@ -1367,7 +1435,7 @@ VID_DescribeCurrentMode_f
 static void VID_DescribeCurrentMode_f (void)
 {
 	if (draw_context)
-		Con_Printf("%dx%dx%d %dHz %s\n",
+		Con_Printf("   %d x %d | %d bit | %d Hz | %s\n",
 			VID_GetCurrentWidth(),
 			VID_GetCurrentHeight(),
 			VID_GetCurrentBPP(),
@@ -1393,7 +1461,7 @@ static void VID_DescribeModes_f (void)
 		{
 			if (count > 0)
 				Con_SafePrintf ("\n");
-			Con_SafePrintf ("   %4i x %4i x %i : %i", modelist[i].width, modelist[i].height, modelist[i].bpp, modelist[i].refreshrate);
+			Con_SafePrintf ("  %5i x %4i | %2i bit | %i Hz", modelist[i].width, modelist[i].height, modelist[i].bpp, modelist[i].refreshrate);
 			lastwidth = modelist[i].width;
 			lastheight = modelist[i].height;
 			lastbpp = modelist[i].bpp;
@@ -1409,6 +1477,59 @@ static void VID_DescribeModes_f (void)
 //
 //==========================================================================
 
+/*
+================
+VID_CompleteModeField
+================
+*/
+static void VID_CompleteModeField (cvar_t *cvar, const char *partial, size_t ofs)
+{
+	int i;
+
+	#define GET_FIELD_FOR_MODE(idx)		*(int*)((uintptr_t)&modelist[idx] + ofs)
+
+	for (i = 0; i < nummodes; i++)
+	{
+		char buf[64];
+		if (i > 0 && GET_FIELD_FOR_MODE (i) == GET_FIELD_FOR_MODE (i-1))
+			continue;
+		q_snprintf (buf, sizeof (buf), "%d", GET_FIELD_FOR_MODE (i));
+		Con_AddToTabList (buf, partial, cvar->value == GET_FIELD_FOR_MODE (i) ? "current" : NULL);
+	}
+
+	#undef GET_FIELD_FOR_MODE
+}
+
+/*
+================
+VID_Width_Completion_f
+================
+*/
+static void VID_Width_Completion_f (cvar_t *cvar, const char *partial)
+{
+	VID_CompleteModeField (cvar, partial, offsetof (vmode_t, width));
+}
+
+/*
+================
+VID_Height_Completion_f
+================
+*/
+static void VID_Height_Completion_f (cvar_t *cvar, const char *partial)
+{
+	VID_CompleteModeField (cvar, partial, offsetof (vmode_t, height));
+}
+
+/*
+================
+VID_Refresh_Completion_f
+================
+*/
+static void VID_Refresh_Completion_f (cvar_t *cvar, const char *partial)
+{
+	VID_CompleteModeField (cvar, partial, offsetof (vmode_t, refreshrate));
+}
+
 /*
 =================
 VID_InitModelist
@@ -1446,7 +1567,7 @@ void	VID_Init (void)
 {
 	static char vid_center[] = "SDL_VIDEO_CENTERED=center";
 	int		p, width, height, refreshrate;
-	int		display_width, display_height, display_refreshrate, display_bpp;
+	int		display_width, display_height, display_refreshrate;
 	qboolean	fullscreen;
 	cmd_function_t	*cmd;
 	const char	*read_vars[] =
@@ -1464,7 +1585,7 @@ void	VID_Init (void)
 		"r_softemu_metric",
 		"scr_pixelaspect",
 	};
-#define num_readvars	( sizeof(read_vars)/sizeof(read_vars[0]) )
+#define num_readvars	Q_COUNTOF(read_vars)
 
 	Con_SafePrintf ("\nVideo initialization\n");
 
@@ -1488,6 +1609,10 @@ void	VID_Init (void)
 	Cvar_SetCallback (&vid_desktopfullscreen, VID_Changed_f);
 	Cvar_SetCallback (&vid_borderless, VID_Changed_f);
 
+	Cvar_SetCompletion (&vid_width, VID_Width_Completion_f);
+	Cvar_SetCompletion (&vid_height, VID_Height_Completion_f);
+	Cvar_SetCompletion (&vid_refreshrate, VID_Refresh_Completion_f);
+
 	Cvar_RegisterVariable (&gl_texture_anisotropy);
 	Cvar_SetCallback (&gl_texture_anisotropy, &TexMgr_Anisotropy_f);
 
@@ -1518,7 +1643,6 @@ void	VID_Init (void)
 		display_width = mode.w;
 		display_height = mode.h;
 		display_refreshrate = mode.refresh_rate;
-		display_bpp = SDL_BITSPERPIXEL(mode.format);
 	}
 
 	Cvar_SetValueQuick (&vid_width, (float)display_width);
@@ -1533,6 +1657,7 @@ void	VID_Init (void)
 	CFG_ReadCvarOverrides(read_vars, num_readvars);
 
 	VID_InitModelist();
+	VID_InitMouseCursors();
 
 	width = (int)vid_width.value;
 	height = (int)vid_height.value;
