@@ -55,6 +55,7 @@ typedef struct particle_s
 	byte		color;
 // drivers never touch the following fields
 	byte		type;
+	float		spawn;
 	float		die;
 	vec3_t		vel;
 	float		ramp;
@@ -102,9 +103,13 @@ extern	cvar_t	r_dynamic;
 extern	cvar_t	r_novis;
 extern	cvar_t	r_scale;
 
+extern	cvar_t	r_oit;
+extern	cvar_t	r_alphasort;
+
 extern	cvar_t	gl_clear;
 extern	cvar_t	gl_polyblend;
 extern	cvar_t	gl_nocolors;
+extern	cvar_t	gl_finish;
 
 extern	cvar_t	gl_playermip;
 
@@ -306,13 +311,27 @@ extern	qboolean	gl_anisotropy_able;
 #define OFFSET_SHOWTRIS -3
 void GL_PolygonOffset (int);
 
-typedef enum {
+typedef enum
+{
 	ZRANGE_FULL,
 	ZRANGE_VIEWMODEL,
 	ZRANGE_NEAR,
 } zrange_t;
 void GL_DepthRange (zrange_t range);
 
+typedef enum
+{
+	ALPHAMODE_BASIC,
+	ALPHAMODE_SORTED,
+	ALPHAMODE_OIT,
+
+	ALPHAMODE_COUNT
+} alphamode_t;
+
+alphamode_t R_GetAlphaMode (void);
+alphamode_t R_GetEffectiveAlphaMode (void);
+void R_SetAlphaMode (alphamode_t mode);
+
 //johnfitz -- rendering statistics
 extern int rs_brushpolys, rs_aliaspolys, rs_skypolys;
 extern int rs_dynamiclightmaps, rs_brushpasses, rs_aliaspasses, rs_skypasses;
@@ -360,6 +379,9 @@ extern qboolean r_fullbright_cheatsafe, r_lightmap_cheatsafe, r_drawworld_cheats
 extern float	map_wateralpha, map_lavaalpha, map_telealpha, map_slimealpha; //ericw
 extern float	map_fallbackalpha; //spike -- because we might want r_wateralpha to apply to teleporters while water itself wasn't watervised
 
+#define NUMVERTEXNORMALS	162
+extern const float	r_avertexnormals[NUMVERTEXNORMALS][3];
+
 //johnfitz -- fog functions called from outside gl_fog.c
 void Fog_ParseServerMessage (void);
 float *Fog_GetColor (void);
@@ -398,12 +420,15 @@ typedef struct gpuframedata_s {
 	float	skyfogdata[4];
 	vec3_t	winddir;
 	float	windphase;
+	float	screendither;
+	float	texturedither;
+	float	_padding1[2];
 	vec3_t	eyepos;
 	float	time;
 	float	zlogscale;
 	float	zlogbias;
 	int		numlights;
-	int		padding;
+	int		_padding2;
 } gpuframedata_t;
 
 extern gpulightbuffer_t r_lightbuffer;
@@ -439,6 +464,8 @@ void R_DrawSpriteModels_ShowTris (entity_t **ents, int count);
 
 entity_t **R_GetVisEntities (modtype_t type, qboolean translucent, int *outcount);
 
+void R_ClearBoundingBoxes (void);
+
 #define MAX_BMODEL_DRAWS		4096
 #define MAX_BMODEL_INSTANCES	1024
 
@@ -450,19 +477,20 @@ typedef struct bmodel_draw_indirect_s {
 	GLuint		baseInstance;
 } bmodel_draw_indirect_t;
 
-typedef struct bmodel_gpu_leaf_s {
-	vec3_t		mins;
-	GLuint		firstsurf;
-	vec3_t		maxs;
-	GLuint		surfcountsky; // bit 0=sky; bits 1..31=surfcount
-} bmodel_gpu_leaf_t;
+typedef struct bmodel_gpu_marksurf_s {
+	GLuint		packedleafsky; // bit 0=sky; bits 1..31=leafindex
+	GLuint		surfindex;
+} bmodel_gpu_marksurf_t;
 
 typedef struct bmodel_gpu_surf_s {
 	vec4_t		plane;
+	vec3_t		mins;
 	GLuint		framecount;
+	vec3_t		maxs;
 	GLuint		texnum;
 	GLuint		numedges;
 	GLuint		firstvert;
+	GLuint		padding[2];
 } bmodel_gpu_surf_t;
 
 void GL_BuildLightmaps (void);
@@ -470,6 +498,7 @@ void GL_BuildLightmaps (void);
 void GL_DeleteBModelBuffers (void);
 void GL_BuildBModelVertexBuffer (void);
 void GL_BuildBModelMarkBuffers (void);
+void GLMesh_LoadVertexBuffer (qmodel_t *m, aliashdr_t *hdr);
 void GLMesh_LoadVertexBuffers (void);
 void GLMesh_DeleteVertexBuffers (void);
 
@@ -492,13 +521,13 @@ typedef struct glprogs_s {
 	GLuint		oit_resolve[2];		// [msaa]
 
 	/* 3d */
-	GLuint		world[2][3][3];		// [OIT][dither][mode:solid/alpha test/water]
+	GLuint		world[2][3][3];		// [OIT][standard/dithered/banded][solid/alpha test/water]
 	GLuint		water[2][2];		// [OIT][dither]
 	GLuint		skystencil;
 	GLuint		skylayers[2];		// [dither]
 	GLuint		skycubemap[2][2];	// [anim][dither]
 	GLuint		skyboxside[2];		// [dither]
-	GLuint		alias[2][3][2];		// [OIT][mode:standard/dithered/noperspective][alpha test]
+	GLuint		alias[2][3][2][2];	// [OIT][mode:standard/dithered/noperspective][alpha test][md5]
 	GLuint		sprites[2];			// [dither]
 	GLuint		particles[2][2];	// [OIT][dither]
 	GLuint		debug3d;
