@@ -24,6 +24,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 #include "quakedef.h"
 #include "bgmusic.h"
+#include "steam.h"
 #include <setjmp.h>
 
 /*
@@ -80,6 +81,7 @@ cvar_t	coop = {"coop","0",CVAR_NONE};			// 0 or 1
 cvar_t	pausable = {"pausable","1",CVAR_NONE};
 
 cvar_t	developer = {"developer","0",CVAR_NONE};
+cvar_t	map_checks = {"map_checks","0",CVAR_NONE};
 
 cvar_t	temp1 = {"temp1","0",CVAR_NONE};
 
@@ -132,6 +134,21 @@ static void Max_Fps_f (cvar_t *var)
 	}
 }
 
+/*
+================
+Map_Checks_f -- called when map_checks cvar changes
+================
+*/
+static void Map_Checks_f (cvar_t *var)
+{
+	static qboolean showed_message = false;
+	if (!var->value || showed_message)
+		return;
+	showed_message = true;
+	Con_SafePrintf ("\x02Note: ");
+	Con_SafePrintf ("%s overrides gl_zfix, r_oit, and r_alphasort\n", var->name);
+}
+
 /*
 ================
 Host_EndGame
@@ -165,12 +182,12 @@ void Host_EndGame (const char *message, ...)
 
 /*
 ================
-Host_Error
+Host_ReportError
 
 This shuts down both the client and server
 ================
 */
-void Host_Error (const char *error, ...)
+void Host_ReportError (const char *error, ...)
 {
 	va_list		argptr;
 	char		string[1024];
@@ -257,10 +274,18 @@ void	Host_FindMaxClients (void)
 
 void Host_Version_f (void)
 {
-	Con_Printf ("Quake Version %1.2f\n", VERSION);
-	Con_Printf ("QuakeSpasm Version " QUAKESPASM_VER_STRING "\n");
-	Con_Printf ("Ironwail Version " IRONWAIL_VER_STRING "\n");
-	Con_Printf ("Exe: " __TIME__ " " __DATE__ " (%s %d-bit)\n", SDL_GetPlatform (), (int)sizeof(void*)*8);
+	SDL_version sdl_linked;
+	SDL_GetVersion (&sdl_linked);
+
+	Con_Printf ("\n");
+	Con_Printf ("Quake      %1.2f\n", VERSION);
+	Con_Printf ("QuakeSpasm " QUAKESPASM_VER_STRING "\n");
+	Con_Printf ("Ironwail   " IRONWAIL_VER_STRING "\n");
+	Con_Printf ("Exe        " __TIME__ " " __DATE__ "\n");
+	Con_Printf ("SDL        " Q_SDL_COMPILED_VERSION_STRING " (compiled)\n");
+	Con_Printf ("           %d.%d.%d (linked)\n", sdl_linked.major, sdl_linked.minor, sdl_linked.patch);
+	Con_Printf ("OS         %s %d-bit\n", SDL_GetPlatform (), (int)sizeof(void*)*8);
+	Con_Printf ("\n");
 }
 
 /* cvar callback functions : */
@@ -285,7 +310,9 @@ void Host_WriteConfigurationToFile (const char *name)
 // config.cfg cvars
 	if (host_initialized && !isDedicated && !host_parms->errstate)
 	{
-		f = Sys_fopen (va("%s/%s", com_gamedir, name), "w");
+		char fullname[MAX_OSPATH];
+		q_snprintf (fullname, sizeof (fullname), "%s/%s", com_gamedir, name);
+		f = Sys_fopen (fullname, "w");
 		if (!f)
 		{
 			Con_Printf ("Couldn't write %s.\n", name);
@@ -304,7 +331,9 @@ void Host_WriteConfigurationToFile (const char *name)
 
 		fclose (f);
 
-		Con_Printf ("Wrote %s.\n", name);
+		Con_SafePrintf ("Wrote ");
+		Con_LinkPrintf (fullname, "%s", name);
+		Con_SafePrintf (".\n");
 	}
 }
 
@@ -361,6 +390,8 @@ void Host_InitLocal (void)
 	Cvar_SetCallback (&noexit, Host_Callback_Notify);
 	Cvar_RegisterVariable (&skill);
 	Cvar_RegisterVariable (&developer);
+	Cvar_RegisterVariable (&map_checks);
+	Cvar_SetCallback (&map_checks, Map_Checks_f);
 	Cvar_RegisterVariable (&coop);
 	Cvar_RegisterVariable (&deathmatch);
 
@@ -609,6 +640,8 @@ not reinitialize anything.
 */
 void Host_ClearMemory (void)
 {
+	R_ClearBoundingBoxes ();
+
 	if (cl.qcvm.extfuncs.CSQC_Shutdown)
 	{
 		PR_SwitchQCVM(&cl.qcvm);
@@ -618,14 +651,13 @@ void Host_ClearMemory (void)
 	}
 
 	Con_DPrintf ("Clearing memory\n");
-	D_FlushCaches ();
 	Mod_ClearAll ();
 	Sky_ClearAll();
 	PR_ClearProgs(&sv.qcvm);
 	PR_ClearProgs(&cl.qcvm);
 /* host_hunklevel MUST be set at this point */
 	Hunk_FreeToLowMark (host_hunklevel);
-	CL_ClearSignons ();
+	cls.signon = 0; // not CL_ClearSignons()
 	memset (&sv, 0, sizeof(sv));
 
 	CL_FreeState ();
@@ -821,11 +853,12 @@ static void Host_CheckAutosave (void)
 {
 	float health_change, speed, elapsed, score;
 
-	if (!sv_autosave.value || sv_autosave_interval.value <= 0.f || svs.maxclients != 1 || sv_player->v.health <= 0.f)
+	if (!sv_autosave.value || sv_autosave_interval.value <= 0.f || svs.maxclients != 1 || sv_player->v.health <= 0.f || cl.intermission)
 		return;
 
 	if (cls.signon == SIGNONS)
 	{
+		// Track new secrets
 		if (pr_global_struct->found_secrets != sv.autosave.prev_secrets)
 		{
 			sv.autosave.prev_secrets = pr_global_struct->found_secrets;
@@ -835,47 +868,71 @@ static void Host_CheckAutosave (void)
 			sv.autosave.secret_boost = q_max (0.f, sv.autosave.secret_boost - host_frametime / 1.5f);
 	}
 
+	// Track health changes
 	if (!sv.autosave.prev_health)
 		sv.autosave.prev_health = sv_player->v.health;
 	health_change = sv_player->v.health - sv.autosave.prev_health;
-	if (health_change < 0.f && sv_player->v.health < 100.f) // megahealth decay doesn't count as getting hurt
-		sv.autosave.hurt_time = qcvm->time;
+	if (health_change < 0.f)
+		if (health_change < -3.f || sv_player->v.health < 100.f || sv_player->v.watertype == CONTENTS_SLIME || sv_player->v.watertype == CONTENTS_LAVA)
+			sv.autosave.hurt_time = qcvm->time;
 	sv.autosave.prev_health = sv_player->v.health;
 
+	// Track attacking
 	if (sv_player->v.button0)
 		sv.autosave.shoot_time = qcvm->time;
 
+	// Time spent with cheats active doesn't count
 	if (sv_player->v.movetype == MOVETYPE_NOCLIP || (int)sv_player->v.flags & (FL_GODMODE|FL_NOTARGET))
 	{
 		sv.autosave.cheat += host_frametime;
 		return;
 	}
 
+	// Don't save if the player has been hurt recently
 	if (qcvm->time - sv.autosave.hurt_time < 3.f)
 		return;
+
+	// Don't save if the player has fired recently
 	if (qcvm->time - sv.autosave.shoot_time < 3.f)
 		return;
 
+	// Only save when the player slows down a bit
 	speed = VectorLength (sv_player->v.velocity);
 	if (speed > 100.f)
 		return;
 
+	// Copper's func_void holds the player at the bottom for a bit before inflicting damage,
+	// so we can't assume it's safe to save just because we're no longer falling
+	if ((int)sv_player->v.movetype == MOVETYPE_NONE)
+		return;
+
+	// Don't save too often
 	elapsed = qcvm->time - sv.autosave.time - sv.autosave.cheat;
 	if (elapsed < 3.f)
 		return;
 
+	// Compute a normalized autosave score
+
+	// Base value is the fraction of the autosave interval already passed
 	score = elapsed / sv_autosave_interval.value;
+	// Scale down the score if health + armor is below 100 (save less often with lower health)
 	score *= q_min (100.f, (sv_player->v.health + sv_player->v.armortype * sv_player->v.armorvalue)) / 100.f;
+	// Boost the score right after picking up health
 	score += q_max (0.f, health_change) / 100.f;
+	// Lower score a bit based on speed (favor standing still/slowing down)
 	score -= (speed / 100.f) * 0.25f;
+	// Boost the score after finding a secret
 	score += sv.autosave.secret_boost * 0.25f;
+	// Boost the score after teleporting
 	score += CLAMP (0.f, 1.f - (qcvm->time - sv_player->v.teleport_time) / 1.5f, 1.f) * 0.5f;
+
+	// Only save if the score is high enough
 	if (score < 1.f)
 		return;
 
 	sv.autosave.time = qcvm->time;
 	sv.autosave.cheat = 0;
-	Cbuf_AddText (va ("save \"autosave/%s\"\n", sv.name));
+	Cbuf_AddText (va ("save \"autosave/%s\" 0\n", sv.name));
 }
 
 /*
@@ -929,6 +986,8 @@ void Host_ServerFrame (void)
 
 typedef struct summary_s {
 	struct {
+		int		players;
+		int		max_players;
 		int		skill;
 		int		monsters;
 		int		total_monsters;
@@ -938,6 +997,20 @@ typedef struct summary_s {
 	char		map[countof (cl.mapname)];
 } summary_t;
 
+/*
+==================
+CountActiveClients
+==================
+*/
+static int CountActiveClients (void)
+{
+	int i, active;
+	for (i = active = 0; i < cl.maxclients; i++)
+		if (cl.scores[i].name[0])
+			active++;
+	return active;
+}
+
 /*
 ==================
 GetGameSummary
@@ -953,6 +1026,8 @@ static void GetGameSummary (summary_t *s)
 	else
 	{
 		q_strlcpy (s->map, cl.mapname, countof (s->map));
+		s->stats.players        = CountActiveClients ();
+		s->stats.max_players    = cl.maxclients;
 		s->stats.skill          = (int) skill.value;
 		s->stats.monsters       = cl.stats[STAT_MONSTERS];
 		s->stats.total_monsters = cl.stats[STAT_TOTALMONSTERS];
@@ -969,7 +1044,7 @@ UpdateWindowTitle
 static void UpdateWindowTitle (void)
 {
 	static float timeleft = 0.f;
-	static summary_t last;
+	static summary_t last = {{-1}}; // negative value to force initial update
 	summary_t current;
 
 	timeleft -= host_frametime;
@@ -984,24 +1059,36 @@ static void UpdateWindowTitle (void)
 
 	if (current.map[0])
 	{
-		char levelname[4 * sizeof (cl.levelname)];
+		char cleanname[sizeof (cl.levelname)];
+		char utf8name[4 * sizeof (cl.levelname)];
 		char title[1024];
 
-		UTF8_FromQuake (levelname, sizeof (levelname), cl.levelname);
+		Mod_SanitizeMapDescription (cleanname, sizeof (cleanname), cl.levelname);
+
+		UTF8_FromQuake (utf8name, sizeof (utf8name), cleanname);
 		q_snprintf (title, sizeof (title),
-			levelname[0] ?
+			utf8name[0] ?
 				"%s (%s)  |  skill %d  |  %d/%d kills  |  %d/%d secrets  -  " WINDOW_TITLE_STRING :
 				"%s%s  |  skill %d  |  %d/%d kills  |  %d/%d secrets  -  " WINDOW_TITLE_STRING,
-			levelname, current.map,
+			utf8name, current.map,
 			current.stats.skill,
 			current.stats.monsters, current.stats.total_monsters,
 			current.stats.secrets, current.stats.total_secrets
 		);
 		VID_SetWindowTitle (title);
+
+		if (current.stats.max_players > 1)
+			Steam_SetStatus_Multiplayer (current.stats.players, current.stats.max_players, utf8name);
+		else
+			Steam_SetStatus_SinglePlayer (utf8name);
 	}
 	else
 	{
 		VID_SetWindowTitle (WINDOW_TITLE_STRING);
+		if (cls.state == ca_connected)
+			Steam_ClearStatus ();
+		else
+			Steam_SetStatus_Menu ();
 	}
 }
 
@@ -1065,6 +1152,43 @@ static void CL_LoadCSProgs (void)
 	}
 }
 
+/*
+==================
+Host_PrintTimes
+==================
+*/
+static void Host_PrintTimes (const double times[], const char *names[], int count, qboolean showtotal)
+{
+	char line[1024];
+	double total = 0.0;
+	int i, worst;
+
+	for (i = 0, worst = -1; i < count; i++)
+	{
+		if (worst == -1 || times[i] > times[worst])
+			worst = i;
+		total += times[i];
+	}
+
+	if (showtotal)
+		q_snprintf (line, sizeof (line), "%5.2f tot | ", total * 1000.0);
+	else
+		line[0] = '\0';
+
+	for (i = 0; i < count; i++)
+	{
+		char entry[256];
+		q_snprintf (entry, sizeof (entry), "%5.2f %s", times[i] * 1000.0, names[i]);
+		if (i == worst)
+			COM_TintString (entry, entry, sizeof (entry));
+		if (i != 0)
+			q_strlcat (line, " | ", sizeof (line));
+		q_strlcat (line, entry, sizeof (line));
+	}
+
+	Con_Printf ("%s\n", line);
+}
+
 /*
 ==================
 Host_Frame
@@ -1178,9 +1302,7 @@ void _Host_Frame (double time)
 
 	if (host_speeds.value)
 	{
-		static double pass1 = 0.0;
-		static double pass2 = 0.0;
-		static double pass3 = 0.0;
+		static double pass[3] = {0.0, 0.0, 0.0};
 		static double elapsed = 0.0;
 		static int numframes = 0;
 		static int numserverframes = 0;
@@ -1191,26 +1313,24 @@ void _Host_Frame (double time)
 
 		if (ranserver || host_speeds.value < 0.f)
 		{
-			pass1 += time1;
+			pass[0] += time1;
 			numserverframes++;
 		}
 		numframes++;
-		pass2 += time2;
-		pass3 += time3;
+		pass[1] += time2;
+		pass[2] += time3;
 		elapsed += time;
 
 		if (elapsed >= host_speeds.value * 0.375)
 		{
-			pass1 /= q_max (numserverframes, 1);
-			pass2 /= numframes;
-			pass3 /= numframes;
-			if (host_speeds.value < 0.f)
-				Con_Printf ("%5.2f tot | %5.2f server | %5.2f gfx | %5.2f snd\n",
-					(pass1+pass2+pass3)*1000.0, pass1*1000.0, pass2*1000.0, pass3*1000.0);
-			else
-				Con_Printf ("%5.2f server | %5.2f gfx | %5.2f snd\n",
-					pass1*1000.0, pass2*1000.0, pass3*1000.0);
-			pass1 = pass2 = pass3 = elapsed = 0.0;
+			const char *names[3] = {"server", "gfx", "snd"};
+			pass[0] /= q_max (numserverframes, 1);
+			pass[1] /= numframes;
+			pass[2] /= numframes;
+
+			Host_PrintTimes (pass, names, countof (pass), host_speeds.value < 0.f);
+
+			pass[0] = pass[1] = pass[2] = elapsed = 0.0;
 			numframes = numserverframes = 0;
 		}
 	}
@@ -1271,10 +1391,8 @@ void Host_Init (void)
 	if (host_parms->memsize < minimum_memory)
 		Sys_Error ("Only %4.1f megs of memory available, can't execute game", host_parms->memsize / (float)0x100000);
 
-	com_argc = host_parms->argc;
-	com_argv = host_parms->argv;
-
 	Memory_Init (host_parms->membase, host_parms->memsize);
+	AsyncQueue_Init (&async_queue, 1024);
 	Cbuf_Init ();
 	Cmd_Init ();
 	LOG_Init (host_parms);
@@ -1292,7 +1410,6 @@ void Host_Init (void)
 	Mod_Init ();
 	NET_Init ();
 	SV_Init ();
-	AsyncQueue_Init (&async_queue, 1024);
 
 	Con_Printf ("Exe: " __TIME__ " " __DATE__ " (%s %d-bit)\n", SDL_GetPlatform (), (int)sizeof(void*)*8);
 	Con_Printf ("%4.1f megabyte heap\n", host_parms->memsize/ (1024*1024.0));
@@ -1317,7 +1434,6 @@ void Host_Init (void)
 		BGM_Init();
 		Sbar_Init ();
 		CL_Init ();
-		Modlist_Init (); //johnfitz
 		ExtraMaps_Init (); //johnfitz
 		DemoList_Init (); //ericw
 		SaveList_Init ();
@@ -1333,6 +1449,15 @@ void Host_Init (void)
 	host_initialized = true;
 	Con_Printf ("\n========= Quake Initialized =========\n\n");
 
+	if (!COM_CheckParm ("-nomapchecks") && Sys_IsStartedFromMapEditor ())
+	{
+		Con_Printf (
+			"Level editing environment detected, enabling map_checks\n"
+			"(pass -nomapchecks or set map_checks to 0 to disable)\n"
+		);
+		Cvar_SetValueQuick (&map_checks, 1.f);
+	}
+
 	if (cls.state != ca_dedicated)
 	{
 		Cbuf_InsertText ("exec quake.rc\n");
@@ -1374,6 +1499,8 @@ void Host_Shutdown(void)
 // keep Con_Printf from trying to update the screen
 	scr_disabled_for_loading = true;
 
+	Steam_Shutdown ();
+
 	AsyncQueue_Destroy (&async_queue);
 
 	Host_ShutdownSave ();
