@@ -24,8 +24,10 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include "quakedef.h"
 #include "q_ctype.h"
 #include "json.h"
-
+#include <time.h>
+#ifndef WITHOUT_CURL
 #include <curl/curl.h>
+#endif
 #define MAX_URL	2048
 
 extern cvar_t	pausable;
@@ -69,12 +71,12 @@ static filelist_item_t *FileList_AddWithData (const char *name, const void *data
 
 	// ignore duplicate
 	for (item = *list; item; item = item->next)
-	{
-		if (!Q_strcmp (name, item->name))
+		if (!q_strcasecmp (name, item->name))
 			return item;
-	}
 
-	item = (filelist_item_t *) Z_Malloc(sizeof(filelist_item_t) + datasize);
+	item = (filelist_item_t *) malloc (sizeof(filelist_item_t) + datasize);
+	if (!item)
+		Sys_Error ("FileList_AddWithData: out of memory on %" SDL_PRIu64 " bytes (%s)", (uint64_t)(sizeof(filelist_item_t) + datasize), name);
 	q_strlcpy (item->name, name, sizeof(item->name));
 	if (datasize)
 	{
@@ -129,7 +131,7 @@ static void FileList_Clear (filelist_item_t **list)
 	while (*list)
 	{
 		blah = (*list)->next;
-		Z_Free(*list);
+		free (*list);
 		*list = blah;
 	}
 }
@@ -177,7 +179,7 @@ static void FileList_Print (filelist_item_t *list, const char *types[2], const c
 	filelist_item_t	*item;
 	const char		*desc;
 	char			buf[256], buf2[256];
-	char			padchar = '.' | 0x80;
+	char			padchar = QCHAR_COLORED ('.');
 	size_t			ofsdesc = list == extralevels ? maxlevelnamelen + 2 : 0;
 
 	if (substr && *substr)
@@ -569,6 +571,10 @@ typedef struct download_s
 
 static qboolean Download (const char *url, download_t *download)
 {
+#ifdef WITHOUT_CURL
+	download->error = "download support disabled at compile time.";
+	return false;
+#else
 	CURL				*curl;
 	CURLM				*multi_handle;
 	CURLMcode			mc;
@@ -602,6 +608,8 @@ static qboolean Download (const char *url, download_t *download)
 	curl_easy_setopt (curl, CURLOPT_WRITEDATA, download->write_data);
 	curl_easy_setopt (curl, CURLOPT_USE_SSL, CURLUSESSL_ALL);
 	curl_easy_setopt (curl, CURLOPT_ACCEPT_ENCODING, "");
+	curl_easy_setopt (curl, CURLOPT_FOLLOWLOCATION, 1L);
+	curl_easy_setopt (curl, CURLOPT_MAXREDIRS, 50L);
 	//curl_easy_setopt (curl, CURLOPT_VERBOSE, 1L);
 
 	mc = curl_multi_add_handle (multi_handle, curl);
@@ -642,9 +650,10 @@ done:
 	if (header)
 		curl_slist_free_all (header);
 	curl_easy_cleanup (curl);
-	curl_multi_cleanup (curl);
+	curl_multi_cleanup (multi_handle);
 
 	return !download->error && !still_running && download->response == 200;
+#endif
 }
 
 typedef struct
@@ -671,7 +680,11 @@ static SDL_Thread*		extramods_install_thread;
 const char *Modlist_GetFullName (const filelist_item_t *item)
 {
 	const modinfo_t *info = (const modinfo_t *) (item + 1);
-	return info->full_name;
+	const char *full_name = info->full_name;
+	// 2021 rerelease episode names are localized
+	if (full_name && full_name[0] == '$')
+		full_name = LOC_GetRawString (full_name);
+	return full_name;
 }
 
 const char *Modlist_GetDescription (const filelist_item_t *item)
@@ -739,7 +752,7 @@ static void Modlist_RegisterAddons (void *param)
 
 	for (entry = addons->firstchild; entry; entry = entry->next)
 	{
-		const char		*download, *gamedir, *name, *author, *date, *description, *version;
+		const char		*download, *gamedir, *name, *author, *date, *description;
 		const double	*size;
 		modinfo_t		*info;
 		filelist_item_t	*item;
@@ -755,7 +768,6 @@ static void Modlist_RegisterAddons (void *param)
 		name		= JSON_FindString (entry, "name");
 		author		= JSON_FindString (entry, "author");
 		date		= JSON_FindString (entry, "date");
-		version		= JSON_FindString (entry, "version");
 		size		= JSON_FindNumber (entry, "size");
 		description	= JSON_FindString (JSON_Find (entry, "description", JSON_OBJECT), "en");
 
@@ -784,8 +796,8 @@ static void Modlist_RegisterAddons (void *param)
 		"Add-on server status:\n"
 		"%3d add-on%s available for download\n"
 		"%3d add-on%s already installed\n\n",
-		total - installed, PLURAL (total - installed),
-		installed, PLURAL (installed)
+		PLURAL (total - installed),
+		PLURAL (installed)
 	);
 
 	extramods_json = json;
@@ -1074,6 +1086,7 @@ static void Modlist_Add (const char *name)
 	filelist_item_t	*item;
 	modinfo_t		*info;
 	int				i;
+	unsigned int	path_id;
 
 	memset (&info, 0, sizeof (info));
 	item = FileList_AddWithData (name, NULL, sizeof (*info), &modlist);
@@ -1111,6 +1124,50 @@ static void Modlist_Add (const char *name)
 		}
 	}
 
+	// look for mapdb.json file
+	if (!info->full_name)
+	{
+		char *mapdb = (char *) COM_LoadMallocFile ("mapdb.json", &path_id);
+		if (mapdb)
+		{
+			qboolean is_base_mapdb = !com_searchpaths || path_id < com_searchpaths->path_id;
+			json_t *json = JSON_Parse (mapdb);
+			free (mapdb);
+			if (json)
+			{
+				const jsonentry_t *episodes = JSON_Find (json->root, "episodes", JSON_ARRAY);
+				if (episodes)
+				{
+					const jsonentry_t *entry;
+					for (entry = episodes->firstchild; entry; entry = entry->next)
+					{
+						const char *mod_name = JSON_FindString (entry, "name");
+						const char *mod_dir = JSON_FindString (entry, "dir");
+						if (!mod_name || !mod_dir)
+							continue;
+
+						// The 2021 rerelease has a single mapdb.json file in id1 with definitions
+						// for all the included episodes (id1, hipnotic, rogue, dopa & mg1).
+						// If the mapdb file comes from a base dir we only use the episode name
+						// if the local mod dir matches the episode dir.
+						// We also perform a dir check if the name of the episode is "copper"
+						// in order to avoid showing all Copper-based mods as "Underdark Overbright"
+						// if they include Copper's mapdb.json unmodified.
+						// In all other cases we skip the dir check so that players can rename mod dirs
+						// as they please without losing their descriptions in the add-on menu.
+						if (is_base_mapdb || q_strcasecmp (mod_dir, "copper") == 0)
+							if (q_strcasecmp (mod_dir, name) != 0)
+								continue;
+
+						info->full_name = strdup (mod_name);
+						break;
+					}
+				}
+				JSON_Free (json);
+			}
+		}
+	}
+
 	// look for mod in hard-coded list
 	if (!info->full_name)
 	{
@@ -1180,7 +1237,11 @@ static void Modlist_FindLocal (void)
 				continue;
 #endif
 			if (Modlist_Check (find->name, com_basedirs[i]))
+			{
+				COM_AddGameDirectory (find->name);
 				Modlist_Add (find->name);
+				COM_ResetGameDirectories ("");
+			}
 		}
 	}
 }
@@ -1244,6 +1305,15 @@ void Modlist_ShutDown (void)
 	}
 }
 
+qboolean Modlist_IsInstalled (const char *game)
+{
+	filelist_item_t *item;
+	for (item = modlist; item; item = item->next)
+		if (q_strcasecmp (item->name, game) == 0 && Modlist_GetStatus (item) == MODSTATUS_INSTALLED)
+			return true;
+	return false;
+}
+
 //==============================================================================
 //ericw -- demo list management
 //==============================================================================
@@ -1374,10 +1444,19 @@ static qboolean SkyList_AddFile (const char *path)
 	char		skyname[MAX_QPATH];
 	size_t		len;
 
+	// Check that the file is in the right directory
+	// We need this for pak files, which are all passed to this function
+	// without any kind of path filtering
+	if (q_strncasecmp (path, prefix, sizeof (prefix) - 1) != 0)
+		return false;
+	path += sizeof (prefix) - 1;
+
+	// Only accept TGA files
 	ext = COM_FileGetExtension (path);
 	if (q_strcasecmp (ext, "tga") != 0)
 		return false;
 
+	// Check that the image has the right suffix
 	COM_StripExtension (path, skyname, sizeof (skyname));
 	len = strlen (skyname);
 	if (len < sizeof (suffix) - 1)
@@ -1387,12 +1466,8 @@ static qboolean SkyList_AddFile (const char *path)
 		return false;
 	skyname[len] = '\0';
 
-	SDL_assert (len > sizeof (prefix) - 1);
-	SDL_assert (!q_strncasecmp (skyname, prefix, sizeof (prefix) - 1));
-	if (len <= sizeof (prefix) - 1)
-		return false;
-
-	FileList_Add (skyname + (sizeof (prefix) - 1), &skylist);
+	// All ok, add skybox to the list
+	FileList_Add (skyname, &skylist);
 
 	return true;
 }
@@ -1696,6 +1771,9 @@ adapted from fteqw, originally by Alex Shadowalker
 */
 static void Host_SetPos_f(void)
 {
+	int		i, numargs;
+	float	args[6];
+
 	if (cmd_source == src_command)
 	{
 		Cmd_ForwardToServer ();
@@ -1705,19 +1783,28 @@ static void Host_SetPos_f(void)
 	if (pr_global_struct->deathmatch)
 		return;
 
-	if (Cmd_Argc() != 7 && Cmd_Argc() != 4)
+	for (i = 1, numargs = 0; i < Cmd_Argc (); i++)
+	{
+		const char *str = Cmd_Argv (i);
+		if (strcmp (str, "(") == 0 || strcmp (str, ")") == 0)
+			continue;
+		if (++numargs <= 6)
+			args[numargs - 1] = atof (str);
+	}
+
+	if (numargs != 6 && numargs != 3)
 	{
 		SV_ClientPrintf("usage:\n");
 		SV_ClientPrintf("   setpos <x> <y> <z>\n");
 		SV_ClientPrintf("   setpos <x> <y> <z> <pitch> <yaw> <roll>\n");
 		SV_ClientPrintf("current values:\n");
 		SV_ClientPrintf("   %i %i %i %i %i %i\n",
-			(int)sv_player->v.origin[0],
-			(int)sv_player->v.origin[1],
-			(int)sv_player->v.origin[2],
-			(int)sv_player->v.v_angle[0],
-			(int)sv_player->v.v_angle[1],
-			(int)sv_player->v.v_angle[2]);
+			Q_rint (sv_player->v.origin[0]),
+			Q_rint (sv_player->v.origin[1]),
+			Q_rint (sv_player->v.origin[2]),
+			Q_rint (sv_player->v.v_angle[0]),
+			Q_rint (sv_player->v.v_angle[1]),
+			Q_rint (sv_player->v.v_angle[2]));
 		return;
 	}
 
@@ -1732,19 +1819,19 @@ static void Host_SetPos_f(void)
 	sv_player->v.velocity[0] = 0;
 	sv_player->v.velocity[1] = 0;
 	sv_player->v.velocity[2] = 0;
-	
-	sv_player->v.origin[0] = atof(Cmd_Argv(1));
-	sv_player->v.origin[1] = atof(Cmd_Argv(2));
-	sv_player->v.origin[2] = atof(Cmd_Argv(3));
-	
-	if (Cmd_Argc() == 7)
+
+	sv_player->v.origin[0] = args[0];
+	sv_player->v.origin[1] = args[1];
+	sv_player->v.origin[2] = args[2];
+
+	if (numargs == 6)
 	{
-		sv_player->v.angles[0] = atof(Cmd_Argv(4));
-		sv_player->v.angles[1] = atof(Cmd_Argv(5));
-		sv_player->v.angles[2] = atof(Cmd_Argv(6));
+		sv_player->v.angles[0] = args[3];
+		sv_player->v.angles[1] = args[4];
+		sv_player->v.angles[2] = args[5];
 		sv_player->v.fixangle = 1;
 	}
-	
+
 	SV_LinkEdict (sv_player, false);
 }
 
@@ -1955,7 +2042,7 @@ Host_AutoLoad
 */
 static qboolean Host_AutoLoad (void)
 {
-	if (!sv_autoload.value || !sv.lastsave[0] || svs.maxclients != 1)
+	if (!sv_autoload.value || !sv.lastsave[0] || svs.maxclients != 1 || cl.intermission)
 		return false;
 
 	if (sv_autoload.value < 2.f)
@@ -2118,38 +2205,39 @@ Host_SavegameComment
 Writes a SAVEGAME_COMMENT_LENGTH character comment describing the current
 ===============
 */
-void Host_SavegameComment (char *text)
+void Host_SavegameComment (char text[SAVEGAME_COMMENT_LENGTH + 1])
 {
 	int		i;
 	char	*levelname;
 	char	kills[20];
-	char	*p1, *p2;
+	char	*p;
 
 	for (i = 0; i < SAVEGAME_COMMENT_LENGTH; i++)
 		text[i] = ' ';
+	text[SAVEGAME_COMMENT_LENGTH] = '\0';
 
-// Remove CR/LFs from level name to avoid broken saves, e.g. with autumn_sp map:
-// https://celephais.net/board/view_thread.php?id=60452&start=3666
 	levelname = cl.levelname[0] ? cl.levelname : cl.mapname;
-	p1 = strchr(levelname, '\n');
-	p2 = strchr(levelname, '\r');
-	if (p1 != NULL) *p1 = 0;
-	if (p2 != NULL) *p2 = 0;
 
 	i = (int) strlen(levelname);
 	if (i > 22) i = 22;
 	memcpy (text, levelname, (size_t)i);
+
+// Remove CR/LFs from level name to avoid broken saves, e.g. with autumn_sp map:
+// https://celephais.net/board/view_thread.php?id=60452&start=3666
+	while ((p = strchr(text, '\n')) != NULL)
+		*p = ' ';
+	while ((p = strchr(text, '\r')) != NULL)
+		*p = ' ';
+
 	sprintf (kills,"kills:%3i/%3i", cl.stats[STAT_MONSTERS], cl.stats[STAT_TOTALMONSTERS]);
 	memcpy (text+22, kills, strlen(kills));
+
 // convert space to _ to make stdio happy
 	for (i = 0; i < SAVEGAME_COMMENT_LENGTH; i++)
 	{
 		if (text[i] == ' ')
 			text[i] = '_';
 	}
-	if (p1 != NULL) *p1 = '\n';
-	if (p2 != NULL) *p2 = '\r';
-	text[SAVEGAME_COMMENT_LENGTH] = '\0';
 }
 
 static void Host_InvalidateSave (const char *relname)
@@ -2275,10 +2363,11 @@ Host_Savegame_f
 */
 static void Host_Savegame_f (void)
 {
-	char	relname[MAX_OSPATH];
-	char	name[MAX_OSPATH];
-	FILE	*f;
-	int		i;
+	char		relname[MAX_OSPATH];
+	char		name[MAX_OSPATH];
+	const char	*skipnotify;
+	FILE		*f;
+	int			i;
 
 	if (cmd_source != src_command)
 		return;
@@ -2330,8 +2419,13 @@ static void Host_Savegame_f (void)
 
 	q_strlcpy (relname, Cmd_Argv(1), sizeof(relname));
 	COM_AddExtension (relname, ".sav", sizeof(relname));
-	if (Cmd_Argc () < 3)
-		Con_Printf ("Saving game to %s...\n", relname);
+	q_snprintf (name, sizeof(name), "%s/%s", com_gamedir, relname);
+
+	// second argument, if present, indicates whether or not text should be printed to the notification area
+	skipnotify = (Cmd_Argc () < 3 || atof (Cmd_Argv (2))) ? "" : "[skipnotify]";
+	Con_SafePrintf ("%sSaving game to ", skipnotify);
+	Con_LinkPrintf (name, "%s%s", skipnotify, relname);
+	Con_SafePrintf ("%s...\n", skipnotify);
 
 	if (!strcmp (relname, sv.lastsave) && Host_IsSaving ())
 	{
@@ -2342,8 +2436,6 @@ static void Host_Savegame_f (void)
 		SDL_UnlockMutex (save_mutex);
 	}
 
-	q_snprintf (name, sizeof(name), "%s/%s", com_gamedir, relname);
-
 	f = Sys_fopen (name, "w");
 	if (!f)
 	{
@@ -2391,11 +2483,12 @@ static void Host_Loadgame_f (void)
 	int	entnum;
 	int	version;
 	float	spawn_parms[NUM_SPAWN_PARMS];
+	qboolean kexonly = false;
 
 	if (cmd_source != src_command)
 		return;
 
-	if (Cmd_Argc() != 2)
+	if (Cmd_Argc() < 2)
 	{
 		Con_Printf ("load <savename> : load a game\n");
 		return;
@@ -2407,6 +2500,10 @@ static void Host_Loadgame_f (void)
 		return;
 	}
 
+	// When loading a file that doesn't belong to a mod dir we only accept KEX saves
+	if (Cmd_Argc () >= 3 && q_strcasecmp (Cmd_Argv (2), "kex") == 0)
+		kexonly = true;
+
 	if (nomonsters.value)
 	{
 		Con_Warning ("\"%s\" disabled automatically.\n", nomonsters.name);
@@ -2419,6 +2516,21 @@ static void Host_Loadgame_f (void)
 	COM_AddExtension (relname, ".sav", sizeof(relname));
 
 	q_snprintf (name, sizeof(name), "%s/%s", com_gamedir, relname);
+
+	// Look for savefile in basedirs instead of gamedir
+	if (kexonly || !Sys_FileExists (name))
+	{
+		for (i = com_numbasedirs - 1; i >= 0; i--)
+		{
+			q_snprintf (name, sizeof(name), "%s/%s", com_basedirs[i], relname);
+			if (Sys_FileExists (name))
+			{
+				kexonly = true;
+				break;
+			}
+		}
+	}
+
 	if (!Sys_FileExists (name))
 	{
 		Con_Printf ("ERROR: %s not found.\n", relname);
@@ -2426,7 +2538,9 @@ static void Host_Loadgame_f (void)
 		return;
 	}
 
-	Con_Printf ("Loading game from %s...\n", relname);
+	Con_SafePrintf ("Loading game from ");
+	Con_LinkPrintf (name, "%s", relname);
+	Con_SafePrintf ("...\n");
 
 	SCR_BeginLoadingPlaque ();
 
@@ -2445,14 +2559,33 @@ static void Host_Loadgame_f (void)
 
 	data = start;
 	data = COM_ParseIntNewline (data, &version);
-	if (version != SAVEGAME_VERSION)
+	if (version == SAVEGAME_VERSION_KEX)
 	{
+		extern char com_gamenames[];
+		const char *game = *com_gamenames ? com_gamenames : GAMENAME;
+		data = COM_ParseStringNewline (data);
+		if (strcmp (game, com_token) != 0)
+		{
+			if (!Modlist_IsInstalled (com_token))
+			{
+				Con_Printf ("ERROR: mod \"%s\" is not installed.\n", com_token);
+				return;
+			}
+			COM_SwitchGame (com_token);
+			Cbuf_Execute ();
+			if (key_dest == key_menu)
+				M_ToggleMenu_f ();
+		}
+	}
+	else if (version != SAVEGAME_VERSION || kexonly)
+	{
+		int expected = kexonly ? SAVEGAME_VERSION_KEX : SAVEGAME_VERSION;
 		free (start);
 		start = NULL;
 		if (sv.autoloading)
-			Con_Printf ("ERROR: Savegame is version %i, not %i\n", version, SAVEGAME_VERSION);
+			Con_Printf ("ERROR: Savegame is version %i, not %i\n", version, expected);
 		else
-			Host_Error ("Savegame is version %i, not %i", version, SAVEGAME_VERSION);
+			Host_Error ("Savegame is version %i, not %i", version, expected);
 		Host_InvalidateSave (relname);
 		SCR_EndLoadingPlaque ();
 		return;
@@ -2469,7 +2602,10 @@ static void Host_Loadgame_f (void)
 	q_strlcpy (mapname, com_token, sizeof(mapname));
 	data = COM_ParseFloatNewline (data, &time);
 
-	CL_Disconnect_f ();
+// Note: calling CL_Disconnect instead of CL_Disconnect_f to avoid stopping the music
+	CL_Disconnect ();
+	if (sv.active)
+		Host_ShutdownServer (false);
 
 	PR_SwitchQCVM(&sv.qcvm);
 	SV_SpawnServer (mapname);
@@ -2531,6 +2667,12 @@ static void Host_Loadgame_f (void)
 		entnum++;
 	}
 
+	// Free edicts allocated during map loading but no longer used after restoring saved game state
+	// Note: we use ED_ClearEdict instead of ED_Free to avoid placing entities >= num_edicts in the free list
+	// This is different from QuakeSpasm, which doesn't use a free list
+	for (i = entnum; i < qcvm->num_edicts; i++)
+		ED_ClearEdict (EDICT_NUM (i));
+
 	qcvm->num_edicts = entnum;
 	qcvm->time = time;
 	sv.autosave.time = time;
@@ -2550,6 +2692,9 @@ static void Host_Loadgame_f (void)
 		CL_EstablishConnection ("local");
 		Host_Reconnect_f ();
 	}
+
+	if (cls.state != ca_dedicated && key_dest == key_game)
+		IN_Activate(); // moved to here from M_Load_Key()
 }
 
 //============================================================================
@@ -2835,7 +2980,6 @@ static void Host_Pause_f (void)
 	if (cls.demoplayback)
 	{
 		cls.demopaused = !cls.demopaused;
-		cl.paused = cls.demopaused;
 		return;
 	}
 
@@ -3009,7 +3153,7 @@ static void Host_Spawn_f (void)
 
 	MSG_WriteByte (&host_client->message, svc_signonnum);
 	MSG_WriteByte (&host_client->message, 3);
-	host_client->sendsignon = true;
+	host_client->sendsignon = PRESPAWN_FLUSH;
 }
 
 /*
