@@ -30,9 +30,10 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #define		HISTORY_FILE_NAME "history.txt"
 
 char		key_lines[CMDLINES][MAXCMDLINE];
+char		key_tabhint[MAXCMDLINE];
 
 int		key_linepos;
-int		key_insert;	//johnfitz -- insert key toggle (for editing)
+int		key_insert = 1;	//johnfitz -- insert key toggle (for editing)
 double		key_blinktime; //johnfitz -- fudge cursor blinking to make it easier to spot in certain cases
 
 int		edit_line = 0;
@@ -51,7 +52,7 @@ typedef struct
 	int		keynum;
 } keyname_t;
 
-keyname_t keynames[] =
+static const keyname_t keynames[] =
 {
 	{"TAB", K_TAB},
 	{"ENTER", K_ENTER},
@@ -119,44 +120,6 @@ keyname_t keynames[] =
 	{"MOUSE4", K_MOUSE4},
 	{"MOUSE5", K_MOUSE5},
 
-	{"JOY1", K_JOY1},
-	{"JOY2", K_JOY2},
-	{"JOY3", K_JOY3},
-	{"JOY4", K_JOY4},
-
-	{"AUX1", K_AUX1},
-	{"AUX2", K_AUX2},
-	{"AUX3", K_AUX3},
-	{"AUX4", K_AUX4},
-	{"AUX5", K_AUX5},
-	{"AUX6", K_AUX6},
-	{"AUX7", K_AUX7},
-	{"AUX8", K_AUX8},
-	{"AUX9", K_AUX9},
-	{"AUX10", K_AUX10},
-	{"AUX11", K_AUX11},
-	{"AUX12", K_AUX12},
-	{"AUX13", K_AUX13},
-	{"AUX14", K_AUX14},
-	{"AUX15", K_AUX15},
-	{"AUX16", K_AUX16},
-	{"AUX17", K_AUX17},
-	{"AUX18", K_AUX18},
-	{"AUX19", K_AUX19},
-	{"AUX20", K_AUX20},
-	{"AUX21", K_AUX21},
-	{"AUX22", K_AUX22},
-	{"AUX23", K_AUX23},
-	{"AUX24", K_AUX24},
-	{"AUX25", K_AUX25},
-	{"AUX26", K_AUX26},
-	{"AUX27", K_AUX27},
-	{"AUX28", K_AUX28},
-	{"AUX29", K_AUX29},
-	{"AUX30", K_AUX30},
-	{"AUX31", K_AUX31},
-	{"AUX32", K_AUX32},
-
 	{"PAUSE", K_PAUSE},
 
 	{"MWHEELUP", K_MWHEELUP},
@@ -171,16 +134,52 @@ keyname_t keynames[] =
 	{"RTHUMB", K_RTHUMB},
 	{"LSHOULDER", K_LSHOULDER},
 	{"RSHOULDER", K_RSHOULDER},
+	{"DPAD_UP", K_DPAD_UP},
+	{"DPAD_DOWN", K_DPAD_DOWN},
+	{"DPAD_LEFT", K_DPAD_LEFT},
+	{"DPAD_RIGHT", K_DPAD_RIGHT},
 	{"ABUTTON", K_ABUTTON},
 	{"BBUTTON", K_BBUTTON},
 	{"XBUTTON", K_XBUTTON},
 	{"YBUTTON", K_YBUTTON},
 	{"LTRIGGER", K_LTRIGGER},
 	{"RTRIGGER", K_RTRIGGER},
+	{"MISC1", K_MISC1},
+	{"PADDLE1", K_PADDLE1},
+	{"PADDLE2", K_PADDLE2},
+	{"PADDLE3", K_PADDLE3},
+	{"PADDLE4", K_PADDLE4},
+	{"TOUCHPAD", K_TOUCHPAD},
+
+	// Gamepad "Start" and "Back" buttons are always mapped to ESC/TAB.
+	// We don't expose the key names in order to avoid having the player bind them in the console.
+	//{"JOY_START", K_START},
+	//{"JOY_BACK", K_BACK},
 
 	{NULL,		0}
 };
 
+static const char *const xbox_names[K_GAMEPAD_COUNT] =
+{
+	#define GAMEPAD_KEY_NAME(keycode, value, xboxname, psname, nintendoname) xboxname,
+	GAMEPAD_KEY_LIST (GAMEPAD_KEY_NAME)
+	#undef GAMEPAD_KEY_NAME
+};
+
+static const char *const ps_names[K_GAMEPAD_COUNT] =
+{
+	#define GAMEPAD_KEY_NAME(keycode, value, xboxname, psname, nintendoname) psname,
+	GAMEPAD_KEY_LIST (GAMEPAD_KEY_NAME)
+	#undef GAMEPAD_KEY_NAME
+};
+
+static const char *const nintendo_names[K_GAMEPAD_COUNT] =
+{
+	#define GAMEPAD_KEY_NAME(keycode, value, xboxname, psname, nintendoname) nintendoname,
+	GAMEPAD_KEY_LIST (GAMEPAD_KEY_NAME)
+	#undef GAMEPAD_KEY_NAME
+};
+
 /*
 ==============================================================================
 
@@ -240,6 +239,7 @@ static qboolean Key_IsWordSeparator (char c)
 	switch (c)
 	{
 	case ' ':
+	case '_':
 	case '\t':
 	case ';':
 		return true;
@@ -293,6 +293,7 @@ void Key_Console (int key)
 	{
 	case K_ENTER:
 	case K_KP_ENTER:
+	case K_ABUTTON:
 		key_tabpartial[0] = 0;
 		Cbuf_AddText (workline + 1);	// skip the prompt
 		Cbuf_AddText ("\n");
@@ -307,12 +308,13 @@ void Key_Console (int key)
 		key_lines[edit_line][0] = ']';
 		key_lines[edit_line][1] = 0; //johnfitz -- otherwise old history items show up in the new edit line
 		key_linepos = 1;
+		key_tabhint[0] = '\0';
 		if (cls.state == ca_disconnected)
 			SCR_UpdateScreen (); // force an update, because the command may take some time
 		return;
 
 	case K_TAB:
-		Con_TabComplete ();
+		Con_TabComplete (TABCOMPLETE_USER);
 		return;
 
 	case K_BACKSPACE:
@@ -326,6 +328,7 @@ void Key_Console (int key)
 			SDL_assert ((int)len >= numchars);
 			memmove (workline, workline + numchars, len + 1 - numchars);
 			key_linepos -= numchars;
+			Con_TabComplete (TABCOMPLETE_AUTOHINT);
 		}
 		return;
 
@@ -339,6 +342,7 @@ void Key_Console (int key)
 			len = strlen (workline);
 			SDL_assert ((int)len >= numchars);
 			memmove (workline, workline + numchars, len + 1 - numchars);
+			Con_TabComplete (TABCOMPLETE_AUTOHINT);
 		}
 		return;
 
@@ -363,29 +367,30 @@ void Key_Console (int key)
 			con_backscroll = CLAMP(0, con_current-i%con_totallines-2, con_totallines-(glheight>>3)-1);
 		}
 		else	key_linepos = 1;
+		Con_TabComplete (TABCOMPLETE_AUTOHINT);
+		Con_ForceMouseMove ();
 		return;
 
 	case K_END:
 		if (keydown[K_CTRL])
 			con_backscroll = 0;
 		else	key_linepos = strlen(workline);
+		Con_TabComplete (TABCOMPLETE_AUTOHINT);
+		Con_ForceMouseMove ();
 		return;
 
 	case K_PGUP:
 	case K_MWHEELUP:
-		con_backscroll += keydown[K_CTRL] ? ((con_vislines>>3) - 4) : 2;
-		if (con_backscroll > con_totallines - (vid.height>>3) - 1)
-			con_backscroll = con_totallines - (vid.height>>3) - 1;
+		Con_Scroll (keydown[K_CTRL] ? ((con_vislines>>3) - 4) : 2);
 		return;
 
 	case K_PGDN:
 	case K_MWHEELDOWN:
-		con_backscroll -= keydown[K_CTRL] ? ((con_vislines>>3) - 4) : 2;
-		if (con_backscroll < 0)
-			con_backscroll = 0;
+		Con_Scroll (keydown[K_CTRL] ? -((con_vislines>>3) - 4) : -2);
 		return;
 
 	case K_LEFTARROW:
+	case K_DPAD_LEFT:
 		if (key_linepos > 1)
 		{
 			if (keydown[K_CTRL])
@@ -393,10 +398,12 @@ void Key_Console (int key)
 			else
 				key_linepos--;
 			key_blinktime = realtime;
+			Con_TabComplete (TABCOMPLETE_AUTOHINT);
 		}
 		return;
 
 	case K_RIGHTARROW:
+	case K_DPAD_RIGHT:
 		len = strlen(workline);
 		if ((int)len == key_linepos)
 		{
@@ -416,9 +423,11 @@ void Key_Console (int key)
 				key_linepos++;
 			key_blinktime = realtime;
 		}
+		Con_TabComplete (TABCOMPLETE_AUTOHINT);
 		return;
 
 	case K_UPARROW:
+	case K_DPAD_UP:
 		if (history_line == edit_line)
 			Q_strcpy(current, workline);
 
@@ -438,9 +447,11 @@ void Key_Console (int key)
 		len = strlen(key_lines[history_line]);
 		memmove(workline, key_lines[history_line], len+1);
 		key_linepos = (int)len;
+		Con_TabComplete (TABCOMPLETE_AUTOHINT);
 		return;
 
 	case K_DOWNARROW:
+	case K_DPAD_DOWN:
 		if (history_line == edit_line)
 			return;
 
@@ -462,12 +473,19 @@ void Key_Console (int key)
 			memmove(workline, key_lines[history_line], len+1);
 		}
 		key_linepos = (int)len;
+		Con_TabComplete (TABCOMPLETE_AUTOHINT);
 		return;
 
 	case K_INS:
 		if (keydown[K_SHIFT])		/* Shift-Ins paste */
 			PasteToConsole();
+		else if (keydown[K_CTRL])
+		{
+			Con_CopySelectionToClipboard ();
+			return;
+		}
 		else	key_insert ^= 1;
+		Con_TabComplete (TABCOMPLETE_AUTOHINT);
 		return;
 
 	case 'v':
@@ -475,11 +493,21 @@ void Key_Console (int key)
 #if defined(PLATFORM_OSX) || defined(PLATFORM_MAC)
 		if (keydown[K_COMMAND]) {	/* Cmd+v paste (Mac-only) */
 			PasteToConsole();
+			Con_TabComplete (TABCOMPLETE_AUTOHINT);
 			return;
 		}
 #endif
 		if (keydown[K_CTRL]) {		/* Ctrl+v paste */
 			PasteToConsole();
+			Con_TabComplete (TABCOMPLETE_AUTOHINT);
+			return;
+		}
+		break;
+
+	case 'a':
+	case 'A':
+		if (keydown[K_CTRL]) {		/* Ctrl+A: select whole buffer */
+			Con_SelectAll();
 			return;
 		}
 		break;
@@ -487,11 +515,14 @@ void Key_Console (int key)
 	case 'c':
 	case 'C':
 		if (keydown[K_CTRL]) {		/* Ctrl+C: abort the line -- S.A */
+			if (Con_CopySelectionToClipboard ())
+				return;
 			Con_Printf ("%s\n", workline);
 			workline[0] = ']';
 			workline[1] = 0;
 			key_linepos = 1;
 			history_line= edit_line;
+			key_tabhint[0] = '\0';
 			return;
 		}
 		break;
@@ -514,7 +545,13 @@ void Char_Console (int key)
 			workline[MAXCMDLINE - 2] = 0;
 			workline += key_linepos;
 			len = strlen(workline) + 1;
+			#if defined(__GNUC__) && (__GNUC__ > 7)
+			#pragma GCC diagnostic ignored "-Warray-bounds"
+			#endif
 			memmove (workline + 1, workline, len);
+			#if defined(__GNUC__) && (__GNUC__ > 7)
+			#pragma GCC diagnostic pop
+			#endif
 			*workline = key;
 		}
 		else
@@ -526,6 +563,8 @@ void Char_Console (int key)
 				workline[1] = 0;
 		}
 		key_linepos++;
+
+		Con_TabComplete (TABCOMPLETE_AUTOHINT);
 	}
 }
 
@@ -602,7 +641,7 @@ the K_* names are matched up.
 */
 int Key_StringToKeynum (const char *str)
 {
-	keyname_t	*kn;
+	const keyname_t *kn;
 
 	if (!str || !str[0])
 		return -1;
@@ -629,7 +668,7 @@ FIXME: handle quote special (general escape sequence?)
 const char *Key_KeynumToString (int keynum)
 {
 	static	char	tinystr[128][2];
-	keyname_t	*kn;
+	const keyname_t	*kn;
 
 	if (keynum == -1)
 		return "<KEY NOT FOUND>";
@@ -649,6 +688,34 @@ const char *Key_KeynumToString (int keynum)
 	return "<UNKNOWN KEYNUM>";
 }
 
+/*
+===================
+Key_KeynumToFriendlyString
+
+Returns a user-facing string for the given keynum.
+===================
+*/
+const char *Key_KeynumToFriendlyString (int keynum)
+{
+	if (keynum >= K_GAMEPAD_BEGIN && keynum < K_GAMEPAD_END)
+	{
+		const char *str = NULL;
+
+		switch (IN_GetGamepadType ())
+		{
+		default:
+		case GAMEPAD_XBOX:			str = xbox_names		[keynum - K_GAMEPAD_BEGIN]; break;
+		case GAMEPAD_PLAYSTATION:	str = ps_names			[keynum - K_GAMEPAD_BEGIN]; break;
+		case GAMEPAD_NINTENDO:		str = nintendo_names	[keynum - K_GAMEPAD_BEGIN]; break;
+		}
+
+		if (str && *str)
+			return str;
+	}
+
+	return Key_KeynumToString (keynum);
+}
+
 
 /*
 ===================
@@ -672,12 +739,39 @@ void Key_SetBinding (int keynum, const char *binding)
 		keybindings[keynum] = Z_Strdup(binding);
 }
 
+/*
+===================
+Key_GetDeviceForKeynum
+===================
+*/
+keydevice_t Key_GetDeviceForKeynum (int keynum)
+{
+	if (keynum < 0 || keynum >= NUM_KEYCODES)
+		return KD_NONE;
+	if (keynum >= K_MOUSE_BEGIN && keynum < K_MOUSE_END)
+		return KD_MOUSE;
+	if (keynum >= K_GAMEPAD_BEGIN && keynum < K_GAMEPAD_END)
+		return KD_GAMEPAD;
+	return KD_KEYBOARD;
+}
+
+/*
+===================
+Key_GetDeviceMaskForKeynum
+===================
+*/
+keydevicemask_t Key_GetDeviceMaskForKeynum (int keynum)
+{
+	keydevice_t device = Key_GetDeviceForKeynum (keynum);
+	return device == KD_NONE ? KDM_NONE : (keydevicemask_t) (1 << (int)device);
+}
+
 /*
 ===================
 Key_GetKeysForCommand
 ===================
 */
-int Key_GetKeysForCommand (const char *command, int *keys, int maxkeys)
+int Key_GetKeysForCommand (const char *command, int *keys, int maxkeys, keydevicemask_t devmask)
 {
 	int i, count;
 
@@ -692,6 +786,8 @@ int Key_GetKeysForCommand (const char *command, int *keys, int maxkeys)
 	{
 		if (keybindings[i] && !strcmp (keybindings[i], command))
 		{
+			if ((Key_GetDeviceMaskForKeynum (i) & devmask) == 0)
+				continue;
 			keys[count++] = i;
 			if (count == maxkeys)
 				break;
@@ -920,6 +1016,10 @@ void Key_Init (void)
 	consolekeys[K_DOWNARROW] = true;
 	consolekeys[K_LEFTARROW] = true;
 	consolekeys[K_RIGHTARROW] = true;
+	consolekeys[K_DPAD_UP] = true;
+	consolekeys[K_DPAD_DOWN] = true;
+	consolekeys[K_DPAD_LEFT] = true;
+	consolekeys[K_DPAD_RIGHT] = true;
 	consolekeys[K_CTRL] = true;
 	consolekeys[K_SHIFT] = true;
 	consolekeys[K_INS] = true;
@@ -945,11 +1045,13 @@ void Key_Init (void)
 	consolekeys[K_KP_ENTER] = true;
 	consolekeys[K_KP_INS] = true;
 	consolekeys[K_KP_DEL] = true;
+	consolekeys[K_MOUSE1] = true;
 #if defined(PLATFORM_OSX) || defined(PLATFORM_MAC)
 	consolekeys[K_COMMAND] = true;
 #endif
 	consolekeys[K_MWHEELUP] = true;
 	consolekeys[K_MWHEELDOWN] = true;
+	consolekeys[K_ABUTTON] = true;
 
 //
 // initialize menubound[]
@@ -1026,6 +1128,21 @@ Should NOT be called during an interrupt!
 ===================
 */
 void Key_Event (int key, qboolean down)
+{
+	Key_EventWithKeycode (key, down, 0);
+}
+
+/*
+===================
+Key_EventWithKeycode
+
+Called by the system between frames for both key up and key down events
+Should NOT be called during an interrupt!
+keycode parameter should have the key's actual keycode using the current keyboard layout,
+not necessarily the US-keyboard-based scancode. Pass 0 if not applicable.
+===================
+*/
+void Key_EventWithKeycode (int key, qboolean down, int keycode)
 {
 	char	*kb;
 	char	cmd[1024];
@@ -1049,13 +1166,13 @@ void Key_Event (int key, qboolean down)
 			if (key_dest == key_game && !con_forcedup)
 				return; // ignore autorepeats in game mode
 		}
-		else if (key >= 200 && !keybindings[key] && key_dest != key_menu)
+		else if (key >= 200 && !keybindings[key] && key_dest == key_game && !cls.demoplayback)
 		{
 			int optkey;
-			if (Key_GetKeysForCommand ("menu_options", &optkey, 1))
-				Con_Printf ("%s is unbound, hit %s to set.\n", Key_KeynumToString (key), Key_KeynumToString (optkey));
+			if (Key_GetKeysForCommand ("menu_options", &optkey, 1, KDM_ANY))
+				Con_Printf ("%s is unbound, hit %s to set.\n", Key_KeynumToFriendlyString (key), Key_KeynumToFriendlyString (optkey));
 			else
-				Con_Printf ("%s is unbound, use Options menu to set.\n", Key_KeynumToString (key));
+				Con_Printf ("%s is unbound, use Options menu to set.\n", Key_KeynumToFriendlyString (key));
 		}
 	}
 	else if (!keydown[key])
@@ -1067,10 +1184,20 @@ void Key_Event (int key, qboolean down)
 	if (key_inputgrab.active)
 	{
 		if (down)
+		{
 			key_inputgrab.lastkey = key;
+			if (keycode > 0)
+				key_inputgrab.lastchar = keycode;
+		}
 		return;
 	}
 
+// generate char events if we want text input without popping up an on-screen keyboard
+// when a physical one isn't present, e.g. when using a searchable menu on the Steam Deck.
+// Note: shift modifier is currently not supported for emulated char events.
+	if (down && !keydown[K_SHIFT] && IN_EmulatedCharEvents ())
+		Char_Event (keycode);
+
 // handle escape specialy, so the user can never unbind it
 	if (key == K_ESCAPE)
 	{
@@ -1110,6 +1237,59 @@ void Key_Event (int key, qboolean down)
 		return;
 	}
 
+// demo controls
+	if (cls.demoplayback && key_dest == key_game)
+	{
+		switch (key)
+		{
+		case K_SPACE:
+		case K_YBUTTON:
+			// Pause
+			if (down > wasdown)
+				cls.demopaused = !cls.demopaused;
+			return;
+
+		case K_UPARROW:
+		case K_DPAD_UP:
+			// Resume/increase speed
+			if (down > wasdown)
+			{
+				if (!cls.demopaused)
+					cls.basedemospeed = CLAMP (0.25f, cls.basedemospeed * 2.f, 8.f);
+				cls.demopaused = false;
+			}
+			return;
+
+		case K_DOWNARROW:
+		case K_DPAD_DOWN:
+			// Decrease speed/pause
+			if (down > wasdown)
+			{
+				cls.basedemospeed *= 0.5f;
+				if (cls.basedemospeed < 0.25f)
+				{
+					cls.basedemospeed = 0.25f;
+					cls.demopaused = true;
+				}
+			}
+			return;
+
+		case K_LEFTARROW:
+		case K_RIGHTARROW:
+		case K_DPAD_LEFT:
+		case K_DPAD_RIGHT:
+		case K_SHIFT:
+		case K_CTRL:
+			// Temporary modifiers: they don't perform their actions on up/down events, but are queried per frame instead
+			// to avoid having to manage state transitions (e.g. pressing esc while still holding left arrow to rewind).
+			return;
+
+		default:
+			// Not a demo control key
+			break;
+		}
+	}
+
 // key up events only generate commands if the game key binding is
 // a button command (leading + sign).  These will occur even in console mode,
 // to keep the character from continuing an action started before a console
@@ -1120,7 +1300,7 @@ void Key_Event (int key, qboolean down)
 		kb = keybindings[key];
 		if (kb && kb[0] == '+')
 		{
-			sprintf (cmd, "-%s %i\n", kb+1, key);
+			q_snprintf (cmd, sizeof (cmd), "-%s %i\n", kb+1, key);
 			Cbuf_AddText (cmd);
 		}
 		return;
@@ -1134,7 +1314,7 @@ void Key_Event (int key, qboolean down)
 	}
 
 // if not a consolekey, send to the interpreter no matter what mode is
-	if ((key_dest == key_menu && menubound[key]) ||
+	if ((key_dest == key_menu && menubound[key] && !M_WaitingForKeyBinding ()) ||
 	    (key_dest == key_console && !consolekeys[key]) ||
 	    (key_dest == key_game && (!con_forcedup || !consolekeys[key])))
 	{
@@ -1143,7 +1323,7 @@ void Key_Event (int key, qboolean down)
 		{
 			if (kb[0] == '+')
 			{	// button commands add keynum as a parm
-				sprintf (cmd, "%s %i\n", kb, key);
+				q_snprintf (cmd, sizeof (cmd), "%s %i\n", kb, key);
 				Cbuf_AddText (cmd);
 			}
 			else
@@ -1226,25 +1406,30 @@ void Char_Event (int key)
 Key_TextEntry
 ===================
 */
-qboolean Key_TextEntry (void)
+textmode_t Key_TextEntry (void)
 {
 	if (key_inputgrab.active)
-		return true;
+	{
+		// This path is used for simple single-letter inputs (y/n prompts) that also
+		// accept controller input, so we don't want an onscreen keyboard for this case.
+		return TEXTMODE_NOPOPUP;
+	}
 
 	switch (key_dest)
 	{
 	case key_message:
-		return true;
+		return TEXTMODE_ON;
 	case key_menu:
 		return M_TextEntry();
 	case key_game:
-		if (!con_forcedup)
-			return false;
-		/* fallthrough */
+		// Don't return true even during con_forcedup, because that happens while starting a
+		// game and we don't to trigger text input (and the onscreen keyboard on some devices)
+		// during this.
+		return TEXTMODE_OFF;
 	case key_console:
-		return true;
+		return TEXTMODE_ON;
 	default:
-		return false;
+		return TEXTMODE_OFF;
 	}
 }
 
