@@ -23,7 +23,9 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 #include "quakedef.h"
 
-int		type_size[8] = {
+extern edict_t **bbox_linked;
+
+const int type_size[NUM_TYPE_SIZES] = {
 	1,					// ev_void
 	1,	// sizeof(string_t) / 4		// ev_string
 	1,					// ev_float
@@ -177,8 +179,11 @@ Sets everything to NULL
 */
 void ED_ClearEdict (edict_t *e)
 {
+	if (!e->free)
+		SV_UnlinkEdict (e);
+	else
+		ED_RemoveFromFreeList (e);
 	memset (&e->v, 0, qcvm->progs->entityfields * 4);
-	ED_RemoveFromFreeList (e);
 }
 
 /*
@@ -258,16 +263,14 @@ ED_GlobalAtOfs
 */
 static ddef_t *ED_GlobalAtOfs (int ofs)
 {
-	ddef_t		*def;
-	int			i;
+	if (ofs < 0 || ofs > qcvm->maxglobalofs)
+		return NULL;
 
-	for (i = 0; i < qcvm->progs->numglobaldefs; i++)
-	{
-		def = &qcvm->globaldefs[i];
-		if (def->ofs == ofs)
-			return def;
-	}
-	return NULL;
+	ofs = qcvm->ofstoglobal[ofs];
+	if (ofs < 0)
+		return NULL;
+
+	return &qcvm->globaldefs[ofs];
 }
 
 /*
@@ -277,16 +280,14 @@ ED_FieldAtOfs
 */
 static ddef_t *ED_FieldAtOfs (int ofs)
 {
-	ddef_t		*def;
-	int			i;
+	if (ofs < 0 || ofs > qcvm->maxfieldofs)
+		return NULL;
 
-	for (i = 0; i < qcvm->progs->numfielddefs; i++)
-	{
-		def = &qcvm->fielddefs[i];
-		if (def->ofs == ofs)
-			return def;
-	}
-	return NULL;
+	ofs = qcvm->ofstofield[ofs];
+	if (ofs < 0)
+		return NULL;
+
+	return &qcvm->fielddefs[ofs];
 }
 
 /*
@@ -408,6 +409,15 @@ eval_t *GetEdictFieldValueByName(edict_t *ed, const char *name)
 	return GetEdictFieldValue(ed, ED_FindFieldOffset(name));
 }
 
+/*
+============
+PR_FloatFormat
+============
+*/
+static const char *PR_FloatFormat (float f)
+{
+	return fabs (f - round (f)) < 0.05f ? "% 5.0f  " : "% 7.1f";
+}
 
 /*
 ============
@@ -420,8 +430,11 @@ Returns a string describing *data in a type specific manner
 static const char *PR_ValueString (int type, eval_t *val)
 {
 	static char	line[512];
+	char		fmt[64];
+	const char	*str;
 	ddef_t		*def;
 	dfunction_t	*f;
+	edict_t		*ed;
 
 	type &= ~DEF_SAVEGLOBAL;
 
@@ -431,7 +444,9 @@ static const char *PR_ValueString (int type, eval_t *val)
 		q_snprintf (line, sizeof(line), "%s", PR_GetString(val->string));
 		break;
 	case ev_entity:
-		q_snprintf (line, sizeof(line), "entity %i", NUM_FOR_EDICT(PROG_TO_EDICT(val->edict)) );
+		ed = PROG_TO_EDICT(val->edict);
+		str = PR_GetString(ed->v.classname);
+		q_snprintf (line, sizeof(line), *str ? "entity %i (%s)" : "entity %i", NUM_FOR_EDICT(ed), PR_GetString(ed->v.classname));
 		break;
 	case ev_function:
 		f = qcvm->functions + val->function;
@@ -445,10 +460,13 @@ static const char *PR_ValueString (int type, eval_t *val)
 		q_snprintf (line, sizeof(line), "void");
 		break;
 	case ev_float:
-		q_snprintf (line, sizeof(line), "%5.1f", val->_float);
+		// Note: leading space, so that float fields are aligned with the first value in vector fields
+		q_snprintf (fmt, sizeof(fmt), " %s", PR_FloatFormat (val->_float));
+		q_snprintf (line, sizeof(line), fmt, val->_float);
 		break;
 	case ev_vector:
-		q_snprintf (line, sizeof(line), "'%5.1f %5.1f %5.1f'", val->vector[0], val->vector[1], val->vector[2]);
+		q_snprintf (fmt, sizeof(fmt), "'%s %s %s'", PR_FloatFormat (val->vector[0]), PR_FloatFormat (val->vector[1]), PR_FloatFormat (val->vector[2]));
+		q_snprintf (line, sizeof(line), fmt, val->vector[0], val->vector[1], val->vector[2]);
 		break;
 	case ev_pointer:
 		q_snprintf (line, sizeof(line), "pointer");
@@ -599,6 +617,7 @@ padded to 20 field width
 const char *PR_GlobalString (int ofs)
 {
 	static char	line[512];
+	static const int lastchari = Q_COUNTOF(line) - 2;
 	const char	*s;
 	int		i;
 	ddef_t		*def;
@@ -617,7 +636,11 @@ const char *PR_GlobalString (int ofs)
 	i = strlen(line);
 	for ( ; i < 20; i++)
 		strcat (line, " ");
-	strcat (line, " ");
+
+	if (i < lastchari)
+		strcat (line, " ");
+	else
+		line[lastchari] = ' ';
 
 	return line;
 }
@@ -625,6 +648,7 @@ const char *PR_GlobalString (int ofs)
 const char *PR_GlobalStringNoContents (int ofs)
 {
 	static char	line[512];
+	static const int lastchari = Q_COUNTOF(line) - 2;
 	int		i;
 	ddef_t		*def;
 
@@ -637,11 +661,205 @@ const char *PR_GlobalStringNoContents (int ofs)
 	i = strlen(line);
 	for ( ; i < 20; i++)
 		strcat (line, " ");
-	strcat (line, " ");
+
+	if (i < lastchari)
+		strcat (line, " ");
+	else
+		line[lastchari] = ' ';
 
 	return line;
 }
 
+/*
+=============
+ED_IsRelevantField
+
+Returns true if the field should be printed by the edict command:
+- not a _x/_y_z variable
+- non-zero contents
+=============
+*/
+qboolean ED_IsRelevantField (edict_t *ed, ddef_t *d)
+{
+	const char	*name;
+	size_t		l;
+	int			*v;
+	int			type;
+	int			i;
+
+	name = PR_GetString (d->s_name);
+	l = strlen (name);
+	if (l > 1 && name[l - 2] == '_')
+		return false;	// skip _x, _y, _z vars
+
+	type = d->type & ~DEF_SAVEGLOBAL;
+	if (type >= NUM_TYPE_SIZES)
+		return false;
+
+	// if the value is still all 0, skip the field
+	v = (int *)((char *)&ed->v + d->ofs*4);
+	for (i = 0; i < type_size[type]; i++)
+		if (v[i])
+			return true;
+
+	return false;
+}
+
+/*
+=============
+ED_AppendFlagString
+=============
+*/
+static void ED_AppendFlagString (char *dst, size_t dstsize, const char *desc)
+{
+	if (*dst)
+		q_strlcat (dst, " | ", dstsize);
+	q_strlcat (dst, desc, dstsize);
+}
+
+/*
+=============
+ED_FieldValueString
+=============
+*/
+const char *ED_FieldValueString (edict_t *ed, ddef_t *d)
+{
+	static char str[1024];
+	int ofs = d->ofs*4;
+	eval_t *val = (eval_t *)((char *)&ed->v + ofs);
+
+	// .movetype
+	if (ofs == offsetof (entvars_t, movetype) && val->_float == (int)val->_float)
+	{
+		switch ((int)val->_float)
+		{
+		#define MOVETYPE_CASE(x)	case x: return #x
+		MOVETYPE_CASE (MOVETYPE_NONE);
+		MOVETYPE_CASE (MOVETYPE_ANGLENOCLIP);
+		MOVETYPE_CASE (MOVETYPE_ANGLECLIP);
+		MOVETYPE_CASE (MOVETYPE_WALK);
+		MOVETYPE_CASE (MOVETYPE_STEP);
+		MOVETYPE_CASE (MOVETYPE_FLY);
+		MOVETYPE_CASE (MOVETYPE_TOSS);
+		MOVETYPE_CASE (MOVETYPE_PUSH);
+		MOVETYPE_CASE (MOVETYPE_NOCLIP);
+		MOVETYPE_CASE (MOVETYPE_FLYMISSILE);
+		MOVETYPE_CASE (MOVETYPE_BOUNCE);
+		MOVETYPE_CASE (MOVETYPE_GIB);
+		#undef MOVETYPE_CASE
+		default:
+			break;
+		}
+	}
+
+	// .solid
+	if (ofs == offsetof (entvars_t, solid) && val->_float == (int)val->_float)
+	{
+		switch ((int)val->_float)
+		{
+		#define SOLID_CASE(x)	case x: return #x
+		SOLID_CASE (SOLID_NOT);
+		SOLID_CASE (SOLID_TRIGGER);
+		SOLID_CASE (SOLID_BBOX);
+		SOLID_CASE (SOLID_SLIDEBOX);
+		SOLID_CASE (SOLID_BSP);
+		#undef SOLID_CASE
+		default:
+			break;
+		}
+	}
+
+	// .deadflag
+	if (ofs == offsetof (entvars_t, deadflag) && val->_float == (int)val->_float)
+	{
+		switch ((int)val->_float)
+		{
+		#define DEAD_CASE(x)	case x: return #x
+		DEAD_CASE (DEAD_NO);
+		DEAD_CASE (DEAD_DYING);
+		DEAD_CASE (DEAD_DEAD);
+		DEAD_CASE (DEAD_RESPAWNABLE);
+		#undef DEAD_CASE
+		default:
+			break;
+		}
+	}
+
+	// .takedamage
+	if (ofs == offsetof (entvars_t, takedamage) && val->_float == (int)val->_float)
+	{
+		switch ((int)val->_float)
+		{
+		#define TAKEDAMAGE_CASE(x)	case x: return #x
+		TAKEDAMAGE_CASE (DAMAGE_NO);
+		TAKEDAMAGE_CASE (DAMAGE_YES);
+		TAKEDAMAGE_CASE (DAMAGE_AIM);
+		#undef TAKEDAMAGE_CASE
+		default:
+			break;
+		}
+	}
+
+	// bitfield: .flags, .spawnflags, .effects
+	if ((ofs == offsetof (entvars_t, flags) || ofs == offsetof (entvars_t, spawnflags) || ofs == offsetof (entvars_t, effects)) && val->_float == (int)val->_float)
+	{
+		int bits = (int)val->_float;
+		str[0] = '\0';
+
+		#define BIT_CASE(f) do { if (bits & (int)f) { bits ^= (int)f; ED_AppendFlagString (str, sizeof (str), #f); } } while (0)
+
+		if (ofs == offsetof (entvars_t, flags))
+		{
+			BIT_CASE (FL_FLY);
+			BIT_CASE (FL_CONVEYOR);
+			BIT_CASE (FL_CLIENT);
+			BIT_CASE (FL_INWATER);
+			BIT_CASE (FL_MONSTER);
+			BIT_CASE (FL_GODMODE);
+			BIT_CASE (FL_NOTARGET);
+			BIT_CASE (FL_ITEM);
+			BIT_CASE (FL_ONGROUND);
+			BIT_CASE (FL_PARTIALGROUND);
+			BIT_CASE (FL_WATERJUMP);
+			BIT_CASE (FL_JUMPRELEASED);
+		}
+		else if (ofs == offsetof (entvars_t, spawnflags))
+		{
+			BIT_CASE (SPAWNFLAG_NOT_EASY);
+			BIT_CASE (SPAWNFLAG_NOT_MEDIUM);
+			BIT_CASE (SPAWNFLAG_NOT_HARD);
+			BIT_CASE (SPAWNFLAG_NOT_DEATHMATCH);
+		}
+		else if (ofs == offsetof (entvars_t, effects))
+		{
+			BIT_CASE (EF_BRIGHTFIELD);
+			BIT_CASE (EF_MUZZLEFLASH);
+			BIT_CASE (EF_BRIGHTLIGHT);
+			BIT_CASE (EF_DIMLIGHT);
+		}
+
+		#undef BIT_CASE
+
+		while (bits)
+		{
+			int lowest = bits & -bits;
+			bits ^= lowest;
+			ED_AppendFlagString (str, sizeof (str), va ("%d", lowest));
+		}
+
+		return str;
+	}
+
+	// .nextthink
+	if (ofs == offsetof (entvars_t, nextthink) && val->_float)
+	{
+		return va (" %7.1f (%+.2f)", val->_float, val->_float - qcvm->time);
+	}
+
+	// generic field
+	return PR_ValueString (d->type, val);
+}
+
 
 /*
 =============
@@ -653,10 +871,7 @@ For debugging
 void ED_Print (edict_t *ed)
 {
 	ddef_t	*d;
-	int		*v;
-	int		i, j, l;
-	const char	*name;
-	int		type;
+	int		i, l;
 	char	field[4096], buf[4096], *p;
 
 	if (ed->free)
@@ -670,25 +885,10 @@ void ED_Print (edict_t *ed)
 	for (i = 1; i < qcvm->progs->numfielddefs; i++)
 	{
 		d = &qcvm->fielddefs[i];
-		name = PR_GetString(d->s_name);
-		l = strlen (name);
-		if (l > 1 && name[l - 2] == '_')
-			continue;	// skip _x, _y, _z vars
-
-		v = (int *)((char *)&ed->v + d->ofs*4);
-
-	// if the value is still all 0, skip the field
-		type = d->type & ~DEF_SAVEGLOBAL;
-
-		for (j = 0; j < type_size[type]; j++)
-		{
-			if (v[j])
-				break;
-		}
-		if (j == type_size[type])
+		if (!ED_IsRelevantField (ed, d))
 			continue;
 
-		q_snprintf (field, sizeof (field), "%-14s %s\n", name, PR_ValueString (d->type, (eval_t *)v)); // johnfitz -- was Con_Printf
+		q_snprintf (field, sizeof (field), "%-14s %s\n", PR_GetString (d->s_name), ED_FieldValueString (ed, d)); // johnfitz -- was Con_Printf
 		l = strlen (field);
 		if (l + 1 > buf + sizeof (buf) - p)
 		{
@@ -734,6 +934,10 @@ void ED_Write (savedata_t *save, edict_t *ed)
 
 	// if the value is still all 0, skip the field
 		type = d->type & ~DEF_SAVEGLOBAL;
+
+		if (type >= NUM_TYPE_SIZES)
+			continue;
+
 		for (j = 0; j < type_size[type]; j++)
 		{
 			if (v[j])
@@ -808,6 +1012,28 @@ static void ED_PrintEdict_f (void)
 	PR_PopQCVM(oldqcvm);
 }
 
+/*
+=============
+ED_PrintEdict_Completion_f -- tab completion for the edict command
+=============
+*/
+static void ED_PrintEdict_Completion_f (const char *partial)
+{
+	qcvm_t	*oldqcvm;
+	int		i;
+
+	if (!sv.active || Cmd_Argc () > 2 || VEC_SIZE (bbox_linked) == 0)
+		return;
+
+	PR_PushQCVM (&sv.qcvm, &oldqcvm);
+	for (i = 0; i < (int) VEC_SIZE (bbox_linked); i++)
+	{
+		edict_t *ed = bbox_linked[i];
+		Con_AddToTabList (va ("%d", NUM_FOR_EDICT (ed)), partial, PR_GetString (ed->v.classname));
+	}
+	PR_PopQCVM (oldqcvm);
+}
+
 /*
 =============
 ED_Count
@@ -1144,7 +1370,10 @@ const char *ED_ParseEdict (const char *data, edict_t *ent)
 		}
 
 		// parse value
-		data = COM_Parse (data);
+		// HACK: we allow truncation when reading the wad field,
+		// otherwise maps using lots of wads with absolute paths
+		// could cause a parse error
+		data = COM_ParseEx (data, !strcmp (keyname, "wad") ? CPE_ALLOWTRUNC : CPE_NOTRUNC);
 		if (!data)
 			Host_Error ("ED_ParseEntity: EOF without closing brace");
 
@@ -1184,11 +1413,29 @@ const char *ED_ParseEdict (const char *data, edict_t *ent)
 	}
 
 	if (!init)
-		ED_AddToFreeList (ent);
+		ED_Free (ent);
 
 	return data;
 }
 
+/*
+================
+ED_IsSkillSelector
+================
+*/
+static qboolean ED_IsSkillSelector (const edict_t *ent)
+{
+	int skill;
+	const char *classname = PR_GetString (ent->v.classname);
+
+	if (strcmp (classname, "trigger_setskill") == 0 || strcmp (classname, "target_setskill") == 0)
+		return true;
+	if (strcmp (classname, "info_command") == 0 && (int)ent->v.message != 0 && sscanf (PR_GetString (ent->v.message), "skill %d", &skill) == 1)
+		return true;
+
+	return false;
+}
+
 
 /*
 ================
@@ -1230,6 +1477,51 @@ void ED_LoadFromFile (const char *data)
 			ent = ED_Alloc ();
 		data = ED_ParseEdict (data, ent);
 
+		if (!ent->v.classname)
+		{
+			Con_SafePrintf ("No classname for:\n"); //johnfitz -- was Con_Printf
+			ED_Print (ent);
+			ED_Free (ent);
+			continue;
+		}
+
+		classname = PR_GetString (ent->v.classname);
+
+		if (sv.mapchecks.active)
+		{
+			int skillflags = (int)ent->v.spawnflags & (SPAWNFLAG_NOT_EASY|SPAWNFLAG_NOT_MEDIUM|SPAWNFLAG_NOT_HARD);
+			if (!(skillflags & SPAWNFLAG_NOT_EASY))
+				sv.mapchecks.skill_ents[0]++;
+			if (!(skillflags & SPAWNFLAG_NOT_MEDIUM))
+				sv.mapchecks.skill_ents[1]++;
+			if (!(skillflags & SPAWNFLAG_NOT_HARD))
+				sv.mapchecks.skill_ents[2]++;
+
+			if (strcmp (classname, "trigger_changelevel") == 0)
+			{
+				ddef_t *mapfield = ED_FindField ("map");
+				sv.mapchecks.trigger_changelevel++;
+				if (mapfield && (mapfield->type & ~DEF_SAVEGLOBAL) == ev_string)
+				{
+					eval_t		*val = GetEdictFieldValue (ent, mapfield->ofs);
+					const char	*map = COM_SkipSpace (PR_GetString (val->string));
+					if (*map)
+					{
+						sv.mapchecks.changelevel = map;
+						sv.mapchecks.valid_changelevel++;
+					}
+				}
+			}
+			else if (ED_IsSkillSelector (ent))
+				sv.mapchecks.skill_triggers++;
+			else if (strcmp (classname, "info_intermission") == 0)
+				sv.mapchecks.intermission++;
+			else if (strcmp (classname, "info_player_coop") == 0)
+				sv.mapchecks.coop_spawns++;
+			else if (strcmp (classname, "info_player_deathmatch") == 0)
+				sv.mapchecks.dm_spawns++;
+		}
+
 		// remove things from different skill levels or deathmatch
 		if (deathmatch.value)
 		{
@@ -1249,18 +1541,7 @@ void ED_LoadFromFile (const char *data)
 			continue;
 		}
 
-//
-// immediately call spawn function
-//
-		if (!ent->v.classname)
-		{
-			Con_SafePrintf ("No classname for:\n"); //johnfitz -- was Con_Printf
-			ED_Print (ent);
-			ED_Free (ent);
-			continue;
-		}
-
-		classname = PR_GetString (ent->v.classname);
+		// remove monsters if nomonsters is set
 		if (sv.nomonsters && !Q_strncmp (classname, "monster_", 8))
 		{
 			ED_Free (ent);
@@ -1268,6 +1549,10 @@ void ED_LoadFromFile (const char *data)
 			continue;
 		}
 
+
+//
+// immediately call spawn function
+//
 	// look for the spawn function
 		func = ED_FindFunction (classname);
 
@@ -1530,6 +1815,8 @@ static void PR_MergeEngineFieldDefs (void)
 	{	//we now know how many entries we need to add...
 		ddef_t *olddefs = qcvm->fielddefs;
 		qcvm->fielddefs = malloc(maxdefs * sizeof(*qcvm->fielddefs));
+		if (!qcvm->fielddefs)
+			Sys_Error ("PR_MergeEngineFieldDefs: out of memory (%d defs)", maxdefs);
 		memcpy(qcvm->fielddefs, olddefs, qcvm->progs->numfielddefs*sizeof(*qcvm->fielddefs));
 		if (olddefs != (ddef_t *)((byte *)qcvm->progs + qcvm->progs->ofs_fielddefs))
 			free(olddefs);
@@ -1682,6 +1969,127 @@ static void PR_FindSavegameFields (void)
 	}
 }
 
+/*
+===============
+PR_FindEntityFields
+
+Finds all the .entity fields (used by r_showbboxes & co for identifying entity links)
+===============
+*/
+static void PR_FindEntityFields (void)
+{
+	int i, count;
+
+	count = 0;
+	for (i = 1; i < qcvm->progs->numfielddefs; i++)
+	{
+		ddef_t *field = &qcvm->fielddefs[i];
+		if ((field->type & ~DEF_SAVEGLOBAL) == ev_entity)
+			count++;
+	}
+
+	qcvm->numentityfields = count;
+	qcvm->entityfieldofs = (int *) Hunk_AllocName (qcvm->numentityfields * sizeof (int), "entityfieldofs");
+	qcvm->entityfields = (ddef_t **) Hunk_AllocName (qcvm->numentityfields * sizeof (ddef_t*), "entityfields");
+
+	count = 0;
+	for (i = 1; i < qcvm->progs->numfielddefs; i++)
+	{
+		ddef_t *field = &qcvm->fielddefs[i];
+		if ((field->type & ~DEF_SAVEGLOBAL) == ev_entity)
+		{
+			qcvm->entityfieldofs[count] = field->ofs*4;
+			qcvm->entityfields[count] = field;
+			count++;
+		}
+	}
+}
+
+/*
+===============
+PR_CompareFunction
+===============
+*/
+static int PR_CompareFunction (const void *pa, const void *pb)
+{
+	const dfunction_t *fa = &qcvm->functions[*(const int *)pa];
+	const dfunction_t *fb = &qcvm->functions[*(const int *)pb];
+	return fa->first_statement - fb->first_statement;
+}
+
+/*
+===============
+PR_FindFunctionRanges
+===============
+*/
+static void PR_FindFunctionRanges (void)
+{
+	int		i, mark;
+	int		*order;
+
+	qcvm->functionsizes = (int *) Hunk_AllocName (qcvm->progs->numfunctions * sizeof (*order), "func_sizes");
+	mark = Hunk_LowMark ();
+
+	order = (int *) Hunk_AllocNoFill (qcvm->progs->numfunctions * sizeof (*order));
+	for (i = 0; i < qcvm->progs->numfunctions; i++)
+		order[i] = i;
+	qsort (order, qcvm->progs->numfunctions, sizeof (*order), &PR_CompareFunction);
+
+	for (i = 0; i < qcvm->progs->numfunctions; i++)
+	{
+		dfunction_t *f = &qcvm->functions[order[i]];
+		if (f->first_statement <= 0)
+			continue;
+		if (i == qcvm->progs->numfunctions - 1)
+			qcvm->functionsizes[order[i]] = qcvm->progs->numstatements - f->first_statement;
+		else
+			qcvm->functionsizes[order[i]] = qcvm->functions[order[i + 1]].first_statement - f->first_statement;
+	}
+
+	Hunk_FreeToLowMark (mark);
+}
+
+/*
+===============
+PR_FillOffsetTables
+===============
+*/
+static void PR_FillOffsetTables (void)
+{
+	int		pass, i, maxofs, *data;
+	struct
+	{
+		int			**offsets;
+		int			*maxofs;
+		int			numdefs;
+		ddef_t		*defs;
+		const char	*allocname;
+	}
+	passes[] =
+	{
+		{ &qcvm->ofstofield,	&qcvm->maxfieldofs,		qcvm->progs->numfielddefs,	qcvm->fielddefs,	"ofstofield"	},
+		{ &qcvm->ofstoglobal,	&qcvm->maxglobalofs,	qcvm->progs->numglobaldefs,	qcvm->globaldefs,	"ofstoglobal"	},
+	};
+
+	for (pass = 0; pass < (int) Q_COUNTOF (passes); pass++)
+	{
+		// find maximum offset
+		for (i = 1, maxofs = 0; i < passes[pass].numdefs; i++)
+			maxofs = q_max (maxofs, passes[pass].defs[i].ofs);
+		*passes[pass].maxofs = maxofs;
+
+		// alloc table and fill it with -1
+		data = *passes[pass].offsets = (int *) Hunk_AllocNameNoFill ((maxofs + 1) * sizeof (int), passes[pass].allocname);
+		for (i = 0; i <= maxofs; i++)
+			data[i] = -1;
+
+		// fill actual offsets in descending order so that earlier defs are written last
+		// this preserves the behavior of ED_FieldAtOfs/ED_GlobalAtOfs, which stopped at the first match
+		for (i = passes[pass].numdefs - 1; i > 0; i--)
+			data[passes[pass].defs[i].ofs] = i;
+	}
+}
+
 /*
 ===============
 PR_LoadProgs
@@ -1696,11 +2104,9 @@ qboolean PR_LoadProgs (const char *filename, qboolean fatal)
 	qcvm->progs = (dprograms_t *)COM_LoadHunkFile (filename, NULL);
 	if (!qcvm->progs)
 		return false;
-	Con_DPrintf ("Programs occupy %iK.\n", com_filesize/1024);
+	Con_DPrintf ("Programs occupy %" SDL_PRIs64 "K.\n", com_filesize/1024);
 
-	CRC_Init (&qcvm->crc);
-	for (i = 0; i < com_filesize; i++)
-		CRC_ProcessByte (&qcvm->crc, ((byte *)qcvm->progs)[i]);
+	qcvm->crc = CRC_Block (qcvm->progs, com_filesize);
 
 	// byte swap the header
 	for (i = 0; i < (int) sizeof(*qcvm->progs) / 4; i++)
@@ -1836,6 +2242,9 @@ qboolean PR_LoadProgs (const char *filename, qboolean fatal)
 	PR_PatchRereleaseBuiltins ();
 	PR_EnableExtensions ();
 	PR_FindSavegameFields ();
+	PR_FindEntityFields ();
+	PR_FindFunctionRanges ();
+	PR_FillOffsetTables ();
 
 	qcvm->effects_mask = PR_FindSupportedEffects ();
 
@@ -1861,7 +2270,11 @@ PR_Init
 */
 void PR_Init (void)
 {
-	Cmd_AddCommand ("edict", ED_PrintEdict_f);
+	cmd_function_t *cmd;
+
+	cmd = Cmd_AddCommand ("edict", ED_PrintEdict_f);
+	if (cmd)
+		cmd->completion = ED_PrintEdict_Completion_f;
 	Cmd_AddCommand ("edicts", ED_PrintEdicts);
 	Cmd_AddCommand ("edictcount", ED_Count);
 	Cmd_AddCommand ("profile", PR_Profile_f);
@@ -2027,7 +2440,7 @@ int PR_AllocString (int size, char **ptr)
 
 void SaveData_Init (savedata_t *save)
 {
-	memset (save, 0, sizeof (save));
+	memset (save, 0, sizeof (*save));
 	save->buffersize = 48 * 1024 * 1024; // ad_sepulcher needs ~32 MB
 	save->buffer = (byte *) malloc (save->buffersize);
 	if (!save->buffer)
