@@ -83,14 +83,7 @@ cvar_t		snd_mixspeed = {"snd_mixspeed", "44100", CVAR_NONE};
 
 cvar_t		snd_waterfx = {"snd_waterfx", "1", CVAR_ARCHIVE};
 
-#if defined(_WIN32)
-#define SND_FILTERQUALITY_DEFAULT "5"
-#else
-#define SND_FILTERQUALITY_DEFAULT "1"
-#endif
-
-cvar_t		snd_filterquality = {"snd_filterquality", SND_FILTERQUALITY_DEFAULT,
-								 CVAR_NONE};
+cvar_t		snd_filterquality = {"snd_filterquality", "5", CVAR_ARCHIVE};
 
 static	cvar_t	nosound = {"nosound", "0", CVAR_NONE};
 static	cvar_t	ambient_level = {"ambient_level", "0.3", CVAR_NONE};
@@ -128,7 +121,7 @@ static void SND_Callback_snd_filterquality (cvar_t *var)
 	if (snd_filterquality.value < 1 || snd_filterquality.value > 5)
 	{
 		Con_Printf ("snd_filterquality must be between 1 and 5\n");
-		Cvar_SetQuick (&snd_filterquality, SND_FILTERQUALITY_DEFAULT);
+		Cvar_SetQuick (&snd_filterquality, snd_filterquality.default_string);
 	}
 }
 
@@ -185,7 +178,7 @@ void S_Init (void)
 	Cvar_RegisterVariable(&snd_mixspeed);
 	Cvar_RegisterVariable(&snd_filterquality);
 	Cvar_RegisterVariable(&snd_waterfx);
-	
+
 	if (safemode || COM_CheckParm("-nosound"))
 		return;
 
@@ -384,9 +377,6 @@ channel_t *SND_PickChannel (int entnum, int entchannel)
 	if (first_to_die == -1)
 		return NULL;
 
-	if (snd_channels[first_to_die].sfx)
-		snd_channels[first_to_die].sfx = NULL;
-
 	return &snd_channels[first_to_die];
 }
 
@@ -449,8 +439,12 @@ void S_StartSound (int entnum, int entchannel, sfx_t *sfx, vec3_t origin, float
 {
 	channel_t	*target_chan, *check;
 	sfxcache_t	*sc;
-	int		ch_idx;
-	int		skip;
+	sfx_t		*old_sfx;
+	vec3_t		old_origin;
+	float		old_vol;
+	float		old_atten;
+	int			ch_idx;
+	int			skip;
 
 	if (!sound_started)
 		return;
@@ -466,6 +460,19 @@ void S_StartSound (int entnum, int entchannel, sfx_t *sfx, vec3_t origin, float
 	if (!target_chan)
 		return;
 
+// keep track of the old sound playing on this channel (for demo rewinding)
+	old_sfx = NULL;
+	VectorCopy (origin, old_origin);
+	old_vol = fvol;
+	old_atten = attenuation;
+	if (entnum > 0 && entchannel > 0 && target_chan->entnum == entnum && target_chan->entchannel == entchannel)
+	{
+		old_sfx = target_chan->sfx;
+		VectorCopy (target_chan->origin, old_origin);
+		old_vol = target_chan->master_vol;
+		old_atten = target_chan->dist_mult * sound_nominal_clip_dist;
+	}
+
 // spatialize
 	memset (target_chan, 0, sizeof(*target_chan));
 	VectorCopy(origin, target_chan->origin);
@@ -486,6 +493,11 @@ void S_StartSound (int entnum, int entchannel, sfx_t *sfx, vec3_t origin, float
 		return;		// couldn't load the sound's data
 	}
 
+// if this is a looping sound and we're not rewinding, keep track of the previous sound playing
+// on the same ent/channel so that when we do rewind past this frame we start playing it instead
+	if (cls.demoplayback && cls.demospeed > 0.f && sc->loopstart != -1)
+		CL_AddDemoRewindSound (entnum, entchannel, old_sfx, old_origin, old_vol, old_atten);
+
 	target_chan->sfx = sfx;
 	target_chan->pos = 0.0;
 	target_chan->end = paintedtime + sc->length;
@@ -577,7 +589,8 @@ void S_ClearBuffer (void)
 	else
 		clear = 0;
 
-	memset(shm->buffer, clear, shm->samples * shm->samplebits / 8);
+	memset (shm->buffer, clear, shm->samples * shm->samplebits / 8);
+	memset (s_rawsamples, 0, sizeof (s_rawsamples));
 
 	SNDDMA_Submit ();
 }
@@ -884,7 +897,10 @@ void S_Update (vec3_t origin, vec3_t forward, vec3_t right, vec3_t up)
 		{
 			if (ch->sfx && (ch->leftvol || ch->rightvol) )
 			{
-			//	Con_Printf ("%3i %3i %s\n", ch->leftvol, ch->rightvol, ch->sfx->name);
+				sfxcache_t *sc = (sfxcache_t *) Cache_Check (&ch->sfx->cache);
+				if (snd_show.value >= 2.f)
+					Con_SafePrintf ("L:%3i R:%3i | ENT:%5i CH:%3i | %s%s\n",
+						ch->leftvol, ch->rightvol, ch->entnum, ch->entchannel, ch->sfx->name, sc && sc->loopstart >= 0 ? " [L]" : "");
 				total++;
 			}
 		}
@@ -1050,7 +1066,7 @@ static void S_SoundList (void)
 	int		i;
 	sfx_t	*sfx;
 	sfxcache_t	*sc;
-	int		size, total;
+	int	size, total;
 
 	total = 0;
 	for (sfx = known_sfx, i = 0; i < num_sfx; i++, sfx++)
@@ -1093,13 +1109,10 @@ void S_ClearPrecache (void)
 {
 }
 
-
 void S_BeginPrecaching (void)
 {
 }
 
-
 void S_EndPrecaching (void)
 {
 }
-
