@@ -42,12 +42,6 @@ void SV_CalcStats(client_t *client, int *statsi, float *statsf, const char **sta
 	size_t i;
 	edict_t *ent = client->edict;
 	//FIXME: string stats!
-	int items;
-	eval_t *val = GetEdictFieldValue(ent, qcvm->extfields.items2);
-	if (val)
-		items = (int)ent->v.items | ((int)val->_float << 23);
-	else
-		items = (int)ent->v.items | ((int)pr_global_struct->serverflags << 28);
 
 	memset(statsi, 0, sizeof(*statsi)*MAX_CL_STATS);
 	memset(statsf, 0, sizeof(*statsf)*MAX_CL_STATS);
@@ -156,6 +150,7 @@ void SV_Init (void)
 	extern	cvar_t	sv_aim;
 	extern	cvar_t	sv_altnoclip; //johnfitz
 	extern	cvar_t	sv_gameplayfix_random;
+	extern	cvar_t	sv_gameplayfix_elevators;
 	extern	cvar_t	sv_autoload;
 	extern	cvar_t	sv_autosave;
 	extern	cvar_t	sv_autosave_interval;
@@ -177,6 +172,7 @@ void SV_Init (void)
 	Cvar_RegisterVariable (&pr_checkextension);
 	Cvar_RegisterVariable (&sv_altnoclip); //johnfitz
 	Cvar_RegisterVariable (&sv_gameplayfix_random);
+	Cvar_RegisterVariable (&sv_gameplayfix_elevators);
 	Cvar_RegisterVariable (&sv_netsort);
 	Cvar_RegisterVariable (&sv_autoload);
 	Cvar_RegisterVariable (&sv_autosave);
@@ -456,7 +452,7 @@ void SV_SendServerinfo (client_t *client)
 	MSG_WriteByte (&client->message, svc_signonnum);
 	MSG_WriteByte (&client->message, 1);
 
-	client->sendsignon = true;
+	client->sendsignon = PRESPAWN_FLUSH;
 	client->spawned = false;		// need prespawn, spawn, etc
 }
 
@@ -632,6 +628,7 @@ given point.
 byte *SV_FatPVS (vec3_t org, qmodel_t *worldmodel) //johnfitz -- added worldmodel as a parameter
 {
 	fatbytes = (worldmodel->numleafs+7)>>3; // ericw -- was +31, assumed to be a bug/typo
+	fatbytes = (fatbytes + VIS_ALIGN_MASK) & ~VIS_ALIGN_MASK; // round up
 	if (fatpvs == NULL || fatbytes > fatpvs_capacity)
 	{
 		fatpvs_capacity = fatbytes;
@@ -1704,6 +1701,201 @@ void SV_SaveSpawnparms (void)
 	}
 }
 
+typedef enum
+{
+	MAPCHECK_FAILED,
+	MAPCHECK_PARTIAL,
+	MAPCHECK_OK,
+} mapcheck_t;
+
+
+/*
+================
+SV_MapCheckThresh
+================
+*/
+static mapcheck_t SV_MapCheckThresh (int current, int target)
+{
+	if (current <= 0)
+		return MAPCHECK_FAILED;
+	if (current >= target)
+		return MAPCHECK_OK;
+	return MAPCHECK_PARTIAL;
+}
+
+/*
+================
+SV_PrintMapCheck
+================
+*/
+static void SV_PrintMapCheck (mapcheck_t status, const char *format, ...)
+{
+	char		str[1024];
+	char		tinted[1024];
+	va_list		argptr;
+
+	va_start (argptr, format);
+	q_vsnprintf (str, sizeof (str), format, argptr);
+	va_end (argptr);
+
+	if (status == MAPCHECK_OK)
+	{
+		Con_SafePrintf ("[x] %s\n", str);
+	}
+	else
+	{
+		Con_SafePrintf ("\xDB%c\xDD %s\n", status == MAPCHECK_PARTIAL ? '-' ^ 0x80 : ' ', COM_TintString (str, tinted, sizeof (tinted)));
+		sv.mapchecks.numwarnings++;
+	}
+}
+
+
+/*
+================
+SV_PrintMapChecklist
+================
+*/
+static void SV_PrintMapChecklist (void)
+{
+	const int	MIN_DM_SPAWN_POINTS = 5;
+	const int	MIN_COOP_SPAWN_POINTS = 3;
+
+	qboolean	skill_levels;
+	char		buf[1024];
+	int			i, track, numskies, count;
+
+	//
+	// header
+	//
+	Con_SafePrintf ("\n");
+	Con_SafePrintf ("=====================================\n");
+	Con_SafePrintf ("\n");
+	Con_SafePrintf ("Map checklist (%s):\n\n", COM_SkipPath (sv.modelname));
+
+	//
+	// light data
+	//
+	SV_PrintMapCheck (sv.worldmodel->lightdata != NULL ? MAPCHECK_OK : MAPCHECK_FAILED, "lightmap data");
+
+	//
+	// vis data
+	//
+	if (!sv.worldmodel->visdata)
+	{
+		char pointfile[MAX_OSPATH];
+		q_snprintf (pointfile, sizeof (pointfile), "maps/%s.pts", sv.name);
+		if (COM_FileExists (pointfile, NULL))
+			SV_PrintMapCheck (MAPCHECK_FAILED, "vis data (unsealed map?)");
+		else
+			SV_PrintMapCheck (MAPCHECK_FAILED, "vis data");
+	}
+	else
+		SV_PrintMapCheck (MAPCHECK_OK, "vis data");
+
+	//
+	// changelevel trigger
+	//
+	if (!sv.mapchecks.trigger_changelevel)
+		SV_PrintMapCheck (MAPCHECK_FAILED, "trigger_changelevel");
+	else if (sv.mapchecks.trigger_changelevel == 1)
+	{
+		if (sv.mapchecks.valid_changelevel == sv.mapchecks.trigger_changelevel)
+			SV_PrintMapCheck (MAPCHECK_OK, "trigger_changelevel (%s)", sv.mapchecks.changelevel);
+		else
+			SV_PrintMapCheck (MAPCHECK_PARTIAL, "trigger_changelevel (missing \"map\" key)");
+	}
+	else
+	{
+		if (sv.mapchecks.valid_changelevel == sv.mapchecks.trigger_changelevel)
+			SV_PrintMapCheck (MAPCHECK_OK, "trigger_changelevel (%d)", sv.mapchecks.trigger_changelevel);
+		else
+			SV_PrintMapCheck (MAPCHECK_PARTIAL, "trigger_changelevel (%d/%d missing \"map\" key)",
+				sv.mapchecks.trigger_changelevel - sv.mapchecks.valid_changelevel,
+				sv.mapchecks.trigger_changelevel
+			);
+	}
+
+	//
+	// intermission camera
+	//
+	if (!sv.mapchecks.intermission)
+		SV_PrintMapCheck (MAPCHECK_FAILED, "info_intermission");
+	else
+		SV_PrintMapCheck (MAPCHECK_OK, "info_intermission (%d)", sv.mapchecks.intermission);
+
+	//
+	// skill levels
+	//
+	skill_levels = sv.mapchecks.skill_triggers > 0 ||
+		(sv.mapchecks.skill_ents[0] != sv.mapchecks.skill_ents[1] || sv.mapchecks.skill_ents[1] != sv.mapchecks.skill_ents[2]);
+	SV_PrintMapCheck (skill_levels ? MAPCHECK_OK : MAPCHECK_FAILED, "skill spawnflags/triggers");
+
+	//
+	// coop spawn points
+	//
+	SV_PrintMapCheck (SV_MapCheckThresh (sv.mapchecks.coop_spawns, MIN_COOP_SPAWN_POINTS),
+		"info_player_coop (%d/%d+)", sv.mapchecks.coop_spawns, MIN_COOP_SPAWN_POINTS);
+
+	//
+	// deathmatch spawn points
+	//
+	SV_PrintMapCheck (SV_MapCheckThresh (sv.mapchecks.dm_spawns, MIN_DM_SPAWN_POINTS),
+		"info_player_deathmatch (%d/%d+)", sv.mapchecks.dm_spawns, MIN_DM_SPAWN_POINTS);
+
+	//
+	// music track
+	//
+	track = (int)qcvm->edicts->v.sounds;
+	if (track == 0)
+		SV_PrintMapCheck (MAPCHECK_FAILED, "music track (worldspawn \"sounds\" field)");
+	else if (track < 2 || track > 255) 
+		SV_PrintMapCheck (MAPCHECK_FAILED, "music track (%d, should be between 2 and 255)");
+	else
+		SV_PrintMapCheck (MAPCHECK_OK, "music track (%d)", track);
+
+	//
+	// map title
+	//
+	Mod_SanitizeMapDescription (buf, sizeof (buf), PR_GetString ((int)qcvm->edicts->v.message));
+	if (buf[0])
+		SV_PrintMapCheck (MAPCHECK_OK, "map title (%s)", buf);
+	else
+		SV_PrintMapCheck (MAPCHECK_FAILED, "map title (worldspawn \"message\" field)");
+
+	//
+	// warn about multiple skies (e.g. "Slip Tripping" / markiesm1.bsp), of which only the last would get rendered in other engines,
+	// and sky textures with non-standard sizes (e.g. "Tears of the False God" / ad_tears.bsp), which could even crash older engines 
+	//
+	numskies = sv.worldmodel->texofs[(int)TEXTYPE_SKY + 1] - sv.worldmodel->texofs[TEXTYPE_SKY];
+	count = 0;
+	for (i = 0; i < numskies; i++)
+	{
+		texture_t *tex = sv.worldmodel->textures[sv.worldmodel->usedtextures[sv.worldmodel->texofs[TEXTYPE_SKY] + i]];
+		if (tex->width != 256 || tex->height != 128)
+			count++;
+	}
+	if (numskies > 1 || count > 0)
+	{
+		SV_PrintMapCheck (MAPCHECK_FAILED, "compat: single %ssky texture (%d found)", count > 0 ? "256 x 128 " : "", numskies);
+		for (i = 0; i < numskies; i++)
+		{
+			texture_t *tex = sv.worldmodel->textures[sv.worldmodel->usedtextures[sv.worldmodel->texofs[TEXTYPE_SKY] + i]];
+			if (tex->width != 256 || tex->height != 128)
+				q_snprintf (buf, sizeof (buf), " (%d x %d)", tex->width, tex->height);
+			else
+				buf[0] = '\0';
+			Con_SafePrintf ("\x02    %d. %s%s\n", i + 1, tex->name, buf);
+		}
+	}
+
+	//
+	// footer
+	//
+	Con_SafePrintf ("\n");
+	Con_SafePrintf ("=====================================\n");
+	Con_SafePrintf ("\n");
+}
+
 
 /*
 ================
@@ -1758,6 +1950,8 @@ void SV_SpawnServer (const char *server)
 	Host_ClearMemory ();
 
 	q_strlcpy (sv.name, server, sizeof(sv.name));
+	if (developer.value || map_checks.value)
+		sv.mapchecks.active = true;
 
 	sv.protocol = sv_protocol; // johnfitz
 	
@@ -1777,6 +1971,8 @@ void SV_SpawnServer (const char *server)
 	/* Host_ClearMemory() called above already cleared the whole sv structure */
 	qcvm->max_edicts = CLAMP (MIN_EDICTS,(int)max_edicts.value,MAX_EDICTS); //johnfitz -- max_edicts cvar
 	qcvm->edicts = (edict_t *) malloc (qcvm->max_edicts*qcvm->edict_size); // ericw -- sv.edicts switched to use malloc()
+	if (!qcvm->edicts)
+		Sys_Error ("SV_SpawnServer: out of memory (%d edicts x %d bytes)", qcvm->max_edicts, qcvm->edict_size);
 	ClearLink (&qcvm->free_edicts);
 
 	sv.datagram.maxsize = sizeof(sv.datagram_buf);
@@ -1879,5 +2075,8 @@ void SV_SpawnServer (const char *server)
 			SV_SendServerinfo (host_client);
 
 	Con_DPrintf ("Server spawned.\n");
+
+	if (sv.mapchecks.active)
+		SV_PrintMapChecklist ();
 }
 
