@@ -123,8 +123,6 @@ Returns false if the entity removed itself.
 qboolean SV_RunThink (edict_t *ent)
 {
 	float	thinktime;
-	float	oldframe; //johnfitz
-	int		i; //johnfitz
 
 	thinktime = ent->v.nextthink;
 	if (thinktime <= 0 || thinktime > qcvm->time + host_frametime)
@@ -135,7 +133,8 @@ qboolean SV_RunThink (edict_t *ent)
 								// it is possible to start that way
 								// by a trigger with a local time.
 
-	oldframe = ent->v.frame; //johnfitz
+	ent->oldthinktime = thinktime;
+	ent->oldframe = ent->v.frame; //johnfitz
 
 	ent->v.nextthink = 0;
 	pr_global_struct->time = thinktime;
@@ -143,18 +142,6 @@ qboolean SV_RunThink (edict_t *ent)
 	pr_global_struct->other = EDICT_TO_PROG(qcvm->edicts);
 	PR_ExecuteProgram (ent->v.think);
 
-//johnfitz -- PROTOCOL_FITZQUAKE
-//capture interval to nextthink here and send it to client for better
-//lerp timing, but only if interval is not 0.1 (which client assumes)
-	ent->sendinterval = false;
-	if (!ent->free && ent->v.nextthink && (ent->v.movetype == MOVETYPE_STEP || ent->v.frame != oldframe))
-	{
-		i = Q_rint((ent->v.nextthink-thinktime)*255);
-		if (i >= 0 && i < 256 && i != 25 && i != 26) //25 and 26 are close enough to 0.1 to not send
-			ent->sendinterval = true;
-	}
-//johnfitz
-
 	return !ent->free;
 }
 
@@ -443,12 +430,14 @@ trace_t SV_PushEntity (edict_t *ent, vec3_t push)
 SV_PushMove
 ============
 */
+cvar_t sv_gameplayfix_elevators = {"sv_gameplayfix_elevators", "2", CVAR_ARCHIVE}; // 0=off; 1=clients only; 2=all entities
 void SV_PushMove (edict_t *pusher, float movetime)
 {
 	int			i, e;
 	edict_t		*check, *block;
-	vec3_t		mins, maxs, move;
+	vec4_t		mins, maxs, move;
 	vec3_t		entorig, pushorig;
+	float		solid_backup;
 	int			num_moved;
 	edict_t		**moved_edict; //johnfitz -- dynamically allocate
 	vec3_t		*moved_from; //johnfitz -- dynamically allocate
@@ -477,8 +466,8 @@ void SV_PushMove (edict_t *pusher, float movetime)
 
 	//johnfitz -- dynamically allocate
 	mark = Hunk_LowMark ();
-	moved_edict = (edict_t **) Hunk_Alloc (qcvm->num_edicts*sizeof(edict_t *));
-	moved_from = (vec3_t *) Hunk_Alloc (qcvm->num_edicts*sizeof(vec3_t));
+	moved_edict = (edict_t **) Hunk_AllocNoFill (qcvm->num_edicts*sizeof(edict_t *));
+	moved_from = (vec3_t *) Hunk_AllocNoFill (qcvm->num_edicts*sizeof(vec3_t));
 	//johnfitz
 
 // see if any solid entities are inside the final position
@@ -486,17 +475,28 @@ void SV_PushMove (edict_t *pusher, float movetime)
 	check = NEXT_EDICT(qcvm->edicts);
 	for (e=1 ; e<qcvm->num_edicts ; e++, check = NEXT_EDICT(check))
 	{
+		qboolean riding;
+		int movemask;
 		if (check->free)
 			continue;
-		if (check->v.movetype == MOVETYPE_PUSH
-		|| check->v.movetype == MOVETYPE_NONE
-		|| check->v.movetype == MOVETYPE_NOCLIP)
+		movemask = 1 << (int)check->v.movetype;
+		if (movemask & ((1<<MOVETYPE_PUSH) | (1<<MOVETYPE_NONE) | (1<<MOVETYPE_NOCLIP)))
 			continue;
 
 	// if the entity is standing on the pusher, it will definately be moved
 		if ( ! ( ((int)check->v.flags & FL_ONGROUND)
 		&& PROG_TO_EDICT(check->v.groundentity) == pusher) )
 		{
+#ifdef USE_SSE2
+			__m128 check_absmin_vec = _mm_loadu_ps (check->v.absmin);
+			__m128 check_absmax_vec = _mm_loadu_ps (check->v.absmax);
+			__m128 maxs_vec = _mm_loadu_ps (maxs);
+			__m128 mins_vec = _mm_loadu_ps (mins);
+			if (_mm_movemask_ps (_mm_cmpnlt_ps (check_absmin_vec, maxs_vec)) & 7)
+				continue;
+			if (_mm_movemask_ps (_mm_cmpngt_ps (check_absmax_vec, mins_vec)) & 7)
+				continue;
+#else
 			if ( check->v.absmin[0] >= maxs[0]
 			|| check->v.absmin[1] >= maxs[1]
 			|| check->v.absmin[2] >= maxs[2]
@@ -504,11 +504,16 @@ void SV_PushMove (edict_t *pusher, float movetime)
 			|| check->v.absmax[1] <= mins[1]
 			|| check->v.absmax[2] <= mins[2] )
 				continue;
+#endif
 
 		// see if the ent's bbox is inside the pusher's final position
 			if (!SV_TestEntityPosition (check))
 				continue;
+
+			riding = false;
 		}
+		else
+			riding = true;
 
 	// remove the onground flag for non-players
 		if (check->v.movetype != MOVETYPE_WALK)
@@ -520,9 +525,16 @@ void SV_PushMove (edict_t *pusher, float movetime)
 		num_moved++;
 
 		// try moving the contacted entity
-		pusher->v.solid = SOLID_NOT;
-		SV_PushEntity (check, move);
-		pusher->v.solid = SOLID_BSP;
+		// https://www.quake-info-pool.net/q1/qfix.htm#movetype_push
+		solid_backup = pusher->v.solid;
+		if (solid_backup == SOLID_BSP ||
+			solid_backup == SOLID_BBOX ||
+			solid_backup == SOLID_SLIDEBOX)
+		{
+			pusher->v.solid = SOLID_NOT;
+			SV_PushEntity (check, move);
+			pusher->v.solid = solid_backup;
+		}
 
 	// if it is still inside the pusher, block
 		block = SV_TestEntityPosition (check);
@@ -537,6 +549,35 @@ void SV_PushMove (edict_t *pusher, float movetime)
 				continue;
 			}
 
+			// try moving the entity up a bit if it's blocked by the pusher while also standing on it
+			if (riding && block == pusher &&
+				(sv_gameplayfix_elevators.value >= 2.f ||
+				(sv_gameplayfix_elevators.value && e <= svs.maxclients)))
+			{
+				check->v.origin[2] += DIST_EPSILON;
+				if (!SV_TestEntityPosition (check))
+				{
+					// notify developer about potential issue
+					if (map_checks.value || developer.value)
+					{
+						vec3_t check_center, pusher_center;
+
+						VectorAdd (check->v.absmin, check->v.absmax, check_center);
+						VectorScale (check_center, 0.5f, check_center);
+						VectorAdd (pusher->v.absmin, pusher->v.absmax, pusher_center);
+						VectorScale (pusher_center, 0.5f, pusher_center);
+
+						Con_Warning ("sv_gameplayfix_elevators nudged %s #%d at (%.0f %.0f %.0f) above %s #%d at (%.0f %.0f %.0f)\n",
+							PR_GetString (check->v.classname), NUM_FOR_EDICT (check), check_center[0], check_center[1], check_center[2],
+							PR_GetString (pusher->v.classname), NUM_FOR_EDICT (pusher), pusher_center[0], pusher_center[1], pusher_center[2]
+						);
+					}
+
+					// move on to next entity
+					continue;
+				}
+			}
+
 			VectorCopy (entorig, check->v.origin);
 			SV_LinkEdict (check, true);
 
@@ -1235,6 +1276,18 @@ void SV_Physics (void)
 			SV_Physics_Toss (ent);
 		else
 			Sys_Error ("SV_Physics: bad movetype %i", (int)ent->v.movetype);
+
+	//johnfitz -- PROTOCOL_FITZQUAKE
+	//capture interval to nextthink here and send it to client for better
+	//lerp timing, but only if interval is not 0.1 (which client assumes)
+		ent->sendinterval = false;
+		if (!ent->free && ent->v.nextthink > qcvm->time && (ent->v.movetype == MOVETYPE_STEP || ent->v.movetype == MOVETYPE_WALK || ent->v.frame != ent->oldframe))
+		{
+			int j = Q_rint((ent->v.nextthink-ent->oldthinktime)*255);
+			if (j >= 0 && j < 256 && j != 25 && j != 26) //25 and 26 are close enough to 0.1 to not send
+				ent->sendinterval = true;
+		}
+	//johnfitz
 	}
 
 	if (pr_global_struct->force_retouch)
