@@ -23,11 +23,13 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 #include "arch_def.h"
 #include "quakedef.h"
+#include "steam.h"
 
 #include <sys/types.h>
 #include <errno.h>
 #include <unistd.h>
-#ifdef PLATFORM_OSX
+#include <string.h>
+#if defined(PLATFORM_OSX) || defined(PLATFORM_HAIKU)
 #include <libgen.h>	/* dirname() and basename() */
 #endif
 #include <sys/stat.h>
@@ -36,6 +38,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include <time.h>
 #include <dirent.h>
 #include <pwd.h>
+#include <dlfcn.h>
 
 #if defined(SDL_FRAMEWORK) || defined(NO_SDL_CONFIG)
 #if defined(USE_SDL2)
@@ -52,6 +55,7 @@ qboolean		isDedicated;
 
 #define	MAX_HANDLES		32	/* johnfitz -- was 10 */
 static FILE		*sys_handles[MAX_HANDLES];
+static qboolean		stdinIsATTY;	/* from ioquake3 source */
 
 static double rcp_counter_freq;
 
@@ -96,6 +100,18 @@ FILE *Sys_fopen (const char *path, const char *mode)
 	return fopen (path, mode);
 }
 
+COMPILE_TIME_ASSERT (CHECK_LARGE_FILE_SUPPORT, sizeof (off_t) >= sizeof (qfileofs_t));
+
+int Sys_fseek (FILE *file, qfileofs_t ofs, int origin)
+{
+	return fseeko (file, ofs, origin);
+}
+
+qfileofs_t Sys_ftell (FILE *file)
+{
+	return ftello (file);
+}
+
 int Sys_remove (const char *path)
 {
 	return remove (path);
@@ -106,22 +122,23 @@ int Sys_rename (const char *oldname, const char *newname)
 	return rename (oldname, newname);
 }
 
-long Sys_filelength (FILE *f)
+qfileofs_t Sys_filelength (FILE *f)
 {
-	long		pos, end;
+	qfileofs_t	pos, end;
 
-	pos = ftell (f);
-	fseek (f, 0, SEEK_END);
-	end = ftell (f);
-	fseek (f, pos, SEEK_SET);
+	pos = Sys_ftell (f);
+	Sys_fseek (f, 0, SEEK_END);
+	end = Sys_ftell (f);
+	Sys_fseek (f, pos, SEEK_SET);
 
 	return end;
 }
 
-int Sys_FileOpenRead (const char *path, int *hndl)
+qfileofs_t Sys_FileOpenRead (const char *path, int *hndl)
 {
-	FILE	*f;
-	int	i, retval;
+	FILE		*f;
+	int			i;
+	qfileofs_t	retval;
 
 	i = findhandle ();
 	f = fopen(path, "rb");
@@ -292,10 +309,8 @@ static int Sys_NumCPUs (void)
 
 qboolean Sys_GetSteamDir (char *path, size_t pathsize)
 {
-	const char	STEAM_DIR[] = ".steam/steam";
-	const char	*home_dir = NULL;
+	const char		*home_dir = NULL;
 	struct passwd	*pwent;
-	struct stat		st;
 
 	pwent = getpwuid( getuid() );
 	if (pwent == NULL)
@@ -307,10 +322,16 @@ qboolean Sys_GetSteamDir (char *path, size_t pathsize)
 	if (home_dir == NULL)
 		return false;
 
-	if ((size_t) q_snprintf (path, pathsize, "%s/%s", home_dir, STEAM_DIR) >= pathsize)
-		return false;
+	if ((size_t) q_snprintf (path, pathsize, "%s/.steam/steam", home_dir) < pathsize && Steam_IsValidPath (path))
+		return true;
+	if ((size_t) q_snprintf (path, pathsize, "%s/.local/share/Steam", home_dir) < pathsize && Steam_IsValidPath (path))
+		return true;
+	if ((size_t) q_snprintf (path, pathsize, "%s/.var/app/com.valvesoftware.Steam/.steam/steam", home_dir) < pathsize && Steam_IsValidPath (path))
+		return true;
+	if ((size_t) q_snprintf (path, pathsize, "%s/.var/app/com.valvesoftware.Steam/.local/share/Steam", home_dir) < pathsize && Steam_IsValidPath (path))
+		return true;
 
-	return stat (path, &st) == 0 && S_ISDIR (st.st_mode);
+	return false;
 }
 
 qboolean Sys_GetSteamQuakeUserDir (char *path, size_t pathsize, const char *library)
@@ -324,6 +345,53 @@ qboolean Sys_GetSteamQuakeUserDir (char *path, size_t pathsize, const char *libr
 	return stat (path, &st) == 0 && S_ISDIR (st.st_mode);
 }
 
+qboolean Sys_GetSteamAPILibraryPath (char *path, size_t pathsize, const steamgame_t *game)
+{
+	char		config_info_path[MAX_OSPATH];
+	char		*line = NULL;
+	size_t		line_size = 0;
+	ssize_t		bytes_read = 0;
+	FILE		*config_info;
+	int			read_lines;
+	qboolean	result;
+
+	if ((size_t) q_snprintf (config_info_path, sizeof (config_info_path), "%s/steamapps/compatdata/%d/config_info", game->library, game->appid) >= pathsize)
+		return false;
+
+	config_info = Sys_fopen (config_info_path, "r");
+	if (!config_info)
+		return false;
+
+	// lib dir is on line 3, lib64 on line 4
+	read_lines = sizeof (void *) == 4 ? 3 : 4;
+	while (read_lines-- > 0)
+	{
+		if ((bytes_read = getline (&line, &line_size, config_info)) == -1)
+		{
+			fclose (config_info);
+			free (line);
+			return false;
+		}
+	}
+
+	fclose (config_info);
+	if (!line)
+		return false;
+
+	line_size = strlen (line);
+
+	if (line_size > 0 && line[line_size - 1] == '\n')
+		line[--line_size] = '\0';
+	if (line_size > 0 && line[line_size - 1] == '/')
+		line[--line_size] = '\0';
+
+	result = (size_t) q_snprintf (path, pathsize, "%s/libsteam_api.so", line) < pathsize;
+
+	free (line);
+
+	return result;
+}
+
 qboolean Sys_GetGOGQuakeDir (char *path, size_t pathsize)
 {
 	return false;
@@ -359,6 +427,23 @@ static char	cwd[MAX_OSPATH];
 #ifdef DO_USERDIRS
 static char	userdir[MAX_OSPATH];
 
+#ifdef PLATFORM_HAIKU
+#include <FindDirectory.h>
+#include <fs_info.h>
+
+static void Sys_GetUserdir (char *dst, size_t dstsize)
+{
+	dev_t volume = dev_for_path("/boot");
+	char buffer[B_PATH_NAME_LENGTH];
+	status_t result;
+
+	result = find_directory(B_USER_NONPACKAGED_DATA_DIRECTORY, volume, false, buffer, sizeof(buffer));
+	if (result != B_OK)
+		Sys_Error ("Couldn't determine userspace directory");
+
+	q_snprintf (dst, dstsize, "%s/%s", buffer, SYS_USERDIR);
+}
+#else
 static void Sys_GetUserdir (char *dst, size_t dstsize)
 {
 	size_t		n;
@@ -385,6 +470,7 @@ static void Sys_GetUserdir (char *dst, size_t dstsize)
 
 	q_snprintf (dst, dstsize, "%s/%s", home_dir, SYS_USERDIR);
 }
+#endif	/* PLATFORM_HAIKU */
 #endif	/* DO_USERDIRS */
 
 qboolean Sys_GetAltUserPrefDir (qboolean remastered, char *dst, size_t dstsize)
@@ -409,6 +495,80 @@ qboolean Sys_GetAltUserPrefDir (qboolean remastered, char *dst, size_t dstsize)
 	return true;
 }
 
+static qboolean Sys_Exec (const char *cmd, ...)
+{
+	pid_t p = fork ();
+	if (p < 0) // fork failed
+		return false;
+	else if (p == 0) // child process
+	{
+		va_list		argptr;
+		const char	*argv[1024];
+		size_t		argc;
+		FILE		*dummy;
+
+		argv[0] = (char *)cmd;
+		argc = 1;
+
+		va_start (argptr, cmd);
+		for (; argc + 1 < countof (argv); argc++)
+		{
+			const char *cur = va_arg (argptr, const char*);
+			if (!cur)
+				break;
+			argv[argc] = cur;
+		}
+		va_end (argptr);
+		argv[argc++] = NULL;
+
+		// Disable stdout/stderr
+		// Note: using a dummy variable to avoid triggering -Wunused-result
+		dummy = freopen ("/dev/null", "w", stdout); (void)dummy;
+		dummy = freopen ("/dev/null", "w", stderr); (void)dummy;
+
+		execvp (cmd, (char **) argv);
+
+		exit (EXIT_FAILURE);
+	}
+	else // original process
+	{
+		return true;
+	}
+}
+
+qboolean Sys_Explore (const char *path)
+{
+	char buf[32768];
+	char *s;
+
+	if (Sys_FileType (path) == FS_ENT_NONE)
+		return false;
+
+	// Try to identify the current desktop so we can open the parent dir in the file manager *and* select the right file in it
+	s = getenv ("XDG_CURRENT_DESKTOP");
+	if (s)
+	{
+		char *cur = NULL;
+		char *desktop = NULL;
+		q_strlcpy (buf, s, sizeof (buf));
+		for (desktop = strtok_r (buf, ":", &cur); desktop; desktop = strtok_r (NULL, ":", &cur))
+		{
+			if (q_strcasecmp (desktop, "gnome") == 0)
+				return Sys_Exec ("nautilus", "--select", path);
+			if (q_strcasecmp (desktop, "kde") == 0)
+				return Sys_Exec ("dolphin", "--select", path);
+		}
+	}
+
+	// Fall back to just opening the parent dir without selecting the file
+	q_strlcpy (buf, path, sizeof (buf));
+	s = Q_strrchr (buf, '/');
+	if (!s)
+		return false;
+	s[1] = '\0'; // terminate after the slash
+	return SDL_OpenURL (buf) == 0;
+}
+
 #ifdef PLATFORM_OSX
 static char *OSX_StripAppBundle (char *dir)
 { /* based on the ioquake3 project at icculus.org. */
@@ -454,6 +614,21 @@ static void Sys_GetBasedir (char *argv0, char *dst, size_t dstsize)
 {
 	char	*tmp;
 
+	#ifdef PLATFORM_HAIKU
+	if (realpath(argv0, dst) == NULL)
+	{
+		perror("realpath");
+		if (getcwd(dst, dstsize - 1) == NULL)
+	_fail:		Sys_Error ("Couldn't determine current directory");
+	}
+	else
+	{
+		/* strip off the binary name */
+		if (! (tmp = strdup (dst))) goto _fail;
+		q_strlcpy (dst, dirname(tmp), dstsize);
+		free (tmp);
+	}
+	#else
 	if (getcwd(dst, dstsize - 1) == NULL)
 		Sys_Error ("Couldn't determine current directory");
 
@@ -466,6 +641,7 @@ static void Sys_GetBasedir (char *argv0, char *dst, size_t dstsize)
 		if (tmp != dst && *tmp == '/')
 			*tmp = 0;
 	}
+	#endif
 }
 #endif
 
@@ -536,6 +712,8 @@ findfile_t *Sys_FindFirst (const char *dir, const char *ext)
 	}
 
 	ret = (unixfindfile_t *) calloc (1, sizeof (unixfindfile_t));
+	if (!ret)
+		Sys_Error ("Sys_FindFirst: out of memory");
 	ret->handle = handle;
 	ret->data = data;
 	q_strlcpy (ret->filter, ext, sizeof (ret->filter));
@@ -567,8 +745,54 @@ void Sys_FindClose (findfile_t *find)
 	}
 }
 
+/*
+=================
+Sys_GetParentProcessName
+=================
+*/
+static qboolean Sys_GetParentProcessName (char *dst, size_t dstsize)
+{
+#ifdef __linux__
+	char link[MAX_OSPATH];
+	q_snprintf (link, sizeof (link), "/proc/%d/exe", getppid ());
+	if (readlink (link, dst, dstsize) < 0)
+		return false;
+	dst[dstsize - 1] = '\0';
+	return true;
+#else
+	return false;
+#endif
+}
+
+/*
+=================
+Sys_IsStartedFromMapEditor
+
+Returns true if the process was started from TrenchBroom
+=================
+*/
+qboolean Sys_IsStartedFromMapEditor (void)
+{
+	char path[MAX_OSPATH];
+	const char *slash, *exe;
+
+	if (!Sys_GetParentProcessName (path, sizeof (path)))
+		return false;
+
+	slash = strrchr (path, '/');
+	exe = slash ? slash + 1 : path;
+
+	return q_strcasestr (exe, "trenchbroom") != NULL || q_strcasestr (exe, "jack") != NULL;
+}
+
 void Sys_Init (void)
 {
+	const char* term = getenv("TERM");
+	stdinIsATTY = isatty(STDIN_FILENO) &&
+			!(term && (!strcmp(term, "raw") || !strcmp(term, "dumb")));
+	if (!stdinIsATTY)
+		Sys_Printf("Terminal input not available.\n");
+
 	memset (cwd, 0, sizeof(cwd));
 	Sys_GetBasedir(host_parms->argv[0], cwd, sizeof(cwd));
 	host_parms->basedir = cwd;
@@ -603,10 +827,15 @@ void Sys_mkdir (const char *path)
 	}
 }
 
+qboolean Sys_IsDebuggerPresent (void)
+{
+	return false;
+}
+
 static const char errortxt1[] = "\nERROR-OUT BEGIN\n\n";
 static const char errortxt2[] = "\nQUAKE ERROR: ";
 
-void Sys_Error (const char *error, ...)
+void Sys_ReportError (const char *error, ...)
 {
 	va_list		argptr;
 	char		text[1024];
@@ -640,6 +869,9 @@ void Sys_Printf (const char *fmt, ...)
 
 	UTF8_FromQuake (u8text, sizeof (u8text), qtext);
 	printf ("%s", u8text);
+
+	// log all messages to file as well if -condebug was specified
+	Con_DebugLog (u8text);
 }
 
 void Sys_Quit (void)
@@ -656,12 +888,16 @@ double Sys_DoubleTime (void)
 
 const char *Sys_ConsoleInput (void)
 {
+	static qboolean	con_eof = false;
 	static char	con_text[256];
 	static int	textlen;
 	char		c;
 	fd_set		set;
 	struct timeval	timeout;
 
+	if (!stdinIsATTY || con_eof)
+		return NULL;
+
 	FD_ZERO (&set);
 	FD_SET (0, &set);	// stdin
 	timeout.tv_sec = 0;
@@ -669,7 +905,13 @@ const char *Sys_ConsoleInput (void)
 
 	while (select (1, &set, NULL, NULL, &timeout))
 	{
-		read (0, &c, 1);
+		if (read(0, &c, 1) <= 0)
+		{
+			// Finish processing whatever is already in the
+			// buffer (if anything), then stop reading
+			con_eof = true;
+			c = '\n';
+		}
 		if (c == '\n' || c == '\r')
 		{
 			con_text[textlen] = '\0';
@@ -717,3 +959,18 @@ void Sys_SendKeyEvents (void)
 void Sys_ActivateKeyFilter (qboolean active)
 {
 }
+
+void *Sys_LoadLibrary (const char *path)
+{
+	return dlopen (path, RTLD_LAZY);
+}
+
+void *Sys_GetLibraryFunction (void *lib, const char *func)
+{
+	return dlsym (lib, func);
+}
+
+void Sys_CloseLibrary (void *lib)
+{
+	dlclose (lib);
+}
