@@ -22,6 +22,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
 #include "quakedef.h"
+#include "steam.h"
 
 #ifndef MICROSOFT_WINDOWS_WINBASE_H_DEFINE_INTERLOCKED_CPLUSPLUS_OVERLOADS
 #define MICROSOFT_WINDOWS_WINBASE_H_DEFINE_INTERLOCKED_CPLUSPLUS_OVERLOADS 0
@@ -29,10 +30,15 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #ifndef WIN32_LEAN_AND_MEAN
 #define WIN32_LEAN_AND_MEAN
 #endif
+#ifdef _WIN32_WINNT
+#undef _WIN32_WINNT
+#endif
+#define _WIN32_WINNT 0x0600
 #include <windows.h>
 #include <mmsystem.h>
 #include <winreg.h>
 #include <versionhelpers.h>
+#include <tlhelp32.h>
 
 #ifdef _MSC_VER
 #pragma warning (push)
@@ -143,6 +149,16 @@ FILE *Sys_fopen (const char *path, const char *mode)
 	return f;
 }
 
+int Sys_fseek (FILE *file, qfileofs_t ofs, int origin)
+{
+	return _fseeki64 (file, ofs, origin);
+}
+
+qfileofs_t Sys_ftell (FILE *file)
+{
+	return _ftelli64 (file);
+}
+
 int Sys_remove (const char *path)
 {
 	wchar_t	wpath[MAX_PATH];
@@ -159,22 +175,23 @@ int Sys_rename (const char *oldname, const char *newname)
 	return _wrename (oldnamew, newnamew);
 }
 
-long Sys_filelength (FILE *f)
+qfileofs_t Sys_filelength (FILE *f)
 {
-	long		pos, end;
+	qfileofs_t	pos, end;
 
-	pos = ftell (f);
-	fseek (f, 0, SEEK_END);
-	end = ftell (f);
-	fseek (f, pos, SEEK_SET);
+	pos = Sys_ftell (f);
+	Sys_fseek (f, 0, SEEK_END);
+	end = Sys_ftell (f);
+	Sys_fseek (f, pos, SEEK_SET);
 
 	return end;
 }
 
-int Sys_FileOpenRead (const char *path, int *hndl)
+qfileofs_t Sys_FileOpenRead (const char *path, int *hndl)
 {
-	FILE	*f;
-	int		i, retval;
+	FILE		*f;
+	int			i;
+	qfileofs_t	retval;
 
 	i = findhandle ();
 	f = Sys_fopen (path, "rb");
@@ -405,6 +422,17 @@ qboolean Sys_GetSteamQuakeUserDir (char *path, size_t pathsize, const char *libr
 	return Sys_GetNightdiveDir (path, pathsize);
 }
 
+qboolean Sys_GetSteamAPILibraryPath (char *path, size_t pathsize, const steamgame_t *game)
+{
+#ifdef _WIN64
+	char installdir[MAX_OSPATH];
+	if (!Steam_ResolvePath (installdir, sizeof (installdir), game))
+		return false;
+	return (size_t) q_snprintf (path, pathsize, "%s/rerelease/steam_api64.dll", installdir) < pathsize;
+#endif
+	return false;
+}
+
 qboolean Sys_GetGOGQuakeEnhancedUserDir (char *path, size_t pathsize)
 {
 	return Sys_GetNightdiveDir (path, pathsize);
@@ -420,7 +448,7 @@ qboolean Sys_GetEGSManifestDir (char *path, size_t pathsize)
 
 const char *Sys_GetEGSLauncherData (void)
 {
-	char		path[MAX_PATH];
+	char		path[MAX_OSPATH];
 	char		*buf;
 	FILE		*file;
 	int64_t		filesize;
@@ -508,6 +536,57 @@ qboolean Sys_GetAltUserPrefDir (qboolean remastered, char *path, size_t pathsize
 	return Sys_GetKnownFolder (&FOLDERID_SavedGames, subdir, path, pathsize);
 }
 
+qboolean Sys_Explore (const char *path)
+{
+	wchar_t			wpath[MAX_PATH];
+	LPITEMIDLIST	file, folder;
+	HRESULT			hr;
+	SFGAOF			sfgaof;
+	int				i, slash;
+	qboolean		result = false;
+
+	if (Sys_FileType (path) == FS_ENT_NONE)
+		return false;
+
+	UTF8ToWideString (path, wpath, countof (wpath));
+	for (i = 0, slash = -1; wpath[i]; i++)
+	{
+		if (wpath[i] == L'/')
+			wpath[i] = L'\\';
+		if (wpath[i] == L'\\')
+			slash = i;
+	}
+
+	if (slash == -1)
+		return false;
+
+	hr = Sys_InitCOM ();
+	if (FAILED (hr))
+		return false;
+
+	wpath[slash] = L'\0';
+	hr = SHParseDisplayName (wpath, NULL, &folder, 0, &sfgaof);
+	if (FAILED (hr))
+		goto cleanup_com;
+
+	wpath[slash] = L'\\';
+	hr = SHParseDisplayName (wpath, NULL, &file, 0, &sfgaof);
+	if (FAILED (hr))
+		goto cleanup_folder;
+
+	hr = SHOpenFolderAndSelectItems (folder, 1, (LPCITEMIDLIST *) &file, 0);
+	if (SUCCEEDED (hr))
+		result = true;
+
+	CoTaskMemFree (file);
+cleanup_folder:
+	CoTaskMemFree (folder);
+cleanup_com:
+	CoUninitialize ();
+
+	return result;
+}
+
 static char	cwd[1024];
 
 static void Sys_GetBasedir (char *argv0, char *dst, size_t dstsize)
@@ -576,7 +655,7 @@ static void Sys_FillFindData (winfindfile_t *find)
 findfile_t *Sys_FindFirst (const char *dir, const char *ext)
 {
 	winfindfile_t		*ret;
-	char				pattern[MAX_PATH];
+	char				pattern[MAX_OSPATH];
 	wchar_t				wpattern[MAX_PATH];
 	HANDLE				handle;
 	WIN32_FIND_DATAW	data;
@@ -594,6 +673,8 @@ findfile_t *Sys_FindFirst (const char *dir, const char *ext)
 		return NULL;
 
 	ret = (winfindfile_t *) calloc (1, sizeof (winfindfile_t));
+	if (!ret)
+		Sys_Error ("Sys_FindFirst: out of memory");
 	ret->handle = handle;
 	ret->data = data;
 	Sys_FillFindData (ret);
@@ -723,7 +804,7 @@ LRESULT CALLBACK KeyFilter (int nCode, WPARAM wParam, LPARAM lParam)
 			qboolean intercept =
 				force_intercept ||
 				key == HK_PRINTSCREEN ||
-				(key_dest == key_game || M_KeyBinding ())
+				(key_dest == key_game || M_WaitingForKeyBinding ())
 			;
 
 			if (intercept)
@@ -746,6 +827,100 @@ LRESULT CALLBACK KeyFilter (int nCode, WPARAM wParam, LPARAM lParam)
 	return CallNextHookEx (NULL, nCode, wParam, lParam);
 }
 
+typedef struct
+{
+	WCHAR name[MAX_PATH];
+	DWORD id;
+	DWORD parent_id;
+} procinfo_t;
+
+/*
+=================
+Sys_GetProcList
+=================
+*/
+static procinfo_t *Sys_GetProcList (void)
+{
+	HANDLE				snapshot;
+	PROCESSENTRY32W		proc_entry;
+	procinfo_t			*result = NULL;
+
+	snapshot = CreateToolhelp32Snapshot (TH32CS_SNAPPROCESS, 0);
+	if (snapshot == INVALID_HANDLE_VALUE)
+		return NULL;
+
+	proc_entry.dwSize = sizeof (proc_entry);
+
+	if (Process32FirstW (snapshot, &proc_entry))
+	{
+		do
+		{
+			procinfo_t info;
+			wcscpy_s (info.name, Q_COUNTOF (info.name), proc_entry.szExeFile);
+			info.id = proc_entry.th32ProcessID;
+			info.parent_id = proc_entry.th32ParentProcessID;
+			VEC_PUSH (result, info);
+		} while (Process32NextW (snapshot, &proc_entry));
+	}
+
+	CloseHandle (snapshot);
+
+	return result;
+}
+
+/*
+=================
+Sys_FindProc
+=================
+*/
+static procinfo_t *Sys_FindProc (DWORD id, procinfo_t *procs)
+{
+	size_t i;
+	if (id == 0)
+		return NULL;
+	for (i = 0; i < VEC_SIZE (procs); i++)
+		if (procs[i].id == id)
+			return &procs[i];
+	return NULL;
+}
+
+/*
+=================
+Sys_IsStartedFromMapEditor
+
+Returns true if the process was started from TrenchBroom, JACK, ne_q1spCompilingGui, or q1compile
+=================
+*/
+qboolean Sys_IsStartedFromMapEditor (void)
+{
+	qboolean from_editor = false;
+	procinfo_t *proclist = Sys_GetProcList ();
+
+	if (proclist)
+	{
+		procinfo_t *current = Sys_FindProc (GetCurrentProcessId (), proclist);
+		procinfo_t *parent = current ? Sys_FindProc (current->parent_id, proclist) : NULL;
+
+		while (parent && _wcsicmp (parent->name, L"cmd.exe") == 0)
+			parent = Sys_FindProc (parent->parent_id, proclist);
+
+		if (parent)
+		{
+			#define PARENT_STARTS_WITH(prefix)	(_wcsnicmp (parent->name, L##prefix, wcslen (L##prefix)) == 0)
+			if (PARENT_STARTS_WITH ("TrenchBroom") ||
+				PARENT_STARTS_WITH ("jack") ||
+				PARENT_STARTS_WITH ("ne_q1spCompilingGui") ||
+				PARENT_STARTS_WITH ("q1compile"))
+				from_editor = true;
+			#undef PARENT_STARTS_WITH
+		}
+
+		VEC_FREE (proclist);
+	}
+
+	return from_editor;
+}
+
 void Sys_Init (void)
 {
 	SYSTEM_INFO info;
@@ -806,7 +981,12 @@ void Sys_mkdir (const char *path)
 static const wchar_t errortxt1[] = L"\nERROR-OUT BEGIN\n\n";
 static const wchar_t errortxt2[] = L"\nQUAKE ERROR: ";
 
-void Sys_Error (const char *error, ...)
+qboolean Sys_IsDebuggerPresent (void)
+{
+	return IsDebuggerPresent ();
+}
+
+void Sys_ReportError (const char *error, ...)
 {
 	va_list		argptr;
 	char		text[1024];
@@ -842,9 +1022,6 @@ void Sys_Error (const char *error, ...)
 		SDL_Delay (3000);	/* show the console 3 more seconds */
 	}
 
-	if (IsDebuggerPresent ())
-		DebugBreak ();
-
 	exit (1);
 }
 
@@ -861,6 +1038,10 @@ void Sys_Printf (const char *fmt, ...)
 	va_end (argptr);
 
 	UTF8_FromQuake (u8text, sizeof (u8text), qtext);
+
+	// log all messages to file as well if -condebug was specified
+	Con_DebugLog (u8text);
+
 	len = MultiByteToWideChar (CP_UTF8, 0, u8text, -1, wtext, countof (wtext));
 	if (!len)
 		return;
@@ -989,3 +1170,20 @@ void Sys_ActivateKeyFilter (qboolean active)
 			Sys_Printf ("Warning: SetWindowsHookExW failed (%lu)\n", GetLastError ());
 	}
 }
+
+void *Sys_LoadLibrary (const char *path)
+{
+	wchar_t wpath[MAX_PATH];
+	UTF8ToWideString (path, wpath, countof (wpath));
+	return (void *) LoadLibraryW (wpath);
+}
+
+void *Sys_GetLibraryFunction (void *lib, const char *func)
+{
+	return GetProcAddress ((HMODULE) lib, func);
+}
+
+void Sys_CloseLibrary (void *lib)
+{
+	FreeLibrary ((HMODULE) lib);
+}
