@@ -60,18 +60,14 @@ cvar_t	v_idlescale = {"v_idlescale", "0", CVAR_NONE};
 cvar_t	crosshair = {"crosshair", "0", CVAR_ARCHIVE};
 char	crosshair_char = '\0';
 
-cvar_t	gl_cshiftpercent = {"gl_cshiftpercent", "100", CVAR_NONE};
-cvar_t	gl_cshiftpercent_contents = {"gl_cshiftpercent_contents", "100", CVAR_NONE}; // QuakeSpasm
-cvar_t	gl_cshiftpercent_damage = {"gl_cshiftpercent_damage", "100", CVAR_NONE}; // QuakeSpasm
-cvar_t	gl_cshiftpercent_bonus = {"gl_cshiftpercent_bonus", "100", CVAR_NONE}; // QuakeSpasm
-cvar_t	gl_cshiftpercent_powerup = {"gl_cshiftpercent_powerup", "100", CVAR_NONE}; // QuakeSpasm
+cvar_t	gl_cshiftpercent = {"gl_cshiftpercent", "100", CVAR_ARCHIVE};
+cvar_t	gl_cshiftpercent_contents = {"gl_cshiftpercent_contents", "100", CVAR_ARCHIVE}; // QuakeSpasm
+cvar_t	gl_cshiftpercent_damage = {"gl_cshiftpercent_damage", "100", CVAR_ARCHIVE}; // QuakeSpasm
+cvar_t	gl_cshiftpercent_bonus = {"gl_cshiftpercent_bonus", "100", CVAR_ARCHIVE}; // QuakeSpasm
+cvar_t	gl_cshiftpercent_powerup = {"gl_cshiftpercent_powerup", "100", CVAR_ARCHIVE}; // QuakeSpasm
 
 cvar_t	r_viewmodel_quake = {"r_viewmodel_quake", "0", CVAR_ARCHIVE};
 
-float	v_dmg_time, v_dmg_roll, v_dmg_pitch;
-
-extern	int			in_forward, in_forward2, in_back;
-
 vec3_t	v_punchangles[2]; //johnfitz -- copied from cl.punchangle.  0 is current, 1 is previous value. never the same unless map just loaded
 
 /*
@@ -150,6 +146,7 @@ cvar_t	v_centerspeed = {"v_centerspeed","500", CVAR_NONE};
 
 void V_StartPitchDrift (void)
 {
+	cl.lastcenterstart = cl.time;
 #if 1
 	if (cl.laststop == cl.time)
 	{
@@ -253,10 +250,12 @@ void V_DriftPitch (void)
 ==============================================================================
 */
 
-cshift_t	cshift_empty = { {130,80,50}, 0 };
-cshift_t	cshift_water = { {130,80,50}, 128 };
-cshift_t	cshift_slime = { {0,25,5}, 150 };
-cshift_t	cshift_lava = { {255,80,0}, 150 };
+cshift_t cshift_empty = { {130,80,50}, 0 };
+static cshift_t cshift_water = { {130,80,50}, 128 };
+static cshift_t cshift_slime = { {0,25,5}, 150 };
+static cshift_t cshift_lava = { {255,80,0}, 150 };
+
+static float v_dmg_time, v_dmg_roll, v_dmg_pitch;
 
 float		v_blend[4];		// rgba 0.0 - 1.0; see note in V_PolyBlend for more info
 
@@ -449,15 +448,17 @@ V_CalcBlend
 */
 void V_CalcBlend (void)
 {
-	float	r, g, b, a, a2;
-	int		j;
-	cvar_t	*cshiftpercent_cvars[NUM_CSHIFTS] = {
+	static const cvar_t	* const cshiftpercent_cvars[NUM_CSHIFTS] =
+	{
 		&gl_cshiftpercent_contents,
 		&gl_cshiftpercent_damage,
 		&gl_cshiftpercent_bonus,
 		&gl_cshiftpercent_powerup
 	};
 
+	float	r, g, b, a, a2;
+	int		j;
+
 	r = 0;
 	g = 0;
 	b = 0;
@@ -504,7 +505,7 @@ V_UpdateBlend -- johnfitz -- V_UpdatePalette cleaned up and renamed
 void V_UpdateBlend (void)
 {
 	int		i, j;
-	float	frametime = cl.time - cl.oldtime;
+	float	frametime = fabs (cl.time - cl.oldtime); // time can go backwards when rewinding demos
 	qboolean	blend_changed;
 
 	V_CalcPowerupCshift ();
@@ -717,7 +718,7 @@ void V_CalcViewRoll (void)
 	{
 		r_refdef.viewangles[ROLL] += v_dmg_time/v_kicktime.value*v_dmg_roll;
 		r_refdef.viewangles[PITCH] += v_dmg_time/v_kicktime.value*v_dmg_pitch;
-		v_dmg_time -= host_frametime;
+		v_dmg_time -= fabs (cl.time - cl.oldtime);
 	}
 
 	if (cl.stats[STAT_HEALTH] <= 0)
@@ -837,6 +838,14 @@ void V_CalcRefdef (void)
 			view->origin[2] += 0.5;
 	}
 
+	if (ent->lerpflags & LERP_FINISH)
+	{
+		view->lerpflags |= LERP_FINISH;
+		view->lerpfinish = ent->lerpfinish;
+	}
+	else
+		view->lerpflags &= ~LERP_FINISH;
+
 	view->model = cl.model_precache[cl.stats[STAT_WEAPON]];
 	view->frame = cl.stats[STAT_WEAPONFRAME];
 	view->colormap = vid.colormap;
