@@ -339,8 +339,10 @@ void SV_TouchLinks (edict_t *ent)
 	edict_t		*touch;
 	int		old_self, old_other;
 	int		i, listcount;
+	int		mark;
 
-	list = alloca (qcvm->num_edicts*sizeof(edict_t *));
+	mark = Hunk_LowMark ();
+	list = (edict_t **) Hunk_AllocNoFill (qcvm->num_edicts*sizeof(edict_t *));
 
 	listcount = 0;
 	SV_AreaTriggerEdicts (ent, sv_areanodes, list, &listcount, qcvm->num_edicts);
@@ -372,6 +374,9 @@ void SV_TouchLinks (edict_t *ent)
 		pr_global_struct->self = old_self;
 		pr_global_struct->other = old_other;
 	}
+
+// free hunk-allocated edicts array
+	Hunk_FreeToLowMark (mark);
 }
 
 
@@ -391,13 +396,13 @@ void SV_FindTouchedLeafs (edict_t *ent, mnode_t *node)
 	if (node->contents == CONTENTS_SOLID)
 		return;
 
+	if (ent->num_leafs == MAX_ENT_LEAFS)
+		return;
+
 // add an efrag if the node is a leaf
 
 	if ( node->contents < 0)
 	{
-		if (ent->num_leafs == MAX_ENT_LEAFS)
-			return;
-
 		leaf = (mleaf_t *)node;
 		leafnum = leaf - sv.worldmodel->leafs - 1;
 
@@ -613,7 +618,7 @@ edict_t	*SV_TestEntityPosition (edict_t *ent)
 	trace = SV_Move (ent->v.origin, ent->v.mins, ent->v.maxs, ent->v.origin, 0, ent);
 
 	if (trace.startsolid)
-		return qcvm->edicts;
+		return trace.ent ? trace.ent : qcvm->edicts;
 
 	return NULL;
 }
